import mordred
from mordred import Calculator
from mordred import ABCIndex, AcidBase, AdjacencyMatrix, Aromatic, AtomCount, Autocorrelation, BCUT, BalabanJ, BaryszMatrix, BertzCT, BondCount, CPSA, CarbonTypes, Chi, Constitutional, DetourMatrix, DistanceMatrix, EState, EccentricConnectivityIndex, ExtendedTopochemicalAtom, FragmentComplexity, Framework, GeometricalIndex, GravitationalIndex, HydrogenBond, InformationContent, KappaShapeIndex, Lipinski, LogS, McGowanVolume, MoRSE, MoeType, MolecularDistanceEdge, MolecularId, MomentOfInertia, PBF, PathCount, Polarizability, RingCount, RotatableBond, SLogP, TopoPSA, TopologicalCharge, TopologicalIndex, VdwVolumeABC, VertexAdjacencyInformation, WalkCount, Weight, WienerIndex, ZagrebIndex
from rdkit import Chem
import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin
import re
import parser

class smilesListToDataframe(BaseEstimator, TransformerMixin):
    def fit(self, X, y=None):
        return self
    def transform(self, X, y=None):
        return pd.DataFrame(index=X)



class calculateDescriptors(BaseEstimator, TransformerMixin):
    descriptors_dict={'ABC': {'function': mordred.ABCIndex.ABCIndex(), 'type': '2D', 'description': 'atom-bond connectivity index', 'module': 'ABCIndex', 'constructor': 'ABCIndex()'}, 'ABCGG': {'function': mordred.ABCIndex.ABCGGIndex(), 'type': '2D', 'description': 'Graovac-Ghorbani atom-bond connectivity index', 'module': 'ABCIndex', 'constructor': 'ABCGGIndex()'}, 'nAcid': {'function': mordred.AcidBase.AcidicGroupCount(), 'type': '2D', 'description': 'acidic group count', 'module': 'AcidBase', 'constructor': 'AcidicGroupCount()'}, 'nBase': {'function': mordred.AcidBase.BasicGroupCount(), 'type': '2D', 'description': 'basic group count', 'module': 'AcidBase', 'constructor': 'BasicGroupCount()'}, 'SpAbs_A': {'function': mordred.AdjacencyMatrix.AdjacencyMatrix('SpAbs'), 'type': '2D', 'description': 'SpAbs of adjacency matrix', 'module': 'AdjacencyMatrix', 'constructor': 'AdjacencyMatrix("SpAbs")'}, 'SpMax_A': {'function': mordred.AdjacencyMatrix.AdjacencyMatrix('SpMax'), 'type': '2D', 'description': 'SpMax of adjacency matrix', 'module': 'AdjacencyMatrix', 'constructor': 'AdjacencyMatrix("SpMax")'}, 'SpDiam_A': {'function': mordred.AdjacencyMatrix.AdjacencyMatrix('SpDiam'), 'type': '2D', 'description': 'SpDiam of adjacency matrix', 'module': 'AdjacencyMatrix', 'constructor': 'AdjacencyMatrix("SpDiam")'}, 'SpAD_A': {'function': mordred.AdjacencyMatrix.AdjacencyMatrix('SpAD'), 'type': '2D', 'description': 'SpAD of adjacency matrix', 'module': 'AdjacencyMatrix', 'constructor': 'AdjacencyMatrix("SpAD")'}, 'SpMAD_A': {'function': mordred.AdjacencyMatrix.AdjacencyMatrix('SpMAD'), 'type': '2D', 'description': 'SpMAD of adjacency matrix', 'module': 'AdjacencyMatrix', 'constructor': 'AdjacencyMatrix("SpMAD")'}, 'LogEE_A': {'function': mordred.AdjacencyMatrix.AdjacencyMatrix('LogEE'), 'type': '2D', 'description': 'LogEE of adjacency matrix', 'module': 'AdjacencyMatrix', 'constructor': 'AdjacencyMatrix("LogEE")'}, 'VE1_A': {'function': mordred.AdjacencyMatrix.AdjacencyMatrix('VE1'), 'type': '2D', 'description': 'VE1 of adjacency matrix', 'module': 'AdjacencyMatrix', 'constructor': 'AdjacencyMatrix("VE1")'}, 'VE2_A': {'function': mordred.AdjacencyMatrix.AdjacencyMatrix('VE2'), 'type': '2D', 'description': 'VE2 of adjacency matrix', 'module': 'AdjacencyMatrix', 'constructor': 'AdjacencyMatrix("VE2")'}, 'VE3_A': {'function': mordred.AdjacencyMatrix.AdjacencyMatrix('VE3'), 'type': '2D', 'description': 'VE3 of adjacency matrix', 'module': 'AdjacencyMatrix', 'constructor': 'AdjacencyMatrix("VE3")'}, 'VR1_A': {'function': mordred.AdjacencyMatrix.AdjacencyMatrix('VR1'), 'type': '2D', 'description': 'VR1 of adjacency matrix', 'module': 'AdjacencyMatrix', 'constructor': 'AdjacencyMatrix("VR1")'}, 'VR2_A': {'function': mordred.AdjacencyMatrix.AdjacencyMatrix('VR2'), 'type': '2D', 'description': 'VR2 of adjacency matrix', 'module': 'AdjacencyMatrix', 'constructor': 'AdjacencyMatrix("VR2")'}, 'VR3_A': {'function': mordred.AdjacencyMatrix.AdjacencyMatrix('VR3'), 'type': '2D', 'description': 'VR3 of adjacency matrix', 'module': 'AdjacencyMatrix', 'constructor': 'AdjacencyMatrix("VR3")'}, 'nAromAtom': {'function': mordred.Aromatic.AromaticAtomsCount(), 'type': '2D', 'description': 'aromatic atoms count', 'module': 'Aromatic', 'constructor': 'AromaticAtomsCount()'}, 'nAromBond': {'function': mordred.Aromatic.AromaticBondsCount(), 'type': '2D', 'description': 'aromatic bonds count', 'module': 'Aromatic', 'constructor': 'AromaticBondsCount()'}, 'nAtom': {'function': mordred.AtomCount.AtomCount('Atom'), 'type': '2D', 'description': 'number of all atoms', 'module': 'AtomCount', 'constructor': 'AtomCount("Atom")'}, 'nHeavyAtom': {'function': mordred.AtomCount.AtomCount('HeavyAtom'), 'type': '2D', 'description': 'number of heavy atoms', 'module': 'AtomCount', 'constructor': 'AtomCount("HeavyAtom")'}, 'nSpiro': {'function': mordred.AtomCount.AtomCount('Spiro'), 'type': '2D', 'description': 'number of spiro atoms', 'module': 'AtomCount', 'constructor': 'AtomCount("Spiro")'}, 'nBridgehead': {'function': mordred.AtomCount.AtomCount('Bridgehead'), 'type': '2D', 'description': 'number of bridgehead atoms', 'module': 'AtomCount', 'constructor': 'AtomCount("Bridgehead")'}, 'nHetero': {'function': mordred.AtomCount.AtomCount('Hetero'), 'type': '2D', 'description': 'number of hetero atoms', 'module': 'AtomCount', 'constructor': 'AtomCount("Hetero")'}, 'nH': {'function': mordred.AtomCount.AtomCount('H'), 'type': '2D', 'description': 'number of H atoms', 'module': 'AtomCount', 'constructor': 'AtomCount("H")'}, 'nB': {'function': mordred.AtomCount.AtomCount('B'), 'type': '2D', 'description': 'number of B atoms', 'module': 'AtomCount', 'constructor': 'AtomCount("B")'}, 'nC': {'function': mordred.AtomCount.AtomCount('C'), 'type': '2D', 'description': 'number of C atoms', 'module': 'AtomCount', 'constructor': 'AtomCount("C")'}, 'nN': {'function': mordred.AtomCount.AtomCount('N'), 'type': '2D', 'description': 'number of N atoms', 'module': 'AtomCount', 'constructor': 'AtomCount("N")'}, 'nO': {'function': mordred.AtomCount.AtomCount('O'), 'type': '2D', 'description': 'number of O atoms', 'module': 'AtomCount', 'constructor': 'AtomCount("O")'}, 'nS': {'function': mordred.AtomCount.AtomCount('S'), 'type': '2D', 'description': 'number of S atoms', 'module': 'AtomCount', 'constructor': 'AtomCount("S")'}, 'nP': {'function': mordred.AtomCount.AtomCount('P'), 'type': '2D', 'description': 'number of P atoms', 'module': 'AtomCount', 'constructor': 'AtomCount("P")'}, 'nF': {'function': mordred.AtomCount.AtomCount('F'), 'type': '2D', 'description': 'number of F atoms', 'module': 'AtomCount', 'constructor': 'AtomCount("F")'}, 'nCl': {'function': mordred.AtomCount.AtomCount('Cl'), 'type': '2D', 'description': 'number of Cl atoms', 'module': 'AtomCount', 'constructor': 'AtomCount("Cl")'}, 'nBr': {'function': mordred.AtomCount.AtomCount('Br'), 'type': '2D', 'description': 'number of Br atoms', 'module': 'AtomCount', 'constructor': 'AtomCount("Br")'}, 'nI': {'function': mordred.AtomCount.AtomCount('I'), 'type': '2D', 'description': 'number of I atoms', 'module': 'AtomCount', 'constructor': 'AtomCount("I")'}, 'nX': {'function': mordred.AtomCount.AtomCount('X'), 'type': '2D', 'description': 'number of halogen atoms', 'module': 'AtomCount', 'constructor': 'AtomCount("X")'}, 'ATS0dv': {'function': mordred.Autocorrelation.ATS(0, 'dv'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 0 weighted by valence electrons', 'module': 'Autocorrelation', 'constructor': 'ATS(0, "dv")'}, 'ATS1dv': {'function': mordred.Autocorrelation.ATS(1, 'dv'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 1 weighted by valence electrons', 'module': 'Autocorrelation', 'constructor': 'ATS(1, "dv")'}, 'ATS2dv': {'function': mordred.Autocorrelation.ATS(2, 'dv'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 2 weighted by valence electrons', 'module': 'Autocorrelation', 'constructor': 'ATS(2, "dv")'}, 'ATS3dv': {'function': mordred.Autocorrelation.ATS(3, 'dv'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 3 weighted by valence electrons', 'module': 'Autocorrelation', 'constructor': 'ATS(3, "dv")'}, 'ATS4dv': {'function': mordred.Autocorrelation.ATS(4, 'dv'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 4 weighted by valence electrons', 'module': 'Autocorrelation', 'constructor': 'ATS(4, "dv")'}, 'ATS5dv': {'function': mordred.Autocorrelation.ATS(5, 'dv'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 5 weighted by valence electrons', 'module': 'Autocorrelation', 'constructor': 'ATS(5, "dv")'}, 'ATS6dv': {'function': mordred.Autocorrelation.ATS(6, 'dv'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 6 weighted by valence electrons', 'module': 'Autocorrelation', 'constructor': 'ATS(6, "dv")'}, 'ATS7dv': {'function': mordred.Autocorrelation.ATS(7, 'dv'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 7 weighted by valence electrons', 'module': 'Autocorrelation', 'constructor': 'ATS(7, "dv")'}, 'ATS8dv': {'function': mordred.Autocorrelation.ATS(8, 'dv'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 8 weighted by valence electrons', 'module': 'Autocorrelation', 'constructor': 'ATS(8, "dv")'}, 'ATS0d': {'function': mordred.Autocorrelation.ATS(0, 'd'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 0 weighted by sigma electrons', 'module': 'Autocorrelation', 'constructor': 'ATS(0, "d")'}, 'ATS1d': {'function': mordred.Autocorrelation.ATS(1, 'd'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 1 weighted by sigma electrons', 'module': 'Autocorrelation', 'constructor': 'ATS(1, "d")'}, 'ATS2d': {'function': mordred.Autocorrelation.ATS(2, 'd'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 2 weighted by sigma electrons', 'module': 'Autocorrelation', 'constructor': 'ATS(2, "d")'}, 'ATS3d': {'function': mordred.Autocorrelation.ATS(3, 'd'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 3 weighted by sigma electrons', 'module': 'Autocorrelation', 'constructor': 'ATS(3, "d")'}, 'ATS4d': {'function': mordred.Autocorrelation.ATS(4, 'd'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 4 weighted by sigma electrons', 'module': 'Autocorrelation', 'constructor': 'ATS(4, "d")'}, 'ATS5d': {'function': mordred.Autocorrelation.ATS(5, 'd'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 5 weighted by sigma electrons', 'module': 'Autocorrelation', 'constructor': 'ATS(5, "d")'}, 'ATS6d': {'function': mordred.Autocorrelation.ATS(6, 'd'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 6 weighted by sigma electrons', 'module': 'Autocorrelation', 'constructor': 'ATS(6, "d")'}, 'ATS7d': {'function': mordred.Autocorrelation.ATS(7, 'd'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 7 weighted by sigma electrons', 'module': 'Autocorrelation', 'constructor': 'ATS(7, "d")'}, 'ATS8d': {'function': mordred.Autocorrelation.ATS(8, 'd'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 8 weighted by sigma electrons', 'module': 'Autocorrelation', 'constructor': 'ATS(8, "d")'}, 'ATS0s': {'function': mordred.Autocorrelation.ATS(0, 's'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 0 weighted by intrinsic state', 'module': 'Autocorrelation', 'constructor': 'ATS(0, "s")'}, 'ATS1s': {'function': mordred.Autocorrelation.ATS(1, 's'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 1 weighted by intrinsic state', 'module': 'Autocorrelation', 'constructor': 'ATS(1, "s")'}, 'ATS2s': {'function': mordred.Autocorrelation.ATS(2, 's'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 2 weighted by intrinsic state', 'module': 'Autocorrelation', 'constructor': 'ATS(2, "s")'}, 'ATS3s': {'function': mordred.Autocorrelation.ATS(3, 's'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 3 weighted by intrinsic state', 'module': 'Autocorrelation', 'constructor': 'ATS(3, "s")'}, 'ATS4s': {'function': mordred.Autocorrelation.ATS(4, 's'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 4 weighted by intrinsic state', 'module': 'Autocorrelation', 'constructor': 'ATS(4, "s")'}, 'ATS5s': {'function': mordred.Autocorrelation.ATS(5, 's'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 5 weighted by intrinsic state', 'module': 'Autocorrelation', 'constructor': 'ATS(5, "s")'}, 'ATS6s': {'function': mordred.Autocorrelation.ATS(6, 's'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 6 weighted by intrinsic state', 'module': 'Autocorrelation', 'constructor': 'ATS(6, "s")'}, 'ATS7s': {'function': mordred.Autocorrelation.ATS(7, 's'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 7 weighted by intrinsic state', 'module': 'Autocorrelation', 'constructor': 'ATS(7, "s")'}, 'ATS8s': {'function': mordred.Autocorrelation.ATS(8, 's'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 8 weighted by intrinsic state', 'module': 'Autocorrelation', 'constructor': 'ATS(8, "s")'}, 'ATS0Z': {'function': mordred.Autocorrelation.ATS(0, 'Z'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 0 weighted by atomic number', 'module': 'Autocorrelation', 'constructor': 'ATS(0, "Z")'}, 'ATS1Z': {'function': mordred.Autocorrelation.ATS(1, 'Z'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 1 weighted by atomic number', 'module': 'Autocorrelation', 'constructor': 'ATS(1, "Z")'}, 'ATS2Z': {'function': mordred.Autocorrelation.ATS(2, 'Z'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 2 weighted by atomic number', 'module': 'Autocorrelation', 'constructor': 'ATS(2, "Z")'}, 'ATS3Z': {'function': mordred.Autocorrelation.ATS(3, 'Z'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 3 weighted by atomic number', 'module': 'Autocorrelation', 'constructor': 'ATS(3, "Z")'}, 'ATS4Z': {'function': mordred.Autocorrelation.ATS(4, 'Z'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 4 weighted by atomic number', 'module': 'Autocorrelation', 'constructor': 'ATS(4, "Z")'}, 'ATS5Z': {'function': mordred.Autocorrelation.ATS(5, 'Z'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 5 weighted by atomic number', 'module': 'Autocorrelation', 'constructor': 'ATS(5, "Z")'}, 'ATS6Z': {'function': mordred.Autocorrelation.ATS(6, 'Z'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 6 weighted by atomic number', 'module': 'Autocorrelation', 'constructor': 'ATS(6, "Z")'}, 'ATS7Z': {'function': mordred.Autocorrelation.ATS(7, 'Z'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 7 weighted by atomic number', 'module': 'Autocorrelation', 'constructor': 'ATS(7, "Z")'}, 'ATS8Z': {'function': mordred.Autocorrelation.ATS(8, 'Z'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 8 weighted by atomic number', 'module': 'Autocorrelation', 'constructor': 'ATS(8, "Z")'}, 'ATS0m': {'function': mordred.Autocorrelation.ATS(0, 'm'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 0 weighted by mass', 'module': 'Autocorrelation', 'constructor': 'ATS(0, "m")'}, 'ATS1m': {'function': mordred.Autocorrelation.ATS(1, 'm'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 1 weighted by mass', 'module': 'Autocorrelation', 'constructor': 'ATS(1, "m")'}, 'ATS2m': {'function': mordred.Autocorrelation.ATS(2, 'm'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 2 weighted by mass', 'module': 'Autocorrelation', 'constructor': 'ATS(2, "m")'}, 'ATS3m': {'function': mordred.Autocorrelation.ATS(3, 'm'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 3 weighted by mass', 'module': 'Autocorrelation', 'constructor': 'ATS(3, "m")'}, 'ATS4m': {'function': mordred.Autocorrelation.ATS(4, 'm'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 4 weighted by mass', 'module': 'Autocorrelation', 'constructor': 'ATS(4, "m")'}, 'ATS5m': {'function': mordred.Autocorrelation.ATS(5, 'm'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 5 weighted by mass', 'module': 'Autocorrelation', 'constructor': 'ATS(5, "m")'}, 'ATS6m': {'function': mordred.Autocorrelation.ATS(6, 'm'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 6 weighted by mass', 'module': 'Autocorrelation', 'constructor': 'ATS(6, "m")'}, 'ATS7m': {'function': mordred.Autocorrelation.ATS(7, 'm'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 7 weighted by mass', 'module': 'Autocorrelation', 'constructor': 'ATS(7, "m")'}, 'ATS8m': {'function': mordred.Autocorrelation.ATS(8, 'm'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 8 weighted by mass', 'module': 'Autocorrelation', 'constructor': 'ATS(8, "m")'}, 'ATS0v': {'function': mordred.Autocorrelation.ATS(0, 'v'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 0 weighted by vdw volume', 'module': 'Autocorrelation', 'constructor': 'ATS(0, "v")'}, 'ATS1v': {'function': mordred.Autocorrelation.ATS(1, 'v'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 1 weighted by vdw volume', 'module': 'Autocorrelation', 'constructor': 'ATS(1, "v")'}, 'ATS2v': {'function': mordred.Autocorrelation.ATS(2, 'v'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 2 weighted by vdw volume', 'module': 'Autocorrelation', 'constructor': 'ATS(2, "v")'}, 'ATS3v': {'function': mordred.Autocorrelation.ATS(3, 'v'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 3 weighted by vdw volume', 'module': 'Autocorrelation', 'constructor': 'ATS(3, "v")'}, 'ATS4v': {'function': mordred.Autocorrelation.ATS(4, 'v'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 4 weighted by vdw volume', 'module': 'Autocorrelation', 'constructor': 'ATS(4, "v")'}, 'ATS5v': {'function': mordred.Autocorrelation.ATS(5, 'v'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 5 weighted by vdw volume', 'module': 'Autocorrelation', 'constructor': 'ATS(5, "v")'}, 'ATS6v': {'function': mordred.Autocorrelation.ATS(6, 'v'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 6 weighted by vdw volume', 'module': 'Autocorrelation', 'constructor': 'ATS(6, "v")'}, 'ATS7v': {'function': mordred.Autocorrelation.ATS(7, 'v'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 7 weighted by vdw volume', 'module': 'Autocorrelation', 'constructor': 'ATS(7, "v")'}, 'ATS8v': {'function': mordred.Autocorrelation.ATS(8, 'v'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 8 weighted by vdw volume', 'module': 'Autocorrelation', 'constructor': 'ATS(8, "v")'}, 'ATS0se': {'function': mordred.Autocorrelation.ATS(0, 'se'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 0 weighted by sanderson EN', 'module': 'Autocorrelation', 'constructor': 'ATS(0, "se")'}, 'ATS1se': {'function': mordred.Autocorrelation.ATS(1, 'se'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 1 weighted by sanderson EN', 'module': 'Autocorrelation', 'constructor': 'ATS(1, "se")'}, 'ATS2se': {'function': mordred.Autocorrelation.ATS(2, 'se'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 2 weighted by sanderson EN', 'module': 'Autocorrelation', 'constructor': 'ATS(2, "se")'}, 'ATS3se': {'function': mordred.Autocorrelation.ATS(3, 'se'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 3 weighted by sanderson EN', 'module': 'Autocorrelation', 'constructor': 'ATS(3, "se")'}, 'ATS4se': {'function': mordred.Autocorrelation.ATS(4, 'se'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 4 weighted by sanderson EN', 'module': 'Autocorrelation', 'constructor': 'ATS(4, "se")'}, 'ATS5se': {'function': mordred.Autocorrelation.ATS(5, 'se'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 5 weighted by sanderson EN', 'module': 'Autocorrelation', 'constructor': 'ATS(5, "se")'}, 'ATS6se': {'function': mordred.Autocorrelation.ATS(6, 'se'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 6 weighted by sanderson EN', 'module': 'Autocorrelation', 'constructor': 'ATS(6, "se")'}, 'ATS7se': {'function': mordred.Autocorrelation.ATS(7, 'se'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 7 weighted by sanderson EN', 'module': 'Autocorrelation', 'constructor': 'ATS(7, "se")'}, 'ATS8se': {'function': mordred.Autocorrelation.ATS(8, 'se'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 8 weighted by sanderson EN', 'module': 'Autocorrelation', 'constructor': 'ATS(8, "se")'}, 'ATS0pe': {'function': mordred.Autocorrelation.ATS(0, 'pe'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 0 weighted by pauling EN', 'module': 'Autocorrelation', 'constructor': 'ATS(0, "pe")'}, 'ATS1pe': {'function': mordred.Autocorrelation.ATS(1, 'pe'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 1 weighted by pauling EN', 'module': 'Autocorrelation', 'constructor': 'ATS(1, "pe")'}, 'ATS2pe': {'function': mordred.Autocorrelation.ATS(2, 'pe'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 2 weighted by pauling EN', 'module': 'Autocorrelation', 'constructor': 'ATS(2, "pe")'}, 'ATS3pe': {'function': mordred.Autocorrelation.ATS(3, 'pe'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 3 weighted by pauling EN', 'module': 'Autocorrelation', 'constructor': 'ATS(3, "pe")'}, 'ATS4pe': {'function': mordred.Autocorrelation.ATS(4, 'pe'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 4 weighted by pauling EN', 'module': 'Autocorrelation', 'constructor': 'ATS(4, "pe")'}, 'ATS5pe': {'function': mordred.Autocorrelation.ATS(5, 'pe'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 5 weighted by pauling EN', 'module': 'Autocorrelation', 'constructor': 'ATS(5, "pe")'}, 'ATS6pe': {'function': mordred.Autocorrelation.ATS(6, 'pe'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 6 weighted by pauling EN', 'module': 'Autocorrelation', 'constructor': 'ATS(6, "pe")'}, 'ATS7pe': {'function': mordred.Autocorrelation.ATS(7, 'pe'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 7 weighted by pauling EN', 'module': 'Autocorrelation', 'constructor': 'ATS(7, "pe")'}, 'ATS8pe': {'function': mordred.Autocorrelation.ATS(8, 'pe'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 8 weighted by pauling EN', 'module': 'Autocorrelation', 'constructor': 'ATS(8, "pe")'}, 'ATS0are': {'function': mordred.Autocorrelation.ATS(0, 'are'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 0 weighted by allred-rocow EN', 'module': 'Autocorrelation', 'constructor': 'ATS(0, "are")'}, 'ATS1are': {'function': mordred.Autocorrelation.ATS(1, 'are'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 1 weighted by allred-rocow EN', 'module': 'Autocorrelation', 'constructor': 'ATS(1, "are")'}, 'ATS2are': {'function': mordred.Autocorrelation.ATS(2, 'are'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 2 weighted by allred-rocow EN', 'module': 'Autocorrelation', 'constructor': 'ATS(2, "are")'}, 'ATS3are': {'function': mordred.Autocorrelation.ATS(3, 'are'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 3 weighted by allred-rocow EN', 'module': 'Autocorrelation', 'constructor': 'ATS(3, "are")'}, 'ATS4are': {'function': mordred.Autocorrelation.ATS(4, 'are'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 4 weighted by allred-rocow EN', 'module': 'Autocorrelation', 'constructor': 'ATS(4, "are")'}, 'ATS5are': {'function': mordred.Autocorrelation.ATS(5, 'are'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 5 weighted by allred-rocow EN', 'module': 'Autocorrelation', 'constructor': 'ATS(5, "are")'}, 'ATS6are': {'function': mordred.Autocorrelation.ATS(6, 'are'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 6 weighted by allred-rocow EN', 'module': 'Autocorrelation', 'constructor': 'ATS(6, "are")'}, 'ATS7are': {'function': mordred.Autocorrelation.ATS(7, 'are'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 7 weighted by allred-rocow EN', 'module': 'Autocorrelation', 'constructor': 'ATS(7, "are")'}, 'ATS8are': {'function': mordred.Autocorrelation.ATS(8, 'are'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 8 weighted by allred-rocow EN', 'module': 'Autocorrelation', 'constructor': 'ATS(8, "are")'}, 'ATS0p': {'function': mordred.Autocorrelation.ATS(0, 'p'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 0 weighted by polarizability', 'module': 'Autocorrelation', 'constructor': 'ATS(0, "p")'}, 'ATS1p': {'function': mordred.Autocorrelation.ATS(1, 'p'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 1 weighted by polarizability', 'module': 'Autocorrelation', 'constructor': 'ATS(1, "p")'}, 'ATS2p': {'function': mordred.Autocorrelation.ATS(2, 'p'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 2 weighted by polarizability', 'module': 'Autocorrelation', 'constructor': 'ATS(2, "p")'}, 'ATS3p': {'function': mordred.Autocorrelation.ATS(3, 'p'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 3 weighted by polarizability', 'module': 'Autocorrelation', 'constructor': 'ATS(3, "p")'}, 'ATS4p': {'function': mordred.Autocorrelation.ATS(4, 'p'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 4 weighted by polarizability', 'module': 'Autocorrelation', 'constructor': 'ATS(4, "p")'}, 'ATS5p': {'function': mordred.Autocorrelation.ATS(5, 'p'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 5 weighted by polarizability', 'module': 'Autocorrelation', 'constructor': 'ATS(5, "p")'}, 'ATS6p': {'function': mordred.Autocorrelation.ATS(6, 'p'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 6 weighted by polarizability', 'module': 'Autocorrelation', 'constructor': 'ATS(6, "p")'}, 'ATS7p': {'function': mordred.Autocorrelation.ATS(7, 'p'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 7 weighted by polarizability', 'module': 'Autocorrelation', 'constructor': 'ATS(7, "p")'}, 'ATS8p': {'function': mordred.Autocorrelation.ATS(8, 'p'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 8 weighted by polarizability', 'module': 'Autocorrelation', 'constructor': 'ATS(8, "p")'}, 'ATS0i': {'function': mordred.Autocorrelation.ATS(0, 'i'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 0 weighted by ionization potential', 'module': 'Autocorrelation', 'constructor': 'ATS(0, "i")'}, 'ATS1i': {'function': mordred.Autocorrelation.ATS(1, 'i'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 1 weighted by ionization potential', 'module': 'Autocorrelation', 'constructor': 'ATS(1, "i")'}, 'ATS2i': {'function': mordred.Autocorrelation.ATS(2, 'i'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 2 weighted by ionization potential', 'module': 'Autocorrelation', 'constructor': 'ATS(2, "i")'}, 'ATS3i': {'function': mordred.Autocorrelation.ATS(3, 'i'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 3 weighted by ionization potential', 'module': 'Autocorrelation', 'constructor': 'ATS(3, "i")'}, 'ATS4i': {'function': mordred.Autocorrelation.ATS(4, 'i'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 4 weighted by ionization potential', 'module': 'Autocorrelation', 'constructor': 'ATS(4, "i")'}, 'ATS5i': {'function': mordred.Autocorrelation.ATS(5, 'i'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 5 weighted by ionization potential', 'module': 'Autocorrelation', 'constructor': 'ATS(5, "i")'}, 'ATS6i': {'function': mordred.Autocorrelation.ATS(6, 'i'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 6 weighted by ionization potential', 'module': 'Autocorrelation', 'constructor': 'ATS(6, "i")'}, 'ATS7i': {'function': mordred.Autocorrelation.ATS(7, 'i'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 7 weighted by ionization potential', 'module': 'Autocorrelation', 'constructor': 'ATS(7, "i")'}, 'ATS8i': {'function': mordred.Autocorrelation.ATS(8, 'i'), 'type': '2D', 'description': 'moreau-broto autocorrelation of lag 8 weighted by ionization potential', 'module': 'Autocorrelation', 'constructor': 'ATS(8, "i")'}, 'AATS0dv': {'function': mordred.Autocorrelation.AATS(0, 'dv'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 0 weighted by valence electrons', 'module': 'Autocorrelation', 'constructor': 'AATS(0, "dv")'}, 'AATS1dv': {'function': mordred.Autocorrelation.AATS(1, 'dv'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 1 weighted by valence electrons', 'module': 'Autocorrelation', 'constructor': 'AATS(1, "dv")'}, 'AATS2dv': {'function': mordred.Autocorrelation.AATS(2, 'dv'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 2 weighted by valence electrons', 'module': 'Autocorrelation', 'constructor': 'AATS(2, "dv")'}, 'AATS3dv': {'function': mordred.Autocorrelation.AATS(3, 'dv'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 3 weighted by valence electrons', 'module': 'Autocorrelation', 'constructor': 'AATS(3, "dv")'}, 'AATS4dv': {'function': mordred.Autocorrelation.AATS(4, 'dv'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 4 weighted by valence electrons', 'module': 'Autocorrelation', 'constructor': 'AATS(4, "dv")'}, 'AATS5dv': {'function': mordred.Autocorrelation.AATS(5, 'dv'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 5 weighted by valence electrons', 'module': 'Autocorrelation', 'constructor': 'AATS(5, "dv")'}, 'AATS6dv': {'function': mordred.Autocorrelation.AATS(6, 'dv'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 6 weighted by valence electrons', 'module': 'Autocorrelation', 'constructor': 'AATS(6, "dv")'}, 'AATS7dv': {'function': mordred.Autocorrelation.AATS(7, 'dv'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 7 weighted by valence electrons', 'module': 'Autocorrelation', 'constructor': 'AATS(7, "dv")'}, 'AATS8dv': {'function': mordred.Autocorrelation.AATS(8, 'dv'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 8 weighted by valence electrons', 'module': 'Autocorrelation', 'constructor': 'AATS(8, "dv")'}, 'AATS0d': {'function': mordred.Autocorrelation.AATS(0, 'd'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 0 weighted by sigma electrons', 'module': 'Autocorrelation', 'constructor': 'AATS(0, "d")'}, 'AATS1d': {'function': mordred.Autocorrelation.AATS(1, 'd'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 1 weighted by sigma electrons', 'module': 'Autocorrelation', 'constructor': 'AATS(1, "d")'}, 'AATS2d': {'function': mordred.Autocorrelation.AATS(2, 'd'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 2 weighted by sigma electrons', 'module': 'Autocorrelation', 'constructor': 'AATS(2, "d")'}, 'AATS3d': {'function': mordred.Autocorrelation.AATS(3, 'd'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 3 weighted by sigma electrons', 'module': 'Autocorrelation', 'constructor': 'AATS(3, "d")'}, 'AATS4d': {'function': mordred.Autocorrelation.AATS(4, 'd'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 4 weighted by sigma electrons', 'module': 'Autocorrelation', 'constructor': 'AATS(4, "d")'}, 'AATS5d': {'function': mordred.Autocorrelation.AATS(5, 'd'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 5 weighted by sigma electrons', 'module': 'Autocorrelation', 'constructor': 'AATS(5, "d")'}, 'AATS6d': {'function': mordred.Autocorrelation.AATS(6, 'd'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 6 weighted by sigma electrons', 'module': 'Autocorrelation', 'constructor': 'AATS(6, "d")'}, 'AATS7d': {'function': mordred.Autocorrelation.AATS(7, 'd'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 7 weighted by sigma electrons', 'module': 'Autocorrelation', 'constructor': 'AATS(7, "d")'}, 'AATS8d': {'function': mordred.Autocorrelation.AATS(8, 'd'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 8 weighted by sigma electrons', 'module': 'Autocorrelation', 'constructor': 'AATS(8, "d")'}, 'AATS0s': {'function': mordred.Autocorrelation.AATS(0, 's'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 0 weighted by intrinsic state', 'module': 'Autocorrelation', 'constructor': 'AATS(0, "s")'}, 'AATS1s': {'function': mordred.Autocorrelation.AATS(1, 's'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 1 weighted by intrinsic state', 'module': 'Autocorrelation', 'constructor': 'AATS(1, "s")'}, 'AATS2s': {'function': mordred.Autocorrelation.AATS(2, 's'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 2 weighted by intrinsic state', 'module': 'Autocorrelation', 'constructor': 'AATS(2, "s")'}, 'AATS3s': {'function': mordred.Autocorrelation.AATS(3, 's'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 3 weighted by intrinsic state', 'module': 'Autocorrelation', 'constructor': 'AATS(3, "s")'}, 'AATS4s': {'function': mordred.Autocorrelation.AATS(4, 's'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 4 weighted by intrinsic state', 'module': 'Autocorrelation', 'constructor': 'AATS(4, "s")'}, 'AATS5s': {'function': mordred.Autocorrelation.AATS(5, 's'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 5 weighted by intrinsic state', 'module': 'Autocorrelation', 'constructor': 'AATS(5, "s")'}, 'AATS6s': {'function': mordred.Autocorrelation.AATS(6, 's'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 6 weighted by intrinsic state', 'module': 'Autocorrelation', 'constructor': 'AATS(6, "s")'}, 'AATS7s': {'function': mordred.Autocorrelation.AATS(7, 's'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 7 weighted by intrinsic state', 'module': 'Autocorrelation', 'constructor': 'AATS(7, "s")'}, 'AATS8s': {'function': mordred.Autocorrelation.AATS(8, 's'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 8 weighted by intrinsic state', 'module': 'Autocorrelation', 'constructor': 'AATS(8, "s")'}, 'AATS0Z': {'function': mordred.Autocorrelation.AATS(0, 'Z'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 0 weighted by atomic number', 'module': 'Autocorrelation', 'constructor': 'AATS(0, "Z")'}, 'AATS1Z': {'function': mordred.Autocorrelation.AATS(1, 'Z'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 1 weighted by atomic number', 'module': 'Autocorrelation', 'constructor': 'AATS(1, "Z")'}, 'AATS2Z': {'function': mordred.Autocorrelation.AATS(2, 'Z'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 2 weighted by atomic number', 'module': 'Autocorrelation', 'constructor': 'AATS(2, "Z")'}, 'AATS3Z': {'function': mordred.Autocorrelation.AATS(3, 'Z'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 3 weighted by atomic number', 'module': 'Autocorrelation', 'constructor': 'AATS(3, "Z")'}, 'AATS4Z': {'function': mordred.Autocorrelation.AATS(4, 'Z'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 4 weighted by atomic number', 'module': 'Autocorrelation', 'constructor': 'AATS(4, "Z")'}, 'AATS5Z': {'function': mordred.Autocorrelation.AATS(5, 'Z'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 5 weighted by atomic number', 'module': 'Autocorrelation', 'constructor': 'AATS(5, "Z")'}, 'AATS6Z': {'function': mordred.Autocorrelation.AATS(6, 'Z'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 6 weighted by atomic number', 'module': 'Autocorrelation', 'constructor': 'AATS(6, "Z")'}, 'AATS7Z': {'function': mordred.Autocorrelation.AATS(7, 'Z'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 7 weighted by atomic number', 'module': 'Autocorrelation', 'constructor': 'AATS(7, "Z")'}, 'AATS8Z': {'function': mordred.Autocorrelation.AATS(8, 'Z'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 8 weighted by atomic number', 'module': 'Autocorrelation', 'constructor': 'AATS(8, "Z")'}, 'AATS0m': {'function': mordred.Autocorrelation.AATS(0, 'm'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 0 weighted by mass', 'module': 'Autocorrelation', 'constructor': 'AATS(0, "m")'}, 'AATS1m': {'function': mordred.Autocorrelation.AATS(1, 'm'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 1 weighted by mass', 'module': 'Autocorrelation', 'constructor': 'AATS(1, "m")'}, 'AATS2m': {'function': mordred.Autocorrelation.AATS(2, 'm'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 2 weighted by mass', 'module': 'Autocorrelation', 'constructor': 'AATS(2, "m")'}, 'AATS3m': {'function': mordred.Autocorrelation.AATS(3, 'm'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 3 weighted by mass', 'module': 'Autocorrelation', 'constructor': 'AATS(3, "m")'}, 'AATS4m': {'function': mordred.Autocorrelation.AATS(4, 'm'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 4 weighted by mass', 'module': 'Autocorrelation', 'constructor': 'AATS(4, "m")'}, 'AATS5m': {'function': mordred.Autocorrelation.AATS(5, 'm'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 5 weighted by mass', 'module': 'Autocorrelation', 'constructor': 'AATS(5, "m")'}, 'AATS6m': {'function': mordred.Autocorrelation.AATS(6, 'm'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 6 weighted by mass', 'module': 'Autocorrelation', 'constructor': 'AATS(6, "m")'}, 'AATS7m': {'function': mordred.Autocorrelation.AATS(7, 'm'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 7 weighted by mass', 'module': 'Autocorrelation', 'constructor': 'AATS(7, "m")'}, 'AATS8m': {'function': mordred.Autocorrelation.AATS(8, 'm'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 8 weighted by mass', 'module': 'Autocorrelation', 'constructor': 'AATS(8, "m")'}, 'AATS0v': {'function': mordred.Autocorrelation.AATS(0, 'v'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 0 weighted by vdw volume', 'module': 'Autocorrelation', 'constructor': 'AATS(0, "v")'}, 'AATS1v': {'function': mordred.Autocorrelation.AATS(1, 'v'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 1 weighted by vdw volume', 'module': 'Autocorrelation', 'constructor': 'AATS(1, "v")'}, 'AATS2v': {'function': mordred.Autocorrelation.AATS(2, 'v'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 2 weighted by vdw volume', 'module': 'Autocorrelation', 'constructor': 'AATS(2, "v")'}, 'AATS3v': {'function': mordred.Autocorrelation.AATS(3, 'v'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 3 weighted by vdw volume', 'module': 'Autocorrelation', 'constructor': 'AATS(3, "v")'}, 'AATS4v': {'function': mordred.Autocorrelation.AATS(4, 'v'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 4 weighted by vdw volume', 'module': 'Autocorrelation', 'constructor': 'AATS(4, "v")'}, 'AATS5v': {'function': mordred.Autocorrelation.AATS(5, 'v'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 5 weighted by vdw volume', 'module': 'Autocorrelation', 'constructor': 'AATS(5, "v")'}, 'AATS6v': {'function': mordred.Autocorrelation.AATS(6, 'v'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 6 weighted by vdw volume', 'module': 'Autocorrelation', 'constructor': 'AATS(6, "v")'}, 'AATS7v': {'function': mordred.Autocorrelation.AATS(7, 'v'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 7 weighted by vdw volume', 'module': 'Autocorrelation', 'constructor': 'AATS(7, "v")'}, 'AATS8v': {'function': mordred.Autocorrelation.AATS(8, 'v'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 8 weighted by vdw volume', 'module': 'Autocorrelation', 'constructor': 'AATS(8, "v")'}, 'AATS0se': {'function': mordred.Autocorrelation.AATS(0, 'se'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 0 weighted by sanderson EN', 'module': 'Autocorrelation', 'constructor': 'AATS(0, "se")'}, 'AATS1se': {'function': mordred.Autocorrelation.AATS(1, 'se'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 1 weighted by sanderson EN', 'module': 'Autocorrelation', 'constructor': 'AATS(1, "se")'}, 'AATS2se': {'function': mordred.Autocorrelation.AATS(2, 'se'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 2 weighted by sanderson EN', 'module': 'Autocorrelation', 'constructor': 'AATS(2, "se")'}, 'AATS3se': {'function': mordred.Autocorrelation.AATS(3, 'se'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 3 weighted by sanderson EN', 'module': 'Autocorrelation', 'constructor': 'AATS(3, "se")'}, 'AATS4se': {'function': mordred.Autocorrelation.AATS(4, 'se'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 4 weighted by sanderson EN', 'module': 'Autocorrelation', 'constructor': 'AATS(4, "se")'}, 'AATS5se': {'function': mordred.Autocorrelation.AATS(5, 'se'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 5 weighted by sanderson EN', 'module': 'Autocorrelation', 'constructor': 'AATS(5, "se")'}, 'AATS6se': {'function': mordred.Autocorrelation.AATS(6, 'se'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 6 weighted by sanderson EN', 'module': 'Autocorrelation', 'constructor': 'AATS(6, "se")'}, 'AATS7se': {'function': mordred.Autocorrelation.AATS(7, 'se'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 7 weighted by sanderson EN', 'module': 'Autocorrelation', 'constructor': 'AATS(7, "se")'}, 'AATS8se': {'function': mordred.Autocorrelation.AATS(8, 'se'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 8 weighted by sanderson EN', 'module': 'Autocorrelation', 'constructor': 'AATS(8, "se")'}, 'AATS0pe': {'function': mordred.Autocorrelation.AATS(0, 'pe'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 0 weighted by pauling EN', 'module': 'Autocorrelation', 'constructor': 'AATS(0, "pe")'}, 'AATS1pe': {'function': mordred.Autocorrelation.AATS(1, 'pe'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 1 weighted by pauling EN', 'module': 'Autocorrelation', 'constructor': 'AATS(1, "pe")'}, 'AATS2pe': {'function': mordred.Autocorrelation.AATS(2, 'pe'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 2 weighted by pauling EN', 'module': 'Autocorrelation', 'constructor': 'AATS(2, "pe")'}, 'AATS3pe': {'function': mordred.Autocorrelation.AATS(3, 'pe'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 3 weighted by pauling EN', 'module': 'Autocorrelation', 'constructor': 'AATS(3, "pe")'}, 'AATS4pe': {'function': mordred.Autocorrelation.AATS(4, 'pe'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 4 weighted by pauling EN', 'module': 'Autocorrelation', 'constructor': 'AATS(4, "pe")'}, 'AATS5pe': {'function': mordred.Autocorrelation.AATS(5, 'pe'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 5 weighted by pauling EN', 'module': 'Autocorrelation', 'constructor': 'AATS(5, "pe")'}, 'AATS6pe': {'function': mordred.Autocorrelation.AATS(6, 'pe'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 6 weighted by pauling EN', 'module': 'Autocorrelation', 'constructor': 'AATS(6, "pe")'}, 'AATS7pe': {'function': mordred.Autocorrelation.AATS(7, 'pe'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 7 weighted by pauling EN', 'module': 'Autocorrelation', 'constructor': 'AATS(7, "pe")'}, 'AATS8pe': {'function': mordred.Autocorrelation.AATS(8, 'pe'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 8 weighted by pauling EN', 'module': 'Autocorrelation', 'constructor': 'AATS(8, "pe")'}, 'AATS0are': {'function': mordred.Autocorrelation.AATS(0, 'are'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 0 weighted by allred-rocow EN', 'module': 'Autocorrelation', 'constructor': 'AATS(0, "are")'}, 'AATS1are': {'function': mordred.Autocorrelation.AATS(1, 'are'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 1 weighted by allred-rocow EN', 'module': 'Autocorrelation', 'constructor': 'AATS(1, "are")'}, 'AATS2are': {'function': mordred.Autocorrelation.AATS(2, 'are'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 2 weighted by allred-rocow EN', 'module': 'Autocorrelation', 'constructor': 'AATS(2, "are")'}, 'AATS3are': {'function': mordred.Autocorrelation.AATS(3, 'are'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 3 weighted by allred-rocow EN', 'module': 'Autocorrelation', 'constructor': 'AATS(3, "are")'}, 'AATS4are': {'function': mordred.Autocorrelation.AATS(4, 'are'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 4 weighted by allred-rocow EN', 'module': 'Autocorrelation', 'constructor': 'AATS(4, "are")'}, 'AATS5are': {'function': mordred.Autocorrelation.AATS(5, 'are'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 5 weighted by allred-rocow EN', 'module': 'Autocorrelation', 'constructor': 'AATS(5, "are")'}, 'AATS6are': {'function': mordred.Autocorrelation.AATS(6, 'are'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 6 weighted by allred-rocow EN', 'module': 'Autocorrelation', 'constructor': 'AATS(6, "are")'}, 'AATS7are': {'function': mordred.Autocorrelation.AATS(7, 'are'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 7 weighted by allred-rocow EN', 'module': 'Autocorrelation', 'constructor': 'AATS(7, "are")'}, 'AATS8are': {'function': mordred.Autocorrelation.AATS(8, 'are'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 8 weighted by allred-rocow EN', 'module': 'Autocorrelation', 'constructor': 'AATS(8, "are")'}, 'AATS0p': {'function': mordred.Autocorrelation.AATS(0, 'p'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 0 weighted by polarizability', 'module': 'Autocorrelation', 'constructor': 'AATS(0, "p")'}, 'AATS1p': {'function': mordred.Autocorrelation.AATS(1, 'p'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 1 weighted by polarizability', 'module': 'Autocorrelation', 'constructor': 'AATS(1, "p")'}, 'AATS2p': {'function': mordred.Autocorrelation.AATS(2, 'p'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 2 weighted by polarizability', 'module': 'Autocorrelation', 'constructor': 'AATS(2, "p")'}, 'AATS3p': {'function': mordred.Autocorrelation.AATS(3, 'p'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 3 weighted by polarizability', 'module': 'Autocorrelation', 'constructor': 'AATS(3, "p")'}, 'AATS4p': {'function': mordred.Autocorrelation.AATS(4, 'p'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 4 weighted by polarizability', 'module': 'Autocorrelation', 'constructor': 'AATS(4, "p")'}, 'AATS5p': {'function': mordred.Autocorrelation.AATS(5, 'p'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 5 weighted by polarizability', 'module': 'Autocorrelation', 'constructor': 'AATS(5, "p")'}, 'AATS6p': {'function': mordred.Autocorrelation.AATS(6, 'p'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 6 weighted by polarizability', 'module': 'Autocorrelation', 'constructor': 'AATS(6, "p")'}, 'AATS7p': {'function': mordred.Autocorrelation.AATS(7, 'p'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 7 weighted by polarizability', 'module': 'Autocorrelation', 'constructor': 'AATS(7, "p")'}, 'AATS8p': {'function': mordred.Autocorrelation.AATS(8, 'p'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 8 weighted by polarizability', 'module': 'Autocorrelation', 'constructor': 'AATS(8, "p")'}, 'AATS0i': {'function': mordred.Autocorrelation.AATS(0, 'i'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 0 weighted by ionization potential', 'module': 'Autocorrelation', 'constructor': 'AATS(0, "i")'}, 'AATS1i': {'function': mordred.Autocorrelation.AATS(1, 'i'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 1 weighted by ionization potential', 'module': 'Autocorrelation', 'constructor': 'AATS(1, "i")'}, 'AATS2i': {'function': mordred.Autocorrelation.AATS(2, 'i'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 2 weighted by ionization potential', 'module': 'Autocorrelation', 'constructor': 'AATS(2, "i")'}, 'AATS3i': {'function': mordred.Autocorrelation.AATS(3, 'i'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 3 weighted by ionization potential', 'module': 'Autocorrelation', 'constructor': 'AATS(3, "i")'}, 'AATS4i': {'function': mordred.Autocorrelation.AATS(4, 'i'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 4 weighted by ionization potential', 'module': 'Autocorrelation', 'constructor': 'AATS(4, "i")'}, 'AATS5i': {'function': mordred.Autocorrelation.AATS(5, 'i'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 5 weighted by ionization potential', 'module': 'Autocorrelation', 'constructor': 'AATS(5, "i")'}, 'AATS6i': {'function': mordred.Autocorrelation.AATS(6, 'i'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 6 weighted by ionization potential', 'module': 'Autocorrelation', 'constructor': 'AATS(6, "i")'}, 'AATS7i': {'function': mordred.Autocorrelation.AATS(7, 'i'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 7 weighted by ionization potential', 'module': 'Autocorrelation', 'constructor': 'AATS(7, "i")'}, 'AATS8i': {'function': mordred.Autocorrelation.AATS(8, 'i'), 'type': '2D', 'description': 'averaged moreau-broto autocorrelation of lag 8 weighted by ionization potential', 'module': 'Autocorrelation', 'constructor': 'AATS(8, "i")'}, 'ATSC0c': {'function': mordred.Autocorrelation.ATSC(0, 'c'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 0 weighted by gasteiger charge', 'module': 'Autocorrelation', 'constructor': 'ATSC(0, "c")'}, 'ATSC1c': {'function': mordred.Autocorrelation.ATSC(1, 'c'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 1 weighted by gasteiger charge', 'module': 'Autocorrelation', 'constructor': 'ATSC(1, "c")'}, 'ATSC2c': {'function': mordred.Autocorrelation.ATSC(2, 'c'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 2 weighted by gasteiger charge', 'module': 'Autocorrelation', 'constructor': 'ATSC(2, "c")'}, 'ATSC3c': {'function': mordred.Autocorrelation.ATSC(3, 'c'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 3 weighted by gasteiger charge', 'module': 'Autocorrelation', 'constructor': 'ATSC(3, "c")'}, 'ATSC4c': {'function': mordred.Autocorrelation.ATSC(4, 'c'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 4 weighted by gasteiger charge', 'module': 'Autocorrelation', 'constructor': 'ATSC(4, "c")'}, 'ATSC5c': {'function': mordred.Autocorrelation.ATSC(5, 'c'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 5 weighted by gasteiger charge', 'module': 'Autocorrelation', 'constructor': 'ATSC(5, "c")'}, 'ATSC6c': {'function': mordred.Autocorrelation.ATSC(6, 'c'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 6 weighted by gasteiger charge', 'module': 'Autocorrelation', 'constructor': 'ATSC(6, "c")'}, 'ATSC7c': {'function': mordred.Autocorrelation.ATSC(7, 'c'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 7 weighted by gasteiger charge', 'module': 'Autocorrelation', 'constructor': 'ATSC(7, "c")'}, 'ATSC8c': {'function': mordred.Autocorrelation.ATSC(8, 'c'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 8 weighted by gasteiger charge', 'module': 'Autocorrelation', 'constructor': 'ATSC(8, "c")'}, 'ATSC0dv': {'function': mordred.Autocorrelation.ATSC(0, 'dv'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 0 weighted by valence electrons', 'module': 'Autocorrelation', 'constructor': 'ATSC(0, "dv")'}, 'ATSC1dv': {'function': mordred.Autocorrelation.ATSC(1, 'dv'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 1 weighted by valence electrons', 'module': 'Autocorrelation', 'constructor': 'ATSC(1, "dv")'}, 'ATSC2dv': {'function': mordred.Autocorrelation.ATSC(2, 'dv'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 2 weighted by valence electrons', 'module': 'Autocorrelation', 'constructor': 'ATSC(2, "dv")'}, 'ATSC3dv': {'function': mordred.Autocorrelation.ATSC(3, 'dv'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 3 weighted by valence electrons', 'module': 'Autocorrelation', 'constructor': 'ATSC(3, "dv")'}, 'ATSC4dv': {'function': mordred.Autocorrelation.ATSC(4, 'dv'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 4 weighted by valence electrons', 'module': 'Autocorrelation', 'constructor': 'ATSC(4, "dv")'}, 'ATSC5dv': {'function': mordred.Autocorrelation.ATSC(5, 'dv'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 5 weighted by valence electrons', 'module': 'Autocorrelation', 'constructor': 'ATSC(5, "dv")'}, 'ATSC6dv': {'function': mordred.Autocorrelation.ATSC(6, 'dv'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 6 weighted by valence electrons', 'module': 'Autocorrelation', 'constructor': 'ATSC(6, "dv")'}, 'ATSC7dv': {'function': mordred.Autocorrelation.ATSC(7, 'dv'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 7 weighted by valence electrons', 'module': 'Autocorrelation', 'constructor': 'ATSC(7, "dv")'}, 'ATSC8dv': {'function': mordred.Autocorrelation.ATSC(8, 'dv'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 8 weighted by valence electrons', 'module': 'Autocorrelation', 'constructor': 'ATSC(8, "dv")'}, 'ATSC0d': {'function': mordred.Autocorrelation.ATSC(0, 'd'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 0 weighted by sigma electrons', 'module': 'Autocorrelation', 'constructor': 'ATSC(0, "d")'}, 'ATSC1d': {'function': mordred.Autocorrelation.ATSC(1, 'd'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 1 weighted by sigma electrons', 'module': 'Autocorrelation', 'constructor': 'ATSC(1, "d")'}, 'ATSC2d': {'function': mordred.Autocorrelation.ATSC(2, 'd'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 2 weighted by sigma electrons', 'module': 'Autocorrelation', 'constructor': 'ATSC(2, "d")'}, 'ATSC3d': {'function': mordred.Autocorrelation.ATSC(3, 'd'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 3 weighted by sigma electrons', 'module': 'Autocorrelation', 'constructor': 'ATSC(3, "d")'}, 'ATSC4d': {'function': mordred.Autocorrelation.ATSC(4, 'd'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 4 weighted by sigma electrons', 'module': 'Autocorrelation', 'constructor': 'ATSC(4, "d")'}, 'ATSC5d': {'function': mordred.Autocorrelation.ATSC(5, 'd'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 5 weighted by sigma electrons', 'module': 'Autocorrelation', 'constructor': 'ATSC(5, "d")'}, 'ATSC6d': {'function': mordred.Autocorrelation.ATSC(6, 'd'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 6 weighted by sigma electrons', 'module': 'Autocorrelation', 'constructor': 'ATSC(6, "d")'}, 'ATSC7d': {'function': mordred.Autocorrelation.ATSC(7, 'd'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 7 weighted by sigma electrons', 'module': 'Autocorrelation', 'constructor': 'ATSC(7, "d")'}, 'ATSC8d': {'function': mordred.Autocorrelation.ATSC(8, 'd'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 8 weighted by sigma electrons', 'module': 'Autocorrelation', 'constructor': 'ATSC(8, "d")'}, 'ATSC0s': {'function': mordred.Autocorrelation.ATSC(0, 's'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 0 weighted by intrinsic state', 'module': 'Autocorrelation', 'constructor': 'ATSC(0, "s")'}, 'ATSC1s': {'function': mordred.Autocorrelation.ATSC(1, 's'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 1 weighted by intrinsic state', 'module': 'Autocorrelation', 'constructor': 'ATSC(1, "s")'}, 'ATSC2s': {'function': mordred.Autocorrelation.ATSC(2, 's'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 2 weighted by intrinsic state', 'module': 'Autocorrelation', 'constructor': 'ATSC(2, "s")'}, 'ATSC3s': {'function': mordred.Autocorrelation.ATSC(3, 's'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 3 weighted by intrinsic state', 'module': 'Autocorrelation', 'constructor': 'ATSC(3, "s")'}, 'ATSC4s': {'function': mordred.Autocorrelation.ATSC(4, 's'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 4 weighted by intrinsic state', 'module': 'Autocorrelation', 'constructor': 'ATSC(4, "s")'}, 'ATSC5s': {'function': mordred.Autocorrelation.ATSC(5, 's'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 5 weighted by intrinsic state', 'module': 'Autocorrelation', 'constructor': 'ATSC(5, "s")'}, 'ATSC6s': {'function': mordred.Autocorrelation.ATSC(6, 's'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 6 weighted by intrinsic state', 'module': 'Autocorrelation', 'constructor': 'ATSC(6, "s")'}, 'ATSC7s': {'function': mordred.Autocorrelation.ATSC(7, 's'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 7 weighted by intrinsic state', 'module': 'Autocorrelation', 'constructor': 'ATSC(7, "s")'}, 'ATSC8s': {'function': mordred.Autocorrelation.ATSC(8, 's'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 8 weighted by intrinsic state', 'module': 'Autocorrelation', 'constructor': 'ATSC(8, "s")'}, 'ATSC0Z': {'function': mordred.Autocorrelation.ATSC(0, 'Z'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 0 weighted by atomic number', 'module': 'Autocorrelation', 'constructor': 'ATSC(0, "Z")'}, 'ATSC1Z': {'function': mordred.Autocorrelation.ATSC(1, 'Z'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 1 weighted by atomic number', 'module': 'Autocorrelation', 'constructor': 'ATSC(1, "Z")'}, 'ATSC2Z': {'function': mordred.Autocorrelation.ATSC(2, 'Z'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 2 weighted by atomic number', 'module': 'Autocorrelation', 'constructor': 'ATSC(2, "Z")'}, 'ATSC3Z': {'function': mordred.Autocorrelation.ATSC(3, 'Z'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 3 weighted by atomic number', 'module': 'Autocorrelation', 'constructor': 'ATSC(3, "Z")'}, 'ATSC4Z': {'function': mordred.Autocorrelation.ATSC(4, 'Z'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 4 weighted by atomic number', 'module': 'Autocorrelation', 'constructor': 'ATSC(4, "Z")'}, 'ATSC5Z': {'function': mordred.Autocorrelation.ATSC(5, 'Z'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 5 weighted by atomic number', 'module': 'Autocorrelation', 'constructor': 'ATSC(5, "Z")'}, 'ATSC6Z': {'function': mordred.Autocorrelation.ATSC(6, 'Z'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 6 weighted by atomic number', 'module': 'Autocorrelation', 'constructor': 'ATSC(6, "Z")'}, 'ATSC7Z': {'function': mordred.Autocorrelation.ATSC(7, 'Z'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 7 weighted by atomic number', 'module': 'Autocorrelation', 'constructor': 'ATSC(7, "Z")'}, 'ATSC8Z': {'function': mordred.Autocorrelation.ATSC(8, 'Z'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 8 weighted by atomic number', 'module': 'Autocorrelation', 'constructor': 'ATSC(8, "Z")'}, 'ATSC0m': {'function': mordred.Autocorrelation.ATSC(0, 'm'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 0 weighted by mass', 'module': 'Autocorrelation', 'constructor': 'ATSC(0, "m")'}, 'ATSC1m': {'function': mordred.Autocorrelation.ATSC(1, 'm'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 1 weighted by mass', 'module': 'Autocorrelation', 'constructor': 'ATSC(1, "m")'}, 'ATSC2m': {'function': mordred.Autocorrelation.ATSC(2, 'm'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 2 weighted by mass', 'module': 'Autocorrelation', 'constructor': 'ATSC(2, "m")'}, 'ATSC3m': {'function': mordred.Autocorrelation.ATSC(3, 'm'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 3 weighted by mass', 'module': 'Autocorrelation', 'constructor': 'ATSC(3, "m")'}, 'ATSC4m': {'function': mordred.Autocorrelation.ATSC(4, 'm'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 4 weighted by mass', 'module': 'Autocorrelation', 'constructor': 'ATSC(4, "m")'}, 'ATSC5m': {'function': mordred.Autocorrelation.ATSC(5, 'm'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 5 weighted by mass', 'module': 'Autocorrelation', 'constructor': 'ATSC(5, "m")'}, 'ATSC6m': {'function': mordred.Autocorrelation.ATSC(6, 'm'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 6 weighted by mass', 'module': 'Autocorrelation', 'constructor': 'ATSC(6, "m")'}, 'ATSC7m': {'function': mordred.Autocorrelation.ATSC(7, 'm'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 7 weighted by mass', 'module': 'Autocorrelation', 'constructor': 'ATSC(7, "m")'}, 'ATSC8m': {'function': mordred.Autocorrelation.ATSC(8, 'm'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 8 weighted by mass', 'module': 'Autocorrelation', 'constructor': 'ATSC(8, "m")'}, 'ATSC0v': {'function': mordred.Autocorrelation.ATSC(0, 'v'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 0 weighted by vdw volume', 'module': 'Autocorrelation', 'constructor': 'ATSC(0, "v")'}, 'ATSC1v': {'function': mordred.Autocorrelation.ATSC(1, 'v'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 1 weighted by vdw volume', 'module': 'Autocorrelation', 'constructor': 'ATSC(1, "v")'}, 'ATSC2v': {'function': mordred.Autocorrelation.ATSC(2, 'v'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 2 weighted by vdw volume', 'module': 'Autocorrelation', 'constructor': 'ATSC(2, "v")'}, 'ATSC3v': {'function': mordred.Autocorrelation.ATSC(3, 'v'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 3 weighted by vdw volume', 'module': 'Autocorrelation', 'constructor': 'ATSC(3, "v")'}, 'ATSC4v': {'function': mordred.Autocorrelation.ATSC(4, 'v'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 4 weighted by vdw volume', 'module': 'Autocorrelation', 'constructor': 'ATSC(4, "v")'}, 'ATSC5v': {'function': mordred.Autocorrelation.ATSC(5, 'v'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 5 weighted by vdw volume', 'module': 'Autocorrelation', 'constructor': 'ATSC(5, "v")'}, 'ATSC6v': {'function': mordred.Autocorrelation.ATSC(6, 'v'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 6 weighted by vdw volume', 'module': 'Autocorrelation', 'constructor': 'ATSC(6, "v")'}, 'ATSC7v': {'function': mordred.Autocorrelation.ATSC(7, 'v'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 7 weighted by vdw volume', 'module': 'Autocorrelation', 'constructor': 'ATSC(7, "v")'}, 'ATSC8v': {'function': mordred.Autocorrelation.ATSC(8, 'v'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 8 weighted by vdw volume', 'module': 'Autocorrelation', 'constructor': 'ATSC(8, "v")'}, 'ATSC0se': {'function': mordred.Autocorrelation.ATSC(0, 'se'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 0 weighted by sanderson EN', 'module': 'Autocorrelation', 'constructor': 'ATSC(0, "se")'}, 'ATSC1se': {'function': mordred.Autocorrelation.ATSC(1, 'se'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 1 weighted by sanderson EN', 'module': 'Autocorrelation', 'constructor': 'ATSC(1, "se")'}, 'ATSC2se': {'function': mordred.Autocorrelation.ATSC(2, 'se'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 2 weighted by sanderson EN', 'module': 'Autocorrelation', 'constructor': 'ATSC(2, "se")'}, 'ATSC3se': {'function': mordred.Autocorrelation.ATSC(3, 'se'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 3 weighted by sanderson EN', 'module': 'Autocorrelation', 'constructor': 'ATSC(3, "se")'}, 'ATSC4se': {'function': mordred.Autocorrelation.ATSC(4, 'se'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 4 weighted by sanderson EN', 'module': 'Autocorrelation', 'constructor': 'ATSC(4, "se")'}, 'ATSC5se': {'function': mordred.Autocorrelation.ATSC(5, 'se'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 5 weighted by sanderson EN', 'module': 'Autocorrelation', 'constructor': 'ATSC(5, "se")'}, 'ATSC6se': {'function': mordred.Autocorrelation.ATSC(6, 'se'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 6 weighted by sanderson EN', 'module': 'Autocorrelation', 'constructor': 'ATSC(6, "se")'}, 'ATSC7se': {'function': mordred.Autocorrelation.ATSC(7, 'se'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 7 weighted by sanderson EN', 'module': 'Autocorrelation', 'constructor': 'ATSC(7, "se")'}, 'ATSC8se': {'function': mordred.Autocorrelation.ATSC(8, 'se'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 8 weighted by sanderson EN', 'module': 'Autocorrelation', 'constructor': 'ATSC(8, "se")'}, 'ATSC0pe': {'function': mordred.Autocorrelation.ATSC(0, 'pe'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 0 weighted by pauling EN', 'module': 'Autocorrelation', 'constructor': 'ATSC(0, "pe")'}, 'ATSC1pe': {'function': mordred.Autocorrelation.ATSC(1, 'pe'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 1 weighted by pauling EN', 'module': 'Autocorrelation', 'constructor': 'ATSC(1, "pe")'}, 'ATSC2pe': {'function': mordred.Autocorrelation.ATSC(2, 'pe'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 2 weighted by pauling EN', 'module': 'Autocorrelation', 'constructor': 'ATSC(2, "pe")'}, 'ATSC3pe': {'function': mordred.Autocorrelation.ATSC(3, 'pe'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 3 weighted by pauling EN', 'module': 'Autocorrelation', 'constructor': 'ATSC(3, "pe")'}, 'ATSC4pe': {'function': mordred.Autocorrelation.ATSC(4, 'pe'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 4 weighted by pauling EN', 'module': 'Autocorrelation', 'constructor': 'ATSC(4, "pe")'}, 'ATSC5pe': {'function': mordred.Autocorrelation.ATSC(5, 'pe'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 5 weighted by pauling EN', 'module': 'Autocorrelation', 'constructor': 'ATSC(5, "pe")'}, 'ATSC6pe': {'function': mordred.Autocorrelation.ATSC(6, 'pe'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 6 weighted by pauling EN', 'module': 'Autocorrelation', 'constructor': 'ATSC(6, "pe")'}, 'ATSC7pe': {'function': mordred.Autocorrelation.ATSC(7, 'pe'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 7 weighted by pauling EN', 'module': 'Autocorrelation', 'constructor': 'ATSC(7, "pe")'}, 'ATSC8pe': {'function': mordred.Autocorrelation.ATSC(8, 'pe'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 8 weighted by pauling EN', 'module': 'Autocorrelation', 'constructor': 'ATSC(8, "pe")'}, 'ATSC0are': {'function': mordred.Autocorrelation.ATSC(0, 'are'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 0 weighted by allred-rocow EN', 'module': 'Autocorrelation', 'constructor': 'ATSC(0, "are")'}, 'ATSC1are': {'function': mordred.Autocorrelation.ATSC(1, 'are'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 1 weighted by allred-rocow EN', 'module': 'Autocorrelation', 'constructor': 'ATSC(1, "are")'}, 'ATSC2are': {'function': mordred.Autocorrelation.ATSC(2, 'are'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 2 weighted by allred-rocow EN', 'module': 'Autocorrelation', 'constructor': 'ATSC(2, "are")'}, 'ATSC3are': {'function': mordred.Autocorrelation.ATSC(3, 'are'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 3 weighted by allred-rocow EN', 'module': 'Autocorrelation', 'constructor': 'ATSC(3, "are")'}, 'ATSC4are': {'function': mordred.Autocorrelation.ATSC(4, 'are'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 4 weighted by allred-rocow EN', 'module': 'Autocorrelation', 'constructor': 'ATSC(4, "are")'}, 'ATSC5are': {'function': mordred.Autocorrelation.ATSC(5, 'are'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 5 weighted by allred-rocow EN', 'module': 'Autocorrelation', 'constructor': 'ATSC(5, "are")'}, 'ATSC6are': {'function': mordred.Autocorrelation.ATSC(6, 'are'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 6 weighted by allred-rocow EN', 'module': 'Autocorrelation', 'constructor': 'ATSC(6, "are")'}, 'ATSC7are': {'function': mordred.Autocorrelation.ATSC(7, 'are'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 7 weighted by allred-rocow EN', 'module': 'Autocorrelation', 'constructor': 'ATSC(7, "are")'}, 'ATSC8are': {'function': mordred.Autocorrelation.ATSC(8, 'are'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 8 weighted by allred-rocow EN', 'module': 'Autocorrelation', 'constructor': 'ATSC(8, "are")'}, 'ATSC0p': {'function': mordred.Autocorrelation.ATSC(0, 'p'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 0 weighted by polarizability', 'module': 'Autocorrelation', 'constructor': 'ATSC(0, "p")'}, 'ATSC1p': {'function': mordred.Autocorrelation.ATSC(1, 'p'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 1 weighted by polarizability', 'module': 'Autocorrelation', 'constructor': 'ATSC(1, "p")'}, 'ATSC2p': {'function': mordred.Autocorrelation.ATSC(2, 'p'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 2 weighted by polarizability', 'module': 'Autocorrelation', 'constructor': 'ATSC(2, "p")'}, 'ATSC3p': {'function': mordred.Autocorrelation.ATSC(3, 'p'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 3 weighted by polarizability', 'module': 'Autocorrelation', 'constructor': 'ATSC(3, "p")'}, 'ATSC4p': {'function': mordred.Autocorrelation.ATSC(4, 'p'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 4 weighted by polarizability', 'module': 'Autocorrelation', 'constructor': 'ATSC(4, "p")'}, 'ATSC5p': {'function': mordred.Autocorrelation.ATSC(5, 'p'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 5 weighted by polarizability', 'module': 'Autocorrelation', 'constructor': 'ATSC(5, "p")'}, 'ATSC6p': {'function': mordred.Autocorrelation.ATSC(6, 'p'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 6 weighted by polarizability', 'module': 'Autocorrelation', 'constructor': 'ATSC(6, "p")'}, 'ATSC7p': {'function': mordred.Autocorrelation.ATSC(7, 'p'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 7 weighted by polarizability', 'module': 'Autocorrelation', 'constructor': 'ATSC(7, "p")'}, 'ATSC8p': {'function': mordred.Autocorrelation.ATSC(8, 'p'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 8 weighted by polarizability', 'module': 'Autocorrelation', 'constructor': 'ATSC(8, "p")'}, 'ATSC0i': {'function': mordred.Autocorrelation.ATSC(0, 'i'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 0 weighted by ionization potential', 'module': 'Autocorrelation', 'constructor': 'ATSC(0, "i")'}, 'ATSC1i': {'function': mordred.Autocorrelation.ATSC(1, 'i'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 1 weighted by ionization potential', 'module': 'Autocorrelation', 'constructor': 'ATSC(1, "i")'}, 'ATSC2i': {'function': mordred.Autocorrelation.ATSC(2, 'i'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 2 weighted by ionization potential', 'module': 'Autocorrelation', 'constructor': 'ATSC(2, "i")'}, 'ATSC3i': {'function': mordred.Autocorrelation.ATSC(3, 'i'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 3 weighted by ionization potential', 'module': 'Autocorrelation', 'constructor': 'ATSC(3, "i")'}, 'ATSC4i': {'function': mordred.Autocorrelation.ATSC(4, 'i'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 4 weighted by ionization potential', 'module': 'Autocorrelation', 'constructor': 'ATSC(4, "i")'}, 'ATSC5i': {'function': mordred.Autocorrelation.ATSC(5, 'i'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 5 weighted by ionization potential', 'module': 'Autocorrelation', 'constructor': 'ATSC(5, "i")'}, 'ATSC6i': {'function': mordred.Autocorrelation.ATSC(6, 'i'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 6 weighted by ionization potential', 'module': 'Autocorrelation', 'constructor': 'ATSC(6, "i")'}, 'ATSC7i': {'function': mordred.Autocorrelation.ATSC(7, 'i'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 7 weighted by ionization potential', 'module': 'Autocorrelation', 'constructor': 'ATSC(7, "i")'}, 'ATSC8i': {'function': mordred.Autocorrelation.ATSC(8, 'i'), 'type': '2D', 'description': 'centered moreau-broto autocorrelation of lag 8 weighted by ionization potential', 'module': 'Autocorrelation', 'constructor': 'ATSC(8, "i")'}, 'AATSC0c': {'function': mordred.Autocorrelation.AATSC(0, 'c'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 0 weighted by gasteiger charge', 'module': 'Autocorrelation', 'constructor': 'AATSC(0, "c")'}, 'AATSC1c': {'function': mordred.Autocorrelation.AATSC(1, 'c'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 1 weighted by gasteiger charge', 'module': 'Autocorrelation', 'constructor': 'AATSC(1, "c")'}, 'AATSC2c': {'function': mordred.Autocorrelation.AATSC(2, 'c'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 2 weighted by gasteiger charge', 'module': 'Autocorrelation', 'constructor': 'AATSC(2, "c")'}, 'AATSC3c': {'function': mordred.Autocorrelation.AATSC(3, 'c'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 3 weighted by gasteiger charge', 'module': 'Autocorrelation', 'constructor': 'AATSC(3, "c")'}, 'AATSC4c': {'function': mordred.Autocorrelation.AATSC(4, 'c'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 4 weighted by gasteiger charge', 'module': 'Autocorrelation', 'constructor': 'AATSC(4, "c")'}, 'AATSC5c': {'function': mordred.Autocorrelation.AATSC(5, 'c'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 5 weighted by gasteiger charge', 'module': 'Autocorrelation', 'constructor': 'AATSC(5, "c")'}, 'AATSC6c': {'function': mordred.Autocorrelation.AATSC(6, 'c'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 6 weighted by gasteiger charge', 'module': 'Autocorrelation', 'constructor': 'AATSC(6, "c")'}, 'AATSC7c': {'function': mordred.Autocorrelation.AATSC(7, 'c'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 7 weighted by gasteiger charge', 'module': 'Autocorrelation', 'constructor': 'AATSC(7, "c")'}, 'AATSC8c': {'function': mordred.Autocorrelation.AATSC(8, 'c'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 8 weighted by gasteiger charge', 'module': 'Autocorrelation', 'constructor': 'AATSC(8, "c")'}, 'AATSC0dv': {'function': mordred.Autocorrelation.AATSC(0, 'dv'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 0 weighted by valence electrons', 'module': 'Autocorrelation', 'constructor': 'AATSC(0, "dv")'}, 'AATSC1dv': {'function': mordred.Autocorrelation.AATSC(1, 'dv'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 1 weighted by valence electrons', 'module': 'Autocorrelation', 'constructor': 'AATSC(1, "dv")'}, 'AATSC2dv': {'function': mordred.Autocorrelation.AATSC(2, 'dv'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 2 weighted by valence electrons', 'module': 'Autocorrelation', 'constructor': 'AATSC(2, "dv")'}, 'AATSC3dv': {'function': mordred.Autocorrelation.AATSC(3, 'dv'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 3 weighted by valence electrons', 'module': 'Autocorrelation', 'constructor': 'AATSC(3, "dv")'}, 'AATSC4dv': {'function': mordred.Autocorrelation.AATSC(4, 'dv'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 4 weighted by valence electrons', 'module': 'Autocorrelation', 'constructor': 'AATSC(4, "dv")'}, 'AATSC5dv': {'function': mordred.Autocorrelation.AATSC(5, 'dv'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 5 weighted by valence electrons', 'module': 'Autocorrelation', 'constructor': 'AATSC(5, "dv")'}, 'AATSC6dv': {'function': mordred.Autocorrelation.AATSC(6, 'dv'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 6 weighted by valence electrons', 'module': 'Autocorrelation', 'constructor': 'AATSC(6, "dv")'}, 'AATSC7dv': {'function': mordred.Autocorrelation.AATSC(7, 'dv'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 7 weighted by valence electrons', 'module': 'Autocorrelation', 'constructor': 'AATSC(7, "dv")'}, 'AATSC8dv': {'function': mordred.Autocorrelation.AATSC(8, 'dv'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 8 weighted by valence electrons', 'module': 'Autocorrelation', 'constructor': 'AATSC(8, "dv")'}, 'AATSC0d': {'function': mordred.Autocorrelation.AATSC(0, 'd'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 0 weighted by sigma electrons', 'module': 'Autocorrelation', 'constructor': 'AATSC(0, "d")'}, 'AATSC1d': {'function': mordred.Autocorrelation.AATSC(1, 'd'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 1 weighted by sigma electrons', 'module': 'Autocorrelation', 'constructor': 'AATSC(1, "d")'}, 'AATSC2d': {'function': mordred.Autocorrelation.AATSC(2, 'd'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 2 weighted by sigma electrons', 'module': 'Autocorrelation', 'constructor': 'AATSC(2, "d")'}, 'AATSC3d': {'function': mordred.Autocorrelation.AATSC(3, 'd'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 3 weighted by sigma electrons', 'module': 'Autocorrelation', 'constructor': 'AATSC(3, "d")'}, 'AATSC4d': {'function': mordred.Autocorrelation.AATSC(4, 'd'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 4 weighted by sigma electrons', 'module': 'Autocorrelation', 'constructor': 'AATSC(4, "d")'}, 'AATSC5d': {'function': mordred.Autocorrelation.AATSC(5, 'd'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 5 weighted by sigma electrons', 'module': 'Autocorrelation', 'constructor': 'AATSC(5, "d")'}, 'AATSC6d': {'function': mordred.Autocorrelation.AATSC(6, 'd'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 6 weighted by sigma electrons', 'module': 'Autocorrelation', 'constructor': 'AATSC(6, "d")'}, 'AATSC7d': {'function': mordred.Autocorrelation.AATSC(7, 'd'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 7 weighted by sigma electrons', 'module': 'Autocorrelation', 'constructor': 'AATSC(7, "d")'}, 'AATSC8d': {'function': mordred.Autocorrelation.AATSC(8, 'd'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 8 weighted by sigma electrons', 'module': 'Autocorrelation', 'constructor': 'AATSC(8, "d")'}, 'AATSC0s': {'function': mordred.Autocorrelation.AATSC(0, 's'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 0 weighted by intrinsic state', 'module': 'Autocorrelation', 'constructor': 'AATSC(0, "s")'}, 'AATSC1s': {'function': mordred.Autocorrelation.AATSC(1, 's'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 1 weighted by intrinsic state', 'module': 'Autocorrelation', 'constructor': 'AATSC(1, "s")'}, 'AATSC2s': {'function': mordred.Autocorrelation.AATSC(2, 's'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 2 weighted by intrinsic state', 'module': 'Autocorrelation', 'constructor': 'AATSC(2, "s")'}, 'AATSC3s': {'function': mordred.Autocorrelation.AATSC(3, 's'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 3 weighted by intrinsic state', 'module': 'Autocorrelation', 'constructor': 'AATSC(3, "s")'}, 'AATSC4s': {'function': mordred.Autocorrelation.AATSC(4, 's'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 4 weighted by intrinsic state', 'module': 'Autocorrelation', 'constructor': 'AATSC(4, "s")'}, 'AATSC5s': {'function': mordred.Autocorrelation.AATSC(5, 's'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 5 weighted by intrinsic state', 'module': 'Autocorrelation', 'constructor': 'AATSC(5, "s")'}, 'AATSC6s': {'function': mordred.Autocorrelation.AATSC(6, 's'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 6 weighted by intrinsic state', 'module': 'Autocorrelation', 'constructor': 'AATSC(6, "s")'}, 'AATSC7s': {'function': mordred.Autocorrelation.AATSC(7, 's'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 7 weighted by intrinsic state', 'module': 'Autocorrelation', 'constructor': 'AATSC(7, "s")'}, 'AATSC8s': {'function': mordred.Autocorrelation.AATSC(8, 's'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 8 weighted by intrinsic state', 'module': 'Autocorrelation', 'constructor': 'AATSC(8, "s")'}, 'AATSC0Z': {'function': mordred.Autocorrelation.AATSC(0, 'Z'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 0 weighted by atomic number', 'module': 'Autocorrelation', 'constructor': 'AATSC(0, "Z")'}, 'AATSC1Z': {'function': mordred.Autocorrelation.AATSC(1, 'Z'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 1 weighted by atomic number', 'module': 'Autocorrelation', 'constructor': 'AATSC(1, "Z")'}, 'AATSC2Z': {'function': mordred.Autocorrelation.AATSC(2, 'Z'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 2 weighted by atomic number', 'module': 'Autocorrelation', 'constructor': 'AATSC(2, "Z")'}, 'AATSC3Z': {'function': mordred.Autocorrelation.AATSC(3, 'Z'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 3 weighted by atomic number', 'module': 'Autocorrelation', 'constructor': 'AATSC(3, "Z")'}, 'AATSC4Z': {'function': mordred.Autocorrelation.AATSC(4, 'Z'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 4 weighted by atomic number', 'module': 'Autocorrelation', 'constructor': 'AATSC(4, "Z")'}, 'AATSC5Z': {'function': mordred.Autocorrelation.AATSC(5, 'Z'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 5 weighted by atomic number', 'module': 'Autocorrelation', 'constructor': 'AATSC(5, "Z")'}, 'AATSC6Z': {'function': mordred.Autocorrelation.AATSC(6, 'Z'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 6 weighted by atomic number', 'module': 'Autocorrelation', 'constructor': 'AATSC(6, "Z")'}, 'AATSC7Z': {'function': mordred.Autocorrelation.AATSC(7, 'Z'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 7 weighted by atomic number', 'module': 'Autocorrelation', 'constructor': 'AATSC(7, "Z")'}, 'AATSC8Z': {'function': mordred.Autocorrelation.AATSC(8, 'Z'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 8 weighted by atomic number', 'module': 'Autocorrelation', 'constructor': 'AATSC(8, "Z")'}, 'AATSC0m': {'function': mordred.Autocorrelation.AATSC(0, 'm'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 0 weighted by mass', 'module': 'Autocorrelation', 'constructor': 'AATSC(0, "m")'}, 'AATSC1m': {'function': mordred.Autocorrelation.AATSC(1, 'm'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 1 weighted by mass', 'module': 'Autocorrelation', 'constructor': 'AATSC(1, "m")'}, 'AATSC2m': {'function': mordred.Autocorrelation.AATSC(2, 'm'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 2 weighted by mass', 'module': 'Autocorrelation', 'constructor': 'AATSC(2, "m")'}, 'AATSC3m': {'function': mordred.Autocorrelation.AATSC(3, 'm'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 3 weighted by mass', 'module': 'Autocorrelation', 'constructor': 'AATSC(3, "m")'}, 'AATSC4m': {'function': mordred.Autocorrelation.AATSC(4, 'm'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 4 weighted by mass', 'module': 'Autocorrelation', 'constructor': 'AATSC(4, "m")'}, 'AATSC5m': {'function': mordred.Autocorrelation.AATSC(5, 'm'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 5 weighted by mass', 'module': 'Autocorrelation', 'constructor': 'AATSC(5, "m")'}, 'AATSC6m': {'function': mordred.Autocorrelation.AATSC(6, 'm'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 6 weighted by mass', 'module': 'Autocorrelation', 'constructor': 'AATSC(6, "m")'}, 'AATSC7m': {'function': mordred.Autocorrelation.AATSC(7, 'm'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 7 weighted by mass', 'module': 'Autocorrelation', 'constructor': 'AATSC(7, "m")'}, 'AATSC8m': {'function': mordred.Autocorrelation.AATSC(8, 'm'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 8 weighted by mass', 'module': 'Autocorrelation', 'constructor': 'AATSC(8, "m")'}, 'AATSC0v': {'function': mordred.Autocorrelation.AATSC(0, 'v'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 0 weighted by vdw volume', 'module': 'Autocorrelation', 'constructor': 'AATSC(0, "v")'}, 'AATSC1v': {'function': mordred.Autocorrelation.AATSC(1, 'v'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 1 weighted by vdw volume', 'module': 'Autocorrelation', 'constructor': 'AATSC(1, "v")'}, 'AATSC2v': {'function': mordred.Autocorrelation.AATSC(2, 'v'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 2 weighted by vdw volume', 'module': 'Autocorrelation', 'constructor': 'AATSC(2, "v")'}, 'AATSC3v': {'function': mordred.Autocorrelation.AATSC(3, 'v'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 3 weighted by vdw volume', 'module': 'Autocorrelation', 'constructor': 'AATSC(3, "v")'}, 'AATSC4v': {'function': mordred.Autocorrelation.AATSC(4, 'v'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 4 weighted by vdw volume', 'module': 'Autocorrelation', 'constructor': 'AATSC(4, "v")'}, 'AATSC5v': {'function': mordred.Autocorrelation.AATSC(5, 'v'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 5 weighted by vdw volume', 'module': 'Autocorrelation', 'constructor': 'AATSC(5, "v")'}, 'AATSC6v': {'function': mordred.Autocorrelation.AATSC(6, 'v'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 6 weighted by vdw volume', 'module': 'Autocorrelation', 'constructor': 'AATSC(6, "v")'}, 'AATSC7v': {'function': mordred.Autocorrelation.AATSC(7, 'v'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 7 weighted by vdw volume', 'module': 'Autocorrelation', 'constructor': 'AATSC(7, "v")'}, 'AATSC8v': {'function': mordred.Autocorrelation.AATSC(8, 'v'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 8 weighted by vdw volume', 'module': 'Autocorrelation', 'constructor': 'AATSC(8, "v")'}, 'AATSC0se': {'function': mordred.Autocorrelation.AATSC(0, 'se'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 0 weighted by sanderson EN', 'module': 'Autocorrelation', 'constructor': 'AATSC(0, "se")'}, 'AATSC1se': {'function': mordred.Autocorrelation.AATSC(1, 'se'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 1 weighted by sanderson EN', 'module': 'Autocorrelation', 'constructor': 'AATSC(1, "se")'}, 'AATSC2se': {'function': mordred.Autocorrelation.AATSC(2, 'se'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 2 weighted by sanderson EN', 'module': 'Autocorrelation', 'constructor': 'AATSC(2, "se")'}, 'AATSC3se': {'function': mordred.Autocorrelation.AATSC(3, 'se'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 3 weighted by sanderson EN', 'module': 'Autocorrelation', 'constructor': 'AATSC(3, "se")'}, 'AATSC4se': {'function': mordred.Autocorrelation.AATSC(4, 'se'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 4 weighted by sanderson EN', 'module': 'Autocorrelation', 'constructor': 'AATSC(4, "se")'}, 'AATSC5se': {'function': mordred.Autocorrelation.AATSC(5, 'se'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 5 weighted by sanderson EN', 'module': 'Autocorrelation', 'constructor': 'AATSC(5, "se")'}, 'AATSC6se': {'function': mordred.Autocorrelation.AATSC(6, 'se'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 6 weighted by sanderson EN', 'module': 'Autocorrelation', 'constructor': 'AATSC(6, "se")'}, 'AATSC7se': {'function': mordred.Autocorrelation.AATSC(7, 'se'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 7 weighted by sanderson EN', 'module': 'Autocorrelation', 'constructor': 'AATSC(7, "se")'}, 'AATSC8se': {'function': mordred.Autocorrelation.AATSC(8, 'se'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 8 weighted by sanderson EN', 'module': 'Autocorrelation', 'constructor': 'AATSC(8, "se")'}, 'AATSC0pe': {'function': mordred.Autocorrelation.AATSC(0, 'pe'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 0 weighted by pauling EN', 'module': 'Autocorrelation', 'constructor': 'AATSC(0, "pe")'}, 'AATSC1pe': {'function': mordred.Autocorrelation.AATSC(1, 'pe'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 1 weighted by pauling EN', 'module': 'Autocorrelation', 'constructor': 'AATSC(1, "pe")'}, 'AATSC2pe': {'function': mordred.Autocorrelation.AATSC(2, 'pe'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 2 weighted by pauling EN', 'module': 'Autocorrelation', 'constructor': 'AATSC(2, "pe")'}, 'AATSC3pe': {'function': mordred.Autocorrelation.AATSC(3, 'pe'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 3 weighted by pauling EN', 'module': 'Autocorrelation', 'constructor': 'AATSC(3, "pe")'}, 'AATSC4pe': {'function': mordred.Autocorrelation.AATSC(4, 'pe'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 4 weighted by pauling EN', 'module': 'Autocorrelation', 'constructor': 'AATSC(4, "pe")'}, 'AATSC5pe': {'function': mordred.Autocorrelation.AATSC(5, 'pe'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 5 weighted by pauling EN', 'module': 'Autocorrelation', 'constructor': 'AATSC(5, "pe")'}, 'AATSC6pe': {'function': mordred.Autocorrelation.AATSC(6, 'pe'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 6 weighted by pauling EN', 'module': 'Autocorrelation', 'constructor': 'AATSC(6, "pe")'}, 'AATSC7pe': {'function': mordred.Autocorrelation.AATSC(7, 'pe'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 7 weighted by pauling EN', 'module': 'Autocorrelation', 'constructor': 'AATSC(7, "pe")'}, 'AATSC8pe': {'function': mordred.Autocorrelation.AATSC(8, 'pe'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 8 weighted by pauling EN', 'module': 'Autocorrelation', 'constructor': 'AATSC(8, "pe")'}, 'AATSC0are': {'function': mordred.Autocorrelation.AATSC(0, 'are'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 0 weighted by allred-rocow EN', 'module': 'Autocorrelation', 'constructor': 'AATSC(0, "are")'}, 'AATSC1are': {'function': mordred.Autocorrelation.AATSC(1, 'are'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 1 weighted by allred-rocow EN', 'module': 'Autocorrelation', 'constructor': 'AATSC(1, "are")'}, 'AATSC2are': {'function': mordred.Autocorrelation.AATSC(2, 'are'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 2 weighted by allred-rocow EN', 'module': 'Autocorrelation', 'constructor': 'AATSC(2, "are")'}, 'AATSC3are': {'function': mordred.Autocorrelation.AATSC(3, 'are'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 3 weighted by allred-rocow EN', 'module': 'Autocorrelation', 'constructor': 'AATSC(3, "are")'}, 'AATSC4are': {'function': mordred.Autocorrelation.AATSC(4, 'are'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 4 weighted by allred-rocow EN', 'module': 'Autocorrelation', 'constructor': 'AATSC(4, "are")'}, 'AATSC5are': {'function': mordred.Autocorrelation.AATSC(5, 'are'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 5 weighted by allred-rocow EN', 'module': 'Autocorrelation', 'constructor': 'AATSC(5, "are")'}, 'AATSC6are': {'function': mordred.Autocorrelation.AATSC(6, 'are'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 6 weighted by allred-rocow EN', 'module': 'Autocorrelation', 'constructor': 'AATSC(6, "are")'}, 'AATSC7are': {'function': mordred.Autocorrelation.AATSC(7, 'are'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 7 weighted by allred-rocow EN', 'module': 'Autocorrelation', 'constructor': 'AATSC(7, "are")'}, 'AATSC8are': {'function': mordred.Autocorrelation.AATSC(8, 'are'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 8 weighted by allred-rocow EN', 'module': 'Autocorrelation', 'constructor': 'AATSC(8, "are")'}, 'AATSC0p': {'function': mordred.Autocorrelation.AATSC(0, 'p'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 0 weighted by polarizability', 'module': 'Autocorrelation', 'constructor': 'AATSC(0, "p")'}, 'AATSC1p': {'function': mordred.Autocorrelation.AATSC(1, 'p'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 1 weighted by polarizability', 'module': 'Autocorrelation', 'constructor': 'AATSC(1, "p")'}, 'AATSC2p': {'function': mordred.Autocorrelation.AATSC(2, 'p'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 2 weighted by polarizability', 'module': 'Autocorrelation', 'constructor': 'AATSC(2, "p")'}, 'AATSC3p': {'function': mordred.Autocorrelation.AATSC(3, 'p'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 3 weighted by polarizability', 'module': 'Autocorrelation', 'constructor': 'AATSC(3, "p")'}, 'AATSC4p': {'function': mordred.Autocorrelation.AATSC(4, 'p'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 4 weighted by polarizability', 'module': 'Autocorrelation', 'constructor': 'AATSC(4, "p")'}, 'AATSC5p': {'function': mordred.Autocorrelation.AATSC(5, 'p'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 5 weighted by polarizability', 'module': 'Autocorrelation', 'constructor': 'AATSC(5, "p")'}, 'AATSC6p': {'function': mordred.Autocorrelation.AATSC(6, 'p'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 6 weighted by polarizability', 'module': 'Autocorrelation', 'constructor': 'AATSC(6, "p")'}, 'AATSC7p': {'function': mordred.Autocorrelation.AATSC(7, 'p'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 7 weighted by polarizability', 'module': 'Autocorrelation', 'constructor': 'AATSC(7, "p")'}, 'AATSC8p': {'function': mordred.Autocorrelation.AATSC(8, 'p'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 8 weighted by polarizability', 'module': 'Autocorrelation', 'constructor': 'AATSC(8, "p")'}, 'AATSC0i': {'function': mordred.Autocorrelation.AATSC(0, 'i'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 0 weighted by ionization potential', 'module': 'Autocorrelation', 'constructor': 'AATSC(0, "i")'}, 'AATSC1i': {'function': mordred.Autocorrelation.AATSC(1, 'i'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 1 weighted by ionization potential', 'module': 'Autocorrelation', 'constructor': 'AATSC(1, "i")'}, 'AATSC2i': {'function': mordred.Autocorrelation.AATSC(2, 'i'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 2 weighted by ionization potential', 'module': 'Autocorrelation', 'constructor': 'AATSC(2, "i")'}, 'AATSC3i': {'function': mordred.Autocorrelation.AATSC(3, 'i'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 3 weighted by ionization potential', 'module': 'Autocorrelation', 'constructor': 'AATSC(3, "i")'}, 'AATSC4i': {'function': mordred.Autocorrelation.AATSC(4, 'i'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 4 weighted by ionization potential', 'module': 'Autocorrelation', 'constructor': 'AATSC(4, "i")'}, 'AATSC5i': {'function': mordred.Autocorrelation.AATSC(5, 'i'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 5 weighted by ionization potential', 'module': 'Autocorrelation', 'constructor': 'AATSC(5, "i")'}, 'AATSC6i': {'function': mordred.Autocorrelation.AATSC(6, 'i'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 6 weighted by ionization potential', 'module': 'Autocorrelation', 'constructor': 'AATSC(6, "i")'}, 'AATSC7i': {'function': mordred.Autocorrelation.AATSC(7, 'i'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 7 weighted by ionization potential', 'module': 'Autocorrelation', 'constructor': 'AATSC(7, "i")'}, 'AATSC8i': {'function': mordred.Autocorrelation.AATSC(8, 'i'), 'type': '2D', 'description': 'averaged and centered moreau-broto autocorrelation of lag 8 weighted by ionization potential', 'module': 'Autocorrelation', 'constructor': 'AATSC(8, "i")'}, 'MATS1c': {'function': mordred.Autocorrelation.MATS(1, 'c'), 'type': '2D', 'description': 'moran coefficient of lag 1 weighted by gasteiger charge', 'module': 'Autocorrelation', 'constructor': 'MATS(1, "c")'}, 'MATS2c': {'function': mordred.Autocorrelation.MATS(2, 'c'), 'type': '2D', 'description': 'moran coefficient of lag 2 weighted by gasteiger charge', 'module': 'Autocorrelation', 'constructor': 'MATS(2, "c")'}, 'MATS3c': {'function': mordred.Autocorrelation.MATS(3, 'c'), 'type': '2D', 'description': 'moran coefficient of lag 3 weighted by gasteiger charge', 'module': 'Autocorrelation', 'constructor': 'MATS(3, "c")'}, 'MATS4c': {'function': mordred.Autocorrelation.MATS(4, 'c'), 'type': '2D', 'description': 'moran coefficient of lag 4 weighted by gasteiger charge', 'module': 'Autocorrelation', 'constructor': 'MATS(4, "c")'}, 'MATS5c': {'function': mordred.Autocorrelation.MATS(5, 'c'), 'type': '2D', 'description': 'moran coefficient of lag 5 weighted by gasteiger charge', 'module': 'Autocorrelation', 'constructor': 'MATS(5, "c")'}, 'MATS6c': {'function': mordred.Autocorrelation.MATS(6, 'c'), 'type': '2D', 'description': 'moran coefficient of lag 6 weighted by gasteiger charge', 'module': 'Autocorrelation', 'constructor': 'MATS(6, "c")'}, 'MATS7c': {'function': mordred.Autocorrelation.MATS(7, 'c'), 'type': '2D', 'description': 'moran coefficient of lag 7 weighted by gasteiger charge', 'module': 'Autocorrelation', 'constructor': 'MATS(7, "c")'}, 'MATS8c': {'function': mordred.Autocorrelation.MATS(8, 'c'), 'type': '2D', 'description': 'moran coefficient of lag 8 weighted by gasteiger charge', 'module': 'Autocorrelation', 'constructor': 'MATS(8, "c")'}, 'MATS1dv': {'function': mordred.Autocorrelation.MATS(1, 'dv'), 'type': '2D', 'description': 'moran coefficient of lag 1 weighted by valence electrons', 'module': 'Autocorrelation', 'constructor': 'MATS(1, "dv")'}, 'MATS2dv': {'function': mordred.Autocorrelation.MATS(2, 'dv'), 'type': '2D', 'description': 'moran coefficient of lag 2 weighted by valence electrons', 'module': 'Autocorrelation', 'constructor': 'MATS(2, "dv")'}, 'MATS3dv': {'function': mordred.Autocorrelation.MATS(3, 'dv'), 'type': '2D', 'description': 'moran coefficient of lag 3 weighted by valence electrons', 'module': 'Autocorrelation', 'constructor': 'MATS(3, "dv")'}, 'MATS4dv': {'function': mordred.Autocorrelation.MATS(4, 'dv'), 'type': '2D', 'description': 'moran coefficient of lag 4 weighted by valence electrons', 'module': 'Autocorrelation', 'constructor': 'MATS(4, "dv")'}, 'MATS5dv': {'function': mordred.Autocorrelation.MATS(5, 'dv'), 'type': '2D', 'description': 'moran coefficient of lag 5 weighted by valence electrons', 'module': 'Autocorrelation', 'constructor': 'MATS(5, "dv")'}, 'MATS6dv': {'function': mordred.Autocorrelation.MATS(6, 'dv'), 'type': '2D', 'description': 'moran coefficient of lag 6 weighted by valence electrons', 'module': 'Autocorrelation', 'constructor': 'MATS(6, "dv")'}, 'MATS7dv': {'function': mordred.Autocorrelation.MATS(7, 'dv'), 'type': '2D', 'description': 'moran coefficient of lag 7 weighted by valence electrons', 'module': 'Autocorrelation', 'constructor': 'MATS(7, "dv")'}, 'MATS8dv': {'function': mordred.Autocorrelation.MATS(8, 'dv'), 'type': '2D', 'description': 'moran coefficient of lag 8 weighted by valence electrons', 'module': 'Autocorrelation', 'constructor': 'MATS(8, "dv")'}, 'MATS1d': {'function': mordred.Autocorrelation.MATS(1, 'd'), 'type': '2D', 'description': 'moran coefficient of lag 1 weighted by sigma electrons', 'module': 'Autocorrelation', 'constructor': 'MATS(1, "d")'}, 'MATS2d': {'function': mordred.Autocorrelation.MATS(2, 'd'), 'type': '2D', 'description': 'moran coefficient of lag 2 weighted by sigma electrons', 'module': 'Autocorrelation', 'constructor': 'MATS(2, "d")'}, 'MATS3d': {'function': mordred.Autocorrelation.MATS(3, 'd'), 'type': '2D', 'description': 'moran coefficient of lag 3 weighted by sigma electrons', 'module': 'Autocorrelation', 'constructor': 'MATS(3, "d")'}, 'MATS4d': {'function': mordred.Autocorrelation.MATS(4, 'd'), 'type': '2D', 'description': 'moran coefficient of lag 4 weighted by sigma electrons', 'module': 'Autocorrelation', 'constructor': 'MATS(4, "d")'}, 'MATS5d': {'function': mordred.Autocorrelation.MATS(5, 'd'), 'type': '2D', 'description': 'moran coefficient of lag 5 weighted by sigma electrons', 'module': 'Autocorrelation', 'constructor': 'MATS(5, "d")'}, 'MATS6d': {'function': mordred.Autocorrelation.MATS(6, 'd'), 'type': '2D', 'description': 'moran coefficient of lag 6 weighted by sigma electrons', 'module': 'Autocorrelation', 'constructor': 'MATS(6, "d")'}, 'MATS7d': {'function': mordred.Autocorrelation.MATS(7, 'd'), 'type': '2D', 'description': 'moran coefficient of lag 7 weighted by sigma electrons', 'module': 'Autocorrelation', 'constructor': 'MATS(7, "d")'}, 'MATS8d': {'function': mordred.Autocorrelation.MATS(8, 'd'), 'type': '2D', 'description': 'moran coefficient of lag 8 weighted by sigma electrons', 'module': 'Autocorrelation', 'constructor': 'MATS(8, "d")'}, 'MATS1s': {'function': mordred.Autocorrelation.MATS(1, 's'), 'type': '2D', 'description': 'moran coefficient of lag 1 weighted by intrinsic state', 'module': 'Autocorrelation', 'constructor': 'MATS(1, "s")'}, 'MATS2s': {'function': mordred.Autocorrelation.MATS(2, 's'), 'type': '2D', 'description': 'moran coefficient of lag 2 weighted by intrinsic state', 'module': 'Autocorrelation', 'constructor': 'MATS(2, "s")'}, 'MATS3s': {'function': mordred.Autocorrelation.MATS(3, 's'), 'type': '2D', 'description': 'moran coefficient of lag 3 weighted by intrinsic state', 'module': 'Autocorrelation', 'constructor': 'MATS(3, "s")'}, 'MATS4s': {'function': mordred.Autocorrelation.MATS(4, 's'), 'type': '2D', 'description': 'moran coefficient of lag 4 weighted by intrinsic state', 'module': 'Autocorrelation', 'constructor': 'MATS(4, "s")'}, 'MATS5s': {'function': mordred.Autocorrelation.MATS(5, 's'), 'type': '2D', 'description': 'moran coefficient of lag 5 weighted by intrinsic state', 'module': 'Autocorrelation', 'constructor': 'MATS(5, "s")'}, 'MATS6s': {'function': mordred.Autocorrelation.MATS(6, 's'), 'type': '2D', 'description': 'moran coefficient of lag 6 weighted by intrinsic state', 'module': 'Autocorrelation', 'constructor': 'MATS(6, "s")'}, 'MATS7s': {'function': mordred.Autocorrelation.MATS(7, 's'), 'type': '2D', 'description': 'moran coefficient of lag 7 weighted by intrinsic state', 'module': 'Autocorrelation', 'constructor': 'MATS(7, "s")'}, 'MATS8s': {'function': mordred.Autocorrelation.MATS(8, 's'), 'type': '2D', 'description': 'moran coefficient of lag 8 weighted by intrinsic state', 'module': 'Autocorrelation', 'constructor': 'MATS(8, "s")'}, 'MATS1Z': {'function': mordred.Autocorrelation.MATS(1, 'Z'), 'type': '2D', 'description': 'moran coefficient of lag 1 weighted by atomic number', 'module': 'Autocorrelation', 'constructor': 'MATS(1, "Z")'}, 'MATS2Z': {'function': mordred.Autocorrelation.MATS(2, 'Z'), 'type': '2D', 'description': 'moran coefficient of lag 2 weighted by atomic number', 'module': 'Autocorrelation', 'constructor': 'MATS(2, "Z")'}, 'MATS3Z': {'function': mordred.Autocorrelation.MATS(3, 'Z'), 'type': '2D', 'description': 'moran coefficient of lag 3 weighted by atomic number', 'module': 'Autocorrelation', 'constructor': 'MATS(3, "Z")'}, 'MATS4Z': {'function': mordred.Autocorrelation.MATS(4, 'Z'), 'type': '2D', 'description': 'moran coefficient of lag 4 weighted by atomic number', 'module': 'Autocorrelation', 'constructor': 'MATS(4, "Z")'}, 'MATS5Z': {'function': mordred.Autocorrelation.MATS(5, 'Z'), 'type': '2D', 'description': 'moran coefficient of lag 5 weighted by atomic number', 'module': 'Autocorrelation', 'constructor': 'MATS(5, "Z")'}, 'MATS6Z': {'function': mordred.Autocorrelation.MATS(6, 'Z'), 'type': '2D', 'description': 'moran coefficient of lag 6 weighted by atomic number', 'module': 'Autocorrelation', 'constructor': 'MATS(6, "Z")'}, 'MATS7Z': {'function': mordred.Autocorrelation.MATS(7, 'Z'), 'type': '2D', 'description': 'moran coefficient of lag 7 weighted by atomic number', 'module': 'Autocorrelation', 'constructor': 'MATS(7, "Z")'}, 'MATS8Z': {'function': mordred.Autocorrelation.MATS(8, 'Z'), 'type': '2D', 'description': 'moran coefficient of lag 8 weighted by atomic number', 'module': 'Autocorrelation', 'constructor': 'MATS(8, "Z")'}, 'MATS1m': {'function': mordred.Autocorrelation.MATS(1, 'm'), 'type': '2D', 'description': 'moran coefficient of lag 1 weighted by mass', 'module': 'Autocorrelation', 'constructor': 'MATS(1, "m")'}, 'MATS2m': {'function': mordred.Autocorrelation.MATS(2, 'm'), 'type': '2D', 'description': 'moran coefficient of lag 2 weighted by mass', 'module': 'Autocorrelation', 'constructor': 'MATS(2, "m")'}, 'MATS3m': {'function': mordred.Autocorrelation.MATS(3, 'm'), 'type': '2D', 'description': 'moran coefficient of lag 3 weighted by mass', 'module': 'Autocorrelation', 'constructor': 'MATS(3, "m")'}, 'MATS4m': {'function': mordred.Autocorrelation.MATS(4, 'm'), 'type': '2D', 'description': 'moran coefficient of lag 4 weighted by mass', 'module': 'Autocorrelation', 'constructor': 'MATS(4, "m")'}, 'MATS5m': {'function': mordred.Autocorrelation.MATS(5, 'm'), 'type': '2D', 'description': 'moran coefficient of lag 5 weighted by mass', 'module': 'Autocorrelation', 'constructor': 'MATS(5, "m")'}, 'MATS6m': {'function': mordred.Autocorrelation.MATS(6, 'm'), 'type': '2D', 'description': 'moran coefficient of lag 6 weighted by mass', 'module': 'Autocorrelation', 'constructor': 'MATS(6, "m")'}, 'MATS7m': {'function': mordred.Autocorrelation.MATS(7, 'm'), 'type': '2D', 'description': 'moran coefficient of lag 7 weighted by mass', 'module': 'Autocorrelation', 'constructor': 'MATS(7, "m")'}, 'MATS8m': {'function': mordred.Autocorrelation.MATS(8, 'm'), 'type': '2D', 'description': 'moran coefficient of lag 8 weighted by mass', 'module': 'Autocorrelation', 'constructor': 'MATS(8, "m")'}, 'MATS1v': {'function': mordred.Autocorrelation.MATS(1, 'v'), 'type': '2D', 'description': 'moran coefficient of lag 1 weighted by vdw volume', 'module': 'Autocorrelation', 'constructor': 'MATS(1, "v")'}, 'MATS2v': {'function': mordred.Autocorrelation.MATS(2, 'v'), 'type': '2D', 'description': 'moran coefficient of lag 2 weighted by vdw volume', 'module': 'Autocorrelation', 'constructor': 'MATS(2, "v")'}, 'MATS3v': {'function': mordred.Autocorrelation.MATS(3, 'v'), 'type': '2D', 'description': 'moran coefficient of lag 3 weighted by vdw volume', 'module': 'Autocorrelation', 'constructor': 'MATS(3, "v")'}, 'MATS4v': {'function': mordred.Autocorrelation.MATS(4, 'v'), 'type': '2D', 'description': 'moran coefficient of lag 4 weighted by vdw volume', 'module': 'Autocorrelation', 'constructor': 'MATS(4, "v")'}, 'MATS5v': {'function': mordred.Autocorrelation.MATS(5, 'v'), 'type': '2D', 'description': 'moran coefficient of lag 5 weighted by vdw volume', 'module': 'Autocorrelation', 'constructor': 'MATS(5, "v")'}, 'MATS6v': {'function': mordred.Autocorrelation.MATS(6, 'v'), 'type': '2D', 'description': 'moran coefficient of lag 6 weighted by vdw volume', 'module': 'Autocorrelation', 'constructor': 'MATS(6, "v")'}, 'MATS7v': {'function': mordred.Autocorrelation.MATS(7, 'v'), 'type': '2D', 'description': 'moran coefficient of lag 7 weighted by vdw volume', 'module': 'Autocorrelation', 'constructor': 'MATS(7, "v")'}, 'MATS8v': {'function': mordred.Autocorrelation.MATS(8, 'v'), 'type': '2D', 'description': 'moran coefficient of lag 8 weighted by vdw volume', 'module': 'Autocorrelation', 'constructor': 'MATS(8, "v")'}, 'MATS1se': {'function': mordred.Autocorrelation.MATS(1, 'se'), 'type': '2D', 'description': 'moran coefficient of lag 1 weighted by sanderson EN', 'module': 'Autocorrelation', 'constructor': 'MATS(1, "se")'}, 'MATS2se': {'function': mordred.Autocorrelation.MATS(2, 'se'), 'type': '2D', 'description': 'moran coefficient of lag 2 weighted by sanderson EN', 'module': 'Autocorrelation', 'constructor': 'MATS(2, "se")'}, 'MATS3se': {'function': mordred.Autocorrelation.MATS(3, 'se'), 'type': '2D', 'description': 'moran coefficient of lag 3 weighted by sanderson EN', 'module': 'Autocorrelation', 'constructor': 'MATS(3, "se")'}, 'MATS4se': {'function': mordred.Autocorrelation.MATS(4, 'se'), 'type': '2D', 'description': 'moran coefficient of lag 4 weighted by sanderson EN', 'module': 'Autocorrelation', 'constructor': 'MATS(4, "se")'}, 'MATS5se': {'function': mordred.Autocorrelation.MATS(5, 'se'), 'type': '2D', 'description': 'moran coefficient of lag 5 weighted by sanderson EN', 'module': 'Autocorrelation', 'constructor': 'MATS(5, "se")'}, 'MATS6se': {'function': mordred.Autocorrelation.MATS(6, 'se'), 'type': '2D', 'description': 'moran coefficient of lag 6 weighted by sanderson EN', 'module': 'Autocorrelation', 'constructor': 'MATS(6, "se")'}, 'MATS7se': {'function': mordred.Autocorrelation.MATS(7, 'se'), 'type': '2D', 'description': 'moran coefficient of lag 7 weighted by sanderson EN', 'module': 'Autocorrelation', 'constructor': 'MATS(7, "se")'}, 'MATS8se': {'function': mordred.Autocorrelation.MATS(8, 'se'), 'type': '2D', 'description': 'moran coefficient of lag 8 weighted by sanderson EN', 'module': 'Autocorrelation', 'constructor': 'MATS(8, "se")'}, 'MATS1pe': {'function': mordred.Autocorrelation.MATS(1, 'pe'), 'type': '2D', 'description': 'moran coefficient of lag 1 weighted by pauling EN', 'module': 'Autocorrelation', 'constructor': 'MATS(1, "pe")'}, 'MATS2pe': {'function': mordred.Autocorrelation.MATS(2, 'pe'), 'type': '2D', 'description': 'moran coefficient of lag 2 weighted by pauling EN', 'module': 'Autocorrelation', 'constructor': 'MATS(2, "pe")'}, 'MATS3pe': {'function': mordred.Autocorrelation.MATS(3, 'pe'), 'type': '2D', 'description': 'moran coefficient of lag 3 weighted by pauling EN', 'module': 'Autocorrelation', 'constructor': 'MATS(3, "pe")'}, 'MATS4pe': {'function': mordred.Autocorrelation.MATS(4, 'pe'), 'type': '2D', 'description': 'moran coefficient of lag 4 weighted by pauling EN', 'module': 'Autocorrelation', 'constructor': 'MATS(4, "pe")'}, 'MATS5pe': {'function': mordred.Autocorrelation.MATS(5, 'pe'), 'type': '2D', 'description': 'moran coefficient of lag 5 weighted by pauling EN', 'module': 'Autocorrelation', 'constructor': 'MATS(5, "pe")'}, 'MATS6pe': {'function': mordred.Autocorrelation.MATS(6, 'pe'), 'type': '2D', 'description': 'moran coefficient of lag 6 weighted by pauling EN', 'module': 'Autocorrelation', 'constructor': 'MATS(6, "pe")'}, 'MATS7pe': {'function': mordred.Autocorrelation.MATS(7, 'pe'), 'type': '2D', 'description': 'moran coefficient of lag 7 weighted by pauling EN', 'module': 'Autocorrelation', 'constructor': 'MATS(7, "pe")'}, 'MATS8pe': {'function': mordred.Autocorrelation.MATS(8, 'pe'), 'type': '2D', 'description': 'moran coefficient of lag 8 weighted by pauling EN', 'module': 'Autocorrelation', 'constructor': 'MATS(8, "pe")'}, 'MATS1are': {'function': mordred.Autocorrelation.MATS(1, 'are'), 'type': '2D', 'description': 'moran coefficient of lag 1 weighted by allred-rocow EN', 'module': 'Autocorrelation', 'constructor': 'MATS(1, "are")'}, 'MATS2are': {'function': mordred.Autocorrelation.MATS(2, 'are'), 'type': '2D', 'description': 'moran coefficient of lag 2 weighted by allred-rocow EN', 'module': 'Autocorrelation', 'constructor': 'MATS(2, "are")'}, 'MATS3are': {'function': mordred.Autocorrelation.MATS(3, 'are'), 'type': '2D', 'description': 'moran coefficient of lag 3 weighted by allred-rocow EN', 'module': 'Autocorrelation', 'constructor': 'MATS(3, "are")'}, 'MATS4are': {'function': mordred.Autocorrelation.MATS(4, 'are'), 'type': '2D', 'description': 'moran coefficient of lag 4 weighted by allred-rocow EN', 'module': 'Autocorrelation', 'constructor': 'MATS(4, "are")'}, 'MATS5are': {'function': mordred.Autocorrelation.MATS(5, 'are'), 'type': '2D', 'description': 'moran coefficient of lag 5 weighted by allred-rocow EN', 'module': 'Autocorrelation', 'constructor': 'MATS(5, "are")'}, 'MATS6are': {'function': mordred.Autocorrelation.MATS(6, 'are'), 'type': '2D', 'description': 'moran coefficient of lag 6 weighted by allred-rocow EN', 'module': 'Autocorrelation', 'constructor': 'MATS(6, "are")'}, 'MATS7are': {'function': mordred.Autocorrelation.MATS(7, 'are'), 'type': '2D', 'description': 'moran coefficient of lag 7 weighted by allred-rocow EN', 'module': 'Autocorrelation', 'constructor': 'MATS(7, "are")'}, 'MATS8are': {'function': mordred.Autocorrelation.MATS(8, 'are'), 'type': '2D', 'description': 'moran coefficient of lag 8 weighted by allred-rocow EN', 'module': 'Autocorrelation', 'constructor': 'MATS(8, "are")'}, 'MATS1p': {'function': mordred.Autocorrelation.MATS(1, 'p'), 'type': '2D', 'description': 'moran coefficient of lag 1 weighted by polarizability', 'module': 'Autocorrelation', 'constructor': 'MATS(1, "p")'}, 'MATS2p': {'function': mordred.Autocorrelation.MATS(2, 'p'), 'type': '2D', 'description': 'moran coefficient of lag 2 weighted by polarizability', 'module': 'Autocorrelation', 'constructor': 'MATS(2, "p")'}, 'MATS3p': {'function': mordred.Autocorrelation.MATS(3, 'p'), 'type': '2D', 'description': 'moran coefficient of lag 3 weighted by polarizability', 'module': 'Autocorrelation', 'constructor': 'MATS(3, "p")'}, 'MATS4p': {'function': mordred.Autocorrelation.MATS(4, 'p'), 'type': '2D', 'description': 'moran coefficient of lag 4 weighted by polarizability', 'module': 'Autocorrelation', 'constructor': 'MATS(4, "p")'}, 'MATS5p': {'function': mordred.Autocorrelation.MATS(5, 'p'), 'type': '2D', 'description': 'moran coefficient of lag 5 weighted by polarizability', 'module': 'Autocorrelation', 'constructor': 'MATS(5, "p")'}, 'MATS6p': {'function': mordred.Autocorrelation.MATS(6, 'p'), 'type': '2D', 'description': 'moran coefficient of lag 6 weighted by polarizability', 'module': 'Autocorrelation', 'constructor': 'MATS(6, "p")'}, 'MATS7p': {'function': mordred.Autocorrelation.MATS(7, 'p'), 'type': '2D', 'description': 'moran coefficient of lag 7 weighted by polarizability', 'module': 'Autocorrelation', 'constructor': 'MATS(7, "p")'}, 'MATS8p': {'function': mordred.Autocorrelation.MATS(8, 'p'), 'type': '2D', 'description': 'moran coefficient of lag 8 weighted by polarizability', 'module': 'Autocorrelation', 'constructor': 'MATS(8, "p")'}, 'MATS1i': {'function': mordred.Autocorrelation.MATS(1, 'i'), 'type': '2D', 'description': 'moran coefficient of lag 1 weighted by ionization potential', 'module': 'Autocorrelation', 'constructor': 'MATS(1, "i")'}, 'MATS2i': {'function': mordred.Autocorrelation.MATS(2, 'i'), 'type': '2D', 'description': 'moran coefficient of lag 2 weighted by ionization potential', 'module': 'Autocorrelation', 'constructor': 'MATS(2, "i")'}, 'MATS3i': {'function': mordred.Autocorrelation.MATS(3, 'i'), 'type': '2D', 'description': 'moran coefficient of lag 3 weighted by ionization potential', 'module': 'Autocorrelation', 'constructor': 'MATS(3, "i")'}, 'MATS4i': {'function': mordred.Autocorrelation.MATS(4, 'i'), 'type': '2D', 'description': 'moran coefficient of lag 4 weighted by ionization potential', 'module': 'Autocorrelation', 'constructor': 'MATS(4, "i")'}, 'MATS5i': {'function': mordred.Autocorrelation.MATS(5, 'i'), 'type': '2D', 'description': 'moran coefficient of lag 5 weighted by ionization potential', 'module': 'Autocorrelation', 'constructor': 'MATS(5, "i")'}, 'MATS6i': {'function': mordred.Autocorrelation.MATS(6, 'i'), 'type': '2D', 'description': 'moran coefficient of lag 6 weighted by ionization potential', 'module': 'Autocorrelation', 'constructor': 'MATS(6, "i")'}, 'MATS7i': {'function': mordred.Autocorrelation.MATS(7, 'i'), 'type': '2D', 'description': 'moran coefficient of lag 7 weighted by ionization potential', 'module': 'Autocorrelation', 'constructor': 'MATS(7, "i")'}, 'MATS8i': {'function': mordred.Autocorrelation.MATS(8, 'i'), 'type': '2D', 'description': 'moran coefficient of lag 8 weighted by ionization potential', 'module': 'Autocorrelation', 'constructor': 'MATS(8, "i")'}, 'GATS1c': {'function': mordred.Autocorrelation.GATS(1, 'c'), 'type': '2D', 'description': 'geary coefficient of lag 1 weighted by gasteiger charge', 'module': 'Autocorrelation', 'constructor': 'GATS(1, "c")'}, 'GATS2c': {'function': mordred.Autocorrelation.GATS(2, 'c'), 'type': '2D', 'description': 'geary coefficient of lag 2 weighted by gasteiger charge', 'module': 'Autocorrelation', 'constructor': 'GATS(2, "c")'}, 'GATS3c': {'function': mordred.Autocorrelation.GATS(3, 'c'), 'type': '2D', 'description': 'geary coefficient of lag 3 weighted by gasteiger charge', 'module': 'Autocorrelation', 'constructor': 'GATS(3, "c")'}, 'GATS4c': {'function': mordred.Autocorrelation.GATS(4, 'c'), 'type': '2D', 'description': 'geary coefficient of lag 4 weighted by gasteiger charge', 'module': 'Autocorrelation', 'constructor': 'GATS(4, "c")'}, 'GATS5c': {'function': mordred.Autocorrelation.GATS(5, 'c'), 'type': '2D', 'description': 'geary coefficient of lag 5 weighted by gasteiger charge', 'module': 'Autocorrelation', 'constructor': 'GATS(5, "c")'}, 'GATS6c': {'function': mordred.Autocorrelation.GATS(6, 'c'), 'type': '2D', 'description': 'geary coefficient of lag 6 weighted by gasteiger charge', 'module': 'Autocorrelation', 'constructor': 'GATS(6, "c")'}, 'GATS7c': {'function': mordred.Autocorrelation.GATS(7, 'c'), 'type': '2D', 'description': 'geary coefficient of lag 7 weighted by gasteiger charge', 'module': 'Autocorrelation', 'constructor': 'GATS(7, "c")'}, 'GATS8c': {'function': mordred.Autocorrelation.GATS(8, 'c'), 'type': '2D', 'description': 'geary coefficient of lag 8 weighted by gasteiger charge', 'module': 'Autocorrelation', 'constructor': 'GATS(8, "c")'}, 'GATS1dv': {'function': mordred.Autocorrelation.GATS(1, 'dv'), 'type': '2D', 'description': 'geary coefficient of lag 1 weighted by valence electrons', 'module': 'Autocorrelation', 'constructor': 'GATS(1, "dv")'}, 'GATS2dv': {'function': mordred.Autocorrelation.GATS(2, 'dv'), 'type': '2D', 'description': 'geary coefficient of lag 2 weighted by valence electrons', 'module': 'Autocorrelation', 'constructor': 'GATS(2, "dv")'}, 'GATS3dv': {'function': mordred.Autocorrelation.GATS(3, 'dv'), 'type': '2D', 'description': 'geary coefficient of lag 3 weighted by valence electrons', 'module': 'Autocorrelation', 'constructor': 'GATS(3, "dv")'}, 'GATS4dv': {'function': mordred.Autocorrelation.GATS(4, 'dv'), 'type': '2D', 'description': 'geary coefficient of lag 4 weighted by valence electrons', 'module': 'Autocorrelation', 'constructor': 'GATS(4, "dv")'}, 'GATS5dv': {'function': mordred.Autocorrelation.GATS(5, 'dv'), 'type': '2D', 'description': 'geary coefficient of lag 5 weighted by valence electrons', 'module': 'Autocorrelation', 'constructor': 'GATS(5, "dv")'}, 'GATS6dv': {'function': mordred.Autocorrelation.GATS(6, 'dv'), 'type': '2D', 'description': 'geary coefficient of lag 6 weighted by valence electrons', 'module': 'Autocorrelation', 'constructor': 'GATS(6, "dv")'}, 'GATS7dv': {'function': mordred.Autocorrelation.GATS(7, 'dv'), 'type': '2D', 'description': 'geary coefficient of lag 7 weighted by valence electrons', 'module': 'Autocorrelation', 'constructor': 'GATS(7, "dv")'}, 'GATS8dv': {'function': mordred.Autocorrelation.GATS(8, 'dv'), 'type': '2D', 'description': 'geary coefficient of lag 8 weighted by valence electrons', 'module': 'Autocorrelation', 'constructor': 'GATS(8, "dv")'}, 'GATS1d': {'function': mordred.Autocorrelation.GATS(1, 'd'), 'type': '2D', 'description': 'geary coefficient of lag 1 weighted by sigma electrons', 'module': 'Autocorrelation', 'constructor': 'GATS(1, "d")'}, 'GATS2d': {'function': mordred.Autocorrelation.GATS(2, 'd'), 'type': '2D', 'description': 'geary coefficient of lag 2 weighted by sigma electrons', 'module': 'Autocorrelation', 'constructor': 'GATS(2, "d")'}, 'GATS3d': {'function': mordred.Autocorrelation.GATS(3, 'd'), 'type': '2D', 'description': 'geary coefficient of lag 3 weighted by sigma electrons', 'module': 'Autocorrelation', 'constructor': 'GATS(3, "d")'}, 'GATS4d': {'function': mordred.Autocorrelation.GATS(4, 'd'), 'type': '2D', 'description': 'geary coefficient of lag 4 weighted by sigma electrons', 'module': 'Autocorrelation', 'constructor': 'GATS(4, "d")'}, 'GATS5d': {'function': mordred.Autocorrelation.GATS(5, 'd'), 'type': '2D', 'description': 'geary coefficient of lag 5 weighted by sigma electrons', 'module': 'Autocorrelation', 'constructor': 'GATS(5, "d")'}, 'GATS6d': {'function': mordred.Autocorrelation.GATS(6, 'd'), 'type': '2D', 'description': 'geary coefficient of lag 6 weighted by sigma electrons', 'module': 'Autocorrelation', 'constructor': 'GATS(6, "d")'}, 'GATS7d': {'function': mordred.Autocorrelation.GATS(7, 'd'), 'type': '2D', 'description': 'geary coefficient of lag 7 weighted by sigma electrons', 'module': 'Autocorrelation', 'constructor': 'GATS(7, "d")'}, 'GATS8d': {'function': mordred.Autocorrelation.GATS(8, 'd'), 'type': '2D', 'description': 'geary coefficient of lag 8 weighted by sigma electrons', 'module': 'Autocorrelation', 'constructor': 'GATS(8, "d")'}, 'GATS1s': {'function': mordred.Autocorrelation.GATS(1, 's'), 'type': '2D', 'description': 'geary coefficient of lag 1 weighted by intrinsic state', 'module': 'Autocorrelation', 'constructor': 'GATS(1, "s")'}, 'GATS2s': {'function': mordred.Autocorrelation.GATS(2, 's'), 'type': '2D', 'description': 'geary coefficient of lag 2 weighted by intrinsic state', 'module': 'Autocorrelation', 'constructor': 'GATS(2, "s")'}, 'GATS3s': {'function': mordred.Autocorrelation.GATS(3, 's'), 'type': '2D', 'description': 'geary coefficient of lag 3 weighted by intrinsic state', 'module': 'Autocorrelation', 'constructor': 'GATS(3, "s")'}, 'GATS4s': {'function': mordred.Autocorrelation.GATS(4, 's'), 'type': '2D', 'description': 'geary coefficient of lag 4 weighted by intrinsic state', 'module': 'Autocorrelation', 'constructor': 'GATS(4, "s")'}, 'GATS5s': {'function': mordred.Autocorrelation.GATS(5, 's'), 'type': '2D', 'description': 'geary coefficient of lag 5 weighted by intrinsic state', 'module': 'Autocorrelation', 'constructor': 'GATS(5, "s")'}, 'GATS6s': {'function': mordred.Autocorrelation.GATS(6, 's'), 'type': '2D', 'description': 'geary coefficient of lag 6 weighted by intrinsic state', 'module': 'Autocorrelation', 'constructor': 'GATS(6, "s")'}, 'GATS7s': {'function': mordred.Autocorrelation.GATS(7, 's'), 'type': '2D', 'description': 'geary coefficient of lag 7 weighted by intrinsic state', 'module': 'Autocorrelation', 'constructor': 'GATS(7, "s")'}, 'GATS8s': {'function': mordred.Autocorrelation.GATS(8, 's'), 'type': '2D', 'description': 'geary coefficient of lag 8 weighted by intrinsic state', 'module': 'Autocorrelation', 'constructor': 'GATS(8, "s")'}, 'GATS1Z': {'function': mordred.Autocorrelation.GATS(1, 'Z'), 'type': '2D', 'description': 'geary coefficient of lag 1 weighted by atomic number', 'module': 'Autocorrelation', 'constructor': 'GATS(1, "Z")'}, 'GATS2Z': {'function': mordred.Autocorrelation.GATS(2, 'Z'), 'type': '2D', 'description': 'geary coefficient of lag 2 weighted by atomic number', 'module': 'Autocorrelation', 'constructor': 'GATS(2, "Z")'}, 'GATS3Z': {'function': mordred.Autocorrelation.GATS(3, 'Z'), 'type': '2D', 'description': 'geary coefficient of lag 3 weighted by atomic number', 'module': 'Autocorrelation', 'constructor': 'GATS(3, "Z")'}, 'GATS4Z': {'function': mordred.Autocorrelation.GATS(4, 'Z'), 'type': '2D', 'description': 'geary coefficient of lag 4 weighted by atomic number', 'module': 'Autocorrelation', 'constructor': 'GATS(4, "Z")'}, 'GATS5Z': {'function': mordred.Autocorrelation.GATS(5, 'Z'), 'type': '2D', 'description': 'geary coefficient of lag 5 weighted by atomic number', 'module': 'Autocorrelation', 'constructor': 'GATS(5, "Z")'}, 'GATS6Z': {'function': mordred.Autocorrelation.GATS(6, 'Z'), 'type': '2D', 'description': 'geary coefficient of lag 6 weighted by atomic number', 'module': 'Autocorrelation', 'constructor': 'GATS(6, "Z")'}, 'GATS7Z': {'function': mordred.Autocorrelation.GATS(7, 'Z'), 'type': '2D', 'description': 'geary coefficient of lag 7 weighted by atomic number', 'module': 'Autocorrelation', 'constructor': 'GATS(7, "Z")'}, 'GATS8Z': {'function': mordred.Autocorrelation.GATS(8, 'Z'), 'type': '2D', 'description': 'geary coefficient of lag 8 weighted by atomic number', 'module': 'Autocorrelation', 'constructor': 'GATS(8, "Z")'}, 'GATS1m': {'function': mordred.Autocorrelation.GATS(1, 'm'), 'type': '2D', 'description': 'geary coefficient of lag 1 weighted by mass', 'module': 'Autocorrelation', 'constructor': 'GATS(1, "m")'}, 'GATS2m': {'function': mordred.Autocorrelation.GATS(2, 'm'), 'type': '2D', 'description': 'geary coefficient of lag 2 weighted by mass', 'module': 'Autocorrelation', 'constructor': 'GATS(2, "m")'}, 'GATS3m': {'function': mordred.Autocorrelation.GATS(3, 'm'), 'type': '2D', 'description': 'geary coefficient of lag 3 weighted by mass', 'module': 'Autocorrelation', 'constructor': 'GATS(3, "m")'}, 'GATS4m': {'function': mordred.Autocorrelation.GATS(4, 'm'), 'type': '2D', 'description': 'geary coefficient of lag 4 weighted by mass', 'module': 'Autocorrelation', 'constructor': 'GATS(4, "m")'}, 'GATS5m': {'function': mordred.Autocorrelation.GATS(5, 'm'), 'type': '2D', 'description': 'geary coefficient of lag 5 weighted by mass', 'module': 'Autocorrelation', 'constructor': 'GATS(5, "m")'}, 'GATS6m': {'function': mordred.Autocorrelation.GATS(6, 'm'), 'type': '2D', 'description': 'geary coefficient of lag 6 weighted by mass', 'module': 'Autocorrelation', 'constructor': 'GATS(6, "m")'}, 'GATS7m': {'function': mordred.Autocorrelation.GATS(7, 'm'), 'type': '2D', 'description': 'geary coefficient of lag 7 weighted by mass', 'module': 'Autocorrelation', 'constructor': 'GATS(7, "m")'}, 'GATS8m': {'function': mordred.Autocorrelation.GATS(8, 'm'), 'type': '2D', 'description': 'geary coefficient of lag 8 weighted by mass', 'module': 'Autocorrelation', 'constructor': 'GATS(8, "m")'}, 'GATS1v': {'function': mordred.Autocorrelation.GATS(1, 'v'), 'type': '2D', 'description': 'geary coefficient of lag 1 weighted by vdw volume', 'module': 'Autocorrelation', 'constructor': 'GATS(1, "v")'}, 'GATS2v': {'function': mordred.Autocorrelation.GATS(2, 'v'), 'type': '2D', 'description': 'geary coefficient of lag 2 weighted by vdw volume', 'module': 'Autocorrelation', 'constructor': 'GATS(2, "v")'}, 'GATS3v': {'function': mordred.Autocorrelation.GATS(3, 'v'), 'type': '2D', 'description': 'geary coefficient of lag 3 weighted by vdw volume', 'module': 'Autocorrelation', 'constructor': 'GATS(3, "v")'}, 'GATS4v': {'function': mordred.Autocorrelation.GATS(4, 'v'), 'type': '2D', 'description': 'geary coefficient of lag 4 weighted by vdw volume', 'module': 'Autocorrelation', 'constructor': 'GATS(4, "v")'}, 'GATS5v': {'function': mordred.Autocorrelation.GATS(5, 'v'), 'type': '2D', 'description': 'geary coefficient of lag 5 weighted by vdw volume', 'module': 'Autocorrelation', 'constructor': 'GATS(5, "v")'}, 'GATS6v': {'function': mordred.Autocorrelation.GATS(6, 'v'), 'type': '2D', 'description': 'geary coefficient of lag 6 weighted by vdw volume', 'module': 'Autocorrelation', 'constructor': 'GATS(6, "v")'}, 'GATS7v': {'function': mordred.Autocorrelation.GATS(7, 'v'), 'type': '2D', 'description': 'geary coefficient of lag 7 weighted by vdw volume', 'module': 'Autocorrelation', 'constructor': 'GATS(7, "v")'}, 'GATS8v': {'function': mordred.Autocorrelation.GATS(8, 'v'), 'type': '2D', 'description': 'geary coefficient of lag 8 weighted by vdw volume', 'module': 'Autocorrelation', 'constructor': 'GATS(8, "v")'}, 'GATS1se': {'function': mordred.Autocorrelation.GATS(1, 'se'), 'type': '2D', 'description': 'geary coefficient of lag 1 weighted by sanderson EN', 'module': 'Autocorrelation', 'constructor': 'GATS(1, "se")'}, 'GATS2se': {'function': mordred.Autocorrelation.GATS(2, 'se'), 'type': '2D', 'description': 'geary coefficient of lag 2 weighted by sanderson EN', 'module': 'Autocorrelation', 'constructor': 'GATS(2, "se")'}, 'GATS3se': {'function': mordred.Autocorrelation.GATS(3, 'se'), 'type': '2D', 'description': 'geary coefficient of lag 3 weighted by sanderson EN', 'module': 'Autocorrelation', 'constructor': 'GATS(3, "se")'}, 'GATS4se': {'function': mordred.Autocorrelation.GATS(4, 'se'), 'type': '2D', 'description': 'geary coefficient of lag 4 weighted by sanderson EN', 'module': 'Autocorrelation', 'constructor': 'GATS(4, "se")'}, 'GATS5se': {'function': mordred.Autocorrelation.GATS(5, 'se'), 'type': '2D', 'description': 'geary coefficient of lag 5 weighted by sanderson EN', 'module': 'Autocorrelation', 'constructor': 'GATS(5, "se")'}, 'GATS6se': {'function': mordred.Autocorrelation.GATS(6, 'se'), 'type': '2D', 'description': 'geary coefficient of lag 6 weighted by sanderson EN', 'module': 'Autocorrelation', 'constructor': 'GATS(6, "se")'}, 'GATS7se': {'function': mordred.Autocorrelation.GATS(7, 'se'), 'type': '2D', 'description': 'geary coefficient of lag 7 weighted by sanderson EN', 'module': 'Autocorrelation', 'constructor': 'GATS(7, "se")'}, 'GATS8se': {'function': mordred.Autocorrelation.GATS(8, 'se'), 'type': '2D', 'description': 'geary coefficient of lag 8 weighted by sanderson EN', 'module': 'Autocorrelation', 'constructor': 'GATS(8, "se")'}, 'GATS1pe': {'function': mordred.Autocorrelation.GATS(1, 'pe'), 'type': '2D', 'description': 'geary coefficient of lag 1 weighted by pauling EN', 'module': 'Autocorrelation', 'constructor': 'GATS(1, "pe")'}, 'GATS2pe': {'function': mordred.Autocorrelation.GATS(2, 'pe'), 'type': '2D', 'description': 'geary coefficient of lag 2 weighted by pauling EN', 'module': 'Autocorrelation', 'constructor': 'GATS(2, "pe")'}, 'GATS3pe': {'function': mordred.Autocorrelation.GATS(3, 'pe'), 'type': '2D', 'description': 'geary coefficient of lag 3 weighted by pauling EN', 'module': 'Autocorrelation', 'constructor': 'GATS(3, "pe")'}, 'GATS4pe': {'function': mordred.Autocorrelation.GATS(4, 'pe'), 'type': '2D', 'description': 'geary coefficient of lag 4 weighted by pauling EN', 'module': 'Autocorrelation', 'constructor': 'GATS(4, "pe")'}, 'GATS5pe': {'function': mordred.Autocorrelation.GATS(5, 'pe'), 'type': '2D', 'description': 'geary coefficient of lag 5 weighted by pauling EN', 'module': 'Autocorrelation', 'constructor': 'GATS(5, "pe")'}, 'GATS6pe': {'function': mordred.Autocorrelation.GATS(6, 'pe'), 'type': '2D', 'description': 'geary coefficient of lag 6 weighted by pauling EN', 'module': 'Autocorrelation', 'constructor': 'GATS(6, "pe")'}, 'GATS7pe': {'function': mordred.Autocorrelation.GATS(7, 'pe'), 'type': '2D', 'description': 'geary coefficient of lag 7 weighted by pauling EN', 'module': 'Autocorrelation', 'constructor': 'GATS(7, "pe")'}, 'GATS8pe': {'function': mordred.Autocorrelation.GATS(8, 'pe'), 'type': '2D', 'description': 'geary coefficient of lag 8 weighted by pauling EN', 'module': 'Autocorrelation', 'constructor': 'GATS(8, "pe")'}, 'GATS1are': {'function': mordred.Autocorrelation.GATS(1, 'are'), 'type': '2D', 'description': 'geary coefficient of lag 1 weighted by allred-rocow EN', 'module': 'Autocorrelation', 'constructor': 'GATS(1, "are")'}, 'GATS2are': {'function': mordred.Autocorrelation.GATS(2, 'are'), 'type': '2D', 'description': 'geary coefficient of lag 2 weighted by allred-rocow EN', 'module': 'Autocorrelation', 'constructor': 'GATS(2, "are")'}, 'GATS3are': {'function': mordred.Autocorrelation.GATS(3, 'are'), 'type': '2D', 'description': 'geary coefficient of lag 3 weighted by allred-rocow EN', 'module': 'Autocorrelation', 'constructor': 'GATS(3, "are")'}, 'GATS4are': {'function': mordred.Autocorrelation.GATS(4, 'are'), 'type': '2D', 'description': 'geary coefficient of lag 4 weighted by allred-rocow EN', 'module': 'Autocorrelation', 'constructor': 'GATS(4, "are")'}, 'GATS5are': {'function': mordred.Autocorrelation.GATS(5, 'are'), 'type': '2D', 'description': 'geary coefficient of lag 5 weighted by allred-rocow EN', 'module': 'Autocorrelation', 'constructor': 'GATS(5, "are")'}, 'GATS6are': {'function': mordred.Autocorrelation.GATS(6, 'are'), 'type': '2D', 'description': 'geary coefficient of lag 6 weighted by allred-rocow EN', 'module': 'Autocorrelation', 'constructor': 'GATS(6, "are")'}, 'GATS7are': {'function': mordred.Autocorrelation.GATS(7, 'are'), 'type': '2D', 'description': 'geary coefficient of lag 7 weighted by allred-rocow EN', 'module': 'Autocorrelation', 'constructor': 'GATS(7, "are")'}, 'GATS8are': {'function': mordred.Autocorrelation.GATS(8, 'are'), 'type': '2D', 'description': 'geary coefficient of lag 8 weighted by allred-rocow EN', 'module': 'Autocorrelation', 'constructor': 'GATS(8, "are")'}, 'GATS1p': {'function': mordred.Autocorrelation.GATS(1, 'p'), 'type': '2D', 'description': 'geary coefficient of lag 1 weighted by polarizability', 'module': 'Autocorrelation', 'constructor': 'GATS(1, "p")'}, 'GATS2p': {'function': mordred.Autocorrelation.GATS(2, 'p'), 'type': '2D', 'description': 'geary coefficient of lag 2 weighted by polarizability', 'module': 'Autocorrelation', 'constructor': 'GATS(2, "p")'}, 'GATS3p': {'function': mordred.Autocorrelation.GATS(3, 'p'), 'type': '2D', 'description': 'geary coefficient of lag 3 weighted by polarizability', 'module': 'Autocorrelation', 'constructor': 'GATS(3, "p")'}, 'GATS4p': {'function': mordred.Autocorrelation.GATS(4, 'p'), 'type': '2D', 'description': 'geary coefficient of lag 4 weighted by polarizability', 'module': 'Autocorrelation', 'constructor': 'GATS(4, "p")'}, 'GATS5p': {'function': mordred.Autocorrelation.GATS(5, 'p'), 'type': '2D', 'description': 'geary coefficient of lag 5 weighted by polarizability', 'module': 'Autocorrelation', 'constructor': 'GATS(5, "p")'}, 'GATS6p': {'function': mordred.Autocorrelation.GATS(6, 'p'), 'type': '2D', 'description': 'geary coefficient of lag 6 weighted by polarizability', 'module': 'Autocorrelation', 'constructor': 'GATS(6, "p")'}, 'GATS7p': {'function': mordred.Autocorrelation.GATS(7, 'p'), 'type': '2D', 'description': 'geary coefficient of lag 7 weighted by polarizability', 'module': 'Autocorrelation', 'constructor': 'GATS(7, "p")'}, 'GATS8p': {'function': mordred.Autocorrelation.GATS(8, 'p'), 'type': '2D', 'description': 'geary coefficient of lag 8 weighted by polarizability', 'module': 'Autocorrelation', 'constructor': 'GATS(8, "p")'}, 'GATS1i': {'function': mordred.Autocorrelation.GATS(1, 'i'), 'type': '2D', 'description': 'geary coefficient of lag 1 weighted by ionization potential', 'module': 'Autocorrelation', 'constructor': 'GATS(1, "i")'}, 'GATS2i': {'function': mordred.Autocorrelation.GATS(2, 'i'), 'type': '2D', 'description': 'geary coefficient of lag 2 weighted by ionization potential', 'module': 'Autocorrelation', 'constructor': 'GATS(2, "i")'}, 'GATS3i': {'function': mordred.Autocorrelation.GATS(3, 'i'), 'type': '2D', 'description': 'geary coefficient of lag 3 weighted by ionization potential', 'module': 'Autocorrelation', 'constructor': 'GATS(3, "i")'}, 'GATS4i': {'function': mordred.Autocorrelation.GATS(4, 'i'), 'type': '2D', 'description': 'geary coefficient of lag 4 weighted by ionization potential', 'module': 'Autocorrelation', 'constructor': 'GATS(4, "i")'}, 'GATS5i': {'function': mordred.Autocorrelation.GATS(5, 'i'), 'type': '2D', 'description': 'geary coefficient of lag 5 weighted by ionization potential', 'module': 'Autocorrelation', 'constructor': 'GATS(5, "i")'}, 'GATS6i': {'function': mordred.Autocorrelation.GATS(6, 'i'), 'type': '2D', 'description': 'geary coefficient of lag 6 weighted by ionization potential', 'module': 'Autocorrelation', 'constructor': 'GATS(6, "i")'}, 'GATS7i': {'function': mordred.Autocorrelation.GATS(7, 'i'), 'type': '2D', 'description': 'geary coefficient of lag 7 weighted by ionization potential', 'module': 'Autocorrelation', 'constructor': 'GATS(7, "i")'}, 'GATS8i': {'function': mordred.Autocorrelation.GATS(8, 'i'), 'type': '2D', 'description': 'geary coefficient of lag 8 weighted by ionization potential', 'module': 'Autocorrelation', 'constructor': 'GATS(8, "i")'}, 'BCUTc-1h': {'function': mordred.BCUT.BCUT('c', 0), 'type': '2D', 'description': 'first heighest eigenvalue of Burden matrix weighted by gasteiger charge', 'module': 'BCUT', 'constructor': 'BCUT("c", 0)'}, 'BCUTc-1l': {'function': mordred.BCUT.BCUT('c', -1), 'type': '2D', 'description': 'first lowest eigenvalue of Burden matrix weighted by gasteiger charge', 'module': 'BCUT', 'constructor': 'BCUT("c", -1)'}, 'BCUTdv-1h': {'function': mordred.BCUT.BCUT('dv', 0), 'type': '2D', 'description': 'first heighest eigenvalue of Burden matrix weighted by valence electrons', 'module': 'BCUT', 'constructor': 'BCUT("dv", 0)'}, 'BCUTdv-1l': {'function': mordred.BCUT.BCUT('dv', -1), 'type': '2D', 'description': 'first lowest eigenvalue of Burden matrix weighted by valence electrons', 'module': 'BCUT', 'constructor': 'BCUT("dv", -1)'}, 'BCUTd-1h': {'function': mordred.BCUT.BCUT('d', 0), 'type': '2D', 'description': 'first heighest eigenvalue of Burden matrix weighted by sigma electrons', 'module': 'BCUT', 'constructor': 'BCUT("d", 0)'}, 'BCUTd-1l': {'function': mordred.BCUT.BCUT('d', -1), 'type': '2D', 'description': 'first lowest eigenvalue of Burden matrix weighted by sigma electrons', 'module': 'BCUT', 'constructor': 'BCUT("d", -1)'}, 'BCUTs-1h': {'function': mordred.BCUT.BCUT('s', 0), 'type': '2D', 'description': 'first heighest eigenvalue of Burden matrix weighted by intrinsic state', 'module': 'BCUT', 'constructor': 'BCUT("s", 0)'}, 'BCUTs-1l': {'function': mordred.BCUT.BCUT('s', -1), 'type': '2D', 'description': 'first lowest eigenvalue of Burden matrix weighted by intrinsic state', 'module': 'BCUT', 'constructor': 'BCUT("s", -1)'}, 'BCUTZ-1h': {'function': mordred.BCUT.BCUT('Z', 0), 'type': '2D', 'description': 'first heighest eigenvalue of Burden matrix weighted by atomic number', 'module': 'BCUT', 'constructor': 'BCUT("Z", 0)'}, 'BCUTZ-1l': {'function': mordred.BCUT.BCUT('Z', -1), 'type': '2D', 'description': 'first lowest eigenvalue of Burden matrix weighted by atomic number', 'module': 'BCUT', 'constructor': 'BCUT("Z", -1)'}, 'BCUTm-1h': {'function': mordred.BCUT.BCUT('m', 0), 'type': '2D', 'description': 'first heighest eigenvalue of Burden matrix weighted by mass', 'module': 'BCUT', 'constructor': 'BCUT("m", 0)'}, 'BCUTm-1l': {'function': mordred.BCUT.BCUT('m', -1), 'type': '2D', 'description': 'first lowest eigenvalue of Burden matrix weighted by mass', 'module': 'BCUT', 'constructor': 'BCUT("m", -1)'}, 'BCUTv-1h': {'function': mordred.BCUT.BCUT('v', 0), 'type': '2D', 'description': 'first heighest eigenvalue of Burden matrix weighted by vdw volume', 'module': 'BCUT', 'constructor': 'BCUT("v", 0)'}, 'BCUTv-1l': {'function': mordred.BCUT.BCUT('v', -1), 'type': '2D', 'description': 'first lowest eigenvalue of Burden matrix weighted by vdw volume', 'module': 'BCUT', 'constructor': 'BCUT("v", -1)'}, 'BCUTse-1h': {'function': mordred.BCUT.BCUT('se', 0), 'type': '2D', 'description': 'first heighest eigenvalue of Burden matrix weighted by sanderson EN', 'module': 'BCUT', 'constructor': 'BCUT("se", 0)'}, 'BCUTse-1l': {'function': mordred.BCUT.BCUT('se', -1), 'type': '2D', 'description': 'first lowest eigenvalue of Burden matrix weighted by sanderson EN', 'module': 'BCUT', 'constructor': 'BCUT("se", -1)'}, 'BCUTpe-1h': {'function': mordred.BCUT.BCUT('pe', 0), 'type': '2D', 'description': 'first heighest eigenvalue of Burden matrix weighted by pauling EN', 'module': 'BCUT', 'constructor': 'BCUT("pe", 0)'}, 'BCUTpe-1l': {'function': mordred.BCUT.BCUT('pe', -1), 'type': '2D', 'description': 'first lowest eigenvalue of Burden matrix weighted by pauling EN', 'module': 'BCUT', 'constructor': 'BCUT("pe", -1)'}, 'BCUTare-1h': {'function': mordred.BCUT.BCUT('are', 0), 'type': '2D', 'description': 'first heighest eigenvalue of Burden matrix weighted by allred-rocow EN', 'module': 'BCUT', 'constructor': 'BCUT("are", 0)'}, 'BCUTare-1l': {'function': mordred.BCUT.BCUT('are', -1), 'type': '2D', 'description': 'first lowest eigenvalue of Burden matrix weighted by allred-rocow EN', 'module': 'BCUT', 'constructor': 'BCUT("are", -1)'}, 'BCUTp-1h': {'function': mordred.BCUT.BCUT('p', 0), 'type': '2D', 'description': 'first heighest eigenvalue of Burden matrix weighted by polarizability', 'module': 'BCUT', 'constructor': 'BCUT("p", 0)'}, 'BCUTp-1l': {'function': mordred.BCUT.BCUT('p', -1), 'type': '2D', 'description': 'first lowest eigenvalue of Burden matrix weighted by polarizability', 'module': 'BCUT', 'constructor': 'BCUT("p", -1)'}, 'BCUTi-1h': {'function': mordred.BCUT.BCUT('i', 0), 'type': '2D', 'description': 'first heighest eigenvalue of Burden matrix weighted by ionization potential', 'module': 'BCUT', 'constructor': 'BCUT("i", 0)'}, 'BCUTi-1l': {'function': mordred.BCUT.BCUT('i', -1), 'type': '2D', 'description': 'first lowest eigenvalue of Burden matrix weighted by ionization potential', 'module': 'BCUT', 'constructor': 'BCUT("i", -1)'}, 'BalabanJ': {'function': mordred.BalabanJ.BalabanJ(), 'type': '2D', 'description': 'Balaban"s J index', 'module': 'BalabanJ', 'constructor': 'BalabanJ()'}, 'SpAbs_DzZ': {'function': mordred.BaryszMatrix.BaryszMatrix('Z', 'SpAbs'), 'type': '2D', 'description': 'graph energy from Barysz matrix weighted by atomic number', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("Z", "SpAbs")'}, 'SpMax_DzZ': {'function': mordred.BaryszMatrix.BaryszMatrix('Z', 'SpMax'), 'type': '2D', 'description': 'leading eigenvalue from Barysz matrix weighted by atomic number', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("Z", "SpMax")'}, 'SpDiam_DzZ': {'function': mordred.BaryszMatrix.BaryszMatrix('Z', 'SpDiam'), 'type': '2D', 'description': 'spectral diamiter from Barysz matrix weighted by atomic number', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("Z", "SpDiam")'}, 'SpAD_DzZ': {'function': mordred.BaryszMatrix.BaryszMatrix('Z', 'SpAD'), 'type': '2D', 'description': 'spectral absolute diviation from Barysz matrix weighted by atomic number', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("Z", "SpAD")'}, 'SpMAD_DzZ': {'function': mordred.BaryszMatrix.BaryszMatrix('Z', 'SpMAD'), 'type': '2D', 'description': 'spectral mean absolute diviation from Barysz matrix weighted by atomic number', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("Z", "SpMAD")'}, 'LogEE_DzZ': {'function': mordred.BaryszMatrix.BaryszMatrix('Z', 'LogEE'), 'type': '2D', 'description': 'Estrada-like index from Barysz matrix weighted by atomic number', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("Z", "LogEE")'}, 'SM1_DzZ': {'function': mordred.BaryszMatrix.BaryszMatrix('Z', 'SM1'), 'type': '2D', 'description': 'spectral moment from Barysz matrix weighted by atomic number', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("Z", "SM1")'}, 'VE1_DzZ': {'function': mordred.BaryszMatrix.BaryszMatrix('Z', 'VE1'), 'type': '2D', 'description': 'coefficient sum of the last eigenvector from Barysz matrix weighted by atomic number', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("Z", "VE1")'}, 'VE2_DzZ': {'function': mordred.BaryszMatrix.BaryszMatrix('Z', 'VE2'), 'type': '2D', 'description': 'average coefficient of the last eigenvector from Barysz matrix weighted by atomic number', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("Z", "VE2")'}, 'VE3_DzZ': {'function': mordred.BaryszMatrix.BaryszMatrix('Z', 'VE3'), 'type': '2D', 'description': 'logarithmic coefficient sum of the last eigenvector from Barysz matrix weighted by atomic number', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("Z", "VE3")'}, 'VR1_DzZ': {'function': mordred.BaryszMatrix.BaryszMatrix('Z', 'VR1'), 'type': '2D', 'description': 'Randic-like eigenvector-based index from Barysz matrix weighted by atomic number', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("Z", "VR1")'}, 'VR2_DzZ': {'function': mordred.BaryszMatrix.BaryszMatrix('Z', 'VR2'), 'type': '2D', 'description': 'normalized Randic-like eigenvector-based index from Barysz matrix weighted by atomic number', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("Z", "VR2")'}, 'VR3_DzZ': {'function': mordred.BaryszMatrix.BaryszMatrix('Z', 'VR3'), 'type': '2D', 'description': 'logarithmic Randic-like eigenvector-based index from Barysz matrix weighted by atomic number', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("Z", "VR3")'}, 'SpAbs_Dzm': {'function': mordred.BaryszMatrix.BaryszMatrix('m', 'SpAbs'), 'type': '2D', 'description': 'graph energy from Barysz matrix weighted by mass', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("m", "SpAbs")'}, 'SpMax_Dzm': {'function': mordred.BaryszMatrix.BaryszMatrix('m', 'SpMax'), 'type': '2D', 'description': 'leading eigenvalue from Barysz matrix weighted by mass', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("m", "SpMax")'}, 'SpDiam_Dzm': {'function': mordred.BaryszMatrix.BaryszMatrix('m', 'SpDiam'), 'type': '2D', 'description': 'spectral diamiter from Barysz matrix weighted by mass', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("m", "SpDiam")'}, 'SpAD_Dzm': {'function': mordred.BaryszMatrix.BaryszMatrix('m', 'SpAD'), 'type': '2D', 'description': 'spectral absolute diviation from Barysz matrix weighted by mass', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("m", "SpAD")'}, 'SpMAD_Dzm': {'function': mordred.BaryszMatrix.BaryszMatrix('m', 'SpMAD'), 'type': '2D', 'description': 'spectral mean absolute diviation from Barysz matrix weighted by mass', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("m", "SpMAD")'}, 'LogEE_Dzm': {'function': mordred.BaryszMatrix.BaryszMatrix('m', 'LogEE'), 'type': '2D', 'description': 'Estrada-like index from Barysz matrix weighted by mass', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("m", "LogEE")'}, 'SM1_Dzm': {'function': mordred.BaryszMatrix.BaryszMatrix('m', 'SM1'), 'type': '2D', 'description': 'spectral moment from Barysz matrix weighted by mass', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("m", "SM1")'}, 'VE1_Dzm': {'function': mordred.BaryszMatrix.BaryszMatrix('m', 'VE1'), 'type': '2D', 'description': 'coefficient sum of the last eigenvector from Barysz matrix weighted by mass', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("m", "VE1")'}, 'VE2_Dzm': {'function': mordred.BaryszMatrix.BaryszMatrix('m', 'VE2'), 'type': '2D', 'description': 'average coefficient of the last eigenvector from Barysz matrix weighted by mass', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("m", "VE2")'}, 'VE3_Dzm': {'function': mordred.BaryszMatrix.BaryszMatrix('m', 'VE3'), 'type': '2D', 'description': 'logarithmic coefficient sum of the last eigenvector from Barysz matrix weighted by mass', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("m", "VE3")'}, 'VR1_Dzm': {'function': mordred.BaryszMatrix.BaryszMatrix('m', 'VR1'), 'type': '2D', 'description': 'Randic-like eigenvector-based index from Barysz matrix weighted by mass', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("m", "VR1")'}, 'VR2_Dzm': {'function': mordred.BaryszMatrix.BaryszMatrix('m', 'VR2'), 'type': '2D', 'description': 'normalized Randic-like eigenvector-based index from Barysz matrix weighted by mass', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("m", "VR2")'}, 'VR3_Dzm': {'function': mordred.BaryszMatrix.BaryszMatrix('m', 'VR3'), 'type': '2D', 'description': 'logarithmic Randic-like eigenvector-based index from Barysz matrix weighted by mass', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("m", "VR3")'}, 'SpAbs_Dzv': {'function': mordred.BaryszMatrix.BaryszMatrix('v', 'SpAbs'), 'type': '2D', 'description': 'graph energy from Barysz matrix weighted by vdw volume', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("v", "SpAbs")'}, 'SpMax_Dzv': {'function': mordred.BaryszMatrix.BaryszMatrix('v', 'SpMax'), 'type': '2D', 'description': 'leading eigenvalue from Barysz matrix weighted by vdw volume', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("v", "SpMax")'}, 'SpDiam_Dzv': {'function': mordred.BaryszMatrix.BaryszMatrix('v', 'SpDiam'), 'type': '2D', 'description': 'spectral diamiter from Barysz matrix weighted by vdw volume', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("v", "SpDiam")'}, 'SpAD_Dzv': {'function': mordred.BaryszMatrix.BaryszMatrix('v', 'SpAD'), 'type': '2D', 'description': 'spectral absolute diviation from Barysz matrix weighted by vdw volume', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("v", "SpAD")'}, 'SpMAD_Dzv': {'function': mordred.BaryszMatrix.BaryszMatrix('v', 'SpMAD'), 'type': '2D', 'description': 'spectral mean absolute diviation from Barysz matrix weighted by vdw volume', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("v", "SpMAD")'}, 'LogEE_Dzv': {'function': mordred.BaryszMatrix.BaryszMatrix('v', 'LogEE'), 'type': '2D', 'description': 'Estrada-like index from Barysz matrix weighted by vdw volume', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("v", "LogEE")'}, 'SM1_Dzv': {'function': mordred.BaryszMatrix.BaryszMatrix('v', 'SM1'), 'type': '2D', 'description': 'spectral moment from Barysz matrix weighted by vdw volume', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("v", "SM1")'}, 'VE1_Dzv': {'function': mordred.BaryszMatrix.BaryszMatrix('v', 'VE1'), 'type': '2D', 'description': 'coefficient sum of the last eigenvector from Barysz matrix weighted by vdw volume', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("v", "VE1")'}, 'VE2_Dzv': {'function': mordred.BaryszMatrix.BaryszMatrix('v', 'VE2'), 'type': '2D', 'description': 'average coefficient of the last eigenvector from Barysz matrix weighted by vdw volume', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("v", "VE2")'}, 'VE3_Dzv': {'function': mordred.BaryszMatrix.BaryszMatrix('v', 'VE3'), 'type': '2D', 'description': 'logarithmic coefficient sum of the last eigenvector from Barysz matrix weighted by vdw volume', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("v", "VE3")'}, 'VR1_Dzv': {'function': mordred.BaryszMatrix.BaryszMatrix('v', 'VR1'), 'type': '2D', 'description': 'Randic-like eigenvector-based index from Barysz matrix weighted by vdw volume', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("v", "VR1")'}, 'VR2_Dzv': {'function': mordred.BaryszMatrix.BaryszMatrix('v', 'VR2'), 'type': '2D', 'description': 'normalized Randic-like eigenvector-based index from Barysz matrix weighted by vdw volume', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("v", "VR2")'}, 'VR3_Dzv': {'function': mordred.BaryszMatrix.BaryszMatrix('v', 'VR3'), 'type': '2D', 'description': 'logarithmic Randic-like eigenvector-based index from Barysz matrix weighted by vdw volume', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("v", "VR3")'}, 'SpAbs_Dzse': {'function': mordred.BaryszMatrix.BaryszMatrix('se', 'SpAbs'), 'type': '2D', 'description': 'graph energy from Barysz matrix weighted by sanderson EN', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("se", "SpAbs")'}, 'SpMax_Dzse': {'function': mordred.BaryszMatrix.BaryszMatrix('se', 'SpMax'), 'type': '2D', 'description': 'leading eigenvalue from Barysz matrix weighted by sanderson EN', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("se", "SpMax")'}, 'SpDiam_Dzse': {'function': mordred.BaryszMatrix.BaryszMatrix('se', 'SpDiam'), 'type': '2D', 'description': 'spectral diamiter from Barysz matrix weighted by sanderson EN', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("se", "SpDiam")'}, 'SpAD_Dzse': {'function': mordred.BaryszMatrix.BaryszMatrix('se', 'SpAD'), 'type': '2D', 'description': 'spectral absolute diviation from Barysz matrix weighted by sanderson EN', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("se", "SpAD")'}, 'SpMAD_Dzse': {'function': mordred.BaryszMatrix.BaryszMatrix('se', 'SpMAD'), 'type': '2D', 'description': 'spectral mean absolute diviation from Barysz matrix weighted by sanderson EN', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("se", "SpMAD")'}, 'LogEE_Dzse': {'function': mordred.BaryszMatrix.BaryszMatrix('se', 'LogEE'), 'type': '2D', 'description': 'Estrada-like index from Barysz matrix weighted by sanderson EN', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("se", "LogEE")'}, 'SM1_Dzse': {'function': mordred.BaryszMatrix.BaryszMatrix('se', 'SM1'), 'type': '2D', 'description': 'spectral moment from Barysz matrix weighted by sanderson EN', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("se", "SM1")'}, 'VE1_Dzse': {'function': mordred.BaryszMatrix.BaryszMatrix('se', 'VE1'), 'type': '2D', 'description': 'coefficient sum of the last eigenvector from Barysz matrix weighted by sanderson EN', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("se", "VE1")'}, 'VE2_Dzse': {'function': mordred.BaryszMatrix.BaryszMatrix('se', 'VE2'), 'type': '2D', 'description': 'average coefficient of the last eigenvector from Barysz matrix weighted by sanderson EN', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("se", "VE2")'}, 'VE3_Dzse': {'function': mordred.BaryszMatrix.BaryszMatrix('se', 'VE3'), 'type': '2D', 'description': 'logarithmic coefficient sum of the last eigenvector from Barysz matrix weighted by sanderson EN', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("se", "VE3")'}, 'VR1_Dzse': {'function': mordred.BaryszMatrix.BaryszMatrix('se', 'VR1'), 'type': '2D', 'description': 'Randic-like eigenvector-based index from Barysz matrix weighted by sanderson EN', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("se", "VR1")'}, 'VR2_Dzse': {'function': mordred.BaryszMatrix.BaryszMatrix('se', 'VR2'), 'type': '2D', 'description': 'normalized Randic-like eigenvector-based index from Barysz matrix weighted by sanderson EN', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("se", "VR2")'}, 'VR3_Dzse': {'function': mordred.BaryszMatrix.BaryszMatrix('se', 'VR3'), 'type': '2D', 'description': 'logarithmic Randic-like eigenvector-based index from Barysz matrix weighted by sanderson EN', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("se", "VR3")'}, 'SpAbs_Dzpe': {'function': mordred.BaryszMatrix.BaryszMatrix('pe', 'SpAbs'), 'type': '2D', 'description': 'graph energy from Barysz matrix weighted by pauling EN', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("pe", "SpAbs")'}, 'SpMax_Dzpe': {'function': mordred.BaryszMatrix.BaryszMatrix('pe', 'SpMax'), 'type': '2D', 'description': 'leading eigenvalue from Barysz matrix weighted by pauling EN', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("pe", "SpMax")'}, 'SpDiam_Dzpe': {'function': mordred.BaryszMatrix.BaryszMatrix('pe', 'SpDiam'), 'type': '2D', 'description': 'spectral diamiter from Barysz matrix weighted by pauling EN', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("pe", "SpDiam")'}, 'SpAD_Dzpe': {'function': mordred.BaryszMatrix.BaryszMatrix('pe', 'SpAD'), 'type': '2D', 'description': 'spectral absolute diviation from Barysz matrix weighted by pauling EN', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("pe", "SpAD")'}, 'SpMAD_Dzpe': {'function': mordred.BaryszMatrix.BaryszMatrix('pe', 'SpMAD'), 'type': '2D', 'description': 'spectral mean absolute diviation from Barysz matrix weighted by pauling EN', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("pe", "SpMAD")'}, 'LogEE_Dzpe': {'function': mordred.BaryszMatrix.BaryszMatrix('pe', 'LogEE'), 'type': '2D', 'description': 'Estrada-like index from Barysz matrix weighted by pauling EN', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("pe", "LogEE")'}, 'SM1_Dzpe': {'function': mordred.BaryszMatrix.BaryszMatrix('pe', 'SM1'), 'type': '2D', 'description': 'spectral moment from Barysz matrix weighted by pauling EN', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("pe", "SM1")'}, 'VE1_Dzpe': {'function': mordred.BaryszMatrix.BaryszMatrix('pe', 'VE1'), 'type': '2D', 'description': 'coefficient sum of the last eigenvector from Barysz matrix weighted by pauling EN', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("pe", "VE1")'}, 'VE2_Dzpe': {'function': mordred.BaryszMatrix.BaryszMatrix('pe', 'VE2'), 'type': '2D', 'description': 'average coefficient of the last eigenvector from Barysz matrix weighted by pauling EN', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("pe", "VE2")'}, 'VE3_Dzpe': {'function': mordred.BaryszMatrix.BaryszMatrix('pe', 'VE3'), 'type': '2D', 'description': 'logarithmic coefficient sum of the last eigenvector from Barysz matrix weighted by pauling EN', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("pe", "VE3")'}, 'VR1_Dzpe': {'function': mordred.BaryszMatrix.BaryszMatrix('pe', 'VR1'), 'type': '2D', 'description': 'Randic-like eigenvector-based index from Barysz matrix weighted by pauling EN', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("pe", "VR1")'}, 'VR2_Dzpe': {'function': mordred.BaryszMatrix.BaryszMatrix('pe', 'VR2'), 'type': '2D', 'description': 'normalized Randic-like eigenvector-based index from Barysz matrix weighted by pauling EN', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("pe", "VR2")'}, 'VR3_Dzpe': {'function': mordred.BaryszMatrix.BaryszMatrix('pe', 'VR3'), 'type': '2D', 'description': 'logarithmic Randic-like eigenvector-based index from Barysz matrix weighted by pauling EN', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("pe", "VR3")'}, 'SpAbs_Dzare': {'function': mordred.BaryszMatrix.BaryszMatrix('are', 'SpAbs'), 'type': '2D', 'description': 'graph energy from Barysz matrix weighted by allred-rocow EN', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("are", "SpAbs")'}, 'SpMax_Dzare': {'function': mordred.BaryszMatrix.BaryszMatrix('are', 'SpMax'), 'type': '2D', 'description': 'leading eigenvalue from Barysz matrix weighted by allred-rocow EN', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("are", "SpMax")'}, 'SpDiam_Dzare': {'function': mordred.BaryszMatrix.BaryszMatrix('are', 'SpDiam'), 'type': '2D', 'description': 'spectral diamiter from Barysz matrix weighted by allred-rocow EN', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("are", "SpDiam")'}, 'SpAD_Dzare': {'function': mordred.BaryszMatrix.BaryszMatrix('are', 'SpAD'), 'type': '2D', 'description': 'spectral absolute diviation from Barysz matrix weighted by allred-rocow EN', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("are", "SpAD")'}, 'SpMAD_Dzare': {'function': mordred.BaryszMatrix.BaryszMatrix('are', 'SpMAD'), 'type': '2D', 'description': 'spectral mean absolute diviation from Barysz matrix weighted by allred-rocow EN', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("are", "SpMAD")'}, 'LogEE_Dzare': {'function': mordred.BaryszMatrix.BaryszMatrix('are', 'LogEE'), 'type': '2D', 'description': 'Estrada-like index from Barysz matrix weighted by allred-rocow EN', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("are", "LogEE")'}, 'SM1_Dzare': {'function': mordred.BaryszMatrix.BaryszMatrix('are', 'SM1'), 'type': '2D', 'description': 'spectral moment from Barysz matrix weighted by allred-rocow EN', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("are", "SM1")'}, 'VE1_Dzare': {'function': mordred.BaryszMatrix.BaryszMatrix('are', 'VE1'), 'type': '2D', 'description': 'coefficient sum of the last eigenvector from Barysz matrix weighted by allred-rocow EN', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("are", "VE1")'}, 'VE2_Dzare': {'function': mordred.BaryszMatrix.BaryszMatrix('are', 'VE2'), 'type': '2D', 'description': 'average coefficient of the last eigenvector from Barysz matrix weighted by allred-rocow EN', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("are", "VE2")'}, 'VE3_Dzare': {'function': mordred.BaryszMatrix.BaryszMatrix('are', 'VE3'), 'type': '2D', 'description': 'logarithmic coefficient sum of the last eigenvector from Barysz matrix weighted by allred-rocow EN', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("are", "VE3")'}, 'VR1_Dzare': {'function': mordred.BaryszMatrix.BaryszMatrix('are', 'VR1'), 'type': '2D', 'description': 'Randic-like eigenvector-based index from Barysz matrix weighted by allred-rocow EN', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("are", "VR1")'}, 'VR2_Dzare': {'function': mordred.BaryszMatrix.BaryszMatrix('are', 'VR2'), 'type': '2D', 'description': 'normalized Randic-like eigenvector-based index from Barysz matrix weighted by allred-rocow EN', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("are", "VR2")'}, 'VR3_Dzare': {'function': mordred.BaryszMatrix.BaryszMatrix('are', 'VR3'), 'type': '2D', 'description': 'logarithmic Randic-like eigenvector-based index from Barysz matrix weighted by allred-rocow EN', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("are", "VR3")'}, 'SpAbs_Dzp': {'function': mordred.BaryszMatrix.BaryszMatrix('p', 'SpAbs'), 'type': '2D', 'description': 'graph energy from Barysz matrix weighted by polarizability', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("p", "SpAbs")'}, 'SpMax_Dzp': {'function': mordred.BaryszMatrix.BaryszMatrix('p', 'SpMax'), 'type': '2D', 'description': 'leading eigenvalue from Barysz matrix weighted by polarizability', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("p", "SpMax")'}, 'SpDiam_Dzp': {'function': mordred.BaryszMatrix.BaryszMatrix('p', 'SpDiam'), 'type': '2D', 'description': 'spectral diamiter from Barysz matrix weighted by polarizability', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("p", "SpDiam")'}, 'SpAD_Dzp': {'function': mordred.BaryszMatrix.BaryszMatrix('p', 'SpAD'), 'type': '2D', 'description': 'spectral absolute diviation from Barysz matrix weighted by polarizability', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("p", "SpAD")'}, 'SpMAD_Dzp': {'function': mordred.BaryszMatrix.BaryszMatrix('p', 'SpMAD'), 'type': '2D', 'description': 'spectral mean absolute diviation from Barysz matrix weighted by polarizability', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("p", "SpMAD")'}, 'LogEE_Dzp': {'function': mordred.BaryszMatrix.BaryszMatrix('p', 'LogEE'), 'type': '2D', 'description': 'Estrada-like index from Barysz matrix weighted by polarizability', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("p", "LogEE")'}, 'SM1_Dzp': {'function': mordred.BaryszMatrix.BaryszMatrix('p', 'SM1'), 'type': '2D', 'description': 'spectral moment from Barysz matrix weighted by polarizability', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("p", "SM1")'}, 'VE1_Dzp': {'function': mordred.BaryszMatrix.BaryszMatrix('p', 'VE1'), 'type': '2D', 'description': 'coefficient sum of the last eigenvector from Barysz matrix weighted by polarizability', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("p", "VE1")'}, 'VE2_Dzp': {'function': mordred.BaryszMatrix.BaryszMatrix('p', 'VE2'), 'type': '2D', 'description': 'average coefficient of the last eigenvector from Barysz matrix weighted by polarizability', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("p", "VE2")'}, 'VE3_Dzp': {'function': mordred.BaryszMatrix.BaryszMatrix('p', 'VE3'), 'type': '2D', 'description': 'logarithmic coefficient sum of the last eigenvector from Barysz matrix weighted by polarizability', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("p", "VE3")'}, 'VR1_Dzp': {'function': mordred.BaryszMatrix.BaryszMatrix('p', 'VR1'), 'type': '2D', 'description': 'Randic-like eigenvector-based index from Barysz matrix weighted by polarizability', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("p", "VR1")'}, 'VR2_Dzp': {'function': mordred.BaryszMatrix.BaryszMatrix('p', 'VR2'), 'type': '2D', 'description': 'normalized Randic-like eigenvector-based index from Barysz matrix weighted by polarizability', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("p", "VR2")'}, 'VR3_Dzp': {'function': mordred.BaryszMatrix.BaryszMatrix('p', 'VR3'), 'type': '2D', 'description': 'logarithmic Randic-like eigenvector-based index from Barysz matrix weighted by polarizability', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("p", "VR3")'}, 'SpAbs_Dzi': {'function': mordred.BaryszMatrix.BaryszMatrix('i', 'SpAbs'), 'type': '2D', 'description': 'graph energy from Barysz matrix weighted by ionization potential', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("i", "SpAbs")'}, 'SpMax_Dzi': {'function': mordred.BaryszMatrix.BaryszMatrix('i', 'SpMax'), 'type': '2D', 'description': 'leading eigenvalue from Barysz matrix weighted by ionization potential', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("i", "SpMax")'}, 'SpDiam_Dzi': {'function': mordred.BaryszMatrix.BaryszMatrix('i', 'SpDiam'), 'type': '2D', 'description': 'spectral diamiter from Barysz matrix weighted by ionization potential', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("i", "SpDiam")'}, 'SpAD_Dzi': {'function': mordred.BaryszMatrix.BaryszMatrix('i', 'SpAD'), 'type': '2D', 'description': 'spectral absolute diviation from Barysz matrix weighted by ionization potential', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("i", "SpAD")'}, 'SpMAD_Dzi': {'function': mordred.BaryszMatrix.BaryszMatrix('i', 'SpMAD'), 'type': '2D', 'description': 'spectral mean absolute diviation from Barysz matrix weighted by ionization potential', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("i", "SpMAD")'}, 'LogEE_Dzi': {'function': mordred.BaryszMatrix.BaryszMatrix('i', 'LogEE'), 'type': '2D', 'description': 'Estrada-like index from Barysz matrix weighted by ionization potential', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("i", "LogEE")'}, 'SM1_Dzi': {'function': mordred.BaryszMatrix.BaryszMatrix('i', 'SM1'), 'type': '2D', 'description': 'spectral moment from Barysz matrix weighted by ionization potential', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("i", "SM1")'}, 'VE1_Dzi': {'function': mordred.BaryszMatrix.BaryszMatrix('i', 'VE1'), 'type': '2D', 'description': 'coefficient sum of the last eigenvector from Barysz matrix weighted by ionization potential', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("i", "VE1")'}, 'VE2_Dzi': {'function': mordred.BaryszMatrix.BaryszMatrix('i', 'VE2'), 'type': '2D', 'description': 'average coefficient of the last eigenvector from Barysz matrix weighted by ionization potential', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("i", "VE2")'}, 'VE3_Dzi': {'function': mordred.BaryszMatrix.BaryszMatrix('i', 'VE3'), 'type': '2D', 'description': 'logarithmic coefficient sum of the last eigenvector from Barysz matrix weighted by ionization potential', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("i", "VE3")'}, 'VR1_Dzi': {'function': mordred.BaryszMatrix.BaryszMatrix('i', 'VR1'), 'type': '2D', 'description': 'Randic-like eigenvector-based index from Barysz matrix weighted by ionization potential', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("i", "VR1")'}, 'VR2_Dzi': {'function': mordred.BaryszMatrix.BaryszMatrix('i', 'VR2'), 'type': '2D', 'description': 'normalized Randic-like eigenvector-based index from Barysz matrix weighted by ionization potential', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("i", "VR2")'}, 'VR3_Dzi': {'function': mordred.BaryszMatrix.BaryszMatrix('i', 'VR3'), 'type': '2D', 'description': 'logarithmic Randic-like eigenvector-based index from Barysz matrix weighted by ionization potential', 'module': 'BaryszMatrix', 'constructor': 'BaryszMatrix("i", "VR3")'}, 'BertzCT': {'function': mordred.BertzCT.BertzCT(), 'type': '2D', 'description': 'Bertz CT', 'module': 'BertzCT', 'constructor': 'BertzCT()'}, 'nBonds': {'function': mordred.BondCount.BondCount('any', False), 'type': '2D', 'description': 'number of all bonds in non-kekulized structure', 'module': 'BondCount', 'constructor': 'BondCount("any", False)'}, 'nBondsO': {'function': mordred.BondCount.BondCount('heavy', False), 'type': '2D', 'description': 'number of bonds connecting to heavy atom in non-kekulized structure', 'module': 'BondCount', 'constructor': 'BondCount("heavy", False)'}, 'nBondsS': {'function': mordred.BondCount.BondCount('single', False), 'type': '2D', 'description': 'number of single bonds in non-kekulized structure', 'module': 'BondCount', 'constructor': 'BondCount("single", False)'}, 'nBondsD': {'function': mordred.BondCount.BondCount('double', False), 'type': '2D', 'description': 'number of double bonds in non-kekulized structure', 'module': 'BondCount', 'constructor': 'BondCount("double", False)'}, 'nBondsT': {'function': mordred.BondCount.BondCount('triple', False), 'type': '2D', 'description': 'number of triple bonds in non-kekulized structure', 'module': 'BondCount', 'constructor': 'BondCount("triple", False)'}, 'nBondsA': {'function': mordred.BondCount.BondCount('aromatic', False), 'type': '2D', 'description': 'number of aromatic bonds in non-kekulized structure', 'module': 'BondCount', 'constructor': 'BondCount("aromatic", False)'}, 'nBondsM': {'function': mordred.BondCount.BondCount('multiple', False), 'type': '2D', 'description': 'number of multiple bonds in non-kekulized structure', 'module': 'BondCount', 'constructor': 'BondCount("multiple", False)'}, 'nBondsKS': {'function': mordred.BondCount.BondCount('single', True), 'type': '2D', 'description': 'number of single bonds in kekulized structure', 'module': 'BondCount', 'constructor': 'BondCount("single", True)'}, 'nBondsKD': {'function': mordred.BondCount.BondCount('double', True), 'type': '2D', 'description': 'number of double bonds in kekulized structure', 'module': 'BondCount', 'constructor': 'BondCount("double", True)'}, 'PNSA1': {'function': mordred.CPSA.PNSA(1), 'type': '3D', 'description': 'partial negative surface area(version 1)', 'module': 'CPSA', 'constructor': 'PNSA(1)'}, 'PNSA2': {'function': mordred.CPSA.PNSA(2), 'type': '3D', 'description': 'partial negative surface area(version 2)', 'module': 'CPSA', 'constructor': 'PNSA(2)'}, 'PNSA3': {'function': mordred.CPSA.PNSA(3), 'type': '3D', 'description': 'partial negative surface area(version 3)', 'module': 'CPSA', 'constructor': 'PNSA(3)'}, 'PNSA4': {'function': mordred.CPSA.PNSA(4), 'type': '3D', 'description': 'partial negative surface area(version 4)', 'module': 'CPSA', 'constructor': 'PNSA(4)'}, 'PNSA5': {'function': mordred.CPSA.PNSA(5), 'type': '3D', 'description': 'partial negative surface area(version 5)', 'module': 'CPSA', 'constructor': 'PNSA(5)'}, 'PPSA1': {'function': mordred.CPSA.PPSA(1), 'type': '3D', 'description': 'partial positive surface area(version 1)', 'module': 'CPSA', 'constructor': 'PPSA(1)'}, 'PPSA2': {'function': mordred.CPSA.PPSA(2), 'type': '3D', 'description': 'partial positive surface area(version 2)', 'module': 'CPSA', 'constructor': 'PPSA(2)'}, 'PPSA3': {'function': mordred.CPSA.PPSA(3), 'type': '3D', 'description': 'partial positive surface area(version 3)', 'module': 'CPSA', 'constructor': 'PPSA(3)'}, 'PPSA4': {'function': mordred.CPSA.PPSA(4), 'type': '3D', 'description': 'partial positive surface area(version 4)', 'module': 'CPSA', 'constructor': 'PPSA(4)'}, 'PPSA5': {'function': mordred.CPSA.PPSA(5), 'type': '3D', 'description': 'partial positive surface area(version 5)', 'module': 'CPSA', 'constructor': 'PPSA(5)'}, 'DPSA1': {'function': mordred.CPSA.DPSA(1), 'type': '3D', 'description': 'difference in charged partial surface area(version 1)', 'module': 'CPSA', 'constructor': 'DPSA(1)'}, 'DPSA2': {'function': mordred.CPSA.DPSA(2), 'type': '3D', 'description': 'difference in charged partial surface area(version 2)', 'module': 'CPSA', 'constructor': 'DPSA(2)'}, 'DPSA3': {'function': mordred.CPSA.DPSA(3), 'type': '3D', 'description': 'difference in charged partial surface area(version 3)', 'module': 'CPSA', 'constructor': 'DPSA(3)'}, 'DPSA4': {'function': mordred.CPSA.DPSA(4), 'type': '3D', 'description': 'difference in charged partial surface area(version 4)', 'module': 'CPSA', 'constructor': 'DPSA(4)'}, 'DPSA5': {'function': mordred.CPSA.DPSA(5), 'type': '3D', 'description': 'difference in charged partial surface area(version 5)', 'module': 'CPSA', 'constructor': 'DPSA(5)'}, 'FNSA1': {'function': mordred.CPSA.FNSA(1), 'type': '3D', 'description': 'fractional charged partial negative surface area(version 1)', 'module': 'CPSA', 'constructor': 'FNSA(1)'}, 'FNSA2': {'function': mordred.CPSA.FNSA(2), 'type': '3D', 'description': 'fractional charged partial negative surface area(version 2)', 'module': 'CPSA', 'constructor': 'FNSA(2)'}, 'FNSA3': {'function': mordred.CPSA.FNSA(3), 'type': '3D', 'description': 'fractional charged partial negative surface area(version 3)', 'module': 'CPSA', 'constructor': 'FNSA(3)'}, 'FNSA4': {'function': mordred.CPSA.FNSA(4), 'type': '3D', 'description': 'fractional charged partial negative surface area(version 4)', 'module': 'CPSA', 'constructor': 'FNSA(4)'}, 'FNSA5': {'function': mordred.CPSA.FNSA(5), 'type': '3D', 'description': 'fractional charged partial negative surface area(version 5)', 'module': 'CPSA', 'constructor': 'FNSA(5)'}, 'FPSA1': {'function': mordred.CPSA.FPSA(1), 'type': '3D', 'description': 'fractional charged partial positive surface area(version 1)', 'module': 'CPSA', 'constructor': 'FPSA(1)'}, 'FPSA2': {'function': mordred.CPSA.FPSA(2), 'type': '3D', 'description': 'fractional charged partial positive surface area(version 2)', 'module': 'CPSA', 'constructor': 'FPSA(2)'}, 'FPSA3': {'function': mordred.CPSA.FPSA(3), 'type': '3D', 'description': 'fractional charged partial positive surface area(version 3)', 'module': 'CPSA', 'constructor': 'FPSA(3)'}, 'FPSA4': {'function': mordred.CPSA.FPSA(4), 'type': '3D', 'description': 'fractional charged partial positive surface area(version 4)', 'module': 'CPSA', 'constructor': 'FPSA(4)'}, 'FPSA5': {'function': mordred.CPSA.FPSA(5), 'type': '3D', 'description': 'fractional charged partial positive surface area(version 5)', 'module': 'CPSA', 'constructor': 'FPSA(5)'}, 'WNSA1': {'function': mordred.CPSA.WNSA(1), 'type': '3D', 'description': 'surface weighted charged partial negative surface area(version 1)', 'module': 'CPSA', 'constructor': 'WNSA(1)'}, 'WNSA2': {'function': mordred.CPSA.WNSA(2), 'type': '3D', 'description': 'surface weighted charged partial negative surface area(version 2)', 'module': 'CPSA', 'constructor': 'WNSA(2)'}, 'WNSA3': {'function': mordred.CPSA.WNSA(3), 'type': '3D', 'description': 'surface weighted charged partial negative surface area(version 3)', 'module': 'CPSA', 'constructor': 'WNSA(3)'}, 'WNSA4': {'function': mordred.CPSA.WNSA(4), 'type': '3D', 'description': 'surface weighted charged partial negative surface area(version 4)', 'module': 'CPSA', 'constructor': 'WNSA(4)'}, 'WNSA5': {'function': mordred.CPSA.WNSA(5), 'type': '3D', 'description': 'surface weighted charged partial negative surface area(version 5)', 'module': 'CPSA', 'constructor': 'WNSA(5)'}, 'WPSA1': {'function': mordred.CPSA.WPSA(1), 'type': '3D', 'description': 'surface weighted charged partial positive surface area(version 1)', 'module': 'CPSA', 'constructor': 'WPSA(1)'}, 'WPSA2': {'function': mordred.CPSA.WPSA(2), 'type': '3D', 'description': 'surface weighted charged partial positive surface area(version 2)', 'module': 'CPSA', 'constructor': 'WPSA(2)'}, 'WPSA3': {'function': mordred.CPSA.WPSA(3), 'type': '3D', 'description': 'surface weighted charged partial positive surface area(version 3)', 'module': 'CPSA', 'constructor': 'WPSA(3)'}, 'WPSA4': {'function': mordred.CPSA.WPSA(4), 'type': '3D', 'description': 'surface weighted charged partial positive surface area(version 4)', 'module': 'CPSA', 'constructor': 'WPSA(4)'}, 'WPSA5': {'function': mordred.CPSA.WPSA(5), 'type': '3D', 'description': 'surface weighted charged partial positive surface area(version 5)', 'module': 'CPSA', 'constructor': 'WPSA(5)'}, 'RNCG': {'function': mordred.CPSA.RNCG(), 'type': '2D', 'description': 'relative negative charge', 'module': 'CPSA', 'constructor': 'RNCG()'}, 'RPCG': {'function': mordred.CPSA.RPCG(), 'type': '2D', 'description': 'relative positive charge', 'module': 'CPSA', 'constructor': 'RPCG()'}, 'RNCS': {'function': mordred.CPSA.RNCS(), 'type': '3D', 'description': 'relative negative charge surface area', 'module': 'CPSA', 'constructor': 'RNCS()'}, 'RPCS': {'function': mordred.CPSA.RPCS(), 'type': '3D', 'description': 'relative positive charge surface area', 'module': 'CPSA', 'constructor': 'RPCS()'}, 'TASA': {'function': mordred.CPSA.TASA(), 'type': '3D', 'description': 'total hydrophobic surface area', 'module': 'CPSA', 'constructor': 'TASA()'}, 'TPSA': {'function': mordred.CPSA.TPSA(), 'type': '3D', 'description': 'total polar surface area', 'module': 'CPSA', 'constructor': 'TPSA()'}, 'RASA': {'function': mordred.CPSA.RASA(), 'type': '3D', 'description': 'relative hydrophobic surface area', 'module': 'CPSA', 'constructor': 'RASA()'}, 'RPSA': {'function': mordred.CPSA.RPSA(), 'type': '3D', 'description': 'relative polar surface area', 'module': 'CPSA', 'constructor': 'RPSA()'}, 'C1SP1': {'function': mordred.CarbonTypes.CarbonTypes(1, 1), 'type': '2D', 'description': 'SP carbon bound to 1 other carbon', 'module': 'CarbonTypes', 'constructor': 'CarbonTypes(1, 1)'}, 'C2SP1': {'function': mordred.CarbonTypes.CarbonTypes(2, 1), 'type': '2D', 'description': 'SP carbon bound to 2 other carbons', 'module': 'CarbonTypes', 'constructor': 'CarbonTypes(2, 1)'}, 'C1SP2': {'function': mordred.CarbonTypes.CarbonTypes(1, 2), 'type': '2D', 'description': 'SP2 carbon bound to 1 other carbon', 'module': 'CarbonTypes', 'constructor': 'CarbonTypes(1, 2)'}, 'C2SP2': {'function': mordred.CarbonTypes.CarbonTypes(2, 2), 'type': '2D', 'description': 'SP2 carbon bound to 2 other carbons', 'module': 'CarbonTypes', 'constructor': 'CarbonTypes(2, 2)'}, 'C3SP2': {'function': mordred.CarbonTypes.CarbonTypes(3, 2), 'type': '2D', 'description': 'SP2 carbon bound to 3 other carbons', 'module': 'CarbonTypes', 'constructor': 'CarbonTypes(3, 2)'}, 'C1SP3': {'function': mordred.CarbonTypes.CarbonTypes(1, 3), 'type': '2D', 'description': 'SP3 carbon bound to 1 other carbon', 'module': 'CarbonTypes', 'constructor': 'CarbonTypes(1, 3)'}, 'C2SP3': {'function': mordred.CarbonTypes.CarbonTypes(2, 3), 'type': '2D', 'description': 'SP3 carbon bound to 2 other carbons', 'module': 'CarbonTypes', 'constructor': 'CarbonTypes(2, 3)'}, 'C3SP3': {'function': mordred.CarbonTypes.CarbonTypes(3, 3), 'type': '2D', 'description': 'SP3 carbon bound to 3 other carbons', 'module': 'CarbonTypes', 'constructor': 'CarbonTypes(3, 3)'}, 'C4SP3': {'function': mordred.CarbonTypes.CarbonTypes(4, 3), 'type': '2D', 'description': 'SP3 carbon bound to 4 other carbons', 'module': 'CarbonTypes', 'constructor': 'CarbonTypes(4, 3)'}, 'HybRatio': {'function': mordred.CarbonTypes.HybridizationRatio(), 'type': '2D', 'description': 'hybridization ratio', 'module': 'CarbonTypes', 'constructor': 'HybridizationRatio()'}, 'FCSP3': {'function': mordred.CarbonTypes.FractionCSP3(), 'type': '2D', 'description': 'the fraction of C atoms that are SP3 hybridized', 'module': 'CarbonTypes', 'constructor': 'FractionCSP3()'}, 'Xch-3d': {'function': mordred.Chi.Chi('chain', 3, 'd', False), 'type': '2D', 'description': '3-ordered Chi chain weighted by sigma electrons', 'module': 'Chi', 'constructor': 'Chi("chain", 3, "d", False)'}, 'Xch-4d': {'function': mordred.Chi.Chi('chain', 4, 'd', False), 'type': '2D', 'description': '4-ordered Chi chain weighted by sigma electrons', 'module': 'Chi', 'constructor': 'Chi("chain", 4, "d", False)'}, 'Xch-5d': {'function': mordred.Chi.Chi('chain', 5, 'd', False), 'type': '2D', 'description': '5-ordered Chi chain weighted by sigma electrons', 'module': 'Chi', 'constructor': 'Chi("chain", 5, "d", False)'}, 'Xch-6d': {'function': mordred.Chi.Chi('chain', 6, 'd', False), 'type': '2D', 'description': '6-ordered Chi chain weighted by sigma electrons', 'module': 'Chi', 'constructor': 'Chi("chain", 6, "d", False)'}, 'Xch-7d': {'function': mordred.Chi.Chi('chain', 7, 'd', False), 'type': '2D', 'description': '7-ordered Chi chain weighted by sigma electrons', 'module': 'Chi', 'constructor': 'Chi("chain", 7, "d", False)'}, 'Xch-3dv': {'function': mordred.Chi.Chi('chain', 3, 'dv', False), 'type': '2D', 'description': '3-ordered Chi chain weighted by valence electrons', 'module': 'Chi', 'constructor': 'Chi("chain", 3, "dv", False)'}, 'Xch-4dv': {'function': mordred.Chi.Chi('chain', 4, 'dv', False), 'type': '2D', 'description': '4-ordered Chi chain weighted by valence electrons', 'module': 'Chi', 'constructor': 'Chi("chain", 4, "dv", False)'}, 'Xch-5dv': {'function': mordred.Chi.Chi('chain', 5, 'dv', False), 'type': '2D', 'description': '5-ordered Chi chain weighted by valence electrons', 'module': 'Chi', 'constructor': 'Chi("chain", 5, "dv", False)'}, 'Xch-6dv': {'function': mordred.Chi.Chi('chain', 6, 'dv', False), 'type': '2D', 'description': '6-ordered Chi chain weighted by valence electrons', 'module': 'Chi', 'constructor': 'Chi("chain", 6, "dv", False)'}, 'Xch-7dv': {'function': mordred.Chi.Chi('chain', 7, 'dv', False), 'type': '2D', 'description': '7-ordered Chi chain weighted by valence electrons', 'module': 'Chi', 'constructor': 'Chi("chain", 7, "dv", False)'}, 'Xc-3d': {'function': mordred.Chi.Chi('cluster', 3, 'd', False), 'type': '2D', 'description': '3-ordered Chi cluster weighted by sigma electrons', 'module': 'Chi', 'constructor': 'Chi("cluster", 3, "d", False)'}, 'Xc-4d': {'function': mordred.Chi.Chi('cluster', 4, 'd', False), 'type': '2D', 'description': '4-ordered Chi cluster weighted by sigma electrons', 'module': 'Chi', 'constructor': 'Chi("cluster", 4, "d", False)'}, 'Xc-5d': {'function': mordred.Chi.Chi('cluster', 5, 'd', False), 'type': '2D', 'description': '5-ordered Chi cluster weighted by sigma electrons', 'module': 'Chi', 'constructor': 'Chi("cluster", 5, "d", False)'}, 'Xc-6d': {'function': mordred.Chi.Chi('cluster', 6, 'd', False), 'type': '2D', 'description': '6-ordered Chi cluster weighted by sigma electrons', 'module': 'Chi', 'constructor': 'Chi("cluster", 6, "d", False)'}, 'Xc-3dv': {'function': mordred.Chi.Chi('cluster', 3, 'dv', False), 'type': '2D', 'description': '3-ordered Chi cluster weighted by valence electrons', 'module': 'Chi', 'constructor': 'Chi("cluster", 3, "dv", False)'}, 'Xc-4dv': {'function': mordred.Chi.Chi('cluster', 4, 'dv', False), 'type': '2D', 'description': '4-ordered Chi cluster weighted by valence electrons', 'module': 'Chi', 'constructor': 'Chi("cluster", 4, "dv", False)'}, 'Xc-5dv': {'function': mordred.Chi.Chi('cluster', 5, 'dv', False), 'type': '2D', 'description': '5-ordered Chi cluster weighted by valence electrons', 'module': 'Chi', 'constructor': 'Chi("cluster", 5, "dv", False)'}, 'Xc-6dv': {'function': mordred.Chi.Chi('cluster', 6, 'dv', False), 'type': '2D', 'description': '6-ordered Chi cluster weighted by valence electrons', 'module': 'Chi', 'constructor': 'Chi("cluster", 6, "dv", False)'}, 'Xpc-4d': {'function': mordred.Chi.Chi('path_cluster', 4, 'd', False), 'type': '2D', 'description': '4-ordered Chi path-cluster weighted by sigma electrons', 'module': 'Chi', 'constructor': 'Chi("path_cluster", 4, "d", False)'}, 'Xpc-5d': {'function': mordred.Chi.Chi('path_cluster', 5, 'd', False), 'type': '2D', 'description': '5-ordered Chi path-cluster weighted by sigma electrons', 'module': 'Chi', 'constructor': 'Chi("path_cluster", 5, "d", False)'}, 'Xpc-6d': {'function': mordred.Chi.Chi('path_cluster', 6, 'd', False), 'type': '2D', 'description': '6-ordered Chi path-cluster weighted by sigma electrons', 'module': 'Chi', 'constructor': 'Chi("path_cluster", 6, "d", False)'}, 'Xpc-4dv': {'function': mordred.Chi.Chi('path_cluster', 4, 'dv', False), 'type': '2D', 'description': '4-ordered Chi path-cluster weighted by valence electrons', 'module': 'Chi', 'constructor': 'Chi("path_cluster", 4, "dv", False)'}, 'Xpc-5dv': {'function': mordred.Chi.Chi('path_cluster', 5, 'dv', False), 'type': '2D', 'description': '5-ordered Chi path-cluster weighted by valence electrons', 'module': 'Chi', 'constructor': 'Chi("path_cluster", 5, "dv", False)'}, 'Xpc-6dv': {'function': mordred.Chi.Chi('path_cluster', 6, 'dv', False), 'type': '2D', 'description': '6-ordered Chi path-cluster weighted by valence electrons', 'module': 'Chi', 'constructor': 'Chi("path_cluster", 6, "dv", False)'}, 'Xp-0d': {'function': mordred.Chi.Chi('path', 0, 'd', False), 'type': '2D', 'description': '0-ordered Chi path weighted by sigma electrons', 'module': 'Chi', 'constructor': 'Chi("path", 0, "d", False)'}, 'Xp-1d': {'function': mordred.Chi.Chi('path', 1, 'd', False), 'type': '2D', 'description': '1-ordered Chi path weighted by sigma electrons', 'module': 'Chi', 'constructor': 'Chi("path", 1, "d", False)'}, 'Xp-2d': {'function': mordred.Chi.Chi('path', 2, 'd', False), 'type': '2D', 'description': '2-ordered Chi path weighted by sigma electrons', 'module': 'Chi', 'constructor': 'Chi("path", 2, "d", False)'}, 'Xp-3d': {'function': mordred.Chi.Chi('path', 3, 'd', False), 'type': '2D', 'description': '3-ordered Chi path weighted by sigma electrons', 'module': 'Chi', 'constructor': 'Chi("path", 3, "d", False)'}, 'Xp-4d': {'function': mordred.Chi.Chi('path', 4, 'd', False), 'type': '2D', 'description': '4-ordered Chi path weighted by sigma electrons', 'module': 'Chi', 'constructor': 'Chi("path", 4, "d", False)'}, 'Xp-5d': {'function': mordred.Chi.Chi('path', 5, 'd', False), 'type': '2D', 'description': '5-ordered Chi path weighted by sigma electrons', 'module': 'Chi', 'constructor': 'Chi("path", 5, "d", False)'}, 'Xp-6d': {'function': mordred.Chi.Chi('path', 6, 'd', False), 'type': '2D', 'description': '6-ordered Chi path weighted by sigma electrons', 'module': 'Chi', 'constructor': 'Chi("path", 6, "d", False)'}, 'Xp-7d': {'function': mordred.Chi.Chi('path', 7, 'd', False), 'type': '2D', 'description': '7-ordered Chi path weighted by sigma electrons', 'module': 'Chi', 'constructor': 'Chi("path", 7, "d", False)'}, 'AXp-0d': {'function': mordred.Chi.Chi('path', 0, 'd', True), 'type': '2D', 'description': '0-ordered averaged Chi path weighted by sigma electrons', 'module': 'Chi', 'constructor': 'Chi("path", 0, "d", True)'}, 'AXp-1d': {'function': mordred.Chi.Chi('path', 1, 'd', True), 'type': '2D', 'description': '1-ordered averaged Chi path weighted by sigma electrons', 'module': 'Chi', 'constructor': 'Chi("path", 1, "d", True)'}, 'AXp-2d': {'function': mordred.Chi.Chi('path', 2, 'd', True), 'type': '2D', 'description': '2-ordered averaged Chi path weighted by sigma electrons', 'module': 'Chi', 'constructor': 'Chi("path", 2, "d", True)'}, 'AXp-3d': {'function': mordred.Chi.Chi('path', 3, 'd', True), 'type': '2D', 'description': '3-ordered averaged Chi path weighted by sigma electrons', 'module': 'Chi', 'constructor': 'Chi("path", 3, "d", True)'}, 'AXp-4d': {'function': mordred.Chi.Chi('path', 4, 'd', True), 'type': '2D', 'description': '4-ordered averaged Chi path weighted by sigma electrons', 'module': 'Chi', 'constructor': 'Chi("path", 4, "d", True)'}, 'AXp-5d': {'function': mordred.Chi.Chi('path', 5, 'd', True), 'type': '2D', 'description': '5-ordered averaged Chi path weighted by sigma electrons', 'module': 'Chi', 'constructor': 'Chi("path", 5, "d", True)'}, 'AXp-6d': {'function': mordred.Chi.Chi('path', 6, 'd', True), 'type': '2D', 'description': '6-ordered averaged Chi path weighted by sigma electrons', 'module': 'Chi', 'constructor': 'Chi("path", 6, "d", True)'}, 'AXp-7d': {'function': mordred.Chi.Chi('path', 7, 'd', True), 'type': '2D', 'description': '7-ordered averaged Chi path weighted by sigma electrons', 'module': 'Chi', 'constructor': 'Chi("path", 7, "d", True)'}, 'Xp-0dv': {'function': mordred.Chi.Chi('path', 0, 'dv', False), 'type': '2D', 'description': '0-ordered Chi path weighted by valence electrons', 'module': 'Chi', 'constructor': 'Chi("path", 0, "dv", False)'}, 'Xp-1dv': {'function': mordred.Chi.Chi('path', 1, 'dv', False), 'type': '2D', 'description': '1-ordered Chi path weighted by valence electrons', 'module': 'Chi', 'constructor': 'Chi("path", 1, "dv", False)'}, 'Xp-2dv': {'function': mordred.Chi.Chi('path', 2, 'dv', False), 'type': '2D', 'description': '2-ordered Chi path weighted by valence electrons', 'module': 'Chi', 'constructor': 'Chi("path", 2, "dv", False)'}, 'Xp-3dv': {'function': mordred.Chi.Chi('path', 3, 'dv', False), 'type': '2D', 'description': '3-ordered Chi path weighted by valence electrons', 'module': 'Chi', 'constructor': 'Chi("path", 3, "dv", False)'}, 'Xp-4dv': {'function': mordred.Chi.Chi('path', 4, 'dv', False), 'type': '2D', 'description': '4-ordered Chi path weighted by valence electrons', 'module': 'Chi', 'constructor': 'Chi("path", 4, "dv", False)'}, 'Xp-5dv': {'function': mordred.Chi.Chi('path', 5, 'dv', False), 'type': '2D', 'description': '5-ordered Chi path weighted by valence electrons', 'module': 'Chi', 'constructor': 'Chi("path", 5, "dv", False)'}, 'Xp-6dv': {'function': mordred.Chi.Chi('path', 6, 'dv', False), 'type': '2D', 'description': '6-ordered Chi path weighted by valence electrons', 'module': 'Chi', 'constructor': 'Chi("path", 6, "dv", False)'}, 'Xp-7dv': {'function': mordred.Chi.Chi('path', 7, 'dv', False), 'type': '2D', 'description': '7-ordered Chi path weighted by valence electrons', 'module': 'Chi', 'constructor': 'Chi("path", 7, "dv", False)'}, 'AXp-0dv': {'function': mordred.Chi.Chi('path', 0, 'dv', True), 'type': '2D', 'description': '0-ordered averaged Chi path weighted by valence electrons', 'module': 'Chi', 'constructor': 'Chi("path", 0, "dv", True)'}, 'AXp-1dv': {'function': mordred.Chi.Chi('path', 1, 'dv', True), 'type': '2D', 'description': '1-ordered averaged Chi path weighted by valence electrons', 'module': 'Chi', 'constructor': 'Chi("path", 1, "dv", True)'}, 'AXp-2dv': {'function': mordred.Chi.Chi('path', 2, 'dv', True), 'type': '2D', 'description': '2-ordered averaged Chi path weighted by valence electrons', 'module': 'Chi', 'constructor': 'Chi("path", 2, "dv", True)'}, 'AXp-3dv': {'function': mordred.Chi.Chi('path', 3, 'dv', True), 'type': '2D', 'description': '3-ordered averaged Chi path weighted by valence electrons', 'module': 'Chi', 'constructor': 'Chi("path", 3, "dv", True)'}, 'AXp-4dv': {'function': mordred.Chi.Chi('path', 4, 'dv', True), 'type': '2D', 'description': '4-ordered averaged Chi path weighted by valence electrons', 'module': 'Chi', 'constructor': 'Chi("path", 4, "dv", True)'}, 'AXp-5dv': {'function': mordred.Chi.Chi('path', 5, 'dv', True), 'type': '2D', 'description': '5-ordered averaged Chi path weighted by valence electrons', 'module': 'Chi', 'constructor': 'Chi("path", 5, "dv", True)'}, 'AXp-6dv': {'function': mordred.Chi.Chi('path', 6, 'dv', True), 'type': '2D', 'description': '6-ordered averaged Chi path weighted by valence electrons', 'module': 'Chi', 'constructor': 'Chi("path", 6, "dv", True)'}, 'AXp-7dv': {'function': mordred.Chi.Chi('path', 7, 'dv', True), 'type': '2D', 'description': '7-ordered averaged Chi path weighted by valence electrons', 'module': 'Chi', 'constructor': 'Chi("path", 7, "dv", True)'}, 'SZ': {'function': mordred.Constitutional.ConstitutionalSum('Z'), 'type': '2D', 'description': 'sum of constitutional weighted by atomic number', 'module': 'Constitutional', 'constructor': 'ConstitutionalSum("Z")'}, 'Sm': {'function': mordred.Constitutional.ConstitutionalSum('m'), 'type': '2D', 'description': 'sum of constitutional weighted by mass', 'module': 'Constitutional', 'constructor': 'ConstitutionalSum("m")'}, 'Sv': {'function': mordred.Constitutional.ConstitutionalSum('v'), 'type': '2D', 'description': 'sum of constitutional weighted by vdw volume', 'module': 'Constitutional', 'constructor': 'ConstitutionalSum("v")'}, 'Sse': {'function': mordred.Constitutional.ConstitutionalSum('se'), 'type': '2D', 'description': 'sum of constitutional weighted by sanderson EN', 'module': 'Constitutional', 'constructor': 'ConstitutionalSum("se")'}, 'Spe': {'function': mordred.Constitutional.ConstitutionalSum('pe'), 'type': '2D', 'description': 'sum of constitutional weighted by pauling EN', 'module': 'Constitutional', 'constructor': 'ConstitutionalSum("pe")'}, 'Sare': {'function': mordred.Constitutional.ConstitutionalSum('are'), 'type': '2D', 'description': 'sum of constitutional weighted by allred-rocow EN', 'module': 'Constitutional', 'constructor': 'ConstitutionalSum("are")'}, 'Sp': {'function': mordred.Constitutional.ConstitutionalSum('p'), 'type': '2D', 'description': 'sum of constitutional weighted by polarizability', 'module': 'Constitutional', 'constructor': 'ConstitutionalSum("p")'}, 'Si': {'function': mordred.Constitutional.ConstitutionalSum('i'), 'type': '2D', 'description': 'sum of constitutional weighted by ionization potential', 'module': 'Constitutional', 'constructor': 'ConstitutionalSum("i")'}, 'MZ': {'function': mordred.Constitutional.ConstitutionalMean('Z'), 'type': '2D', 'description': 'mean of constitutional weighted by atomic number', 'module': 'Constitutional', 'constructor': 'ConstitutionalMean("Z")'}, 'Mm': {'function': mordred.Constitutional.ConstitutionalMean('m'), 'type': '2D', 'description': 'mean of constitutional weighted by mass', 'module': 'Constitutional', 'constructor': 'ConstitutionalMean("m")'}, 'Mv': {'function': mordred.Constitutional.ConstitutionalMean('v'), 'type': '2D', 'description': 'mean of constitutional weighted by vdw volume', 'module': 'Constitutional', 'constructor': 'ConstitutionalMean("v")'}, 'Mse': {'function': mordred.Constitutional.ConstitutionalMean('se'), 'type': '2D', 'description': 'mean of constitutional weighted by sanderson EN', 'module': 'Constitutional', 'constructor': 'ConstitutionalMean("se")'}, 'Mpe': {'function': mordred.Constitutional.ConstitutionalMean('pe'), 'type': '2D', 'description': 'mean of constitutional weighted by pauling EN', 'module': 'Constitutional', 'constructor': 'ConstitutionalMean("pe")'}, 'Mare': {'function': mordred.Constitutional.ConstitutionalMean('are'), 'type': '2D', 'description': 'mean of constitutional weighted by allred-rocow EN', 'module': 'Constitutional', 'constructor': 'ConstitutionalMean("are")'}, 'Mp': {'function': mordred.Constitutional.ConstitutionalMean('p'), 'type': '2D', 'description': 'mean of constitutional weighted by polarizability', 'module': 'Constitutional', 'constructor': 'ConstitutionalMean("p")'}, 'Mi': {'function': mordred.Constitutional.ConstitutionalMean('i'), 'type': '2D', 'description': 'mean of constitutional weighted by ionization potential', 'module': 'Constitutional', 'constructor': 'ConstitutionalMean("i")'}, 'SpAbs_Dt': {'function': mordred.DetourMatrix.DetourMatrix('SpAbs'), 'type': '2D', 'description': 'graph energy from detourn matrix', 'module': 'DetourMatrix', 'constructor': 'DetourMatrix("SpAbs")'}, 'SpMax_Dt': {'function': mordred.DetourMatrix.DetourMatrix('SpMax'), 'type': '2D', 'description': 'leading eigenvalue from detourn matrix', 'module': 'DetourMatrix', 'constructor': 'DetourMatrix("SpMax")'}, 'SpDiam_Dt': {'function': mordred.DetourMatrix.DetourMatrix('SpDiam'), 'type': '2D', 'description': 'spectral diamiter from detourn matrix', 'module': 'DetourMatrix', 'constructor': 'DetourMatrix("SpDiam")'}, 'SpAD_Dt': {'function': mordred.DetourMatrix.DetourMatrix('SpAD'), 'type': '2D', 'description': 'spectral absolute diviation from detourn matrix', 'module': 'DetourMatrix', 'constructor': 'DetourMatrix("SpAD")'}, 'SpMAD_Dt': {'function': mordred.DetourMatrix.DetourMatrix('SpMAD'), 'type': '2D', 'description': 'spectral mean absolute diviation from detourn matrix', 'module': 'DetourMatrix', 'constructor': 'DetourMatrix("SpMAD")'}, 'LogEE_Dt': {'function': mordred.DetourMatrix.DetourMatrix('LogEE'), 'type': '2D', 'description': 'Estrada-like index from detourn matrix', 'module': 'DetourMatrix', 'constructor': 'DetourMatrix("LogEE")'}, 'SM1_Dt': {'function': mordred.DetourMatrix.DetourMatrix('SM1'), 'type': '2D', 'description': 'spectral moment from detourn matrix', 'module': 'DetourMatrix', 'constructor': 'DetourMatrix("SM1")'}, 'VE1_Dt': {'function': mordred.DetourMatrix.DetourMatrix('VE1'), 'type': '2D', 'description': 'coefficient sum of the last eigenvector from detourn matrix', 'module': 'DetourMatrix', 'constructor': 'DetourMatrix("VE1")'}, 'VE2_Dt': {'function': mordred.DetourMatrix.DetourMatrix('VE2'), 'type': '2D', 'description': 'average coefficient of the last eigenvector from detourn matrix', 'module': 'DetourMatrix', 'constructor': 'DetourMatrix("VE2")'}, 'VE3_Dt': {'function': mordred.DetourMatrix.DetourMatrix('VE3'), 'type': '2D', 'description': 'logarithmic coefficient sum of the last eigenvector from detourn matrix', 'module': 'DetourMatrix', 'constructor': 'DetourMatrix("VE3")'}, 'VR1_Dt': {'function': mordred.DetourMatrix.DetourMatrix('VR1'), 'type': '2D', 'description': 'Randic-like eigenvector-based index from detourn matrix', 'module': 'DetourMatrix', 'constructor': 'DetourMatrix("VR1")'}, 'VR2_Dt': {'function': mordred.DetourMatrix.DetourMatrix('VR2'), 'type': '2D', 'description': 'normalized Randic-like eigenvector-based index from detourn matrix', 'module': 'DetourMatrix', 'constructor': 'DetourMatrix("VR2")'}, 'VR3_Dt': {'function': mordred.DetourMatrix.DetourMatrix('VR3'), 'type': '2D', 'description': 'logarithmic Randic-like eigenvector-based index from detourn matrix', 'module': 'DetourMatrix', 'constructor': 'DetourMatrix("VR3")'}, 'DetourIndex': {'function': mordred.DetourMatrix.DetourIndex(), 'type': '2D', 'description': 'detour index', 'module': 'DetourMatrix', 'constructor': 'DetourIndex()'}, 'SpAbs_D': {'function': mordred.DistanceMatrix.DistanceMatrix('SpAbs'), 'type': '2D', 'description': 'graph energy from distance matrix', 'module': 'DistanceMatrix', 'constructor': 'DistanceMatrix("SpAbs")'}, 'SpMax_D': {'function': mordred.DistanceMatrix.DistanceMatrix('SpMax'), 'type': '2D', 'description': 'leading eigenvalue from distance matrix', 'module': 'DistanceMatrix', 'constructor': 'DistanceMatrix("SpMax")'}, 'SpDiam_D': {'function': mordred.DistanceMatrix.DistanceMatrix('SpDiam'), 'type': '2D', 'description': 'spectral diamiter from distance matrix', 'module': 'DistanceMatrix', 'constructor': 'DistanceMatrix("SpDiam")'}, 'SpAD_D': {'function': mordred.DistanceMatrix.DistanceMatrix('SpAD'), 'type': '2D', 'description': 'spectral absolute diviation from distance matrix', 'module': 'DistanceMatrix', 'constructor': 'DistanceMatrix("SpAD")'}, 'SpMAD_D': {'function': mordred.DistanceMatrix.DistanceMatrix('SpMAD'), 'type': '2D', 'description': 'spectral mean absolute diviation from distance matrix', 'module': 'DistanceMatrix', 'constructor': 'DistanceMatrix("SpMAD")'}, 'LogEE_D': {'function': mordred.DistanceMatrix.DistanceMatrix('LogEE'), 'type': '2D', 'description': 'Estrada-like index from distance matrix', 'module': 'DistanceMatrix', 'constructor': 'DistanceMatrix("LogEE")'}, 'VE1_D': {'function': mordred.DistanceMatrix.DistanceMatrix('VE1'), 'type': '2D', 'description': 'coefficient sum of the last eigenvector from distance matrix', 'module': 'DistanceMatrix', 'constructor': 'DistanceMatrix("VE1")'}, 'VE2_D': {'function': mordred.DistanceMatrix.DistanceMatrix('VE2'), 'type': '2D', 'description': 'average coefficient of the last eigenvector from distance matrix', 'module': 'DistanceMatrix', 'constructor': 'DistanceMatrix("VE2")'}, 'VE3_D': {'function': mordred.DistanceMatrix.DistanceMatrix('VE3'), 'type': '2D', 'description': 'logarithmic coefficient sum of the last eigenvector from distance matrix', 'module': 'DistanceMatrix', 'constructor': 'DistanceMatrix("VE3")'}, 'VR1_D': {'function': mordred.DistanceMatrix.DistanceMatrix('VR1'), 'type': '2D', 'description': 'Randic-like eigenvector-based index from distance matrix', 'module': 'DistanceMatrix', 'constructor': 'DistanceMatrix("VR1")'}, 'VR2_D': {'function': mordred.DistanceMatrix.DistanceMatrix('VR2'), 'type': '2D', 'description': 'normalized Randic-like eigenvector-based index from distance matrix', 'module': 'DistanceMatrix', 'constructor': 'DistanceMatrix("VR2")'}, 'VR3_D': {'function': mordred.DistanceMatrix.DistanceMatrix('VR3'), 'type': '2D', 'description': 'logarithmic Randic-like eigenvector-based index from distance matrix', 'module': 'DistanceMatrix', 'constructor': 'DistanceMatrix("VR3")'}, 'NsLi': {'function': mordred.EState.AtomTypeEState('count', 'sLi'), 'type': '2D', 'description': 'number of sLi', 'module': 'EState', 'constructor': 'AtomTypeEState("count", "sLi")'}, 'NssBe': {'function': mordred.EState.AtomTypeEState('count', 'ssBe'), 'type': '2D', 'description': 'number of ssBe', 'module': 'EState', 'constructor': 'AtomTypeEState("count", "ssBe")'}, 'NssssBe': {'function': mordred.EState.AtomTypeEState('count', 'ssssBe'), 'type': '2D', 'description': 'number of ssssBe', 'module': 'EState', 'constructor': 'AtomTypeEState("count", "ssssBe")'}, 'NssBH': {'function': mordred.EState.AtomTypeEState('count', 'ssBH'), 'type': '2D', 'description': 'number of ssBH', 'module': 'EState', 'constructor': 'AtomTypeEState("count", "ssBH")'}, 'NsssB': {'function': mordred.EState.AtomTypeEState('count', 'sssB'), 'type': '2D', 'description': 'number of sssB', 'module': 'EState', 'constructor': 'AtomTypeEState("count", "sssB")'}, 'NssssB': {'function': mordred.EState.AtomTypeEState('count', 'ssssB'), 'type': '2D', 'description': 'number of ssssB', 'module': 'EState', 'constructor': 'AtomTypeEState("count", "ssssB")'}, 'NsCH3': {'function': mordred.EState.AtomTypeEState('count', 'sCH3'), 'type': '2D', 'description': 'number of sCH3', 'module': 'EState', 'constructor': 'AtomTypeEState("count", "sCH3")'}, 'NdCH2': {'function': mordred.EState.AtomTypeEState('count', 'dCH2'), 'type': '2D', 'description': 'number of dCH2', 'module': 'EState', 'constructor': 'AtomTypeEState("count", "dCH2")'}, 'NssCH2': {'function': mordred.EState.AtomTypeEState('count', 'ssCH2'), 'type': '2D', 'description': 'number of ssCH2', 'module': 'EState', 'constructor': 'AtomTypeEState("count", "ssCH2")'}, 'NtCH': {'function': mordred.EState.AtomTypeEState('count', 'tCH'), 'type': '2D', 'description': 'number of tCH', 'module': 'EState', 'constructor': 'AtomTypeEState("count", "tCH")'}, 'NdsCH': {'function': mordred.EState.AtomTypeEState('count', 'dsCH'), 'type': '2D', 'description': 'number of dsCH', 'module': 'EState', 'constructor': 'AtomTypeEState("count", "dsCH")'}, 'NaaCH': {'function': mordred.EState.AtomTypeEState('count', 'aaCH'), 'type': '2D', 'description': 'number of aaCH', 'module': 'EState', 'constructor': 'AtomTypeEState("count", "aaCH")'}, 'NsssCH': {'function': mordred.EState.AtomTypeEState('count', 'sssCH'), 'type': '2D', 'description': 'number of sssCH', 'module': 'EState', 'constructor': 'AtomTypeEState("count", "sssCH")'}, 'NddC': {'function': mordred.EState.AtomTypeEState('count', 'ddC'), 'type': '2D', 'description': 'number of ddC', 'module': 'EState', 'constructor': 'AtomTypeEState("count", "ddC")'}, 'NtsC': {'function': mordred.EState.AtomTypeEState('count', 'tsC'), 'type': '2D', 'description': 'number of tsC', 'module': 'EState', 'constructor': 'AtomTypeEState("count", "tsC")'}, 'NdssC': {'function': mordred.EState.AtomTypeEState('count', 'dssC'), 'type': '2D', 'description': 'number of dssC', 'module': 'EState', 'constructor': 'AtomTypeEState("count", "dssC")'}, 'NaasC': {'function': mordred.EState.AtomTypeEState('count', 'aasC'), 'type': '2D', 'description': 'number of aasC', 'module': 'EState', 'constructor': 'AtomTypeEState("count", "aasC")'}, 'NaaaC': {'function': mordred.EState.AtomTypeEState('count', 'aaaC'), 'type': '2D', 'description': 'number of aaaC', 'module': 'EState', 'constructor': 'AtomTypeEState("count", "aaaC")'}, 'NssssC': {'function': mordred.EState.AtomTypeEState('count', 'ssssC'), 'type': '2D', 'description': 'number of ssssC', 'module': 'EState', 'constructor': 'AtomTypeEState("count", "ssssC")'}, 'NsNH3': {'function': mordred.EState.AtomTypeEState('count', 'sNH3'), 'type': '2D', 'description': 'number of sNH3', 'module': 'EState', 'constructor': 'AtomTypeEState("count", "sNH3")'}, 'NsNH2': {'function': mordred.EState.AtomTypeEState('count', 'sNH2'), 'type': '2D', 'description': 'number of sNH2', 'module': 'EState', 'constructor': 'AtomTypeEState("count", "sNH2")'}, 'NssNH2': {'function': mordred.EState.AtomTypeEState('count', 'ssNH2'), 'type': '2D', 'description': 'number of ssNH2', 'module': 'EState', 'constructor': 'AtomTypeEState("count", "ssNH2")'}, 'NdNH': {'function': mordred.EState.AtomTypeEState('count', 'dNH'), 'type': '2D', 'description': 'number of dNH', 'module': 'EState', 'constructor': 'AtomTypeEState("count", "dNH")'}, 'NssNH': {'function': mordred.EState.AtomTypeEState('count', 'ssNH'), 'type': '2D', 'description': 'number of ssNH', 'module': 'EState', 'constructor': 'AtomTypeEState("count", "ssNH")'}, 'NaaNH': {'function': mordred.EState.AtomTypeEState('count', 'aaNH'), 'type': '2D', 'description': 'number of aaNH', 'module': 'EState', 'constructor': 'AtomTypeEState("count", "aaNH")'}, 'NtN': {'function': mordred.EState.AtomTypeEState('count', 'tN'), 'type': '2D', 'description': 'number of tN', 'module': 'EState', 'constructor': 'AtomTypeEState("count", "tN")'}, 'NsssNH': {'function': mordred.EState.AtomTypeEState('count', 'sssNH'), 'type': '2D', 'description': 'number of sssNH', 'module': 'EState', 'constructor': 'AtomTypeEState("count", "sssNH")'}, 'NdsN': {'function': mordred.EState.AtomTypeEState('count', 'dsN'), 'type': '2D', 'description': 'number of dsN', 'module': 'EState', 'constructor': 'AtomTypeEState("count", "dsN")'}, 'NaaN': {'function': mordred.EState.AtomTypeEState('count', 'aaN'), 'type': '2D', 'description': 'number of aaN', 'module': 'EState', 'constructor': 'AtomTypeEState("count", "aaN")'}, 'NsssN': {'function': mordred.EState.AtomTypeEState('count', 'sssN'), 'type': '2D', 'description': 'number of sssN', 'module': 'EState', 'constructor': 'AtomTypeEState("count", "sssN")'}, 'NddsN': {'function': mordred.EState.AtomTypeEState('count', 'ddsN'), 'type': '2D', 'description': 'number of ddsN', 'module': 'EState', 'constructor': 'AtomTypeEState("count", "ddsN")'}, 'NaasN': {'function': mordred.EState.AtomTypeEState('count', 'aasN'), 'type': '2D', 'description': 'number of aasN', 'module': 'EState', 'constructor': 'AtomTypeEState("count", "aasN")'}, 'NssssN': {'function': mordred.EState.AtomTypeEState('count', 'ssssN'), 'type': '2D', 'description': 'number of ssssN', 'module': 'EState', 'constructor': 'AtomTypeEState("count", "ssssN")'}, 'NsOH': {'function': mordred.EState.AtomTypeEState('count', 'sOH'), 'type': '2D', 'description': 'number of sOH', 'module': 'EState', 'constructor': 'AtomTypeEState("count", "sOH")'}, 'NdO': {'function': mordred.EState.AtomTypeEState('count', 'dO'), 'type': '2D', 'description': 'number of dO', 'module': 'EState', 'constructor': 'AtomTypeEState("count", "dO")'}, 'NssO': {'function': mordred.EState.AtomTypeEState('count', 'ssO'), 'type': '2D', 'description': 'number of ssO', 'module': 'EState', 'constructor': 'AtomTypeEState("count", "ssO")'}, 'NaaO': {'function': mordred.EState.AtomTypeEState('count', 'aaO'), 'type': '2D', 'description': 'number of aaO', 'module': 'EState', 'constructor': 'AtomTypeEState("count", "aaO")'}, 'NsF': {'function': mordred.EState.AtomTypeEState('count', 'sF'), 'type': '2D', 'description': 'number of sF', 'module': 'EState', 'constructor': 'AtomTypeEState("count", "sF")'}, 'NsSiH3': {'function': mordred.EState.AtomTypeEState('count', 'sSiH3'), 'type': '2D', 'description': 'number of sSiH3', 'module': 'EState', 'constructor': 'AtomTypeEState("count", "sSiH3")'}, 'NssSiH2': {'function': mordred.EState.AtomTypeEState('count', 'ssSiH2'), 'type': '2D', 'description': 'number of ssSiH2', 'module': 'EState', 'constructor': 'AtomTypeEState("count", "ssSiH2")'}, 'NsssSiH': {'function': mordred.EState.AtomTypeEState('count', 'sssSiH'), 'type': '2D', 'description': 'number of sssSiH', 'module': 'EState', 'constructor': 'AtomTypeEState("count", "sssSiH")'}, 'NssssSi': {'function': mordred.EState.AtomTypeEState('count', 'ssssSi'), 'type': '2D', 'description': 'number of ssssSi', 'module': 'EState', 'constructor': 'AtomTypeEState("count", "ssssSi")'}, 'NsPH2': {'function': mordred.EState.AtomTypeEState('count', 'sPH2'), 'type': '2D', 'description': 'number of sPH2', 'module': 'EState', 'constructor': 'AtomTypeEState("count", "sPH2")'}, 'NssPH': {'function': mordred.EState.AtomTypeEState('count', 'ssPH'), 'type': '2D', 'description': 'number of ssPH', 'module': 'EState', 'constructor': 'AtomTypeEState("count", "ssPH")'}, 'NsssP': {'function': mordred.EState.AtomTypeEState('count', 'sssP'), 'type': '2D', 'description': 'number of sssP', 'module': 'EState', 'constructor': 'AtomTypeEState("count", "sssP")'}, 'NdsssP': {'function': mordred.EState.AtomTypeEState('count', 'dsssP'), 'type': '2D', 'description': 'number of dsssP', 'module': 'EState', 'constructor': 'AtomTypeEState("count", "dsssP")'}, 'NsssssP': {'function': mordred.EState.AtomTypeEState('count', 'sssssP'), 'type': '2D', 'description': 'number of sssssP', 'module': 'EState', 'constructor': 'AtomTypeEState("count", "sssssP")'}, 'NsSH': {'function': mordred.EState.AtomTypeEState('count', 'sSH'), 'type': '2D', 'description': 'number of sSH', 'module': 'EState', 'constructor': 'AtomTypeEState("count", "sSH")'}, 'NdS': {'function': mordred.EState.AtomTypeEState('count', 'dS'), 'type': '2D', 'description': 'number of dS', 'module': 'EState', 'constructor': 'AtomTypeEState("count", "dS")'}, 'NssS': {'function': mordred.EState.AtomTypeEState('count', 'ssS'), 'type': '2D', 'description': 'number of ssS', 'module': 'EState', 'constructor': 'AtomTypeEState("count", "ssS")'}, 'NaaS': {'function': mordred.EState.AtomTypeEState('count', 'aaS'), 'type': '2D', 'description': 'number of aaS', 'module': 'EState', 'constructor': 'AtomTypeEState("count", "aaS")'}, 'NdssS': {'function': mordred.EState.AtomTypeEState('count', 'dssS'), 'type': '2D', 'description': 'number of dssS', 'module': 'EState', 'constructor': 'AtomTypeEState("count", "dssS")'}, 'NddssS': {'function': mordred.EState.AtomTypeEState('count', 'ddssS'), 'type': '2D', 'description': 'number of ddssS', 'module': 'EState', 'constructor': 'AtomTypeEState("count", "ddssS")'}, 'NsCl': {'function': mordred.EState.AtomTypeEState('count', 'sCl'), 'type': '2D', 'description': 'number of sCl', 'module': 'EState', 'constructor': 'AtomTypeEState("count", "sCl")'}, 'NsGeH3': {'function': mordred.EState.AtomTypeEState('count', 'sGeH3'), 'type': '2D', 'description': 'number of sGeH3', 'module': 'EState', 'constructor': 'AtomTypeEState("count", "sGeH3")'}, 'NssGeH2': {'function': mordred.EState.AtomTypeEState('count', 'ssGeH2'), 'type': '2D', 'description': 'number of ssGeH2', 'module': 'EState', 'constructor': 'AtomTypeEState("count", "ssGeH2")'}, 'NsssGeH': {'function': mordred.EState.AtomTypeEState('count', 'sssGeH'), 'type': '2D', 'description': 'number of sssGeH', 'module': 'EState', 'constructor': 'AtomTypeEState("count", "sssGeH")'}, 'NssssGe': {'function': mordred.EState.AtomTypeEState('count', 'ssssGe'), 'type': '2D', 'description': 'number of ssssGe', 'module': 'EState', 'constructor': 'AtomTypeEState("count", "ssssGe")'}, 'NsAsH2': {'function': mordred.EState.AtomTypeEState('count', 'sAsH2'), 'type': '2D', 'description': 'number of sAsH2', 'module': 'EState', 'constructor': 'AtomTypeEState("count", "sAsH2")'}, 'NssAsH': {'function': mordred.EState.AtomTypeEState('count', 'ssAsH'), 'type': '2D', 'description': 'number of ssAsH', 'module': 'EState', 'constructor': 'AtomTypeEState("count", "ssAsH")'}, 'NsssAs': {'function': mordred.EState.AtomTypeEState('count', 'sssAs'), 'type': '2D', 'description': 'number of sssAs', 'module': 'EState', 'constructor': 'AtomTypeEState("count", "sssAs")'}, 'NsssdAs': {'function': mordred.EState.AtomTypeEState('count', 'sssdAs'), 'type': '2D', 'description': 'number of sssdAs', 'module': 'EState', 'constructor': 'AtomTypeEState("count", "sssdAs")'}, 'NsssssAs': {'function': mordred.EState.AtomTypeEState('count', 'sssssAs'), 'type': '2D', 'description': 'number of sssssAs', 'module': 'EState', 'constructor': 'AtomTypeEState("count", "sssssAs")'}, 'NsSeH': {'function': mordred.EState.AtomTypeEState('count', 'sSeH'), 'type': '2D', 'description': 'number of sSeH', 'module': 'EState', 'constructor': 'AtomTypeEState("count", "sSeH")'}, 'NdSe': {'function': mordred.EState.AtomTypeEState('count', 'dSe'), 'type': '2D', 'description': 'number of dSe', 'module': 'EState', 'constructor': 'AtomTypeEState("count", "dSe")'}, 'NssSe': {'function': mordred.EState.AtomTypeEState('count', 'ssSe'), 'type': '2D', 'description': 'number of ssSe', 'module': 'EState', 'constructor': 'AtomTypeEState("count", "ssSe")'}, 'NaaSe': {'function': mordred.EState.AtomTypeEState('count', 'aaSe'), 'type': '2D', 'description': 'number of aaSe', 'module': 'EState', 'constructor': 'AtomTypeEState("count", "aaSe")'}, 'NdssSe': {'function': mordred.EState.AtomTypeEState('count', 'dssSe'), 'type': '2D', 'description': 'number of dssSe', 'module': 'EState', 'constructor': 'AtomTypeEState("count", "dssSe")'}, 'NddssSe': {'function': mordred.EState.AtomTypeEState('count', 'ddssSe'), 'type': '2D', 'description': 'number of ddssSe', 'module': 'EState', 'constructor': 'AtomTypeEState("count", "ddssSe")'}, 'NsBr': {'function': mordred.EState.AtomTypeEState('count', 'sBr'), 'type': '2D', 'description': 'number of sBr', 'module': 'EState', 'constructor': 'AtomTypeEState("count", "sBr")'}, 'NsSnH3': {'function': mordred.EState.AtomTypeEState('count', 'sSnH3'), 'type': '2D', 'description': 'number of sSnH3', 'module': 'EState', 'constructor': 'AtomTypeEState("count", "sSnH3")'}, 'NssSnH2': {'function': mordred.EState.AtomTypeEState('count', 'ssSnH2'), 'type': '2D', 'description': 'number of ssSnH2', 'module': 'EState', 'constructor': 'AtomTypeEState("count", "ssSnH2")'}, 'NsssSnH': {'function': mordred.EState.AtomTypeEState('count', 'sssSnH'), 'type': '2D', 'description': 'number of sssSnH', 'module': 'EState', 'constructor': 'AtomTypeEState("count", "sssSnH")'}, 'NssssSn': {'function': mordred.EState.AtomTypeEState('count', 'ssssSn'), 'type': '2D', 'description': 'number of ssssSn', 'module': 'EState', 'constructor': 'AtomTypeEState("count", "ssssSn")'}, 'NsI': {'function': mordred.EState.AtomTypeEState('count', 'sI'), 'type': '2D', 'description': 'number of sI', 'module': 'EState', 'constructor': 'AtomTypeEState("count", "sI")'}, 'NsPbH3': {'function': mordred.EState.AtomTypeEState('count', 'sPbH3'), 'type': '2D', 'description': 'number of sPbH3', 'module': 'EState', 'constructor': 'AtomTypeEState("count", "sPbH3")'}, 'NssPbH2': {'function': mordred.EState.AtomTypeEState('count', 'ssPbH2'), 'type': '2D', 'description': 'number of ssPbH2', 'module': 'EState', 'constructor': 'AtomTypeEState("count", "ssPbH2")'}, 'NsssPbH': {'function': mordred.EState.AtomTypeEState('count', 'sssPbH'), 'type': '2D', 'description': 'number of sssPbH', 'module': 'EState', 'constructor': 'AtomTypeEState("count", "sssPbH")'}, 'NssssPb': {'function': mordred.EState.AtomTypeEState('count', 'ssssPb'), 'type': '2D', 'description': 'number of ssssPb', 'module': 'EState', 'constructor': 'AtomTypeEState("count", "ssssPb")'}, 'SsLi': {'function': mordred.EState.AtomTypeEState('sum', 'sLi'), 'type': '2D', 'description': 'sum of sLi', 'module': 'EState', 'constructor': 'AtomTypeEState("sum", "sLi")'}, 'SssBe': {'function': mordred.EState.AtomTypeEState('sum', 'ssBe'), 'type': '2D', 'description': 'sum of ssBe', 'module': 'EState', 'constructor': 'AtomTypeEState("sum", "ssBe")'}, 'SssssBe': {'function': mordred.EState.AtomTypeEState('sum', 'ssssBe'), 'type': '2D', 'description': 'sum of ssssBe', 'module': 'EState', 'constructor': 'AtomTypeEState("sum", "ssssBe")'}, 'SssBH': {'function': mordred.EState.AtomTypeEState('sum', 'ssBH'), 'type': '2D', 'description': 'sum of ssBH', 'module': 'EState', 'constructor': 'AtomTypeEState("sum", "ssBH")'}, 'SsssB': {'function': mordred.EState.AtomTypeEState('sum', 'sssB'), 'type': '2D', 'description': 'sum of sssB', 'module': 'EState', 'constructor': 'AtomTypeEState("sum", "sssB")'}, 'SssssB': {'function': mordred.EState.AtomTypeEState('sum', 'ssssB'), 'type': '2D', 'description': 'sum of ssssB', 'module': 'EState', 'constructor': 'AtomTypeEState("sum", "ssssB")'}, 'SsCH3': {'function': mordred.EState.AtomTypeEState('sum', 'sCH3'), 'type': '2D', 'description': 'sum of sCH3', 'module': 'EState', 'constructor': 'AtomTypeEState("sum", "sCH3")'}, 'SdCH2': {'function': mordred.EState.AtomTypeEState('sum', 'dCH2'), 'type': '2D', 'description': 'sum of dCH2', 'module': 'EState', 'constructor': 'AtomTypeEState("sum", "dCH2")'}, 'SssCH2': {'function': mordred.EState.AtomTypeEState('sum', 'ssCH2'), 'type': '2D', 'description': 'sum of ssCH2', 'module': 'EState', 'constructor': 'AtomTypeEState("sum", "ssCH2")'}, 'StCH': {'function': mordred.EState.AtomTypeEState('sum', 'tCH'), 'type': '2D', 'description': 'sum of tCH', 'module': 'EState', 'constructor': 'AtomTypeEState("sum", "tCH")'}, 'SdsCH': {'function': mordred.EState.AtomTypeEState('sum', 'dsCH'), 'type': '2D', 'description': 'sum of dsCH', 'module': 'EState', 'constructor': 'AtomTypeEState("sum", "dsCH")'}, 'SaaCH': {'function': mordred.EState.AtomTypeEState('sum', 'aaCH'), 'type': '2D', 'description': 'sum of aaCH', 'module': 'EState', 'constructor': 'AtomTypeEState("sum", "aaCH")'}, 'SsssCH': {'function': mordred.EState.AtomTypeEState('sum', 'sssCH'), 'type': '2D', 'description': 'sum of sssCH', 'module': 'EState', 'constructor': 'AtomTypeEState("sum", "sssCH")'}, 'SddC': {'function': mordred.EState.AtomTypeEState('sum', 'ddC'), 'type': '2D', 'description': 'sum of ddC', 'module': 'EState', 'constructor': 'AtomTypeEState("sum", "ddC")'}, 'StsC': {'function': mordred.EState.AtomTypeEState('sum', 'tsC'), 'type': '2D', 'description': 'sum of tsC', 'module': 'EState', 'constructor': 'AtomTypeEState("sum", "tsC")'}, 'SdssC': {'function': mordred.EState.AtomTypeEState('sum', 'dssC'), 'type': '2D', 'description': 'sum of dssC', 'module': 'EState', 'constructor': 'AtomTypeEState("sum", "dssC")'}, 'SaasC': {'function': mordred.EState.AtomTypeEState('sum', 'aasC'), 'type': '2D', 'description': 'sum of aasC', 'module': 'EState', 'constructor': 'AtomTypeEState("sum", "aasC")'}, 'SaaaC': {'function': mordred.EState.AtomTypeEState('sum', 'aaaC'), 'type': '2D', 'description': 'sum of aaaC', 'module': 'EState', 'constructor': 'AtomTypeEState("sum", "aaaC")'}, 'SssssC': {'function': mordred.EState.AtomTypeEState('sum', 'ssssC'), 'type': '2D', 'description': 'sum of ssssC', 'module': 'EState', 'constructor': 'AtomTypeEState("sum", "ssssC")'}, 'SsNH3': {'function': mordred.EState.AtomTypeEState('sum', 'sNH3'), 'type': '2D', 'description': 'sum of sNH3', 'module': 'EState', 'constructor': 'AtomTypeEState("sum", "sNH3")'}, 'SsNH2': {'function': mordred.EState.AtomTypeEState('sum', 'sNH2'), 'type': '2D', 'description': 'sum of sNH2', 'module': 'EState', 'constructor': 'AtomTypeEState("sum", "sNH2")'}, 'SssNH2': {'function': mordred.EState.AtomTypeEState('sum', 'ssNH2'), 'type': '2D', 'description': 'sum of ssNH2', 'module': 'EState', 'constructor': 'AtomTypeEState("sum", "ssNH2")'}, 'SdNH': {'function': mordred.EState.AtomTypeEState('sum', 'dNH'), 'type': '2D', 'description': 'sum of dNH', 'module': 'EState', 'constructor': 'AtomTypeEState("sum", "dNH")'}, 'SssNH': {'function': mordred.EState.AtomTypeEState('sum', 'ssNH'), 'type': '2D', 'description': 'sum of ssNH', 'module': 'EState', 'constructor': 'AtomTypeEState("sum", "ssNH")'}, 'SaaNH': {'function': mordred.EState.AtomTypeEState('sum', 'aaNH'), 'type': '2D', 'description': 'sum of aaNH', 'module': 'EState', 'constructor': 'AtomTypeEState("sum", "aaNH")'}, 'StN': {'function': mordred.EState.AtomTypeEState('sum', 'tN'), 'type': '2D', 'description': 'sum of tN', 'module': 'EState', 'constructor': 'AtomTypeEState("sum", "tN")'}, 'SsssNH': {'function': mordred.EState.AtomTypeEState('sum', 'sssNH'), 'type': '2D', 'description': 'sum of sssNH', 'module': 'EState', 'constructor': 'AtomTypeEState("sum", "sssNH")'}, 'SdsN': {'function': mordred.EState.AtomTypeEState('sum', 'dsN'), 'type': '2D', 'description': 'sum of dsN', 'module': 'EState', 'constructor': 'AtomTypeEState("sum", "dsN")'}, 'SaaN': {'function': mordred.EState.AtomTypeEState('sum', 'aaN'), 'type': '2D', 'description': 'sum of aaN', 'module': 'EState', 'constructor': 'AtomTypeEState("sum", "aaN")'}, 'SsssN': {'function': mordred.EState.AtomTypeEState('sum', 'sssN'), 'type': '2D', 'description': 'sum of sssN', 'module': 'EState', 'constructor': 'AtomTypeEState("sum", "sssN")'}, 'SddsN': {'function': mordred.EState.AtomTypeEState('sum', 'ddsN'), 'type': '2D', 'description': 'sum of ddsN', 'module': 'EState', 'constructor': 'AtomTypeEState("sum", "ddsN")'}, 'SaasN': {'function': mordred.EState.AtomTypeEState('sum', 'aasN'), 'type': '2D', 'description': 'sum of aasN', 'module': 'EState', 'constructor': 'AtomTypeEState("sum", "aasN")'}, 'SssssN': {'function': mordred.EState.AtomTypeEState('sum', 'ssssN'), 'type': '2D', 'description': 'sum of ssssN', 'module': 'EState', 'constructor': 'AtomTypeEState("sum", "ssssN")'}, 'SsOH': {'function': mordred.EState.AtomTypeEState('sum', 'sOH'), 'type': '2D', 'description': 'sum of sOH', 'module': 'EState', 'constructor': 'AtomTypeEState("sum", "sOH")'}, 'SdO': {'function': mordred.EState.AtomTypeEState('sum', 'dO'), 'type': '2D', 'description': 'sum of dO', 'module': 'EState', 'constructor': 'AtomTypeEState("sum", "dO")'}, 'SssO': {'function': mordred.EState.AtomTypeEState('sum', 'ssO'), 'type': '2D', 'description': 'sum of ssO', 'module': 'EState', 'constructor': 'AtomTypeEState("sum", "ssO")'}, 'SaaO': {'function': mordred.EState.AtomTypeEState('sum', 'aaO'), 'type': '2D', 'description': 'sum of aaO', 'module': 'EState', 'constructor': 'AtomTypeEState("sum", "aaO")'}, 'SsF': {'function': mordred.EState.AtomTypeEState('sum', 'sF'), 'type': '2D', 'description': 'sum of sF', 'module': 'EState', 'constructor': 'AtomTypeEState("sum", "sF")'}, 'SsSiH3': {'function': mordred.EState.AtomTypeEState('sum', 'sSiH3'), 'type': '2D', 'description': 'sum of sSiH3', 'module': 'EState', 'constructor': 'AtomTypeEState("sum", "sSiH3")'}, 'SssSiH2': {'function': mordred.EState.AtomTypeEState('sum', 'ssSiH2'), 'type': '2D', 'description': 'sum of ssSiH2', 'module': 'EState', 'constructor': 'AtomTypeEState("sum", "ssSiH2")'}, 'SsssSiH': {'function': mordred.EState.AtomTypeEState('sum', 'sssSiH'), 'type': '2D', 'description': 'sum of sssSiH', 'module': 'EState', 'constructor': 'AtomTypeEState("sum", "sssSiH")'}, 'SssssSi': {'function': mordred.EState.AtomTypeEState('sum', 'ssssSi'), 'type': '2D', 'description': 'sum of ssssSi', 'module': 'EState', 'constructor': 'AtomTypeEState("sum", "ssssSi")'}, 'SsPH2': {'function': mordred.EState.AtomTypeEState('sum', 'sPH2'), 'type': '2D', 'description': 'sum of sPH2', 'module': 'EState', 'constructor': 'AtomTypeEState("sum", "sPH2")'}, 'SssPH': {'function': mordred.EState.AtomTypeEState('sum', 'ssPH'), 'type': '2D', 'description': 'sum of ssPH', 'module': 'EState', 'constructor': 'AtomTypeEState("sum", "ssPH")'}, 'SsssP': {'function': mordred.EState.AtomTypeEState('sum', 'sssP'), 'type': '2D', 'description': 'sum of sssP', 'module': 'EState', 'constructor': 'AtomTypeEState("sum", "sssP")'}, 'SdsssP': {'function': mordred.EState.AtomTypeEState('sum', 'dsssP'), 'type': '2D', 'description': 'sum of dsssP', 'module': 'EState', 'constructor': 'AtomTypeEState("sum", "dsssP")'}, 'SsssssP': {'function': mordred.EState.AtomTypeEState('sum', 'sssssP'), 'type': '2D', 'description': 'sum of sssssP', 'module': 'EState', 'constructor': 'AtomTypeEState("sum", "sssssP")'}, 'SsSH': {'function': mordred.EState.AtomTypeEState('sum', 'sSH'), 'type': '2D', 'description': 'sum of sSH', 'module': 'EState', 'constructor': 'AtomTypeEState("sum", "sSH")'}, 'SdS': {'function': mordred.EState.AtomTypeEState('sum', 'dS'), 'type': '2D', 'description': 'sum of dS', 'module': 'EState', 'constructor': 'AtomTypeEState("sum", "dS")'}, 'SssS': {'function': mordred.EState.AtomTypeEState('sum', 'ssS'), 'type': '2D', 'description': 'sum of ssS', 'module': 'EState', 'constructor': 'AtomTypeEState("sum", "ssS")'}, 'SaaS': {'function': mordred.EState.AtomTypeEState('sum', 'aaS'), 'type': '2D', 'description': 'sum of aaS', 'module': 'EState', 'constructor': 'AtomTypeEState("sum", "aaS")'}, 'SdssS': {'function': mordred.EState.AtomTypeEState('sum', 'dssS'), 'type': '2D', 'description': 'sum of dssS', 'module': 'EState', 'constructor': 'AtomTypeEState("sum", "dssS")'}, 'SddssS': {'function': mordred.EState.AtomTypeEState('sum', 'ddssS'), 'type': '2D', 'description': 'sum of ddssS', 'module': 'EState', 'constructor': 'AtomTypeEState("sum", "ddssS")'}, 'SsCl': {'function': mordred.EState.AtomTypeEState('sum', 'sCl'), 'type': '2D', 'description': 'sum of sCl', 'module': 'EState', 'constructor': 'AtomTypeEState("sum", "sCl")'}, 'SsGeH3': {'function': mordred.EState.AtomTypeEState('sum', 'sGeH3'), 'type': '2D', 'description': 'sum of sGeH3', 'module': 'EState', 'constructor': 'AtomTypeEState("sum", "sGeH3")'}, 'SssGeH2': {'function': mordred.EState.AtomTypeEState('sum', 'ssGeH2'), 'type': '2D', 'description': 'sum of ssGeH2', 'module': 'EState', 'constructor': 'AtomTypeEState("sum", "ssGeH2")'}, 'SsssGeH': {'function': mordred.EState.AtomTypeEState('sum', 'sssGeH'), 'type': '2D', 'description': 'sum of sssGeH', 'module': 'EState', 'constructor': 'AtomTypeEState("sum", "sssGeH")'}, 'SssssGe': {'function': mordred.EState.AtomTypeEState('sum', 'ssssGe'), 'type': '2D', 'description': 'sum of ssssGe', 'module': 'EState', 'constructor': 'AtomTypeEState("sum", "ssssGe")'}, 'SsAsH2': {'function': mordred.EState.AtomTypeEState('sum', 'sAsH2'), 'type': '2D', 'description': 'sum of sAsH2', 'module': 'EState', 'constructor': 'AtomTypeEState("sum", "sAsH2")'}, 'SssAsH': {'function': mordred.EState.AtomTypeEState('sum', 'ssAsH'), 'type': '2D', 'description': 'sum of ssAsH', 'module': 'EState', 'constructor': 'AtomTypeEState("sum", "ssAsH")'}, 'SsssAs': {'function': mordred.EState.AtomTypeEState('sum', 'sssAs'), 'type': '2D', 'description': 'sum of sssAs', 'module': 'EState', 'constructor': 'AtomTypeEState("sum", "sssAs")'}, 'SsssdAs': {'function': mordred.EState.AtomTypeEState('sum', 'sssdAs'), 'type': '2D', 'description': 'sum of sssdAs', 'module': 'EState', 'constructor': 'AtomTypeEState("sum", "sssdAs")'}, 'SsssssAs': {'function': mordred.EState.AtomTypeEState('sum', 'sssssAs'), 'type': '2D', 'description': 'sum of sssssAs', 'module': 'EState', 'constructor': 'AtomTypeEState("sum", "sssssAs")'}, 'SsSeH': {'function': mordred.EState.AtomTypeEState('sum', 'sSeH'), 'type': '2D', 'description': 'sum of sSeH', 'module': 'EState', 'constructor': 'AtomTypeEState("sum", "sSeH")'}, 'SdSe': {'function': mordred.EState.AtomTypeEState('sum', 'dSe'), 'type': '2D', 'description': 'sum of dSe', 'module': 'EState', 'constructor': 'AtomTypeEState("sum", "dSe")'}, 'SssSe': {'function': mordred.EState.AtomTypeEState('sum', 'ssSe'), 'type': '2D', 'description': 'sum of ssSe', 'module': 'EState', 'constructor': 'AtomTypeEState("sum", "ssSe")'}, 'SaaSe': {'function': mordred.EState.AtomTypeEState('sum', 'aaSe'), 'type': '2D', 'description': 'sum of aaSe', 'module': 'EState', 'constructor': 'AtomTypeEState("sum", "aaSe")'}, 'SdssSe': {'function': mordred.EState.AtomTypeEState('sum', 'dssSe'), 'type': '2D', 'description': 'sum of dssSe', 'module': 'EState', 'constructor': 'AtomTypeEState("sum", "dssSe")'}, 'SddssSe': {'function': mordred.EState.AtomTypeEState('sum', 'ddssSe'), 'type': '2D', 'description': 'sum of ddssSe', 'module': 'EState', 'constructor': 'AtomTypeEState("sum", "ddssSe")'}, 'SsBr': {'function': mordred.EState.AtomTypeEState('sum', 'sBr'), 'type': '2D', 'description': 'sum of sBr', 'module': 'EState', 'constructor': 'AtomTypeEState("sum", "sBr")'}, 'SsSnH3': {'function': mordred.EState.AtomTypeEState('sum', 'sSnH3'), 'type': '2D', 'description': 'sum of sSnH3', 'module': 'EState', 'constructor': 'AtomTypeEState("sum", "sSnH3")'}, 'SssSnH2': {'function': mordred.EState.AtomTypeEState('sum', 'ssSnH2'), 'type': '2D', 'description': 'sum of ssSnH2', 'module': 'EState', 'constructor': 'AtomTypeEState("sum", "ssSnH2")'}, 'SsssSnH': {'function': mordred.EState.AtomTypeEState('sum', 'sssSnH'), 'type': '2D', 'description': 'sum of sssSnH', 'module': 'EState', 'constructor': 'AtomTypeEState("sum", "sssSnH")'}, 'SssssSn': {'function': mordred.EState.AtomTypeEState('sum', 'ssssSn'), 'type': '2D', 'description': 'sum of ssssSn', 'module': 'EState', 'constructor': 'AtomTypeEState("sum", "ssssSn")'}, 'SsI': {'function': mordred.EState.AtomTypeEState('sum', 'sI'), 'type': '2D', 'description': 'sum of sI', 'module': 'EState', 'constructor': 'AtomTypeEState("sum", "sI")'}, 'SsPbH3': {'function': mordred.EState.AtomTypeEState('sum', 'sPbH3'), 'type': '2D', 'description': 'sum of sPbH3', 'module': 'EState', 'constructor': 'AtomTypeEState("sum", "sPbH3")'}, 'SssPbH2': {'function': mordred.EState.AtomTypeEState('sum', 'ssPbH2'), 'type': '2D', 'description': 'sum of ssPbH2', 'module': 'EState', 'constructor': 'AtomTypeEState("sum", "ssPbH2")'}, 'SsssPbH': {'function': mordred.EState.AtomTypeEState('sum', 'sssPbH'), 'type': '2D', 'description': 'sum of sssPbH', 'module': 'EState', 'constructor': 'AtomTypeEState("sum", "sssPbH")'}, 'SssssPb': {'function': mordred.EState.AtomTypeEState('sum', 'ssssPb'), 'type': '2D', 'description': 'sum of ssssPb', 'module': 'EState', 'constructor': 'AtomTypeEState("sum", "ssssPb")'}, 'MAXsLi': {'function': mordred.EState.AtomTypeEState('max', 'sLi'), 'type': '2D', 'description': 'max of sLi', 'module': 'EState', 'constructor': 'AtomTypeEState("max", "sLi")'}, 'MAXssBe': {'function': mordred.EState.AtomTypeEState('max', 'ssBe'), 'type': '2D', 'description': 'max of ssBe', 'module': 'EState', 'constructor': 'AtomTypeEState("max", "ssBe")'}, 'MAXssssBe': {'function': mordred.EState.AtomTypeEState('max', 'ssssBe'), 'type': '2D', 'description': 'max of ssssBe', 'module': 'EState', 'constructor': 'AtomTypeEState("max", "ssssBe")'}, 'MAXssBH': {'function': mordred.EState.AtomTypeEState('max', 'ssBH'), 'type': '2D', 'description': 'max of ssBH', 'module': 'EState', 'constructor': 'AtomTypeEState("max", "ssBH")'}, 'MAXsssB': {'function': mordred.EState.AtomTypeEState('max', 'sssB'), 'type': '2D', 'description': 'max of sssB', 'module': 'EState', 'constructor': 'AtomTypeEState("max", "sssB")'}, 'MAXssssB': {'function': mordred.EState.AtomTypeEState('max', 'ssssB'), 'type': '2D', 'description': 'max of ssssB', 'module': 'EState', 'constructor': 'AtomTypeEState("max", "ssssB")'}, 'MAXsCH3': {'function': mordred.EState.AtomTypeEState('max', 'sCH3'), 'type': '2D', 'description': 'max of sCH3', 'module': 'EState', 'constructor': 'AtomTypeEState("max", "sCH3")'}, 'MAXdCH2': {'function': mordred.EState.AtomTypeEState('max', 'dCH2'), 'type': '2D', 'description': 'max of dCH2', 'module': 'EState', 'constructor': 'AtomTypeEState("max", "dCH2")'}, 'MAXssCH2': {'function': mordred.EState.AtomTypeEState('max', 'ssCH2'), 'type': '2D', 'description': 'max of ssCH2', 'module': 'EState', 'constructor': 'AtomTypeEState("max", "ssCH2")'}, 'MAXtCH': {'function': mordred.EState.AtomTypeEState('max', 'tCH'), 'type': '2D', 'description': 'max of tCH', 'module': 'EState', 'constructor': 'AtomTypeEState("max", "tCH")'}, 'MAXdsCH': {'function': mordred.EState.AtomTypeEState('max', 'dsCH'), 'type': '2D', 'description': 'max of dsCH', 'module': 'EState', 'constructor': 'AtomTypeEState("max", "dsCH")'}, 'MAXaaCH': {'function': mordred.EState.AtomTypeEState('max', 'aaCH'), 'type': '2D', 'description': 'max of aaCH', 'module': 'EState', 'constructor': 'AtomTypeEState("max", "aaCH")'}, 'MAXsssCH': {'function': mordred.EState.AtomTypeEState('max', 'sssCH'), 'type': '2D', 'description': 'max of sssCH', 'module': 'EState', 'constructor': 'AtomTypeEState("max", "sssCH")'}, 'MAXddC': {'function': mordred.EState.AtomTypeEState('max', 'ddC'), 'type': '2D', 'description': 'max of ddC', 'module': 'EState', 'constructor': 'AtomTypeEState("max", "ddC")'}, 'MAXtsC': {'function': mordred.EState.AtomTypeEState('max', 'tsC'), 'type': '2D', 'description': 'max of tsC', 'module': 'EState', 'constructor': 'AtomTypeEState("max", "tsC")'}, 'MAXdssC': {'function': mordred.EState.AtomTypeEState('max', 'dssC'), 'type': '2D', 'description': 'max of dssC', 'module': 'EState', 'constructor': 'AtomTypeEState("max", "dssC")'}, 'MAXaasC': {'function': mordred.EState.AtomTypeEState('max', 'aasC'), 'type': '2D', 'description': 'max of aasC', 'module': 'EState', 'constructor': 'AtomTypeEState("max", "aasC")'}, 'MAXaaaC': {'function': mordred.EState.AtomTypeEState('max', 'aaaC'), 'type': '2D', 'description': 'max of aaaC', 'module': 'EState', 'constructor': 'AtomTypeEState("max", "aaaC")'}, 'MAXssssC': {'function': mordred.EState.AtomTypeEState('max', 'ssssC'), 'type': '2D', 'description': 'max of ssssC', 'module': 'EState', 'constructor': 'AtomTypeEState("max", "ssssC")'}, 'MAXsNH3': {'function': mordred.EState.AtomTypeEState('max', 'sNH3'), 'type': '2D', 'description': 'max of sNH3', 'module': 'EState', 'constructor': 'AtomTypeEState("max", "sNH3")'}, 'MAXsNH2': {'function': mordred.EState.AtomTypeEState('max', 'sNH2'), 'type': '2D', 'description': 'max of sNH2', 'module': 'EState', 'constructor': 'AtomTypeEState("max", "sNH2")'}, 'MAXssNH2': {'function': mordred.EState.AtomTypeEState('max', 'ssNH2'), 'type': '2D', 'description': 'max of ssNH2', 'module': 'EState', 'constructor': 'AtomTypeEState("max", "ssNH2")'}, 'MAXdNH': {'function': mordred.EState.AtomTypeEState('max', 'dNH'), 'type': '2D', 'description': 'max of dNH', 'module': 'EState', 'constructor': 'AtomTypeEState("max", "dNH")'}, 'MAXssNH': {'function': mordred.EState.AtomTypeEState('max', 'ssNH'), 'type': '2D', 'description': 'max of ssNH', 'module': 'EState', 'constructor': 'AtomTypeEState("max", "ssNH")'}, 'MAXaaNH': {'function': mordred.EState.AtomTypeEState('max', 'aaNH'), 'type': '2D', 'description': 'max of aaNH', 'module': 'EState', 'constructor': 'AtomTypeEState("max", "aaNH")'}, 'MAXtN': {'function': mordred.EState.AtomTypeEState('max', 'tN'), 'type': '2D', 'description': 'max of tN', 'module': 'EState', 'constructor': 'AtomTypeEState("max", "tN")'}, 'MAXsssNH': {'function': mordred.EState.AtomTypeEState('max', 'sssNH'), 'type': '2D', 'description': 'max of sssNH', 'module': 'EState', 'constructor': 'AtomTypeEState("max", "sssNH")'}, 'MAXdsN': {'function': mordred.EState.AtomTypeEState('max', 'dsN'), 'type': '2D', 'description': 'max of dsN', 'module': 'EState', 'constructor': 'AtomTypeEState("max", "dsN")'}, 'MAXaaN': {'function': mordred.EState.AtomTypeEState('max', 'aaN'), 'type': '2D', 'description': 'max of aaN', 'module': 'EState', 'constructor': 'AtomTypeEState("max", "aaN")'}, 'MAXsssN': {'function': mordred.EState.AtomTypeEState('max', 'sssN'), 'type': '2D', 'description': 'max of sssN', 'module': 'EState', 'constructor': 'AtomTypeEState("max", "sssN")'}, 'MAXddsN': {'function': mordred.EState.AtomTypeEState('max', 'ddsN'), 'type': '2D', 'description': 'max of ddsN', 'module': 'EState', 'constructor': 'AtomTypeEState("max", "ddsN")'}, 'MAXaasN': {'function': mordred.EState.AtomTypeEState('max', 'aasN'), 'type': '2D', 'description': 'max of aasN', 'module': 'EState', 'constructor': 'AtomTypeEState("max", "aasN")'}, 'MAXssssN': {'function': mordred.EState.AtomTypeEState('max', 'ssssN'), 'type': '2D', 'description': 'max of ssssN', 'module': 'EState', 'constructor': 'AtomTypeEState("max", "ssssN")'}, 'MAXsOH': {'function': mordred.EState.AtomTypeEState('max', 'sOH'), 'type': '2D', 'description': 'max of sOH', 'module': 'EState', 'constructor': 'AtomTypeEState("max", "sOH")'}, 'MAXdO': {'function': mordred.EState.AtomTypeEState('max', 'dO'), 'type': '2D', 'description': 'max of dO', 'module': 'EState', 'constructor': 'AtomTypeEState("max", "dO")'}, 'MAXssO': {'function': mordred.EState.AtomTypeEState('max', 'ssO'), 'type': '2D', 'description': 'max of ssO', 'module': 'EState', 'constructor': 'AtomTypeEState("max", "ssO")'}, 'MAXaaO': {'function': mordred.EState.AtomTypeEState('max', 'aaO'), 'type': '2D', 'description': 'max of aaO', 'module': 'EState', 'constructor': 'AtomTypeEState("max", "aaO")'}, 'MAXsF': {'function': mordred.EState.AtomTypeEState('max', 'sF'), 'type': '2D', 'description': 'max of sF', 'module': 'EState', 'constructor': 'AtomTypeEState("max", "sF")'}, 'MAXsSiH3': {'function': mordred.EState.AtomTypeEState('max', 'sSiH3'), 'type': '2D', 'description': 'max of sSiH3', 'module': 'EState', 'constructor': 'AtomTypeEState("max", "sSiH3")'}, 'MAXssSiH2': {'function': mordred.EState.AtomTypeEState('max', 'ssSiH2'), 'type': '2D', 'description': 'max of ssSiH2', 'module': 'EState', 'constructor': 'AtomTypeEState("max", "ssSiH2")'}, 'MAXsssSiH': {'function': mordred.EState.AtomTypeEState('max', 'sssSiH'), 'type': '2D', 'description': 'max of sssSiH', 'module': 'EState', 'constructor': 'AtomTypeEState("max", "sssSiH")'}, 'MAXssssSi': {'function': mordred.EState.AtomTypeEState('max', 'ssssSi'), 'type': '2D', 'description': 'max of ssssSi', 'module': 'EState', 'constructor': 'AtomTypeEState("max", "ssssSi")'}, 'MAXsPH2': {'function': mordred.EState.AtomTypeEState('max', 'sPH2'), 'type': '2D', 'description': 'max of sPH2', 'module': 'EState', 'constructor': 'AtomTypeEState("max", "sPH2")'}, 'MAXssPH': {'function': mordred.EState.AtomTypeEState('max', 'ssPH'), 'type': '2D', 'description': 'max of ssPH', 'module': 'EState', 'constructor': 'AtomTypeEState("max", "ssPH")'}, 'MAXsssP': {'function': mordred.EState.AtomTypeEState('max', 'sssP'), 'type': '2D', 'description': 'max of sssP', 'module': 'EState', 'constructor': 'AtomTypeEState("max", "sssP")'}, 'MAXdsssP': {'function': mordred.EState.AtomTypeEState('max', 'dsssP'), 'type': '2D', 'description': 'max of dsssP', 'module': 'EState', 'constructor': 'AtomTypeEState("max", "dsssP")'}, 'MAXsssssP': {'function': mordred.EState.AtomTypeEState('max', 'sssssP'), 'type': '2D', 'description': 'max of sssssP', 'module': 'EState', 'constructor': 'AtomTypeEState("max", "sssssP")'}, 'MAXsSH': {'function': mordred.EState.AtomTypeEState('max', 'sSH'), 'type': '2D', 'description': 'max of sSH', 'module': 'EState', 'constructor': 'AtomTypeEState("max", "sSH")'}, 'MAXdS': {'function': mordred.EState.AtomTypeEState('max', 'dS'), 'type': '2D', 'description': 'max of dS', 'module': 'EState', 'constructor': 'AtomTypeEState("max", "dS")'}, 'MAXssS': {'function': mordred.EState.AtomTypeEState('max', 'ssS'), 'type': '2D', 'description': 'max of ssS', 'module': 'EState', 'constructor': 'AtomTypeEState("max", "ssS")'}, 'MAXaaS': {'function': mordred.EState.AtomTypeEState('max', 'aaS'), 'type': '2D', 'description': 'max of aaS', 'module': 'EState', 'constructor': 'AtomTypeEState("max", "aaS")'}, 'MAXdssS': {'function': mordred.EState.AtomTypeEState('max', 'dssS'), 'type': '2D', 'description': 'max of dssS', 'module': 'EState', 'constructor': 'AtomTypeEState("max", "dssS")'}, 'MAXddssS': {'function': mordred.EState.AtomTypeEState('max', 'ddssS'), 'type': '2D', 'description': 'max of ddssS', 'module': 'EState', 'constructor': 'AtomTypeEState("max", "ddssS")'}, 'MAXsCl': {'function': mordred.EState.AtomTypeEState('max', 'sCl'), 'type': '2D', 'description': 'max of sCl', 'module': 'EState', 'constructor': 'AtomTypeEState("max", "sCl")'}, 'MAXsGeH3': {'function': mordred.EState.AtomTypeEState('max', 'sGeH3'), 'type': '2D', 'description': 'max of sGeH3', 'module': 'EState', 'constructor': 'AtomTypeEState("max", "sGeH3")'}, 'MAXssGeH2': {'function': mordred.EState.AtomTypeEState('max', 'ssGeH2'), 'type': '2D', 'description': 'max of ssGeH2', 'module': 'EState', 'constructor': 'AtomTypeEState("max", "ssGeH2")'}, 'MAXsssGeH': {'function': mordred.EState.AtomTypeEState('max', 'sssGeH'), 'type': '2D', 'description': 'max of sssGeH', 'module': 'EState', 'constructor': 'AtomTypeEState("max", "sssGeH")'}, 'MAXssssGe': {'function': mordred.EState.AtomTypeEState('max', 'ssssGe'), 'type': '2D', 'description': 'max of ssssGe', 'module': 'EState', 'constructor': 'AtomTypeEState("max", "ssssGe")'}, 'MAXsAsH2': {'function': mordred.EState.AtomTypeEState('max', 'sAsH2'), 'type': '2D', 'description': 'max of sAsH2', 'module': 'EState', 'constructor': 'AtomTypeEState("max", "sAsH2")'}, 'MAXssAsH': {'function': mordred.EState.AtomTypeEState('max', 'ssAsH'), 'type': '2D', 'description': 'max of ssAsH', 'module': 'EState', 'constructor': 'AtomTypeEState("max", "ssAsH")'}, 'MAXsssAs': {'function': mordred.EState.AtomTypeEState('max', 'sssAs'), 'type': '2D', 'description': 'max of sssAs', 'module': 'EState', 'constructor': 'AtomTypeEState("max", "sssAs")'}, 'MAXsssdAs': {'function': mordred.EState.AtomTypeEState('max', 'sssdAs'), 'type': '2D', 'description': 'max of sssdAs', 'module': 'EState', 'constructor': 'AtomTypeEState("max", "sssdAs")'}, 'MAXsssssAs': {'function': mordred.EState.AtomTypeEState('max', 'sssssAs'), 'type': '2D', 'description': 'max of sssssAs', 'module': 'EState', 'constructor': 'AtomTypeEState("max", "sssssAs")'}, 'MAXsSeH': {'function': mordred.EState.AtomTypeEState('max', 'sSeH'), 'type': '2D', 'description': 'max of sSeH', 'module': 'EState', 'constructor': 'AtomTypeEState("max", "sSeH")'}, 'MAXdSe': {'function': mordred.EState.AtomTypeEState('max', 'dSe'), 'type': '2D', 'description': 'max of dSe', 'module': 'EState', 'constructor': 'AtomTypeEState("max", "dSe")'}, 'MAXssSe': {'function': mordred.EState.AtomTypeEState('max', 'ssSe'), 'type': '2D', 'description': 'max of ssSe', 'module': 'EState', 'constructor': 'AtomTypeEState("max", "ssSe")'}, 'MAXaaSe': {'function': mordred.EState.AtomTypeEState('max', 'aaSe'), 'type': '2D', 'description': 'max of aaSe', 'module': 'EState', 'constructor': 'AtomTypeEState("max", "aaSe")'}, 'MAXdssSe': {'function': mordred.EState.AtomTypeEState('max', 'dssSe'), 'type': '2D', 'description': 'max of dssSe', 'module': 'EState', 'constructor': 'AtomTypeEState("max", "dssSe")'}, 'MAXddssSe': {'function': mordred.EState.AtomTypeEState('max', 'ddssSe'), 'type': '2D', 'description': 'max of ddssSe', 'module': 'EState', 'constructor': 'AtomTypeEState("max", "ddssSe")'}, 'MAXsBr': {'function': mordred.EState.AtomTypeEState('max', 'sBr'), 'type': '2D', 'description': 'max of sBr', 'module': 'EState', 'constructor': 'AtomTypeEState("max", "sBr")'}, 'MAXsSnH3': {'function': mordred.EState.AtomTypeEState('max', 'sSnH3'), 'type': '2D', 'description': 'max of sSnH3', 'module': 'EState', 'constructor': 'AtomTypeEState("max", "sSnH3")'}, 'MAXssSnH2': {'function': mordred.EState.AtomTypeEState('max', 'ssSnH2'), 'type': '2D', 'description': 'max of ssSnH2', 'module': 'EState', 'constructor': 'AtomTypeEState("max", "ssSnH2")'}, 'MAXsssSnH': {'function': mordred.EState.AtomTypeEState('max', 'sssSnH'), 'type': '2D', 'description': 'max of sssSnH', 'module': 'EState', 'constructor': 'AtomTypeEState("max", "sssSnH")'}, 'MAXssssSn': {'function': mordred.EState.AtomTypeEState('max', 'ssssSn'), 'type': '2D', 'description': 'max of ssssSn', 'module': 'EState', 'constructor': 'AtomTypeEState("max", "ssssSn")'}, 'MAXsI': {'function': mordred.EState.AtomTypeEState('max', 'sI'), 'type': '2D', 'description': 'max of sI', 'module': 'EState', 'constructor': 'AtomTypeEState("max", "sI")'}, 'MAXsPbH3': {'function': mordred.EState.AtomTypeEState('max', 'sPbH3'), 'type': '2D', 'description': 'max of sPbH3', 'module': 'EState', 'constructor': 'AtomTypeEState("max", "sPbH3")'}, 'MAXssPbH2': {'function': mordred.EState.AtomTypeEState('max', 'ssPbH2'), 'type': '2D', 'description': 'max of ssPbH2', 'module': 'EState', 'constructor': 'AtomTypeEState("max", "ssPbH2")'}, 'MAXsssPbH': {'function': mordred.EState.AtomTypeEState('max', 'sssPbH'), 'type': '2D', 'description': 'max of sssPbH', 'module': 'EState', 'constructor': 'AtomTypeEState("max", "sssPbH")'}, 'MAXssssPb': {'function': mordred.EState.AtomTypeEState('max', 'ssssPb'), 'type': '2D', 'description': 'max of ssssPb', 'module': 'EState', 'constructor': 'AtomTypeEState("max", "ssssPb")'}, 'MINsLi': {'function': mordred.EState.AtomTypeEState('min', 'sLi'), 'type': '2D', 'description': 'min of sLi', 'module': 'EState', 'constructor': 'AtomTypeEState("min", "sLi")'}, 'MINssBe': {'function': mordred.EState.AtomTypeEState('min', 'ssBe'), 'type': '2D', 'description': 'min of ssBe', 'module': 'EState', 'constructor': 'AtomTypeEState("min", "ssBe")'}, 'MINssssBe': {'function': mordred.EState.AtomTypeEState('min', 'ssssBe'), 'type': '2D', 'description': 'min of ssssBe', 'module': 'EState', 'constructor': 'AtomTypeEState("min", "ssssBe")'}, 'MINssBH': {'function': mordred.EState.AtomTypeEState('min', 'ssBH'), 'type': '2D', 'description': 'min of ssBH', 'module': 'EState', 'constructor': 'AtomTypeEState("min", "ssBH")'}, 'MINsssB': {'function': mordred.EState.AtomTypeEState('min', 'sssB'), 'type': '2D', 'description': 'min of sssB', 'module': 'EState', 'constructor': 'AtomTypeEState("min", "sssB")'}, 'MINssssB': {'function': mordred.EState.AtomTypeEState('min', 'ssssB'), 'type': '2D', 'description': 'min of ssssB', 'module': 'EState', 'constructor': 'AtomTypeEState("min", "ssssB")'}, 'MINsCH3': {'function': mordred.EState.AtomTypeEState('min', 'sCH3'), 'type': '2D', 'description': 'min of sCH3', 'module': 'EState', 'constructor': 'AtomTypeEState("min", "sCH3")'}, 'MINdCH2': {'function': mordred.EState.AtomTypeEState('min', 'dCH2'), 'type': '2D', 'description': 'min of dCH2', 'module': 'EState', 'constructor': 'AtomTypeEState("min", "dCH2")'}, 'MINssCH2': {'function': mordred.EState.AtomTypeEState('min', 'ssCH2'), 'type': '2D', 'description': 'min of ssCH2', 'module': 'EState', 'constructor': 'AtomTypeEState("min", "ssCH2")'}, 'MINtCH': {'function': mordred.EState.AtomTypeEState('min', 'tCH'), 'type': '2D', 'description': 'min of tCH', 'module': 'EState', 'constructor': 'AtomTypeEState("min", "tCH")'}, 'MINdsCH': {'function': mordred.EState.AtomTypeEState('min', 'dsCH'), 'type': '2D', 'description': 'min of dsCH', 'module': 'EState', 'constructor': 'AtomTypeEState("min", "dsCH")'}, 'MINaaCH': {'function': mordred.EState.AtomTypeEState('min', 'aaCH'), 'type': '2D', 'description': 'min of aaCH', 'module': 'EState', 'constructor': 'AtomTypeEState("min", "aaCH")'}, 'MINsssCH': {'function': mordred.EState.AtomTypeEState('min', 'sssCH'), 'type': '2D', 'description': 'min of sssCH', 'module': 'EState', 'constructor': 'AtomTypeEState("min", "sssCH")'}, 'MINddC': {'function': mordred.EState.AtomTypeEState('min', 'ddC'), 'type': '2D', 'description': 'min of ddC', 'module': 'EState', 'constructor': 'AtomTypeEState("min", "ddC")'}, 'MINtsC': {'function': mordred.EState.AtomTypeEState('min', 'tsC'), 'type': '2D', 'description': 'min of tsC', 'module': 'EState', 'constructor': 'AtomTypeEState("min", "tsC")'}, 'MINdssC': {'function': mordred.EState.AtomTypeEState('min', 'dssC'), 'type': '2D', 'description': 'min of dssC', 'module': 'EState', 'constructor': 'AtomTypeEState("min", "dssC")'}, 'MINaasC': {'function': mordred.EState.AtomTypeEState('min', 'aasC'), 'type': '2D', 'description': 'min of aasC', 'module': 'EState', 'constructor': 'AtomTypeEState("min", "aasC")'}, 'MINaaaC': {'function': mordred.EState.AtomTypeEState('min', 'aaaC'), 'type': '2D', 'description': 'min of aaaC', 'module': 'EState', 'constructor': 'AtomTypeEState("min", "aaaC")'}, 'MINssssC': {'function': mordred.EState.AtomTypeEState('min', 'ssssC'), 'type': '2D', 'description': 'min of ssssC', 'module': 'EState', 'constructor': 'AtomTypeEState("min", "ssssC")'}, 'MINsNH3': {'function': mordred.EState.AtomTypeEState('min', 'sNH3'), 'type': '2D', 'description': 'min of sNH3', 'module': 'EState', 'constructor': 'AtomTypeEState("min", "sNH3")'}, 'MINsNH2': {'function': mordred.EState.AtomTypeEState('min', 'sNH2'), 'type': '2D', 'description': 'min of sNH2', 'module': 'EState', 'constructor': 'AtomTypeEState("min", "sNH2")'}, 'MINssNH2': {'function': mordred.EState.AtomTypeEState('min', 'ssNH2'), 'type': '2D', 'description': 'min of ssNH2', 'module': 'EState', 'constructor': 'AtomTypeEState("min", "ssNH2")'}, 'MINdNH': {'function': mordred.EState.AtomTypeEState('min', 'dNH'), 'type': '2D', 'description': 'min of dNH', 'module': 'EState', 'constructor': 'AtomTypeEState("min", "dNH")'}, 'MINssNH': {'function': mordred.EState.AtomTypeEState('min', 'ssNH'), 'type': '2D', 'description': 'min of ssNH', 'module': 'EState', 'constructor': 'AtomTypeEState("min", "ssNH")'}, 'MINaaNH': {'function': mordred.EState.AtomTypeEState('min', 'aaNH'), 'type': '2D', 'description': 'min of aaNH', 'module': 'EState', 'constructor': 'AtomTypeEState("min", "aaNH")'}, 'MINtN': {'function': mordred.EState.AtomTypeEState('min', 'tN'), 'type': '2D', 'description': 'min of tN', 'module': 'EState', 'constructor': 'AtomTypeEState("min", "tN")'}, 'MINsssNH': {'function': mordred.EState.AtomTypeEState('min', 'sssNH'), 'type': '2D', 'description': 'min of sssNH', 'module': 'EState', 'constructor': 'AtomTypeEState("min", "sssNH")'}, 'MINdsN': {'function': mordred.EState.AtomTypeEState('min', 'dsN'), 'type': '2D', 'description': 'min of dsN', 'module': 'EState', 'constructor': 'AtomTypeEState("min", "dsN")'}, 'MINaaN': {'function': mordred.EState.AtomTypeEState('min', 'aaN'), 'type': '2D', 'description': 'min of aaN', 'module': 'EState', 'constructor': 'AtomTypeEState("min", "aaN")'}, 'MINsssN': {'function': mordred.EState.AtomTypeEState('min', 'sssN'), 'type': '2D', 'description': 'min of sssN', 'module': 'EState', 'constructor': 'AtomTypeEState("min", "sssN")'}, 'MINddsN': {'function': mordred.EState.AtomTypeEState('min', 'ddsN'), 'type': '2D', 'description': 'min of ddsN', 'module': 'EState', 'constructor': 'AtomTypeEState("min", "ddsN")'}, 'MINaasN': {'function': mordred.EState.AtomTypeEState('min', 'aasN'), 'type': '2D', 'description': 'min of aasN', 'module': 'EState', 'constructor': 'AtomTypeEState("min", "aasN")'}, 'MINssssN': {'function': mordred.EState.AtomTypeEState('min', 'ssssN'), 'type': '2D', 'description': 'min of ssssN', 'module': 'EState', 'constructor': 'AtomTypeEState("min", "ssssN")'}, 'MINsOH': {'function': mordred.EState.AtomTypeEState('min', 'sOH'), 'type': '2D', 'description': 'min of sOH', 'module': 'EState', 'constructor': 'AtomTypeEState("min", "sOH")'}, 'MINdO': {'function': mordred.EState.AtomTypeEState('min', 'dO'), 'type': '2D', 'description': 'min of dO', 'module': 'EState', 'constructor': 'AtomTypeEState("min", "dO")'}, 'MINssO': {'function': mordred.EState.AtomTypeEState('min', 'ssO'), 'type': '2D', 'description': 'min of ssO', 'module': 'EState', 'constructor': 'AtomTypeEState("min", "ssO")'}, 'MINaaO': {'function': mordred.EState.AtomTypeEState('min', 'aaO'), 'type': '2D', 'description': 'min of aaO', 'module': 'EState', 'constructor': 'AtomTypeEState("min", "aaO")'}, 'MINsF': {'function': mordred.EState.AtomTypeEState('min', 'sF'), 'type': '2D', 'description': 'min of sF', 'module': 'EState', 'constructor': 'AtomTypeEState("min", "sF")'}, 'MINsSiH3': {'function': mordred.EState.AtomTypeEState('min', 'sSiH3'), 'type': '2D', 'description': 'min of sSiH3', 'module': 'EState', 'constructor': 'AtomTypeEState("min", "sSiH3")'}, 'MINssSiH2': {'function': mordred.EState.AtomTypeEState('min', 'ssSiH2'), 'type': '2D', 'description': 'min of ssSiH2', 'module': 'EState', 'constructor': 'AtomTypeEState("min", "ssSiH2")'}, 'MINsssSiH': {'function': mordred.EState.AtomTypeEState('min', 'sssSiH'), 'type': '2D', 'description': 'min of sssSiH', 'module': 'EState', 'constructor': 'AtomTypeEState("min", "sssSiH")'}, 'MINssssSi': {'function': mordred.EState.AtomTypeEState('min', 'ssssSi'), 'type': '2D', 'description': 'min of ssssSi', 'module': 'EState', 'constructor': 'AtomTypeEState("min", "ssssSi")'}, 'MINsPH2': {'function': mordred.EState.AtomTypeEState('min', 'sPH2'), 'type': '2D', 'description': 'min of sPH2', 'module': 'EState', 'constructor': 'AtomTypeEState("min", "sPH2")'}, 'MINssPH': {'function': mordred.EState.AtomTypeEState('min', 'ssPH'), 'type': '2D', 'description': 'min of ssPH', 'module': 'EState', 'constructor': 'AtomTypeEState("min", "ssPH")'}, 'MINsssP': {'function': mordred.EState.AtomTypeEState('min', 'sssP'), 'type': '2D', 'description': 'min of sssP', 'module': 'EState', 'constructor': 'AtomTypeEState("min", "sssP")'}, 'MINdsssP': {'function': mordred.EState.AtomTypeEState('min', 'dsssP'), 'type': '2D', 'description': 'min of dsssP', 'module': 'EState', 'constructor': 'AtomTypeEState("min", "dsssP")'}, 'MINsssssP': {'function': mordred.EState.AtomTypeEState('min', 'sssssP'), 'type': '2D', 'description': 'min of sssssP', 'module': 'EState', 'constructor': 'AtomTypeEState("min", "sssssP")'}, 'MINsSH': {'function': mordred.EState.AtomTypeEState('min', 'sSH'), 'type': '2D', 'description': 'min of sSH', 'module': 'EState', 'constructor': 'AtomTypeEState("min", "sSH")'}, 'MINdS': {'function': mordred.EState.AtomTypeEState('min', 'dS'), 'type': '2D', 'description': 'min of dS', 'module': 'EState', 'constructor': 'AtomTypeEState("min", "dS")'}, 'MINssS': {'function': mordred.EState.AtomTypeEState('min', 'ssS'), 'type': '2D', 'description': 'min of ssS', 'module': 'EState', 'constructor': 'AtomTypeEState("min", "ssS")'}, 'MINaaS': {'function': mordred.EState.AtomTypeEState('min', 'aaS'), 'type': '2D', 'description': 'min of aaS', 'module': 'EState', 'constructor': 'AtomTypeEState("min", "aaS")'}, 'MINdssS': {'function': mordred.EState.AtomTypeEState('min', 'dssS'), 'type': '2D', 'description': 'min of dssS', 'module': 'EState', 'constructor': 'AtomTypeEState("min", "dssS")'}, 'MINddssS': {'function': mordred.EState.AtomTypeEState('min', 'ddssS'), 'type': '2D', 'description': 'min of ddssS', 'module': 'EState', 'constructor': 'AtomTypeEState("min", "ddssS")'}, 'MINsCl': {'function': mordred.EState.AtomTypeEState('min', 'sCl'), 'type': '2D', 'description': 'min of sCl', 'module': 'EState', 'constructor': 'AtomTypeEState("min", "sCl")'}, 'MINsGeH3': {'function': mordred.EState.AtomTypeEState('min', 'sGeH3'), 'type': '2D', 'description': 'min of sGeH3', 'module': 'EState', 'constructor': 'AtomTypeEState("min", "sGeH3")'}, 'MINssGeH2': {'function': mordred.EState.AtomTypeEState('min', 'ssGeH2'), 'type': '2D', 'description': 'min of ssGeH2', 'module': 'EState', 'constructor': 'AtomTypeEState("min", "ssGeH2")'}, 'MINsssGeH': {'function': mordred.EState.AtomTypeEState('min', 'sssGeH'), 'type': '2D', 'description': 'min of sssGeH', 'module': 'EState', 'constructor': 'AtomTypeEState("min", "sssGeH")'}, 'MINssssGe': {'function': mordred.EState.AtomTypeEState('min', 'ssssGe'), 'type': '2D', 'description': 'min of ssssGe', 'module': 'EState', 'constructor': 'AtomTypeEState("min", "ssssGe")'}, 'MINsAsH2': {'function': mordred.EState.AtomTypeEState('min', 'sAsH2'), 'type': '2D', 'description': 'min of sAsH2', 'module': 'EState', 'constructor': 'AtomTypeEState("min", "sAsH2")'}, 'MINssAsH': {'function': mordred.EState.AtomTypeEState('min', 'ssAsH'), 'type': '2D', 'description': 'min of ssAsH', 'module': 'EState', 'constructor': 'AtomTypeEState("min", "ssAsH")'}, 'MINsssAs': {'function': mordred.EState.AtomTypeEState('min', 'sssAs'), 'type': '2D', 'description': 'min of sssAs', 'module': 'EState', 'constructor': 'AtomTypeEState("min", "sssAs")'}, 'MINsssdAs': {'function': mordred.EState.AtomTypeEState('min', 'sssdAs'), 'type': '2D', 'description': 'min of sssdAs', 'module': 'EState', 'constructor': 'AtomTypeEState("min", "sssdAs")'}, 'MINsssssAs': {'function': mordred.EState.AtomTypeEState('min', 'sssssAs'), 'type': '2D', 'description': 'min of sssssAs', 'module': 'EState', 'constructor': 'AtomTypeEState("min", "sssssAs")'}, 'MINsSeH': {'function': mordred.EState.AtomTypeEState('min', 'sSeH'), 'type': '2D', 'description': 'min of sSeH', 'module': 'EState', 'constructor': 'AtomTypeEState("min", "sSeH")'}, 'MINdSe': {'function': mordred.EState.AtomTypeEState('min', 'dSe'), 'type': '2D', 'description': 'min of dSe', 'module': 'EState', 'constructor': 'AtomTypeEState("min", "dSe")'}, 'MINssSe': {'function': mordred.EState.AtomTypeEState('min', 'ssSe'), 'type': '2D', 'description': 'min of ssSe', 'module': 'EState', 'constructor': 'AtomTypeEState("min", "ssSe")'}, 'MINaaSe': {'function': mordred.EState.AtomTypeEState('min', 'aaSe'), 'type': '2D', 'description': 'min of aaSe', 'module': 'EState', 'constructor': 'AtomTypeEState("min", "aaSe")'}, 'MINdssSe': {'function': mordred.EState.AtomTypeEState('min', 'dssSe'), 'type': '2D', 'description': 'min of dssSe', 'module': 'EState', 'constructor': 'AtomTypeEState("min", "dssSe")'}, 'MINddssSe': {'function': mordred.EState.AtomTypeEState('min', 'ddssSe'), 'type': '2D', 'description': 'min of ddssSe', 'module': 'EState', 'constructor': 'AtomTypeEState("min", "ddssSe")'}, 'MINsBr': {'function': mordred.EState.AtomTypeEState('min', 'sBr'), 'type': '2D', 'description': 'min of sBr', 'module': 'EState', 'constructor': 'AtomTypeEState("min", "sBr")'}, 'MINsSnH3': {'function': mordred.EState.AtomTypeEState('min', 'sSnH3'), 'type': '2D', 'description': 'min of sSnH3', 'module': 'EState', 'constructor': 'AtomTypeEState("min", "sSnH3")'}, 'MINssSnH2': {'function': mordred.EState.AtomTypeEState('min', 'ssSnH2'), 'type': '2D', 'description': 'min of ssSnH2', 'module': 'EState', 'constructor': 'AtomTypeEState("min", "ssSnH2")'}, 'MINsssSnH': {'function': mordred.EState.AtomTypeEState('min', 'sssSnH'), 'type': '2D', 'description': 'min of sssSnH', 'module': 'EState', 'constructor': 'AtomTypeEState("min", "sssSnH")'}, 'MINssssSn': {'function': mordred.EState.AtomTypeEState('min', 'ssssSn'), 'type': '2D', 'description': 'min of ssssSn', 'module': 'EState', 'constructor': 'AtomTypeEState("min", "ssssSn")'}, 'MINsI': {'function': mordred.EState.AtomTypeEState('min', 'sI'), 'type': '2D', 'description': 'min of sI', 'module': 'EState', 'constructor': 'AtomTypeEState("min", "sI")'}, 'MINsPbH3': {'function': mordred.EState.AtomTypeEState('min', 'sPbH3'), 'type': '2D', 'description': 'min of sPbH3', 'module': 'EState', 'constructor': 'AtomTypeEState("min", "sPbH3")'}, 'MINssPbH2': {'function': mordred.EState.AtomTypeEState('min', 'ssPbH2'), 'type': '2D', 'description': 'min of ssPbH2', 'module': 'EState', 'constructor': 'AtomTypeEState("min", "ssPbH2")'}, 'MINsssPbH': {'function': mordred.EState.AtomTypeEState('min', 'sssPbH'), 'type': '2D', 'description': 'min of sssPbH', 'module': 'EState', 'constructor': 'AtomTypeEState("min", "sssPbH")'}, 'MINssssPb': {'function': mordred.EState.AtomTypeEState('min', 'ssssPb'), 'type': '2D', 'description': 'min of ssssPb', 'module': 'EState', 'constructor': 'AtomTypeEState("min", "ssssPb")'}, 'ECIndex': {'function': mordred.EccentricConnectivityIndex.EccentricConnectivityIndex(), 'type': '2D', 'description': 'eccentric connectivity index', 'module': 'EccentricConnectivityIndex', 'constructor': 'EccentricConnectivityIndex()'}, 'ETA_alpha': {'function': mordred.ExtendedTopochemicalAtom.EtaCoreCount(False, False), 'type': '2D', 'description': 'ETA core count', 'module': 'ExtendedTopochemicalAtom', 'constructor': 'EtaCoreCount(False, False)'}, 'AETA_alpha': {'function': mordred.ExtendedTopochemicalAtom.EtaCoreCount(True, False), 'type': '2D', 'description': 'averaged ETA core count', 'module': 'ExtendedTopochemicalAtom', 'constructor': 'EtaCoreCount(True, False)'}, 'ETA_shape_p': {'function': mordred.ExtendedTopochemicalAtom.EtaShapeIndex('p'), 'type': '2D', 'description': 'ETA shape index(type: p)', 'module': 'ExtendedTopochemicalAtom', 'constructor': 'EtaShapeIndex("p")'}, 'ETA_shape_y': {'function': mordred.ExtendedTopochemicalAtom.EtaShapeIndex('y'), 'type': '2D', 'description': 'ETA shape index(type: y)', 'module': 'ExtendedTopochemicalAtom', 'constructor': 'EtaShapeIndex("y")'}, 'ETA_shape_x': {'function': mordred.ExtendedTopochemicalAtom.EtaShapeIndex('x'), 'type': '2D', 'description': 'ETA shape index(type: x)', 'module': 'ExtendedTopochemicalAtom', 'constructor': 'EtaShapeIndex("x")'}, 'ETA_beta': {'function': mordred.ExtendedTopochemicalAtom.EtaVEMCount('', False), 'type': '2D', 'description': 'valence electron mobile count', 'module': 'ExtendedTopochemicalAtom', 'constructor': 'EtaVEMCount("", False)'}, 'AETA_beta': {'function': mordred.ExtendedTopochemicalAtom.EtaVEMCount('', True), 'type': '2D', 'description': 'averaged valence electron mobile count', 'module': 'ExtendedTopochemicalAtom', 'constructor': 'EtaVEMCount("", True)'}, 'ETA_beta_s': {'function': mordred.ExtendedTopochemicalAtom.EtaVEMCount('s', False), 'type': '2D', 'description': 'sigma contribution to valence electron mobile count', 'module': 'ExtendedTopochemicalAtom', 'constructor': 'EtaVEMCount("s", False)'}, 'AETA_beta_s': {'function': mordred.ExtendedTopochemicalAtom.EtaVEMCount('s', True), 'type': '2D', 'description': 'averaged sigma contribution to valence electron mobile count', 'module': 'ExtendedTopochemicalAtom', 'constructor': 'EtaVEMCount("s", True)'}, 'ETA_beta_ns': {'function': mordred.ExtendedTopochemicalAtom.EtaVEMCount('ns', False), 'type': '2D', 'description': 'nonsigma contribution to valence electron mobile count', 'module': 'ExtendedTopochemicalAtom', 'constructor': 'EtaVEMCount("ns", False)'}, 'AETA_beta_ns': {'function': mordred.ExtendedTopochemicalAtom.EtaVEMCount('ns', True), 'type': '2D', 'description': 'averaged nonsigma contribution to valence electron mobile count', 'module': 'ExtendedTopochemicalAtom', 'constructor': 'EtaVEMCount("ns", True)'}, 'ETA_beta_ns_d': {'function': mordred.ExtendedTopochemicalAtom.EtaVEMCount('ns_d', False), 'type': '2D', 'description': 'delta contribution to valence electron mobile count', 'module': 'ExtendedTopochemicalAtom', 'constructor': 'EtaVEMCount("ns_d", False)'}, 'AETA_beta_ns_d': {'function': mordred.ExtendedTopochemicalAtom.EtaVEMCount('ns_d', True), 'type': '2D', 'description': 'averaged delta contribution to valence electron mobile count', 'module': 'ExtendedTopochemicalAtom', 'constructor': 'EtaVEMCount("ns_d", True)'}, 'ETA_eta': {'function': mordred.ExtendedTopochemicalAtom.EtaCompositeIndex(False, False, False), 'type': '2D', 'description': 'ETA composite index for reference graph', 'module': 'ExtendedTopochemicalAtom', 'constructor': 'EtaCompositeIndex(False, False, False)'}, 'AETA_eta': {'function': mordred.ExtendedTopochemicalAtom.EtaCompositeIndex(False, False, True), 'type': '2D', 'description': 'averaged ETA composite index for reference graph', 'module': 'ExtendedTopochemicalAtom', 'constructor': 'EtaCompositeIndex(False, False, True)'}, 'ETA_eta_L': {'function': mordred.ExtendedTopochemicalAtom.EtaCompositeIndex(False, True, False), 'type': '2D', 'description': 'local ETA composite index for reference graph', 'module': 'ExtendedTopochemicalAtom', 'constructor': 'EtaCompositeIndex(False, True, False)'}, 'AETA_eta_L': {'function': mordred.ExtendedTopochemicalAtom.EtaCompositeIndex(False, True, True), 'type': '2D', 'description': 'averaged local ETA composite index for reference graph', 'module': 'ExtendedTopochemicalAtom', 'constructor': 'EtaCompositeIndex(False, True, True)'}, 'ETA_eta_R': {'function': mordred.ExtendedTopochemicalAtom.EtaCompositeIndex(True, False, False), 'type': '2D', 'description': 'ETA composite index for reference graph', 'module': 'ExtendedTopochemicalAtom', 'constructor': 'EtaCompositeIndex(True, False, False)'}, 'AETA_eta_R': {'function': mordred.ExtendedTopochemicalAtom.EtaCompositeIndex(True, False, True), 'type': '2D', 'description': 'averaged ETA composite index for reference graph', 'module': 'ExtendedTopochemicalAtom', 'constructor': 'EtaCompositeIndex(True, False, True)'}, 'ETA_eta_RL': {'function': mordred.ExtendedTopochemicalAtom.EtaCompositeIndex(True, True, False), 'type': '2D', 'description': 'local ETA composite index for reference graph', 'module': 'ExtendedTopochemicalAtom', 'constructor': 'EtaCompositeIndex(True, True, False)'}, 'AETA_eta_RL': {'function': mordred.ExtendedTopochemicalAtom.EtaCompositeIndex(True, True, True), 'type': '2D', 'description': 'averaged local ETA composite index for reference graph', 'module': 'ExtendedTopochemicalAtom', 'constructor': 'EtaCompositeIndex(True, True, True)'}, 'ETA_eta_F': {'function': mordred.ExtendedTopochemicalAtom.EtaFunctionalityIndex(False, False), 'type': '2D', 'description': 'ETA functionality index', 'module': 'ExtendedTopochemicalAtom', 'constructor': 'EtaFunctionalityIndex(False, False)'}, 'AETA_eta_F': {'function': mordred.ExtendedTopochemicalAtom.EtaFunctionalityIndex(False, True), 'type': '2D', 'description': 'averaged ETA functionality index', 'module': 'ExtendedTopochemicalAtom', 'constructor': 'EtaFunctionalityIndex(False, True)'}, 'ETA_eta_FL': {'function': mordred.ExtendedTopochemicalAtom.EtaFunctionalityIndex(True, False), 'type': '2D', 'description': 'local ETA functionality index', 'module': 'ExtendedTopochemicalAtom', 'constructor': 'EtaFunctionalityIndex(True, False)'}, 'AETA_eta_FL': {'function': mordred.ExtendedTopochemicalAtom.EtaFunctionalityIndex(True, True), 'type': '2D', 'description': 'averaged local ETA functionality index', 'module': 'ExtendedTopochemicalAtom', 'constructor': 'EtaFunctionalityIndex(True, True)'}, 'ETA_eta_B': {'function': mordred.ExtendedTopochemicalAtom.EtaBranchingIndex(False, False), 'type': '2D', 'description': 'ETA branching index', 'module': 'ExtendedTopochemicalAtom', 'constructor': 'EtaBranchingIndex(False, False)'}, 'AETA_eta_B': {'function': mordred.ExtendedTopochemicalAtom.EtaBranchingIndex(False, True), 'type': '2D', 'description': 'averaged ETA branching index', 'module': 'ExtendedTopochemicalAtom', 'constructor': 'EtaBranchingIndex(False, True)'}, 'ETA_eta_BR': {'function': mordred.ExtendedTopochemicalAtom.EtaBranchingIndex(True, False), 'type': '2D', 'description': 'ETA branching index(use ring count)', 'module': 'ExtendedTopochemicalAtom', 'constructor': 'EtaBranchingIndex(True, False)'}, 'AETA_eta_BR': {'function': mordred.ExtendedTopochemicalAtom.EtaBranchingIndex(True, True), 'type': '2D', 'description': 'averaged ETA branching index(use ring count)', 'module': 'ExtendedTopochemicalAtom', 'constructor': 'EtaBranchingIndex(True, True)'}, 'ETA_dAlpha_A': {'function': mordred.ExtendedTopochemicalAtom.EtaDeltaAlpha('A'), 'type': '2D', 'description': 'ETA delta alpha(type: A)', 'module': 'ExtendedTopochemicalAtom', 'constructor': 'EtaDeltaAlpha("A")'}, 'ETA_dAlpha_B': {'function': mordred.ExtendedTopochemicalAtom.EtaDeltaAlpha('B'), 'type': '2D', 'description': 'ETA delta alpha(type: B)', 'module': 'ExtendedTopochemicalAtom', 'constructor': 'EtaDeltaAlpha("B")'}, 'ETA_epsilon_1': {'function': mordred.ExtendedTopochemicalAtom.EtaEpsilon(1), 'type': '2D', 'description': 'ETA epsilon(type: 1)', 'module': 'ExtendedTopochemicalAtom', 'constructor': 'EtaEpsilon(1)'}, 'ETA_epsilon_2': {'function': mordred.ExtendedTopochemicalAtom.EtaEpsilon(2), 'type': '2D', 'description': 'ETA epsilon(type: 2)', 'module': 'ExtendedTopochemicalAtom', 'constructor': 'EtaEpsilon(2)'}, 'ETA_epsilon_3': {'function': mordred.ExtendedTopochemicalAtom.EtaEpsilon(3), 'type': '2D', 'description': 'ETA epsilon(type: 3)', 'module': 'ExtendedTopochemicalAtom', 'constructor': 'EtaEpsilon(3)'}, 'ETA_epsilon_4': {'function': mordred.ExtendedTopochemicalAtom.EtaEpsilon(4), 'type': '2D', 'description': 'ETA epsilon(type: 4)', 'module': 'ExtendedTopochemicalAtom', 'constructor': 'EtaEpsilon(4)'}, 'ETA_epsilon_5': {'function': mordred.ExtendedTopochemicalAtom.EtaEpsilon(5), 'type': '2D', 'description': 'ETA epsilon(type: 5)', 'module': 'ExtendedTopochemicalAtom', 'constructor': 'EtaEpsilon(5)'}, 'ETA_dEpsilon_A': {'function': mordred.ExtendedTopochemicalAtom.EtaDeltaEpsilon('A'), 'type': '2D', 'description': 'ETA delta epsilon(type: A)', 'module': 'ExtendedTopochemicalAtom', 'constructor': 'EtaDeltaEpsilon("A")'}, 'ETA_dEpsilon_B': {'function': mordred.ExtendedTopochemicalAtom.EtaDeltaEpsilon('B'), 'type': '2D', 'description': 'ETA delta epsilon(type: B)', 'module': 'ExtendedTopochemicalAtom', 'constructor': 'EtaDeltaEpsilon("B")'}, 'ETA_dEpsilon_C': {'function': mordred.ExtendedTopochemicalAtom.EtaDeltaEpsilon('C'), 'type': '2D', 'description': 'ETA delta epsilon(type: C)', 'module': 'ExtendedTopochemicalAtom', 'constructor': 'EtaDeltaEpsilon("C")'}, 'ETA_dEpsilon_D': {'function': mordred.ExtendedTopochemicalAtom.EtaDeltaEpsilon('D'), 'type': '2D', 'description': 'ETA delta epsilon(type: D)', 'module': 'ExtendedTopochemicalAtom', 'constructor': 'EtaDeltaEpsilon("D")'}, 'ETA_dBeta': {'function': mordred.ExtendedTopochemicalAtom.EtaDeltaBeta(False), 'type': '2D', 'description': 'ETA delta beta', 'module': 'ExtendedTopochemicalAtom', 'constructor': 'EtaDeltaBeta(False)'}, 'AETA_dBeta': {'function': mordred.ExtendedTopochemicalAtom.EtaDeltaBeta(True), 'type': '2D', 'description': 'averaged ETA delta beta', 'module': 'ExtendedTopochemicalAtom', 'constructor': 'EtaDeltaBeta(True)'}, 'ETA_psi_1': {'function': mordred.ExtendedTopochemicalAtom.EtaPsi(), 'type': '2D', 'description': 'ETA psi', 'module': 'ExtendedTopochemicalAtom', 'constructor': 'EtaPsi()'}, 'ETA_dPsi_A': {'function': mordred.ExtendedTopochemicalAtom.EtaDeltaPsi('A'), 'type': '2D', 'description': 'ETA delta psi(type: A)', 'module': 'ExtendedTopochemicalAtom', 'constructor': 'EtaDeltaPsi("A")'}, 'ETA_dPsi_B': {'function': mordred.ExtendedTopochemicalAtom.EtaDeltaPsi('B'), 'type': '2D', 'description': 'ETA delta psi(type: B)', 'module': 'ExtendedTopochemicalAtom', 'constructor': 'EtaDeltaPsi("B")'}, 'fragCpx': {'function': mordred.FragmentComplexity.FragmentComplexity(), 'type': '2D', 'description': 'fragment complexity', 'module': 'FragmentComplexity', 'constructor': 'FragmentComplexity()'}, 'fMF': {'function': mordred.Framework.Framework(), 'type': '2D', 'description': 'molecular framework ratio', 'module': 'Framework', 'constructor': 'Framework()'}, 'GeomDiameter': {'function': mordred.GeometricalIndex.Diameter3D(), 'type': '3D', 'description': 'geometric diameter', 'module': 'GeometricalIndex', 'constructor': 'Diameter3D()'}, 'GeomRadius': {'function': mordred.GeometricalIndex.Radius3D(), 'type': '3D', 'description': 'geometric radius', 'module': 'GeometricalIndex', 'constructor': 'Radius3D()'}, 'GeomShapeIndex': {'function': mordred.GeometricalIndex.GeometricalShapeIndex(), 'type': '3D', 'description': 'geometrical shape index', 'module': 'GeometricalIndex', 'constructor': 'GeometricalShapeIndex()'}, 'GeomPetitjeanIndex': {'function': mordred.GeometricalIndex.PetitjeanIndex3D(), 'type': '3D', 'description': 'geometric Petitjean index', 'module': 'GeometricalIndex', 'constructor': 'PetitjeanIndex3D()'}, 'GRAV': {'function': mordred.GravitationalIndex.GravitationalIndex(True, False), 'type': '3D', 'description': 'heavy atom gravitational index', 'module': 'GravitationalIndex', 'constructor': 'GravitationalIndex(True, False)'}, 'GRAVH': {'function': mordred.GravitationalIndex.GravitationalIndex(False, False), 'type': '3D', 'description': 'gravitational index', 'module': 'GravitationalIndex', 'constructor': 'GravitationalIndex(False, False)'}, 'GRAVp': {'function': mordred.GravitationalIndex.GravitationalIndex(True, True), 'type': '3D', 'description': 'heavy atom pair gravitational index', 'module': 'GravitationalIndex', 'constructor': 'GravitationalIndex(True, True)'}, 'GRAVHp': {'function': mordred.GravitationalIndex.GravitationalIndex(False, True), 'type': '3D', 'description': 'pair gravitational index', 'module': 'GravitationalIndex', 'constructor': 'GravitationalIndex(False, True)'}, 'nHBAcc': {'function': mordred.HydrogenBond.HBondAcceptor(), 'type': '2D', 'description': 'number of hydrogen bond acceptor', 'module': 'HydrogenBond', 'constructor': 'HBondAcceptor()'}, 'nHBDon': {'function': mordred.HydrogenBond.HBondDonor(), 'type': '2D', 'description': 'number of hydrogen bond donor', 'module': 'HydrogenBond', 'constructor': 'HBondDonor()'}, 'IC0': {'function': mordred.InformationContent.InformationContent(0), 'type': '2D', 'description': '0-ordered neighborhood information content', 'module': 'InformationContent', 'constructor': 'InformationContent(0)'}, 'IC1': {'function': mordred.InformationContent.InformationContent(1), 'type': '2D', 'description': '1-ordered neighborhood information content', 'module': 'InformationContent', 'constructor': 'InformationContent(1)'}, 'IC2': {'function': mordred.InformationContent.InformationContent(2), 'type': '2D', 'description': '2-ordered neighborhood information content', 'module': 'InformationContent', 'constructor': 'InformationContent(2)'}, 'IC3': {'function': mordred.InformationContent.InformationContent(3), 'type': '2D', 'description': '3-ordered neighborhood information content', 'module': 'InformationContent', 'constructor': 'InformationContent(3)'}, 'IC4': {'function': mordred.InformationContent.InformationContent(4), 'type': '2D', 'description': '4-ordered neighborhood information content', 'module': 'InformationContent', 'constructor': 'InformationContent(4)'}, 'IC5': {'function': mordred.InformationContent.InformationContent(5), 'type': '2D', 'description': '5-ordered neighborhood information content', 'module': 'InformationContent', 'constructor': 'InformationContent(5)'}, 'TIC0': {'function': mordred.InformationContent.TotalIC(0), 'type': '2D', 'description': '0-ordered neighborhood total information content', 'module': 'InformationContent', 'constructor': 'TotalIC(0)'}, 'TIC1': {'function': mordred.InformationContent.TotalIC(1), 'type': '2D', 'description': '1-ordered neighborhood total information content', 'module': 'InformationContent', 'constructor': 'TotalIC(1)'}, 'TIC2': {'function': mordred.InformationContent.TotalIC(2), 'type': '2D', 'description': '2-ordered neighborhood total information content', 'module': 'InformationContent', 'constructor': 'TotalIC(2)'}, 'TIC3': {'function': mordred.InformationContent.TotalIC(3), 'type': '2D', 'description': '3-ordered neighborhood total information content', 'module': 'InformationContent', 'constructor': 'TotalIC(3)'}, 'TIC4': {'function': mordred.InformationContent.TotalIC(4), 'type': '2D', 'description': '4-ordered neighborhood total information content', 'module': 'InformationContent', 'constructor': 'TotalIC(4)'}, 'TIC5': {'function': mordred.InformationContent.TotalIC(5), 'type': '2D', 'description': '5-ordered neighborhood total information content', 'module': 'InformationContent', 'constructor': 'TotalIC(5)'}, 'SIC0': {'function': mordred.InformationContent.StructuralIC(0), 'type': '2D', 'description': '0-ordered structural information content', 'module': 'InformationContent', 'constructor': 'StructuralIC(0)'}, 'SIC1': {'function': mordred.InformationContent.StructuralIC(1), 'type': '2D', 'description': '1-ordered structural information content', 'module': 'InformationContent', 'constructor': 'StructuralIC(1)'}, 'SIC2': {'function': mordred.InformationContent.StructuralIC(2), 'type': '2D', 'description': '2-ordered structural information content', 'module': 'InformationContent', 'constructor': 'StructuralIC(2)'}, 'SIC3': {'function': mordred.InformationContent.StructuralIC(3), 'type': '2D', 'description': '3-ordered structural information content', 'module': 'InformationContent', 'constructor': 'StructuralIC(3)'}, 'SIC4': {'function': mordred.InformationContent.StructuralIC(4), 'type': '2D', 'description': '4-ordered structural information content', 'module': 'InformationContent', 'constructor': 'StructuralIC(4)'}, 'SIC5': {'function': mordred.InformationContent.StructuralIC(5), 'type': '2D', 'description': '5-ordered structural information content', 'module': 'InformationContent', 'constructor': 'StructuralIC(5)'}, 'BIC0': {'function': mordred.InformationContent.BondingIC(0), 'type': '2D', 'description': '0-ordered bonding information content', 'module': 'InformationContent', 'constructor': 'BondingIC(0)'}, 'BIC1': {'function': mordred.InformationContent.BondingIC(1), 'type': '2D', 'description': '1-ordered bonding information content', 'module': 'InformationContent', 'constructor': 'BondingIC(1)'}, 'BIC2': {'function': mordred.InformationContent.BondingIC(2), 'type': '2D', 'description': '2-ordered bonding information content', 'module': 'InformationContent', 'constructor': 'BondingIC(2)'}, 'BIC3': {'function': mordred.InformationContent.BondingIC(3), 'type': '2D', 'description': '3-ordered bonding information content', 'module': 'InformationContent', 'constructor': 'BondingIC(3)'}, 'BIC4': {'function': mordred.InformationContent.BondingIC(4), 'type': '2D', 'description': '4-ordered bonding information content', 'module': 'InformationContent', 'constructor': 'BondingIC(4)'}, 'BIC5': {'function': mordred.InformationContent.BondingIC(5), 'type': '2D', 'description': '5-ordered bonding information content', 'module': 'InformationContent', 'constructor': 'BondingIC(5)'}, 'CIC0': {'function': mordred.InformationContent.ComplementaryIC(0), 'type': '2D', 'description': '0-ordered complementary information content', 'module': 'InformationContent', 'constructor': 'ComplementaryIC(0)'}, 'CIC1': {'function': mordred.InformationContent.ComplementaryIC(1), 'type': '2D', 'description': '1-ordered complementary information content', 'module': 'InformationContent', 'constructor': 'ComplementaryIC(1)'}, 'CIC2': {'function': mordred.InformationContent.ComplementaryIC(2), 'type': '2D', 'description': '2-ordered complementary information content', 'module': 'InformationContent', 'constructor': 'ComplementaryIC(2)'}, 'CIC3': {'function': mordred.InformationContent.ComplementaryIC(3), 'type': '2D', 'description': '3-ordered complementary information content', 'module': 'InformationContent', 'constructor': 'ComplementaryIC(3)'}, 'CIC4': {'function': mordred.InformationContent.ComplementaryIC(4), 'type': '2D', 'description': '4-ordered complementary information content', 'module': 'InformationContent', 'constructor': 'ComplementaryIC(4)'}, 'CIC5': {'function': mordred.InformationContent.ComplementaryIC(5), 'type': '2D', 'description': '5-ordered complementary information content', 'module': 'InformationContent', 'constructor': 'ComplementaryIC(5)'}, 'MIC0': {'function': mordred.InformationContent.ModifiedIC(0), 'type': '2D', 'description': '0-ordered modified information content', 'module': 'InformationContent', 'constructor': 'ModifiedIC(0)'}, 'MIC1': {'function': mordred.InformationContent.ModifiedIC(1), 'type': '2D', 'description': '1-ordered modified information content', 'module': 'InformationContent', 'constructor': 'ModifiedIC(1)'}, 'MIC2': {'function': mordred.InformationContent.ModifiedIC(2), 'type': '2D', 'description': '2-ordered modified information content', 'module': 'InformationContent', 'constructor': 'ModifiedIC(2)'}, 'MIC3': {'function': mordred.InformationContent.ModifiedIC(3), 'type': '2D', 'description': '3-ordered modified information content', 'module': 'InformationContent', 'constructor': 'ModifiedIC(3)'}, 'MIC4': {'function': mordred.InformationContent.ModifiedIC(4), 'type': '2D', 'description': '4-ordered modified information content', 'module': 'InformationContent', 'constructor': 'ModifiedIC(4)'}, 'MIC5': {'function': mordred.InformationContent.ModifiedIC(5), 'type': '2D', 'description': '5-ordered modified information content', 'module': 'InformationContent', 'constructor': 'ModifiedIC(5)'}, 'ZMIC0': {'function': mordred.InformationContent.ZModifiedIC(0), 'type': '2D', 'description': '0-ordered Z-modified information content', 'module': 'InformationContent', 'constructor': 'ZModifiedIC(0)'}, 'ZMIC1': {'function': mordred.InformationContent.ZModifiedIC(1), 'type': '2D', 'description': '1-ordered Z-modified information content', 'module': 'InformationContent', 'constructor': 'ZModifiedIC(1)'}, 'ZMIC2': {'function': mordred.InformationContent.ZModifiedIC(2), 'type': '2D', 'description': '2-ordered Z-modified information content', 'module': 'InformationContent', 'constructor': 'ZModifiedIC(2)'}, 'ZMIC3': {'function': mordred.InformationContent.ZModifiedIC(3), 'type': '2D', 'description': '3-ordered Z-modified information content', 'module': 'InformationContent', 'constructor': 'ZModifiedIC(3)'}, 'ZMIC4': {'function': mordred.InformationContent.ZModifiedIC(4), 'type': '2D', 'description': '4-ordered Z-modified information content', 'module': 'InformationContent', 'constructor': 'ZModifiedIC(4)'}, 'ZMIC5': {'function': mordred.InformationContent.ZModifiedIC(5), 'type': '2D', 'description': '5-ordered Z-modified information content', 'module': 'InformationContent', 'constructor': 'ZModifiedIC(5)'}, 'Kier1': {'function': mordred.KappaShapeIndex.KappaShapeIndex1(), 'type': '2D', 'description': 'kappa shape index 1', 'module': 'KappaShapeIndex', 'constructor': 'KappaShapeIndex1()'}, 'Kier2': {'function': mordred.KappaShapeIndex.KappaShapeIndex2(), 'type': '2D', 'description': 'kappa shape index 2', 'module': 'KappaShapeIndex', 'constructor': 'KappaShapeIndex2()'}, 'Kier3': {'function': mordred.KappaShapeIndex.KappaShapeIndex3(), 'type': '2D', 'description': 'kappa shape index 3', 'module': 'KappaShapeIndex', 'constructor': 'KappaShapeIndex3()'}, 'Lipinski': {'function': mordred.Lipinski.Lipinski(), 'type': '2D', 'description': 'Lipinski rule of five', 'module': 'Lipinski', 'constructor': 'Lipinski()'}, 'GhoseFilter': {'function': mordred.Lipinski.GhoseFilter(), 'type': '2D', 'description': 'Ghose filter', 'module': 'Lipinski', 'constructor': 'GhoseFilter()'}, 'FilterItLogS': {'function': mordred.LogS.LogS(), 'type': '2D', 'description': 'Filter-it™ LogS', 'module': 'LogS', 'constructor': 'LogS()'}, 'VMcGowan': {'function': mordred.McGowanVolume.McGowanVolume(), 'type': '2D', 'description': 'McGowan volume', 'module': 'McGowanVolume', 'constructor': 'McGowanVolume()'}, 'Mor01': {'function': mordred.MoRSE.MoRSE(None, 1), 'type': '3D', 'description': '3D-MoRSE(distance = 1)', 'module': 'MoRSE', 'constructor': 'MoRSE(None, 1)'}, 'Mor02': {'function': mordred.MoRSE.MoRSE(None, 2), 'type': '3D', 'description': '3D-MoRSE(distance = 2)', 'module': 'MoRSE', 'constructor': 'MoRSE(None, 2)'}, 'Mor03': {'function': mordred.MoRSE.MoRSE(None, 3), 'type': '3D', 'description': '3D-MoRSE(distance = 3)', 'module': 'MoRSE', 'constructor': 'MoRSE(None, 3)'}, 'Mor04': {'function': mordred.MoRSE.MoRSE(None, 4), 'type': '3D', 'description': '3D-MoRSE(distance = 4)', 'module': 'MoRSE', 'constructor': 'MoRSE(None, 4)'}, 'Mor05': {'function': mordred.MoRSE.MoRSE(None, 5), 'type': '3D', 'description': '3D-MoRSE(distance = 5)', 'module': 'MoRSE', 'constructor': 'MoRSE(None, 5)'}, 'Mor06': {'function': mordred.MoRSE.MoRSE(None, 6), 'type': '3D', 'description': '3D-MoRSE(distance = 6)', 'module': 'MoRSE', 'constructor': 'MoRSE(None, 6)'}, 'Mor07': {'function': mordred.MoRSE.MoRSE(None, 7), 'type': '3D', 'description': '3D-MoRSE(distance = 7)', 'module': 'MoRSE', 'constructor': 'MoRSE(None, 7)'}, 'Mor08': {'function': mordred.MoRSE.MoRSE(None, 8), 'type': '3D', 'description': '3D-MoRSE(distance = 8)', 'module': 'MoRSE', 'constructor': 'MoRSE(None, 8)'}, 'Mor09': {'function': mordred.MoRSE.MoRSE(None, 9), 'type': '3D', 'description': '3D-MoRSE(distance = 9)', 'module': 'MoRSE', 'constructor': 'MoRSE(None, 9)'}, 'Mor10': {'function': mordred.MoRSE.MoRSE(None, 10), 'type': '3D', 'description': '3D-MoRSE(distance = 10)', 'module': 'MoRSE', 'constructor': 'MoRSE(None, 10)'}, 'Mor11': {'function': mordred.MoRSE.MoRSE(None, 11), 'type': '3D', 'description': '3D-MoRSE(distance = 11)', 'module': 'MoRSE', 'constructor': 'MoRSE(None, 11)'}, 'Mor12': {'function': mordred.MoRSE.MoRSE(None, 12), 'type': '3D', 'description': '3D-MoRSE(distance = 12)', 'module': 'MoRSE', 'constructor': 'MoRSE(None, 12)'}, 'Mor13': {'function': mordred.MoRSE.MoRSE(None, 13), 'type': '3D', 'description': '3D-MoRSE(distance = 13)', 'module': 'MoRSE', 'constructor': 'MoRSE(None, 13)'}, 'Mor14': {'function': mordred.MoRSE.MoRSE(None, 14), 'type': '3D', 'description': '3D-MoRSE(distance = 14)', 'module': 'MoRSE', 'constructor': 'MoRSE(None, 14)'}, 'Mor15': {'function': mordred.MoRSE.MoRSE(None, 15), 'type': '3D', 'description': '3D-MoRSE(distance = 15)', 'module': 'MoRSE', 'constructor': 'MoRSE(None, 15)'}, 'Mor16': {'function': mordred.MoRSE.MoRSE(None, 16), 'type': '3D', 'description': '3D-MoRSE(distance = 16)', 'module': 'MoRSE', 'constructor': 'MoRSE(None, 16)'}, 'Mor17': {'function': mordred.MoRSE.MoRSE(None, 17), 'type': '3D', 'description': '3D-MoRSE(distance = 17)', 'module': 'MoRSE', 'constructor': 'MoRSE(None, 17)'}, 'Mor18': {'function': mordred.MoRSE.MoRSE(None, 18), 'type': '3D', 'description': '3D-MoRSE(distance = 18)', 'module': 'MoRSE', 'constructor': 'MoRSE(None, 18)'}, 'Mor19': {'function': mordred.MoRSE.MoRSE(None, 19), 'type': '3D', 'description': '3D-MoRSE(distance = 19)', 'module': 'MoRSE', 'constructor': 'MoRSE(None, 19)'}, 'Mor20': {'function': mordred.MoRSE.MoRSE(None, 20), 'type': '3D', 'description': '3D-MoRSE(distance = 20)', 'module': 'MoRSE', 'constructor': 'MoRSE(None, 20)'}, 'Mor21': {'function': mordred.MoRSE.MoRSE(None, 21), 'type': '3D', 'description': '3D-MoRSE(distance = 21)', 'module': 'MoRSE', 'constructor': 'MoRSE(None, 21)'}, 'Mor22': {'function': mordred.MoRSE.MoRSE(None, 22), 'type': '3D', 'description': '3D-MoRSE(distance = 22)', 'module': 'MoRSE', 'constructor': 'MoRSE(None, 22)'}, 'Mor23': {'function': mordred.MoRSE.MoRSE(None, 23), 'type': '3D', 'description': '3D-MoRSE(distance = 23)', 'module': 'MoRSE', 'constructor': 'MoRSE(None, 23)'}, 'Mor24': {'function': mordred.MoRSE.MoRSE(None, 24), 'type': '3D', 'description': '3D-MoRSE(distance = 24)', 'module': 'MoRSE', 'constructor': 'MoRSE(None, 24)'}, 'Mor25': {'function': mordred.MoRSE.MoRSE(None, 25), 'type': '3D', 'description': '3D-MoRSE(distance = 25)', 'module': 'MoRSE', 'constructor': 'MoRSE(None, 25)'}, 'Mor26': {'function': mordred.MoRSE.MoRSE(None, 26), 'type': '3D', 'description': '3D-MoRSE(distance = 26)', 'module': 'MoRSE', 'constructor': 'MoRSE(None, 26)'}, 'Mor27': {'function': mordred.MoRSE.MoRSE(None, 27), 'type': '3D', 'description': '3D-MoRSE(distance = 27)', 'module': 'MoRSE', 'constructor': 'MoRSE(None, 27)'}, 'Mor28': {'function': mordred.MoRSE.MoRSE(None, 28), 'type': '3D', 'description': '3D-MoRSE(distance = 28)', 'module': 'MoRSE', 'constructor': 'MoRSE(None, 28)'}, 'Mor29': {'function': mordred.MoRSE.MoRSE(None, 29), 'type': '3D', 'description': '3D-MoRSE(distance = 29)', 'module': 'MoRSE', 'constructor': 'MoRSE(None, 29)'}, 'Mor30': {'function': mordred.MoRSE.MoRSE(None, 30), 'type': '3D', 'description': '3D-MoRSE(distance = 30)', 'module': 'MoRSE', 'constructor': 'MoRSE(None, 30)'}, 'Mor31': {'function': mordred.MoRSE.MoRSE(None, 31), 'type': '3D', 'description': '3D-MoRSE(distance = 31)', 'module': 'MoRSE', 'constructor': 'MoRSE(None, 31)'}, 'Mor32': {'function': mordred.MoRSE.MoRSE(None, 32), 'type': '3D', 'description': '3D-MoRSE(distance = 32)', 'module': 'MoRSE', 'constructor': 'MoRSE(None, 32)'}, 'Mor01m': {'function': mordred.MoRSE.MoRSE('m', 1), 'type': '3D', 'description': '3D-MoRSE weighted by mass(distance = 1)', 'module': 'MoRSE', 'constructor': 'MoRSE("m", 1)'}, 'Mor02m': {'function': mordred.MoRSE.MoRSE('m', 2), 'type': '3D', 'description': '3D-MoRSE weighted by mass(distance = 2)', 'module': 'MoRSE', 'constructor': 'MoRSE("m", 2)'}, 'Mor03m': {'function': mordred.MoRSE.MoRSE('m', 3), 'type': '3D', 'description': '3D-MoRSE weighted by mass(distance = 3)', 'module': 'MoRSE', 'constructor': 'MoRSE("m", 3)'}, 'Mor04m': {'function': mordred.MoRSE.MoRSE('m', 4), 'type': '3D', 'description': '3D-MoRSE weighted by mass(distance = 4)', 'module': 'MoRSE', 'constructor': 'MoRSE("m", 4)'}, 'Mor05m': {'function': mordred.MoRSE.MoRSE('m', 5), 'type': '3D', 'description': '3D-MoRSE weighted by mass(distance = 5)', 'module': 'MoRSE', 'constructor': 'MoRSE("m", 5)'}, 'Mor06m': {'function': mordred.MoRSE.MoRSE('m', 6), 'type': '3D', 'description': '3D-MoRSE weighted by mass(distance = 6)', 'module': 'MoRSE', 'constructor': 'MoRSE("m", 6)'}, 'Mor07m': {'function': mordred.MoRSE.MoRSE('m', 7), 'type': '3D', 'description': '3D-MoRSE weighted by mass(distance = 7)', 'module': 'MoRSE', 'constructor': 'MoRSE("m", 7)'}, 'Mor08m': {'function': mordred.MoRSE.MoRSE('m', 8), 'type': '3D', 'description': '3D-MoRSE weighted by mass(distance = 8)', 'module': 'MoRSE', 'constructor': 'MoRSE("m", 8)'}, 'Mor09m': {'function': mordred.MoRSE.MoRSE('m', 9), 'type': '3D', 'description': '3D-MoRSE weighted by mass(distance = 9)', 'module': 'MoRSE', 'constructor': 'MoRSE("m", 9)'}, 'Mor10m': {'function': mordred.MoRSE.MoRSE('m', 10), 'type': '3D', 'description': '3D-MoRSE weighted by mass(distance = 10)', 'module': 'MoRSE', 'constructor': 'MoRSE("m", 10)'}, 'Mor11m': {'function': mordred.MoRSE.MoRSE('m', 11), 'type': '3D', 'description': '3D-MoRSE weighted by mass(distance = 11)', 'module': 'MoRSE', 'constructor': 'MoRSE("m", 11)'}, 'Mor12m': {'function': mordred.MoRSE.MoRSE('m', 12), 'type': '3D', 'description': '3D-MoRSE weighted by mass(distance = 12)', 'module': 'MoRSE', 'constructor': 'MoRSE("m", 12)'}, 'Mor13m': {'function': mordred.MoRSE.MoRSE('m', 13), 'type': '3D', 'description': '3D-MoRSE weighted by mass(distance = 13)', 'module': 'MoRSE', 'constructor': 'MoRSE("m", 13)'}, 'Mor14m': {'function': mordred.MoRSE.MoRSE('m', 14), 'type': '3D', 'description': '3D-MoRSE weighted by mass(distance = 14)', 'module': 'MoRSE', 'constructor': 'MoRSE("m", 14)'}, 'Mor15m': {'function': mordred.MoRSE.MoRSE('m', 15), 'type': '3D', 'description': '3D-MoRSE weighted by mass(distance = 15)', 'module': 'MoRSE', 'constructor': 'MoRSE("m", 15)'}, 'Mor16m': {'function': mordred.MoRSE.MoRSE('m', 16), 'type': '3D', 'description': '3D-MoRSE weighted by mass(distance = 16)', 'module': 'MoRSE', 'constructor': 'MoRSE("m", 16)'}, 'Mor17m': {'function': mordred.MoRSE.MoRSE('m', 17), 'type': '3D', 'description': '3D-MoRSE weighted by mass(distance = 17)', 'module': 'MoRSE', 'constructor': 'MoRSE("m", 17)'}, 'Mor18m': {'function': mordred.MoRSE.MoRSE('m', 18), 'type': '3D', 'description': '3D-MoRSE weighted by mass(distance = 18)', 'module': 'MoRSE', 'constructor': 'MoRSE("m", 18)'}, 'Mor19m': {'function': mordred.MoRSE.MoRSE('m', 19), 'type': '3D', 'description': '3D-MoRSE weighted by mass(distance = 19)', 'module': 'MoRSE', 'constructor': 'MoRSE("m", 19)'}, 'Mor20m': {'function': mordred.MoRSE.MoRSE('m', 20), 'type': '3D', 'description': '3D-MoRSE weighted by mass(distance = 20)', 'module': 'MoRSE', 'constructor': 'MoRSE("m", 20)'}, 'Mor21m': {'function': mordred.MoRSE.MoRSE('m', 21), 'type': '3D', 'description': '3D-MoRSE weighted by mass(distance = 21)', 'module': 'MoRSE', 'constructor': 'MoRSE("m", 21)'}, 'Mor22m': {'function': mordred.MoRSE.MoRSE('m', 22), 'type': '3D', 'description': '3D-MoRSE weighted by mass(distance = 22)', 'module': 'MoRSE', 'constructor': 'MoRSE("m", 22)'}, 'Mor23m': {'function': mordred.MoRSE.MoRSE('m', 23), 'type': '3D', 'description': '3D-MoRSE weighted by mass(distance = 23)', 'module': 'MoRSE', 'constructor': 'MoRSE("m", 23)'}, 'Mor24m': {'function': mordred.MoRSE.MoRSE('m', 24), 'type': '3D', 'description': '3D-MoRSE weighted by mass(distance = 24)', 'module': 'MoRSE', 'constructor': 'MoRSE("m", 24)'}, 'Mor25m': {'function': mordred.MoRSE.MoRSE('m', 25), 'type': '3D', 'description': '3D-MoRSE weighted by mass(distance = 25)', 'module': 'MoRSE', 'constructor': 'MoRSE("m", 25)'}, 'Mor26m': {'function': mordred.MoRSE.MoRSE('m', 26), 'type': '3D', 'description': '3D-MoRSE weighted by mass(distance = 26)', 'module': 'MoRSE', 'constructor': 'MoRSE("m", 26)'}, 'Mor27m': {'function': mordred.MoRSE.MoRSE('m', 27), 'type': '3D', 'description': '3D-MoRSE weighted by mass(distance = 27)', 'module': 'MoRSE', 'constructor': 'MoRSE("m", 27)'}, 'Mor28m': {'function': mordred.MoRSE.MoRSE('m', 28), 'type': '3D', 'description': '3D-MoRSE weighted by mass(distance = 28)', 'module': 'MoRSE', 'constructor': 'MoRSE("m", 28)'}, 'Mor29m': {'function': mordred.MoRSE.MoRSE('m', 29), 'type': '3D', 'description': '3D-MoRSE weighted by mass(distance = 29)', 'module': 'MoRSE', 'constructor': 'MoRSE("m", 29)'}, 'Mor30m': {'function': mordred.MoRSE.MoRSE('m', 30), 'type': '3D', 'description': '3D-MoRSE weighted by mass(distance = 30)', 'module': 'MoRSE', 'constructor': 'MoRSE("m", 30)'}, 'Mor31m': {'function': mordred.MoRSE.MoRSE('m', 31), 'type': '3D', 'description': '3D-MoRSE weighted by mass(distance = 31)', 'module': 'MoRSE', 'constructor': 'MoRSE("m", 31)'}, 'Mor32m': {'function': mordred.MoRSE.MoRSE('m', 32), 'type': '3D', 'description': '3D-MoRSE weighted by mass(distance = 32)', 'module': 'MoRSE', 'constructor': 'MoRSE("m", 32)'}, 'Mor01v': {'function': mordred.MoRSE.MoRSE('v', 1), 'type': '3D', 'description': '3D-MoRSE weighted by vdw volume(distance = 1)', 'module': 'MoRSE', 'constructor': 'MoRSE("v", 1)'}, 'Mor02v': {'function': mordred.MoRSE.MoRSE('v', 2), 'type': '3D', 'description': '3D-MoRSE weighted by vdw volume(distance = 2)', 'module': 'MoRSE', 'constructor': 'MoRSE("v", 2)'}, 'Mor03v': {'function': mordred.MoRSE.MoRSE('v', 3), 'type': '3D', 'description': '3D-MoRSE weighted by vdw volume(distance = 3)', 'module': 'MoRSE', 'constructor': 'MoRSE("v", 3)'}, 'Mor04v': {'function': mordred.MoRSE.MoRSE('v', 4), 'type': '3D', 'description': '3D-MoRSE weighted by vdw volume(distance = 4)', 'module': 'MoRSE', 'constructor': 'MoRSE("v", 4)'}, 'Mor05v': {'function': mordred.MoRSE.MoRSE('v', 5), 'type': '3D', 'description': '3D-MoRSE weighted by vdw volume(distance = 5)', 'module': 'MoRSE', 'constructor': 'MoRSE("v", 5)'}, 'Mor06v': {'function': mordred.MoRSE.MoRSE('v', 6), 'type': '3D', 'description': '3D-MoRSE weighted by vdw volume(distance = 6)', 'module': 'MoRSE', 'constructor': 'MoRSE("v", 6)'}, 'Mor07v': {'function': mordred.MoRSE.MoRSE('v', 7), 'type': '3D', 'description': '3D-MoRSE weighted by vdw volume(distance = 7)', 'module': 'MoRSE', 'constructor': 'MoRSE("v", 7)'}, 'Mor08v': {'function': mordred.MoRSE.MoRSE('v', 8), 'type': '3D', 'description': '3D-MoRSE weighted by vdw volume(distance = 8)', 'module': 'MoRSE', 'constructor': 'MoRSE("v", 8)'}, 'Mor09v': {'function': mordred.MoRSE.MoRSE('v', 9), 'type': '3D', 'description': '3D-MoRSE weighted by vdw volume(distance = 9)', 'module': 'MoRSE', 'constructor': 'MoRSE("v", 9)'}, 'Mor10v': {'function': mordred.MoRSE.MoRSE('v', 10), 'type': '3D', 'description': '3D-MoRSE weighted by vdw volume(distance = 10)', 'module': 'MoRSE', 'constructor': 'MoRSE("v", 10)'}, 'Mor11v': {'function': mordred.MoRSE.MoRSE('v', 11), 'type': '3D', 'description': '3D-MoRSE weighted by vdw volume(distance = 11)', 'module': 'MoRSE', 'constructor': 'MoRSE("v", 11)'}, 'Mor12v': {'function': mordred.MoRSE.MoRSE('v', 12), 'type': '3D', 'description': '3D-MoRSE weighted by vdw volume(distance = 12)', 'module': 'MoRSE', 'constructor': 'MoRSE("v", 12)'}, 'Mor13v': {'function': mordred.MoRSE.MoRSE('v', 13), 'type': '3D', 'description': '3D-MoRSE weighted by vdw volume(distance = 13)', 'module': 'MoRSE', 'constructor': 'MoRSE("v", 13)'}, 'Mor14v': {'function': mordred.MoRSE.MoRSE('v', 14), 'type': '3D', 'description': '3D-MoRSE weighted by vdw volume(distance = 14)', 'module': 'MoRSE', 'constructor': 'MoRSE("v", 14)'}, 'Mor15v': {'function': mordred.MoRSE.MoRSE('v', 15), 'type': '3D', 'description': '3D-MoRSE weighted by vdw volume(distance = 15)', 'module': 'MoRSE', 'constructor': 'MoRSE("v", 15)'}, 'Mor16v': {'function': mordred.MoRSE.MoRSE('v', 16), 'type': '3D', 'description': '3D-MoRSE weighted by vdw volume(distance = 16)', 'module': 'MoRSE', 'constructor': 'MoRSE("v", 16)'}, 'Mor17v': {'function': mordred.MoRSE.MoRSE('v', 17), 'type': '3D', 'description': '3D-MoRSE weighted by vdw volume(distance = 17)', 'module': 'MoRSE', 'constructor': 'MoRSE("v", 17)'}, 'Mor18v': {'function': mordred.MoRSE.MoRSE('v', 18), 'type': '3D', 'description': '3D-MoRSE weighted by vdw volume(distance = 18)', 'module': 'MoRSE', 'constructor': 'MoRSE("v", 18)'}, 'Mor19v': {'function': mordred.MoRSE.MoRSE('v', 19), 'type': '3D', 'description': '3D-MoRSE weighted by vdw volume(distance = 19)', 'module': 'MoRSE', 'constructor': 'MoRSE("v", 19)'}, 'Mor20v': {'function': mordred.MoRSE.MoRSE('v', 20), 'type': '3D', 'description': '3D-MoRSE weighted by vdw volume(distance = 20)', 'module': 'MoRSE', 'constructor': 'MoRSE("v", 20)'}, 'Mor21v': {'function': mordred.MoRSE.MoRSE('v', 21), 'type': '3D', 'description': '3D-MoRSE weighted by vdw volume(distance = 21)', 'module': 'MoRSE', 'constructor': 'MoRSE("v", 21)'}, 'Mor22v': {'function': mordred.MoRSE.MoRSE('v', 22), 'type': '3D', 'description': '3D-MoRSE weighted by vdw volume(distance = 22)', 'module': 'MoRSE', 'constructor': 'MoRSE("v", 22)'}, 'Mor23v': {'function': mordred.MoRSE.MoRSE('v', 23), 'type': '3D', 'description': '3D-MoRSE weighted by vdw volume(distance = 23)', 'module': 'MoRSE', 'constructor': 'MoRSE("v", 23)'}, 'Mor24v': {'function': mordred.MoRSE.MoRSE('v', 24), 'type': '3D', 'description': '3D-MoRSE weighted by vdw volume(distance = 24)', 'module': 'MoRSE', 'constructor': 'MoRSE("v", 24)'}, 'Mor25v': {'function': mordred.MoRSE.MoRSE('v', 25), 'type': '3D', 'description': '3D-MoRSE weighted by vdw volume(distance = 25)', 'module': 'MoRSE', 'constructor': 'MoRSE("v", 25)'}, 'Mor26v': {'function': mordred.MoRSE.MoRSE('v', 26), 'type': '3D', 'description': '3D-MoRSE weighted by vdw volume(distance = 26)', 'module': 'MoRSE', 'constructor': 'MoRSE("v", 26)'}, 'Mor27v': {'function': mordred.MoRSE.MoRSE('v', 27), 'type': '3D', 'description': '3D-MoRSE weighted by vdw volume(distance = 27)', 'module': 'MoRSE', 'constructor': 'MoRSE("v", 27)'}, 'Mor28v': {'function': mordred.MoRSE.MoRSE('v', 28), 'type': '3D', 'description': '3D-MoRSE weighted by vdw volume(distance = 28)', 'module': 'MoRSE', 'constructor': 'MoRSE("v", 28)'}, 'Mor29v': {'function': mordred.MoRSE.MoRSE('v', 29), 'type': '3D', 'description': '3D-MoRSE weighted by vdw volume(distance = 29)', 'module': 'MoRSE', 'constructor': 'MoRSE("v", 29)'}, 'Mor30v': {'function': mordred.MoRSE.MoRSE('v', 30), 'type': '3D', 'description': '3D-MoRSE weighted by vdw volume(distance = 30)', 'module': 'MoRSE', 'constructor': 'MoRSE("v", 30)'}, 'Mor31v': {'function': mordred.MoRSE.MoRSE('v', 31), 'type': '3D', 'description': '3D-MoRSE weighted by vdw volume(distance = 31)', 'module': 'MoRSE', 'constructor': 'MoRSE("v", 31)'}, 'Mor32v': {'function': mordred.MoRSE.MoRSE('v', 32), 'type': '3D', 'description': '3D-MoRSE weighted by vdw volume(distance = 32)', 'module': 'MoRSE', 'constructor': 'MoRSE("v", 32)'}, 'Mor01se': {'function': mordred.MoRSE.MoRSE('se', 1), 'type': '3D', 'description': '3D-MoRSE weighted by sanderson EN(distance = 1)', 'module': 'MoRSE', 'constructor': 'MoRSE("se", 1)'}, 'Mor02se': {'function': mordred.MoRSE.MoRSE('se', 2), 'type': '3D', 'description': '3D-MoRSE weighted by sanderson EN(distance = 2)', 'module': 'MoRSE', 'constructor': 'MoRSE("se", 2)'}, 'Mor03se': {'function': mordred.MoRSE.MoRSE('se', 3), 'type': '3D', 'description': '3D-MoRSE weighted by sanderson EN(distance = 3)', 'module': 'MoRSE', 'constructor': 'MoRSE("se", 3)'}, 'Mor04se': {'function': mordred.MoRSE.MoRSE('se', 4), 'type': '3D', 'description': '3D-MoRSE weighted by sanderson EN(distance = 4)', 'module': 'MoRSE', 'constructor': 'MoRSE("se", 4)'}, 'Mor05se': {'function': mordred.MoRSE.MoRSE('se', 5), 'type': '3D', 'description': '3D-MoRSE weighted by sanderson EN(distance = 5)', 'module': 'MoRSE', 'constructor': 'MoRSE("se", 5)'}, 'Mor06se': {'function': mordred.MoRSE.MoRSE('se', 6), 'type': '3D', 'description': '3D-MoRSE weighted by sanderson EN(distance = 6)', 'module': 'MoRSE', 'constructor': 'MoRSE("se", 6)'}, 'Mor07se': {'function': mordred.MoRSE.MoRSE('se', 7), 'type': '3D', 'description': '3D-MoRSE weighted by sanderson EN(distance = 7)', 'module': 'MoRSE', 'constructor': 'MoRSE("se", 7)'}, 'Mor08se': {'function': mordred.MoRSE.MoRSE('se', 8), 'type': '3D', 'description': '3D-MoRSE weighted by sanderson EN(distance = 8)', 'module': 'MoRSE', 'constructor': 'MoRSE("se", 8)'}, 'Mor09se': {'function': mordred.MoRSE.MoRSE('se', 9), 'type': '3D', 'description': '3D-MoRSE weighted by sanderson EN(distance = 9)', 'module': 'MoRSE', 'constructor': 'MoRSE("se", 9)'}, 'Mor10se': {'function': mordred.MoRSE.MoRSE('se', 10), 'type': '3D', 'description': '3D-MoRSE weighted by sanderson EN(distance = 10)', 'module': 'MoRSE', 'constructor': 'MoRSE("se", 10)'}, 'Mor11se': {'function': mordred.MoRSE.MoRSE('se', 11), 'type': '3D', 'description': '3D-MoRSE weighted by sanderson EN(distance = 11)', 'module': 'MoRSE', 'constructor': 'MoRSE("se", 11)'}, 'Mor12se': {'function': mordred.MoRSE.MoRSE('se', 12), 'type': '3D', 'description': '3D-MoRSE weighted by sanderson EN(distance = 12)', 'module': 'MoRSE', 'constructor': 'MoRSE("se", 12)'}, 'Mor13se': {'function': mordred.MoRSE.MoRSE('se', 13), 'type': '3D', 'description': '3D-MoRSE weighted by sanderson EN(distance = 13)', 'module': 'MoRSE', 'constructor': 'MoRSE("se", 13)'}, 'Mor14se': {'function': mordred.MoRSE.MoRSE('se', 14), 'type': '3D', 'description': '3D-MoRSE weighted by sanderson EN(distance = 14)', 'module': 'MoRSE', 'constructor': 'MoRSE("se", 14)'}, 'Mor15se': {'function': mordred.MoRSE.MoRSE('se', 15), 'type': '3D', 'description': '3D-MoRSE weighted by sanderson EN(distance = 15)', 'module': 'MoRSE', 'constructor': 'MoRSE("se", 15)'}, 'Mor16se': {'function': mordred.MoRSE.MoRSE('se', 16), 'type': '3D', 'description': '3D-MoRSE weighted by sanderson EN(distance = 16)', 'module': 'MoRSE', 'constructor': 'MoRSE("se", 16)'}, 'Mor17se': {'function': mordred.MoRSE.MoRSE('se', 17), 'type': '3D', 'description': '3D-MoRSE weighted by sanderson EN(distance = 17)', 'module': 'MoRSE', 'constructor': 'MoRSE("se", 17)'}, 'Mor18se': {'function': mordred.MoRSE.MoRSE('se', 18), 'type': '3D', 'description': '3D-MoRSE weighted by sanderson EN(distance = 18)', 'module': 'MoRSE', 'constructor': 'MoRSE("se", 18)'}, 'Mor19se': {'function': mordred.MoRSE.MoRSE('se', 19), 'type': '3D', 'description': '3D-MoRSE weighted by sanderson EN(distance = 19)', 'module': 'MoRSE', 'constructor': 'MoRSE("se", 19)'}, 'Mor20se': {'function': mordred.MoRSE.MoRSE('se', 20), 'type': '3D', 'description': '3D-MoRSE weighted by sanderson EN(distance = 20)', 'module': 'MoRSE', 'constructor': 'MoRSE("se", 20)'}, 'Mor21se': {'function': mordred.MoRSE.MoRSE('se', 21), 'type': '3D', 'description': '3D-MoRSE weighted by sanderson EN(distance = 21)', 'module': 'MoRSE', 'constructor': 'MoRSE("se", 21)'}, 'Mor22se': {'function': mordred.MoRSE.MoRSE('se', 22), 'type': '3D', 'description': '3D-MoRSE weighted by sanderson EN(distance = 22)', 'module': 'MoRSE', 'constructor': 'MoRSE("se", 22)'}, 'Mor23se': {'function': mordred.MoRSE.MoRSE('se', 23), 'type': '3D', 'description': '3D-MoRSE weighted by sanderson EN(distance = 23)', 'module': 'MoRSE', 'constructor': 'MoRSE("se", 23)'}, 'Mor24se': {'function': mordred.MoRSE.MoRSE('se', 24), 'type': '3D', 'description': '3D-MoRSE weighted by sanderson EN(distance = 24)', 'module': 'MoRSE', 'constructor': 'MoRSE("se", 24)'}, 'Mor25se': {'function': mordred.MoRSE.MoRSE('se', 25), 'type': '3D', 'description': '3D-MoRSE weighted by sanderson EN(distance = 25)', 'module': 'MoRSE', 'constructor': 'MoRSE("se", 25)'}, 'Mor26se': {'function': mordred.MoRSE.MoRSE('se', 26), 'type': '3D', 'description': '3D-MoRSE weighted by sanderson EN(distance = 26)', 'module': 'MoRSE', 'constructor': 'MoRSE("se", 26)'}, 'Mor27se': {'function': mordred.MoRSE.MoRSE('se', 27), 'type': '3D', 'description': '3D-MoRSE weighted by sanderson EN(distance = 27)', 'module': 'MoRSE', 'constructor': 'MoRSE("se", 27)'}, 'Mor28se': {'function': mordred.MoRSE.MoRSE('se', 28), 'type': '3D', 'description': '3D-MoRSE weighted by sanderson EN(distance = 28)', 'module': 'MoRSE', 'constructor': 'MoRSE("se", 28)'}, 'Mor29se': {'function': mordred.MoRSE.MoRSE('se', 29), 'type': '3D', 'description': '3D-MoRSE weighted by sanderson EN(distance = 29)', 'module': 'MoRSE', 'constructor': 'MoRSE("se", 29)'}, 'Mor30se': {'function': mordred.MoRSE.MoRSE('se', 30), 'type': '3D', 'description': '3D-MoRSE weighted by sanderson EN(distance = 30)', 'module': 'MoRSE', 'constructor': 'MoRSE("se", 30)'}, 'Mor31se': {'function': mordred.MoRSE.MoRSE('se', 31), 'type': '3D', 'description': '3D-MoRSE weighted by sanderson EN(distance = 31)', 'module': 'MoRSE', 'constructor': 'MoRSE("se", 31)'}, 'Mor32se': {'function': mordred.MoRSE.MoRSE('se', 32), 'type': '3D', 'description': '3D-MoRSE weighted by sanderson EN(distance = 32)', 'module': 'MoRSE', 'constructor': 'MoRSE("se", 32)'}, 'Mor01p': {'function': mordred.MoRSE.MoRSE('p', 1), 'type': '3D', 'description': '3D-MoRSE weighted by polarizability(distance = 1)', 'module': 'MoRSE', 'constructor': 'MoRSE("p", 1)'}, 'Mor02p': {'function': mordred.MoRSE.MoRSE('p', 2), 'type': '3D', 'description': '3D-MoRSE weighted by polarizability(distance = 2)', 'module': 'MoRSE', 'constructor': 'MoRSE("p", 2)'}, 'Mor03p': {'function': mordred.MoRSE.MoRSE('p', 3), 'type': '3D', 'description': '3D-MoRSE weighted by polarizability(distance = 3)', 'module': 'MoRSE', 'constructor': 'MoRSE("p", 3)'}, 'Mor04p': {'function': mordred.MoRSE.MoRSE('p', 4), 'type': '3D', 'description': '3D-MoRSE weighted by polarizability(distance = 4)', 'module': 'MoRSE', 'constructor': 'MoRSE("p", 4)'}, 'Mor05p': {'function': mordred.MoRSE.MoRSE('p', 5), 'type': '3D', 'description': '3D-MoRSE weighted by polarizability(distance = 5)', 'module': 'MoRSE', 'constructor': 'MoRSE("p", 5)'}, 'Mor06p': {'function': mordred.MoRSE.MoRSE('p', 6), 'type': '3D', 'description': '3D-MoRSE weighted by polarizability(distance = 6)', 'module': 'MoRSE', 'constructor': 'MoRSE("p", 6)'}, 'Mor07p': {'function': mordred.MoRSE.MoRSE('p', 7), 'type': '3D', 'description': '3D-MoRSE weighted by polarizability(distance = 7)', 'module': 'MoRSE', 'constructor': 'MoRSE("p", 7)'}, 'Mor08p': {'function': mordred.MoRSE.MoRSE('p', 8), 'type': '3D', 'description': '3D-MoRSE weighted by polarizability(distance = 8)', 'module': 'MoRSE', 'constructor': 'MoRSE("p", 8)'}, 'Mor09p': {'function': mordred.MoRSE.MoRSE('p', 9), 'type': '3D', 'description': '3D-MoRSE weighted by polarizability(distance = 9)', 'module': 'MoRSE', 'constructor': 'MoRSE("p", 9)'}, 'Mor10p': {'function': mordred.MoRSE.MoRSE('p', 10), 'type': '3D', 'description': '3D-MoRSE weighted by polarizability(distance = 10)', 'module': 'MoRSE', 'constructor': 'MoRSE("p", 10)'}, 'Mor11p': {'function': mordred.MoRSE.MoRSE('p', 11), 'type': '3D', 'description': '3D-MoRSE weighted by polarizability(distance = 11)', 'module': 'MoRSE', 'constructor': 'MoRSE("p", 11)'}, 'Mor12p': {'function': mordred.MoRSE.MoRSE('p', 12), 'type': '3D', 'description': '3D-MoRSE weighted by polarizability(distance = 12)', 'module': 'MoRSE', 'constructor': 'MoRSE("p", 12)'}, 'Mor13p': {'function': mordred.MoRSE.MoRSE('p', 13), 'type': '3D', 'description': '3D-MoRSE weighted by polarizability(distance = 13)', 'module': 'MoRSE', 'constructor': 'MoRSE("p", 13)'}, 'Mor14p': {'function': mordred.MoRSE.MoRSE('p', 14), 'type': '3D', 'description': '3D-MoRSE weighted by polarizability(distance = 14)', 'module': 'MoRSE', 'constructor': 'MoRSE("p", 14)'}, 'Mor15p': {'function': mordred.MoRSE.MoRSE('p', 15), 'type': '3D', 'description': '3D-MoRSE weighted by polarizability(distance = 15)', 'module': 'MoRSE', 'constructor': 'MoRSE("p", 15)'}, 'Mor16p': {'function': mordred.MoRSE.MoRSE('p', 16), 'type': '3D', 'description': '3D-MoRSE weighted by polarizability(distance = 16)', 'module': 'MoRSE', 'constructor': 'MoRSE("p", 16)'}, 'Mor17p': {'function': mordred.MoRSE.MoRSE('p', 17), 'type': '3D', 'description': '3D-MoRSE weighted by polarizability(distance = 17)', 'module': 'MoRSE', 'constructor': 'MoRSE("p", 17)'}, 'Mor18p': {'function': mordred.MoRSE.MoRSE('p', 18), 'type': '3D', 'description': '3D-MoRSE weighted by polarizability(distance = 18)', 'module': 'MoRSE', 'constructor': 'MoRSE("p", 18)'}, 'Mor19p': {'function': mordred.MoRSE.MoRSE('p', 19), 'type': '3D', 'description': '3D-MoRSE weighted by polarizability(distance = 19)', 'module': 'MoRSE', 'constructor': 'MoRSE("p", 19)'}, 'Mor20p': {'function': mordred.MoRSE.MoRSE('p', 20), 'type': '3D', 'description': '3D-MoRSE weighted by polarizability(distance = 20)', 'module': 'MoRSE', 'constructor': 'MoRSE("p", 20)'}, 'Mor21p': {'function': mordred.MoRSE.MoRSE('p', 21), 'type': '3D', 'description': '3D-MoRSE weighted by polarizability(distance = 21)', 'module': 'MoRSE', 'constructor': 'MoRSE("p", 21)'}, 'Mor22p': {'function': mordred.MoRSE.MoRSE('p', 22), 'type': '3D', 'description': '3D-MoRSE weighted by polarizability(distance = 22)', 'module': 'MoRSE', 'constructor': 'MoRSE("p", 22)'}, 'Mor23p': {'function': mordred.MoRSE.MoRSE('p', 23), 'type': '3D', 'description': '3D-MoRSE weighted by polarizability(distance = 23)', 'module': 'MoRSE', 'constructor': 'MoRSE("p", 23)'}, 'Mor24p': {'function': mordred.MoRSE.MoRSE('p', 24), 'type': '3D', 'description': '3D-MoRSE weighted by polarizability(distance = 24)', 'module': 'MoRSE', 'constructor': 'MoRSE("p", 24)'}, 'Mor25p': {'function': mordred.MoRSE.MoRSE('p', 25), 'type': '3D', 'description': '3D-MoRSE weighted by polarizability(distance = 25)', 'module': 'MoRSE', 'constructor': 'MoRSE("p", 25)'}, 'Mor26p': {'function': mordred.MoRSE.MoRSE('p', 26), 'type': '3D', 'description': '3D-MoRSE weighted by polarizability(distance = 26)', 'module': 'MoRSE', 'constructor': 'MoRSE("p", 26)'}, 'Mor27p': {'function': mordred.MoRSE.MoRSE('p', 27), 'type': '3D', 'description': '3D-MoRSE weighted by polarizability(distance = 27)', 'module': 'MoRSE', 'constructor': 'MoRSE("p", 27)'}, 'Mor28p': {'function': mordred.MoRSE.MoRSE('p', 28), 'type': '3D', 'description': '3D-MoRSE weighted by polarizability(distance = 28)', 'module': 'MoRSE', 'constructor': 'MoRSE("p", 28)'}, 'Mor29p': {'function': mordred.MoRSE.MoRSE('p', 29), 'type': '3D', 'description': '3D-MoRSE weighted by polarizability(distance = 29)', 'module': 'MoRSE', 'constructor': 'MoRSE("p", 29)'}, 'Mor30p': {'function': mordred.MoRSE.MoRSE('p', 30), 'type': '3D', 'description': '3D-MoRSE weighted by polarizability(distance = 30)', 'module': 'MoRSE', 'constructor': 'MoRSE("p", 30)'}, 'Mor31p': {'function': mordred.MoRSE.MoRSE('p', 31), 'type': '3D', 'description': '3D-MoRSE weighted by polarizability(distance = 31)', 'module': 'MoRSE', 'constructor': 'MoRSE("p", 31)'}, 'Mor32p': {'function': mordred.MoRSE.MoRSE('p', 32), 'type': '3D', 'description': '3D-MoRSE weighted by polarizability(distance = 32)', 'module': 'MoRSE', 'constructor': 'MoRSE("p", 32)'}, 'LabuteASA': {'function': mordred.MoeType.LabuteASA(), 'type': '2D', 'description': 'Labute"s Approximate Surface Area', 'module': 'MoeType', 'constructor': 'LabuteASA()'}, 'PEOE_VSA1': {'function': mordred.MoeType.PEOE_VSA(1), 'type': '2D', 'description': 'MOE Charge VSA Descriptor 1(-inf < x < -0.30)', 'module': 'MoeType', 'constructor': 'PEOE_VSA(1)'}, 'PEOE_VSA2': {'function': mordred.MoeType.PEOE_VSA(2), 'type': '2D', 'description': 'MOE Charge VSA Descriptor 2(-0.30 <= x < -0.25)', 'module': 'MoeType', 'constructor': 'PEOE_VSA(2)'}, 'PEOE_VSA3': {'function': mordred.MoeType.PEOE_VSA(3), 'type': '2D', 'description': 'MOE Charge VSA Descriptor 3(-0.25 <= x < -0.20)', 'module': 'MoeType', 'constructor': 'PEOE_VSA(3)'}, 'PEOE_VSA4': {'function': mordred.MoeType.PEOE_VSA(4), 'type': '2D', 'description': 'MOE Charge VSA Descriptor 4(-0.20 <= x < -0.15)', 'module': 'MoeType', 'constructor': 'PEOE_VSA(4)'}, 'PEOE_VSA5': {'function': mordred.MoeType.PEOE_VSA(5), 'type': '2D', 'description': 'MOE Charge VSA Descriptor 5(-0.15 <= x < -0.10)', 'module': 'MoeType', 'constructor': 'PEOE_VSA(5)'}, 'PEOE_VSA6': {'function': mordred.MoeType.PEOE_VSA(6), 'type': '2D', 'description': 'MOE Charge VSA Descriptor 6(-0.10 <= x < -0.05)', 'module': 'MoeType', 'constructor': 'PEOE_VSA(6)'}, 'PEOE_VSA7': {'function': mordred.MoeType.PEOE_VSA(7), 'type': '2D', 'description': 'MOE Charge VSA Descriptor 7(-0.05 <= x < 0.00)', 'module': 'MoeType', 'constructor': 'PEOE_VSA(7)'}, 'PEOE_VSA8': {'function': mordred.MoeType.PEOE_VSA(8), 'type': '2D', 'description': 'MOE Charge VSA Descriptor 8( 0.00 <= x < 0.05)', 'module': 'MoeType', 'constructor': 'PEOE_VSA(8)'}, 'PEOE_VSA9': {'function': mordred.MoeType.PEOE_VSA(9), 'type': '2D', 'description': 'MOE Charge VSA Descriptor 9( 0.05 <= x < 0.10)', 'module': 'MoeType', 'constructor': 'PEOE_VSA(9)'}, 'PEOE_VSA10': {'function': mordred.MoeType.PEOE_VSA(10), 'type': '2D', 'description': 'MOE Charge VSA Descriptor 10( 0.10 <= x < 0.15)', 'module': 'MoeType', 'constructor': 'PEOE_VSA(10)'}, 'PEOE_VSA11': {'function': mordred.MoeType.PEOE_VSA(11), 'type': '2D', 'description': 'MOE Charge VSA Descriptor 11( 0.15 <= x < 0.20)', 'module': 'MoeType', 'constructor': 'PEOE_VSA(11)'}, 'PEOE_VSA12': {'function': mordred.MoeType.PEOE_VSA(12), 'type': '2D', 'description': 'MOE Charge VSA Descriptor 12( 0.20 <= x < 0.25)', 'module': 'MoeType', 'constructor': 'PEOE_VSA(12)'}, 'PEOE_VSA13': {'function': mordred.MoeType.PEOE_VSA(13), 'type': '2D', 'description': 'MOE Charge VSA Descriptor 13( 0.25 <= x < 0.30)', 'module': 'MoeType', 'constructor': 'PEOE_VSA(13)'}, 'SMR_VSA1': {'function': mordred.MoeType.SMR_VSA(1), 'type': '2D', 'description': 'MOE MR VSA Descriptor 1(-inf < x < 1.29)', 'module': 'MoeType', 'constructor': 'SMR_VSA(1)'}, 'SMR_VSA2': {'function': mordred.MoeType.SMR_VSA(2), 'type': '2D', 'description': 'MOE MR VSA Descriptor 2( 1.29 <= x < 1.82)', 'module': 'MoeType', 'constructor': 'SMR_VSA(2)'}, 'SMR_VSA3': {'function': mordred.MoeType.SMR_VSA(3), 'type': '2D', 'description': 'MOE MR VSA Descriptor 3( 1.82 <= x < 2.24)', 'module': 'MoeType', 'constructor': 'SMR_VSA(3)'}, 'SMR_VSA4': {'function': mordred.MoeType.SMR_VSA(4), 'type': '2D', 'description': 'MOE MR VSA Descriptor 4( 2.24 <= x < 2.45)', 'module': 'MoeType', 'constructor': 'SMR_VSA(4)'}, 'SMR_VSA5': {'function': mordred.MoeType.SMR_VSA(5), 'type': '2D', 'description': 'MOE MR VSA Descriptor 5( 2.45 <= x < 2.75)', 'module': 'MoeType', 'constructor': 'SMR_VSA(5)'}, 'SMR_VSA6': {'function': mordred.MoeType.SMR_VSA(6), 'type': '2D', 'description': 'MOE MR VSA Descriptor 6( 2.75 <= x < 3.05)', 'module': 'MoeType', 'constructor': 'SMR_VSA(6)'}, 'SMR_VSA7': {'function': mordred.MoeType.SMR_VSA(7), 'type': '2D', 'description': 'MOE MR VSA Descriptor 7( 3.05 <= x < 3.63)', 'module': 'MoeType', 'constructor': 'SMR_VSA(7)'}, 'SMR_VSA8': {'function': mordred.MoeType.SMR_VSA(8), 'type': '2D', 'description': 'MOE MR VSA Descriptor 8( 3.63 <= x < 3.80)', 'module': 'MoeType', 'constructor': 'SMR_VSA(8)'}, 'SMR_VSA9': {'function': mordred.MoeType.SMR_VSA(9), 'type': '2D', 'description': 'MOE MR VSA Descriptor 9( 3.80 <= x < 4.00)', 'module': 'MoeType', 'constructor': 'SMR_VSA(9)'}, 'SlogP_VSA1': {'function': mordred.MoeType.SlogP_VSA(1), 'type': '2D', 'description': 'MOE logP VSA Descriptor 1(-inf < x < -0.40)', 'module': 'MoeType', 'constructor': 'SlogP_VSA(1)'}, 'SlogP_VSA2': {'function': mordred.MoeType.SlogP_VSA(2), 'type': '2D', 'description': 'MOE logP VSA Descriptor 2(-0.40 <= x < -0.20)', 'module': 'MoeType', 'constructor': 'SlogP_VSA(2)'}, 'SlogP_VSA3': {'function': mordred.MoeType.SlogP_VSA(3), 'type': '2D', 'description': 'MOE logP VSA Descriptor 3(-0.20 <= x < 0.00)', 'module': 'MoeType', 'constructor': 'SlogP_VSA(3)'}, 'SlogP_VSA4': {'function': mordred.MoeType.SlogP_VSA(4), 'type': '2D', 'description': 'MOE logP VSA Descriptor 4( 0.00 <= x < 0.10)', 'module': 'MoeType', 'constructor': 'SlogP_VSA(4)'}, 'SlogP_VSA5': {'function': mordred.MoeType.SlogP_VSA(5), 'type': '2D', 'description': 'MOE logP VSA Descriptor 5( 0.10 <= x < 0.15)', 'module': 'MoeType', 'constructor': 'SlogP_VSA(5)'}, 'SlogP_VSA6': {'function': mordred.MoeType.SlogP_VSA(6), 'type': '2D', 'description': 'MOE logP VSA Descriptor 6( 0.15 <= x < 0.20)', 'module': 'MoeType', 'constructor': 'SlogP_VSA(6)'}, 'SlogP_VSA7': {'function': mordred.MoeType.SlogP_VSA(7), 'type': '2D', 'description': 'MOE logP VSA Descriptor 7( 0.20 <= x < 0.25)', 'module': 'MoeType', 'constructor': 'SlogP_VSA(7)'}, 'SlogP_VSA8': {'function': mordred.MoeType.SlogP_VSA(8), 'type': '2D', 'description': 'MOE logP VSA Descriptor 8( 0.25 <= x < 0.30)', 'module': 'MoeType', 'constructor': 'SlogP_VSA(8)'}, 'SlogP_VSA9': {'function': mordred.MoeType.SlogP_VSA(9), 'type': '2D', 'description': 'MOE logP VSA Descriptor 9( 0.30 <= x < 0.40)', 'module': 'MoeType', 'constructor': 'SlogP_VSA(9)'}, 'SlogP_VSA10': {'function': mordred.MoeType.SlogP_VSA(10), 'type': '2D', 'description': 'MOE logP VSA Descriptor 10( 0.40 <= x < 0.50)', 'module': 'MoeType', 'constructor': 'SlogP_VSA(10)'}, 'SlogP_VSA11': {'function': mordred.MoeType.SlogP_VSA(11), 'type': '2D', 'description': 'MOE logP VSA Descriptor 11( 0.50 <= x < 0.60)', 'module': 'MoeType', 'constructor': 'SlogP_VSA(11)'}, 'EState_VSA1': {'function': mordred.MoeType.EState_VSA(1), 'type': '2D', 'description': 'EState VSA Descriptor 1(-inf < x < -0.39)', 'module': 'MoeType', 'constructor': 'EState_VSA(1)'}, 'EState_VSA2': {'function': mordred.MoeType.EState_VSA(2), 'type': '2D', 'description': 'EState VSA Descriptor 2( -0.39 <= x < 0.29)', 'module': 'MoeType', 'constructor': 'EState_VSA(2)'}, 'EState_VSA3': {'function': mordred.MoeType.EState_VSA(3), 'type': '2D', 'description': 'EState VSA Descriptor 3( 0.29 <= x < 0.72)', 'module': 'MoeType', 'constructor': 'EState_VSA(3)'}, 'EState_VSA4': {'function': mordred.MoeType.EState_VSA(4), 'type': '2D', 'description': 'EState VSA Descriptor 4( 0.72 <= x < 1.17)', 'module': 'MoeType', 'constructor': 'EState_VSA(4)'}, 'EState_VSA5': {'function': mordred.MoeType.EState_VSA(5), 'type': '2D', 'description': 'EState VSA Descriptor 5( 1.17 <= x < 1.54)', 'module': 'MoeType', 'constructor': 'EState_VSA(5)'}, 'EState_VSA6': {'function': mordred.MoeType.EState_VSA(6), 'type': '2D', 'description': 'EState VSA Descriptor 6( 1.54 <= x < 1.81)', 'module': 'MoeType', 'constructor': 'EState_VSA(6)'}, 'EState_VSA7': {'function': mordred.MoeType.EState_VSA(7), 'type': '2D', 'description': 'EState VSA Descriptor 7( 1.81 <= x < 2.05)', 'module': 'MoeType', 'constructor': 'EState_VSA(7)'}, 'EState_VSA8': {'function': mordred.MoeType.EState_VSA(8), 'type': '2D', 'description': 'EState VSA Descriptor 8( 2.05 <= x < 4.69)', 'module': 'MoeType', 'constructor': 'EState_VSA(8)'}, 'EState_VSA9': {'function': mordred.MoeType.EState_VSA(9), 'type': '2D', 'description': 'EState VSA Descriptor 9( 4.69 <= x < 9.17)', 'module': 'MoeType', 'constructor': 'EState_VSA(9)'}, 'EState_VSA10': {'function': mordred.MoeType.EState_VSA(10), 'type': '2D', 'description': 'EState VSA Descriptor 10( 9.17 <= x < 15.00)', 'module': 'MoeType', 'constructor': 'EState_VSA(10)'}, 'VSA_EState1': {'function': mordred.MoeType.VSA_EState(1), 'type': '2D', 'description': 'VSA EState Descriptor 1(-inf < x < 4.78)', 'module': 'MoeType', 'constructor': 'VSA_EState(1)'}, 'VSA_EState2': {'function': mordred.MoeType.VSA_EState(2), 'type': '2D', 'description': 'VSA EState Descriptor 2( 4.78 <= x < 5.00)', 'module': 'MoeType', 'constructor': 'VSA_EState(2)'}, 'VSA_EState3': {'function': mordred.MoeType.VSA_EState(3), 'type': '2D', 'description': 'VSA EState Descriptor 3( 5.00 <= x < 5.41)', 'module': 'MoeType', 'constructor': 'VSA_EState(3)'}, 'VSA_EState4': {'function': mordred.MoeType.VSA_EState(4), 'type': '2D', 'description': 'VSA EState Descriptor 4( 5.41 <= x < 5.74)', 'module': 'MoeType', 'constructor': 'VSA_EState(4)'}, 'VSA_EState5': {'function': mordred.MoeType.VSA_EState(5), 'type': '2D', 'description': 'VSA EState Descriptor 5( 5.74 <= x < 6.00)', 'module': 'MoeType', 'constructor': 'VSA_EState(5)'}, 'VSA_EState6': {'function': mordred.MoeType.VSA_EState(6), 'type': '2D', 'description': 'VSA EState Descriptor 6( 6.00 <= x < 6.07)', 'module': 'MoeType', 'constructor': 'VSA_EState(6)'}, 'VSA_EState7': {'function': mordred.MoeType.VSA_EState(7), 'type': '2D', 'description': 'VSA EState Descriptor 7( 6.07 <= x < 6.45)', 'module': 'MoeType', 'constructor': 'VSA_EState(7)'}, 'VSA_EState8': {'function': mordred.MoeType.VSA_EState(8), 'type': '2D', 'description': 'VSA EState Descriptor 8( 6.45 <= x < 7.00)', 'module': 'MoeType', 'constructor': 'VSA_EState(8)'}, 'VSA_EState9': {'function': mordred.MoeType.VSA_EState(9), 'type': '2D', 'description': 'VSA EState Descriptor 9( 7.00 <= x < 11.00)', 'module': 'MoeType', 'constructor': 'VSA_EState(9)'}, 'MDEC-11': {'function': mordred.MolecularDistanceEdge.MolecularDistanceEdge(1, 1, 'C'), 'type': '2D', 'description': 'molecular distance edge between primary C and primary C', 'module': 'MolecularDistanceEdge', 'constructor': 'MolecularDistanceEdge(1, 1, "C")'}, 'MDEC-12': {'function': mordred.MolecularDistanceEdge.MolecularDistanceEdge(1, 2, 'C'), 'type': '2D', 'description': 'molecular distance edge between primary C and secondary C', 'module': 'MolecularDistanceEdge', 'constructor': 'MolecularDistanceEdge(1, 2, "C")'}, 'MDEC-13': {'function': mordred.MolecularDistanceEdge.MolecularDistanceEdge(1, 3, 'C'), 'type': '2D', 'description': 'molecular distance edge between primary C and tertiary C', 'module': 'MolecularDistanceEdge', 'constructor': 'MolecularDistanceEdge(1, 3, "C")'}, 'MDEC-14': {'function': mordred.MolecularDistanceEdge.MolecularDistanceEdge(1, 4, 'C'), 'type': '2D', 'description': 'molecular distance edge between primary C and quaternary C', 'module': 'MolecularDistanceEdge', 'constructor': 'MolecularDistanceEdge(1, 4, "C")'}, 'MDEC-22': {'function': mordred.MolecularDistanceEdge.MolecularDistanceEdge(2, 2, 'C'), 'type': '2D', 'description': 'molecular distance edge between secondary C and secondary C', 'module': 'MolecularDistanceEdge', 'constructor': 'MolecularDistanceEdge(2, 2, "C")'}, 'MDEC-23': {'function': mordred.MolecularDistanceEdge.MolecularDistanceEdge(2, 3, 'C'), 'type': '2D', 'description': 'molecular distance edge between secondary C and tertiary C', 'module': 'MolecularDistanceEdge', 'constructor': 'MolecularDistanceEdge(2, 3, "C")'}, 'MDEC-24': {'function': mordred.MolecularDistanceEdge.MolecularDistanceEdge(2, 4, 'C'), 'type': '2D', 'description': 'molecular distance edge between secondary C and quaternary C', 'module': 'MolecularDistanceEdge', 'constructor': 'MolecularDistanceEdge(2, 4, "C")'}, 'MDEC-33': {'function': mordred.MolecularDistanceEdge.MolecularDistanceEdge(3, 3, 'C'), 'type': '2D', 'description': 'molecular distance edge between tertiary C and tertiary C', 'module': 'MolecularDistanceEdge', 'constructor': 'MolecularDistanceEdge(3, 3, "C")'}, 'MDEC-34': {'function': mordred.MolecularDistanceEdge.MolecularDistanceEdge(3, 4, 'C'), 'type': '2D', 'description': 'molecular distance edge between tertiary C and quaternary C', 'module': 'MolecularDistanceEdge', 'constructor': 'MolecularDistanceEdge(3, 4, "C")'}, 'MDEC-44': {'function': mordred.MolecularDistanceEdge.MolecularDistanceEdge(4, 4, 'C'), 'type': '2D', 'description': 'molecular distance edge between quaternary C and quaternary C', 'module': 'MolecularDistanceEdge', 'constructor': 'MolecularDistanceEdge(4, 4, "C")'}, 'MDEO-11': {'function': mordred.MolecularDistanceEdge.MolecularDistanceEdge(1, 1, 'O'), 'type': '2D', 'description': 'molecular distance edge between primary O and primary O', 'module': 'MolecularDistanceEdge', 'constructor': 'MolecularDistanceEdge(1, 1, "O")'}, 'MDEO-12': {'function': mordred.MolecularDistanceEdge.MolecularDistanceEdge(1, 2, 'O'), 'type': '2D', 'description': 'molecular distance edge between primary O and secondary O', 'module': 'MolecularDistanceEdge', 'constructor': 'MolecularDistanceEdge(1, 2, "O")'}, 'MDEO-22': {'function': mordred.MolecularDistanceEdge.MolecularDistanceEdge(2, 2, 'O'), 'type': '2D', 'description': 'molecular distance edge between secondary O and secondary O', 'module': 'MolecularDistanceEdge', 'constructor': 'MolecularDistanceEdge(2, 2, "O")'}, 'MDEN-11': {'function': mordred.MolecularDistanceEdge.MolecularDistanceEdge(1, 1, 'N'), 'type': '2D', 'description': 'molecular distance edge between primary N and primary N', 'module': 'MolecularDistanceEdge', 'constructor': 'MolecularDistanceEdge(1, 1, "N")'}, 'MDEN-12': {'function': mordred.MolecularDistanceEdge.MolecularDistanceEdge(1, 2, 'N'), 'type': '2D', 'description': 'molecular distance edge between primary N and secondary N', 'module': 'MolecularDistanceEdge', 'constructor': 'MolecularDistanceEdge(1, 2, "N")'}, 'MDEN-13': {'function': mordred.MolecularDistanceEdge.MolecularDistanceEdge(1, 3, 'N'), 'type': '2D', 'description': 'molecular distance edge between primary N and tertiary N', 'module': 'MolecularDistanceEdge', 'constructor': 'MolecularDistanceEdge(1, 3, "N")'}, 'MDEN-22': {'function': mordred.MolecularDistanceEdge.MolecularDistanceEdge(2, 2, 'N'), 'type': '2D', 'description': 'molecular distance edge between secondary N and secondary N', 'module': 'MolecularDistanceEdge', 'constructor': 'MolecularDistanceEdge(2, 2, "N")'}, 'MDEN-23': {'function': mordred.MolecularDistanceEdge.MolecularDistanceEdge(2, 3, 'N'), 'type': '2D', 'description': 'molecular distance edge between secondary N and tertiary N', 'module': 'MolecularDistanceEdge', 'constructor': 'MolecularDistanceEdge(2, 3, "N")'}, 'MDEN-33': {'function': mordred.MolecularDistanceEdge.MolecularDistanceEdge(3, 3, 'N'), 'type': '2D', 'description': 'molecular distance edge between tertiary N and tertiary N', 'module': 'MolecularDistanceEdge', 'constructor': 'MolecularDistanceEdge(3, 3, "N")'}, 'MID': {'function': mordred.MolecularId.MolecularId('any', False, 1e-10), 'type': '2D', 'description': 'molecular ID', 'module': 'MolecularId', 'constructor': 'MolecularId("any", False, 1e-10)'}, 'AMID': {'function': mordred.MolecularId.MolecularId('any', True, 1e-10), 'type': '2D', 'description': 'averaged molecular ID', 'module': 'MolecularId', 'constructor': 'MolecularId("any", True, 1e-10)'}, 'MID_h': {'function': mordred.MolecularId.MolecularId('hetero', False, 1e-10), 'type': '2D', 'description': 'molecular ID on h atoms', 'module': 'MolecularId', 'constructor': 'MolecularId("hetero", False, 1e-10)'}, 'AMID_h': {'function': mordred.MolecularId.MolecularId('hetero', True, 1e-10), 'type': '2D', 'description': 'averaged molecular ID on h atoms', 'module': 'MolecularId', 'constructor': 'MolecularId("hetero", True, 1e-10)'}, 'MID_C': {'function': mordred.MolecularId.MolecularId('C', False, 1e-10), 'type': '2D', 'description': 'molecular ID on C atoms', 'module': 'MolecularId', 'constructor': 'MolecularId("C", False, 1e-10)'}, 'AMID_C': {'function': mordred.MolecularId.MolecularId('C', True, 1e-10), 'type': '2D', 'description': 'averaged molecular ID on C atoms', 'module': 'MolecularId', 'constructor': 'MolecularId("C", True, 1e-10)'}, 'MID_N': {'function': mordred.MolecularId.MolecularId('N', False, 1e-10), 'type': '2D', 'description': 'molecular ID on N atoms', 'module': 'MolecularId', 'constructor': 'MolecularId("N", False, 1e-10)'}, 'AMID_N': {'function': mordred.MolecularId.MolecularId('N', True, 1e-10), 'type': '2D', 'description': 'averaged molecular ID on N atoms', 'module': 'MolecularId', 'constructor': 'MolecularId("N", True, 1e-10)'}, 'MID_O': {'function': mordred.MolecularId.MolecularId('O', False, 1e-10), 'type': '2D', 'description': 'molecular ID on O atoms', 'module': 'MolecularId', 'constructor': 'MolecularId("O", False, 1e-10)'}, 'AMID_O': {'function': mordred.MolecularId.MolecularId('O', True, 1e-10), 'type': '2D', 'description': 'averaged molecular ID on O atoms', 'module': 'MolecularId', 'constructor': 'MolecularId("O", True, 1e-10)'}, 'MID_X': {'function': mordred.MolecularId.MolecularId('X', False, 1e-10), 'type': '2D', 'description': 'molecular ID on halogen atoms', 'module': 'MolecularId', 'constructor': 'MolecularId("X", False, 1e-10)'}, 'AMID_X': {'function': mordred.MolecularId.MolecularId('X', True, 1e-10), 'type': '2D', 'description': 'averaged molecular ID on halogen atoms', 'module': 'MolecularId', 'constructor': 'MolecularId("X", True, 1e-10)'}, 'MOMI-X': {'function': mordred.MomentOfInertia.MomentOfInertia('X'), 'type': '3D', 'description': 'moment of inertia(axis = X)', 'module': 'MomentOfInertia', 'constructor': 'MomentOfInertia("X")'}, 'MOMI-Y': {'function': mordred.MomentOfInertia.MomentOfInertia('Y'), 'type': '3D', 'description': 'moment of inertia(axis = Y)', 'module': 'MomentOfInertia', 'constructor': 'MomentOfInertia("Y")'}, 'MOMI-Z': {'function': mordred.MomentOfInertia.MomentOfInertia('Z'), 'type': '3D', 'description': 'moment of inertia(axis = Z)', 'module': 'MomentOfInertia', 'constructor': 'MomentOfInertia("Z")'}, 'PBF': {'function': mordred.PBF.PBF(), 'type': '3D', 'description': 'PBF', 'module': 'PBF', 'constructor': 'PBF()'}, 'MPC2': {'function': mordred.PathCount.PathCount(2, False, False, False), 'type': '2D', 'description': '2-ordered path count', 'module': 'PathCount', 'constructor': 'PathCount(2, False, False, False)'}, 'MPC3': {'function': mordred.PathCount.PathCount(3, False, False, False), 'type': '2D', 'description': '3-ordered path count', 'module': 'PathCount', 'constructor': 'PathCount(3, False, False, False)'}, 'MPC4': {'function': mordred.PathCount.PathCount(4, False, False, False), 'type': '2D', 'description': '4-ordered path count', 'module': 'PathCount', 'constructor': 'PathCount(4, False, False, False)'}, 'MPC5': {'function': mordred.PathCount.PathCount(5, False, False, False), 'type': '2D', 'description': '5-ordered path count', 'module': 'PathCount', 'constructor': 'PathCount(5, False, False, False)'}, 'MPC6': {'function': mordred.PathCount.PathCount(6, False, False, False), 'type': '2D', 'description': '6-ordered path count', 'module': 'PathCount', 'constructor': 'PathCount(6, False, False, False)'}, 'MPC7': {'function': mordred.PathCount.PathCount(7, False, False, False), 'type': '2D', 'description': '7-ordered path count', 'module': 'PathCount', 'constructor': 'PathCount(7, False, False, False)'}, 'MPC8': {'function': mordred.PathCount.PathCount(8, False, False, False), 'type': '2D', 'description': '8-ordered path count', 'module': 'PathCount', 'constructor': 'PathCount(8, False, False, False)'}, 'MPC9': {'function': mordred.PathCount.PathCount(9, False, False, False), 'type': '2D', 'description': '9-ordered path count', 'module': 'PathCount', 'constructor': 'PathCount(9, False, False, False)'}, 'MPC10': {'function': mordred.PathCount.PathCount(10, False, False, False), 'type': '2D', 'description': '10-ordered path count', 'module': 'PathCount', 'constructor': 'PathCount(10, False, False, False)'}, 'TMPC10': {'function': mordred.PathCount.PathCount(10, False, True, False), 'type': '2D', 'description': '10-ordered total path count', 'module': 'PathCount', 'constructor': 'PathCount(10, False, True, False)'}, 'piPC1': {'function': mordred.PathCount.PathCount(1, True, False, True), 'type': '2D', 'description': '1-ordered pi-path count(log scale)', 'module': 'PathCount', 'constructor': 'PathCount(1, True, False, True)'}, 'piPC2': {'function': mordred.PathCount.PathCount(2, True, False, True), 'type': '2D', 'description': '2-ordered pi-path count(log scale)', 'module': 'PathCount', 'constructor': 'PathCount(2, True, False, True)'}, 'piPC3': {'function': mordred.PathCount.PathCount(3, True, False, True), 'type': '2D', 'description': '3-ordered pi-path count(log scale)', 'module': 'PathCount', 'constructor': 'PathCount(3, True, False, True)'}, 'piPC4': {'function': mordred.PathCount.PathCount(4, True, False, True), 'type': '2D', 'description': '4-ordered pi-path count(log scale)', 'module': 'PathCount', 'constructor': 'PathCount(4, True, False, True)'}, 'piPC5': {'function': mordred.PathCount.PathCount(5, True, False, True), 'type': '2D', 'description': '5-ordered pi-path count(log scale)', 'module': 'PathCount', 'constructor': 'PathCount(5, True, False, True)'}, 'piPC6': {'function': mordred.PathCount.PathCount(6, True, False, True), 'type': '2D', 'description': '6-ordered pi-path count(log scale)', 'module': 'PathCount', 'constructor': 'PathCount(6, True, False, True)'}, 'piPC7': {'function': mordred.PathCount.PathCount(7, True, False, True), 'type': '2D', 'description': '7-ordered pi-path count(log scale)', 'module': 'PathCount', 'constructor': 'PathCount(7, True, False, True)'}, 'piPC8': {'function': mordred.PathCount.PathCount(8, True, False, True), 'type': '2D', 'description': '8-ordered pi-path count(log scale)', 'module': 'PathCount', 'constructor': 'PathCount(8, True, False, True)'}, 'piPC9': {'function': mordred.PathCount.PathCount(9, True, False, True), 'type': '2D', 'description': '9-ordered pi-path count(log scale)', 'module': 'PathCount', 'constructor': 'PathCount(9, True, False, True)'}, 'piPC10': {'function': mordred.PathCount.PathCount(10, True, False, True), 'type': '2D', 'description': '10-ordered pi-path count(log scale)', 'module': 'PathCount', 'constructor': 'PathCount(10, True, False, True)'}, 'TpiPC10': {'function': mordred.PathCount.PathCount(10, True, True, True), 'type': '2D', 'description': '10-ordered total pi-path count(log scale)', 'module': 'PathCount', 'constructor': 'PathCount(10, True, True, True)'}, 'apol': {'function': mordred.Polarizability.APol(False), 'type': '2D', 'description': 'atomic polarizability', 'module': 'Polarizability', 'constructor': 'APol(False)'}, 'bpol': {'function': mordred.Polarizability.BPol(False), 'type': '2D', 'description': 'bond polarizability', 'module': 'Polarizability', 'constructor': 'BPol(False)'}, 'nRing': {'function': mordred.RingCount.RingCount(None, False, False, None, None), 'type': '2D', 'description': 'ring count', 'module': 'RingCount', 'constructor': 'RingCount(None, False, False, None, None)'}, 'n3Ring': {'function': mordred.RingCount.RingCount(3, False, False, None, None), 'type': '2D', 'description': '3-membered ring count', 'module': 'RingCount', 'constructor': 'RingCount(3, False, False, None, None)'}, 'n4Ring': {'function': mordred.RingCount.RingCount(4, False, False, None, None), 'type': '2D', 'description': '4-membered ring count', 'module': 'RingCount', 'constructor': 'RingCount(4, False, False, None, None)'}, 'n5Ring': {'function': mordred.RingCount.RingCount(5, False, False, None, None), 'type': '2D', 'description': '5-membered ring count', 'module': 'RingCount', 'constructor': 'RingCount(5, False, False, None, None)'}, 'n6Ring': {'function': mordred.RingCount.RingCount(6, False, False, None, None), 'type': '2D', 'description': '6-membered ring count', 'module': 'RingCount', 'constructor': 'RingCount(6, False, False, None, None)'}, 'n7Ring': {'function': mordred.RingCount.RingCount(7, False, False, None, None), 'type': '2D', 'description': '7-membered ring count', 'module': 'RingCount', 'constructor': 'RingCount(7, False, False, None, None)'}, 'n8Ring': {'function': mordred.RingCount.RingCount(8, False, False, None, None), 'type': '2D', 'description': '8-membered ring count', 'module': 'RingCount', 'constructor': 'RingCount(8, False, False, None, None)'}, 'n9Ring': {'function': mordred.RingCount.RingCount(9, False, False, None, None), 'type': '2D', 'description': '9-membered ring count', 'module': 'RingCount', 'constructor': 'RingCount(9, False, False, None, None)'}, 'n10Ring': {'function': mordred.RingCount.RingCount(10, False, False, None, None), 'type': '2D', 'description': '10-membered ring count', 'module': 'RingCount', 'constructor': 'RingCount(10, False, False, None, None)'}, 'n11Ring': {'function': mordred.RingCount.RingCount(11, False, False, None, None), 'type': '2D', 'description': '11-membered ring count', 'module': 'RingCount', 'constructor': 'RingCount(11, False, False, None, None)'}, 'n12Ring': {'function': mordred.RingCount.RingCount(12, False, False, None, None), 'type': '2D', 'description': '12-membered ring count', 'module': 'RingCount', 'constructor': 'RingCount(12, False, False, None, None)'}, 'nG12Ring': {'function': mordred.RingCount.RingCount(12, True, False, None, None), 'type': '2D', 'description': '12-or-greater-membered ring count', 'module': 'RingCount', 'constructor': 'RingCount(12, True, False, None, None)'}, 'nHRing': {'function': mordred.RingCount.RingCount(None, False, False, None, True), 'type': '2D', 'description': 'hetero ring count', 'module': 'RingCount', 'constructor': 'RingCount(None, False, False, None, True)'}, 'n3HRing': {'function': mordred.RingCount.RingCount(3, False, False, None, True), 'type': '2D', 'description': '3-membered hetero ring count', 'module': 'RingCount', 'constructor': 'RingCount(3, False, False, None, True)'}, 'n4HRing': {'function': mordred.RingCount.RingCount(4, False, False, None, True), 'type': '2D', 'description': '4-membered hetero ring count', 'module': 'RingCount', 'constructor': 'RingCount(4, False, False, None, True)'}, 'n5HRing': {'function': mordred.RingCount.RingCount(5, False, False, None, True), 'type': '2D', 'description': '5-membered hetero ring count', 'module': 'RingCount', 'constructor': 'RingCount(5, False, False, None, True)'}, 'n6HRing': {'function': mordred.RingCount.RingCount(6, False, False, None, True), 'type': '2D', 'description': '6-membered hetero ring count', 'module': 'RingCount', 'constructor': 'RingCount(6, False, False, None, True)'}, 'n7HRing': {'function': mordred.RingCount.RingCount(7, False, False, None, True), 'type': '2D', 'description': '7-membered hetero ring count', 'module': 'RingCount', 'constructor': 'RingCount(7, False, False, None, True)'}, 'n8HRing': {'function': mordred.RingCount.RingCount(8, False, False, None, True), 'type': '2D', 'description': '8-membered hetero ring count', 'module': 'RingCount', 'constructor': 'RingCount(8, False, False, None, True)'}, 'n9HRing': {'function': mordred.RingCount.RingCount(9, False, False, None, True), 'type': '2D', 'description': '9-membered hetero ring count', 'module': 'RingCount', 'constructor': 'RingCount(9, False, False, None, True)'}, 'n10HRing': {'function': mordred.RingCount.RingCount(10, False, False, None, True), 'type': '2D', 'description': '10-membered hetero ring count', 'module': 'RingCount', 'constructor': 'RingCount(10, False, False, None, True)'}, 'n11HRing': {'function': mordred.RingCount.RingCount(11, False, False, None, True), 'type': '2D', 'description': '11-membered hetero ring count', 'module': 'RingCount', 'constructor': 'RingCount(11, False, False, None, True)'}, 'n12HRing': {'function': mordred.RingCount.RingCount(12, False, False, None, True), 'type': '2D', 'description': '12-membered hetero ring count', 'module': 'RingCount', 'constructor': 'RingCount(12, False, False, None, True)'}, 'nG12HRing': {'function': mordred.RingCount.RingCount(12, True, False, None, True), 'type': '2D', 'description': '12-or-greater-membered hetero ring count', 'module': 'RingCount', 'constructor': 'RingCount(12, True, False, None, True)'}, 'naRing': {'function': mordred.RingCount.RingCount(None, False, False, True, None), 'type': '2D', 'description': 'aromatic ring count', 'module': 'RingCount', 'constructor': 'RingCount(None, False, False, True, None)'}, 'n3aRing': {'function': mordred.RingCount.RingCount(3, False, False, True, None), 'type': '2D', 'description': '3-membered aromatic ring count', 'module': 'RingCount', 'constructor': 'RingCount(3, False, False, True, None)'}, 'n4aRing': {'function': mordred.RingCount.RingCount(4, False, False, True, None), 'type': '2D', 'description': '4-membered aromatic ring count', 'module': 'RingCount', 'constructor': 'RingCount(4, False, False, True, None)'}, 'n5aRing': {'function': mordred.RingCount.RingCount(5, False, False, True, None), 'type': '2D', 'description': '5-membered aromatic ring count', 'module': 'RingCount', 'constructor': 'RingCount(5, False, False, True, None)'}, 'n6aRing': {'function': mordred.RingCount.RingCount(6, False, False, True, None), 'type': '2D', 'description': '6-membered aromatic ring count', 'module': 'RingCount', 'constructor': 'RingCount(6, False, False, True, None)'}, 'n7aRing': {'function': mordred.RingCount.RingCount(7, False, False, True, None), 'type': '2D', 'description': '7-membered aromatic ring count', 'module': 'RingCount', 'constructor': 'RingCount(7, False, False, True, None)'}, 'n8aRing': {'function': mordred.RingCount.RingCount(8, False, False, True, None), 'type': '2D', 'description': '8-membered aromatic ring count', 'module': 'RingCount', 'constructor': 'RingCount(8, False, False, True, None)'}, 'n9aRing': {'function': mordred.RingCount.RingCount(9, False, False, True, None), 'type': '2D', 'description': '9-membered aromatic ring count', 'module': 'RingCount', 'constructor': 'RingCount(9, False, False, True, None)'}, 'n10aRing': {'function': mordred.RingCount.RingCount(10, False, False, True, None), 'type': '2D', 'description': '10-membered aromatic ring count', 'module': 'RingCount', 'constructor': 'RingCount(10, False, False, True, None)'}, 'n11aRing': {'function': mordred.RingCount.RingCount(11, False, False, True, None), 'type': '2D', 'description': '11-membered aromatic ring count', 'module': 'RingCount', 'constructor': 'RingCount(11, False, False, True, None)'}, 'n12aRing': {'function': mordred.RingCount.RingCount(12, False, False, True, None), 'type': '2D', 'description': '12-membered aromatic ring count', 'module': 'RingCount', 'constructor': 'RingCount(12, False, False, True, None)'}, 'nG12aRing': {'function': mordred.RingCount.RingCount(12, True, False, True, None), 'type': '2D', 'description': '12-or-greater-membered aromatic ring count', 'module': 'RingCount', 'constructor': 'RingCount(12, True, False, True, None)'}, 'naHRing': {'function': mordred.RingCount.RingCount(None, False, False, True, True), 'type': '2D', 'description': 'aromatic hetero ring count', 'module': 'RingCount', 'constructor': 'RingCount(None, False, False, True, True)'}, 'n3aHRing': {'function': mordred.RingCount.RingCount(3, False, False, True, True), 'type': '2D', 'description': '3-membered aromatic hetero ring count', 'module': 'RingCount', 'constructor': 'RingCount(3, False, False, True, True)'}, 'n4aHRing': {'function': mordred.RingCount.RingCount(4, False, False, True, True), 'type': '2D', 'description': '4-membered aromatic hetero ring count', 'module': 'RingCount', 'constructor': 'RingCount(4, False, False, True, True)'}, 'n5aHRing': {'function': mordred.RingCount.RingCount(5, False, False, True, True), 'type': '2D', 'description': '5-membered aromatic hetero ring count', 'module': 'RingCount', 'constructor': 'RingCount(5, False, False, True, True)'}, 'n6aHRing': {'function': mordred.RingCount.RingCount(6, False, False, True, True), 'type': '2D', 'description': '6-membered aromatic hetero ring count', 'module': 'RingCount', 'constructor': 'RingCount(6, False, False, True, True)'}, 'n7aHRing': {'function': mordred.RingCount.RingCount(7, False, False, True, True), 'type': '2D', 'description': '7-membered aromatic hetero ring count', 'module': 'RingCount', 'constructor': 'RingCount(7, False, False, True, True)'}, 'n8aHRing': {'function': mordred.RingCount.RingCount(8, False, False, True, True), 'type': '2D', 'description': '8-membered aromatic hetero ring count', 'module': 'RingCount', 'constructor': 'RingCount(8, False, False, True, True)'}, 'n9aHRing': {'function': mordred.RingCount.RingCount(9, False, False, True, True), 'type': '2D', 'description': '9-membered aromatic hetero ring count', 'module': 'RingCount', 'constructor': 'RingCount(9, False, False, True, True)'}, 'n10aHRing': {'function': mordred.RingCount.RingCount(10, False, False, True, True), 'type': '2D', 'description': '10-membered aromatic hetero ring count', 'module': 'RingCount', 'constructor': 'RingCount(10, False, False, True, True)'}, 'n11aHRing': {'function': mordred.RingCount.RingCount(11, False, False, True, True), 'type': '2D', 'description': '11-membered aromatic hetero ring count', 'module': 'RingCount', 'constructor': 'RingCount(11, False, False, True, True)'}, 'n12aHRing': {'function': mordred.RingCount.RingCount(12, False, False, True, True), 'type': '2D', 'description': '12-membered aromatic hetero ring count', 'module': 'RingCount', 'constructor': 'RingCount(12, False, False, True, True)'}, 'nG12aHRing': {'function': mordred.RingCount.RingCount(12, True, False, True, True), 'type': '2D', 'description': '12-or-greater-membered aromatic hetero ring count', 'module': 'RingCount', 'constructor': 'RingCount(12, True, False, True, True)'}, 'nARing': {'function': mordred.RingCount.RingCount(None, False, False, False, None), 'type': '2D', 'description': 'aliphatic ring count', 'module': 'RingCount', 'constructor': 'RingCount(None, False, False, False, None)'}, 'n3ARing': {'function': mordred.RingCount.RingCount(3, False, False, False, None), 'type': '2D', 'description': '3-membered aliphatic ring count', 'module': 'RingCount', 'constructor': 'RingCount(3, False, False, False, None)'}, 'n4ARing': {'function': mordred.RingCount.RingCount(4, False, False, False, None), 'type': '2D', 'description': '4-membered aliphatic ring count', 'module': 'RingCount', 'constructor': 'RingCount(4, False, False, False, None)'}, 'n5ARing': {'function': mordred.RingCount.RingCount(5, False, False, False, None), 'type': '2D', 'description': '5-membered aliphatic ring count', 'module': 'RingCount', 'constructor': 'RingCount(5, False, False, False, None)'}, 'n6ARing': {'function': mordred.RingCount.RingCount(6, False, False, False, None), 'type': '2D', 'description': '6-membered aliphatic ring count', 'module': 'RingCount', 'constructor': 'RingCount(6, False, False, False, None)'}, 'n7ARing': {'function': mordred.RingCount.RingCount(7, False, False, False, None), 'type': '2D', 'description': '7-membered aliphatic ring count', 'module': 'RingCount', 'constructor': 'RingCount(7, False, False, False, None)'}, 'n8ARing': {'function': mordred.RingCount.RingCount(8, False, False, False, None), 'type': '2D', 'description': '8-membered aliphatic ring count', 'module': 'RingCount', 'constructor': 'RingCount(8, False, False, False, None)'}, 'n9ARing': {'function': mordred.RingCount.RingCount(9, False, False, False, None), 'type': '2D', 'description': '9-membered aliphatic ring count', 'module': 'RingCount', 'constructor': 'RingCount(9, False, False, False, None)'}, 'n10ARing': {'function': mordred.RingCount.RingCount(10, False, False, False, None), 'type': '2D', 'description': '10-membered aliphatic ring count', 'module': 'RingCount', 'constructor': 'RingCount(10, False, False, False, None)'}, 'n11ARing': {'function': mordred.RingCount.RingCount(11, False, False, False, None), 'type': '2D', 'description': '11-membered aliphatic ring count', 'module': 'RingCount', 'constructor': 'RingCount(11, False, False, False, None)'}, 'n12ARing': {'function': mordred.RingCount.RingCount(12, False, False, False, None), 'type': '2D', 'description': '12-membered aliphatic ring count', 'module': 'RingCount', 'constructor': 'RingCount(12, False, False, False, None)'}, 'nG12ARing': {'function': mordred.RingCount.RingCount(12, True, False, False, None), 'type': '2D', 'description': '12-or-greater-membered aliphatic ring count', 'module': 'RingCount', 'constructor': 'RingCount(12, True, False, False, None)'}, 'nAHRing': {'function': mordred.RingCount.RingCount(None, False, False, False, True), 'type': '2D', 'description': 'aliphatic hetero ring count', 'module': 'RingCount', 'constructor': 'RingCount(None, False, False, False, True)'}, 'n3AHRing': {'function': mordred.RingCount.RingCount(3, False, False, False, True), 'type': '2D', 'description': '3-membered aliphatic hetero ring count', 'module': 'RingCount', 'constructor': 'RingCount(3, False, False, False, True)'}, 'n4AHRing': {'function': mordred.RingCount.RingCount(4, False, False, False, True), 'type': '2D', 'description': '4-membered aliphatic hetero ring count', 'module': 'RingCount', 'constructor': 'RingCount(4, False, False, False, True)'}, 'n5AHRing': {'function': mordred.RingCount.RingCount(5, False, False, False, True), 'type': '2D', 'description': '5-membered aliphatic hetero ring count', 'module': 'RingCount', 'constructor': 'RingCount(5, False, False, False, True)'}, 'n6AHRing': {'function': mordred.RingCount.RingCount(6, False, False, False, True), 'type': '2D', 'description': '6-membered aliphatic hetero ring count', 'module': 'RingCount', 'constructor': 'RingCount(6, False, False, False, True)'}, 'n7AHRing': {'function': mordred.RingCount.RingCount(7, False, False, False, True), 'type': '2D', 'description': '7-membered aliphatic hetero ring count', 'module': 'RingCount', 'constructor': 'RingCount(7, False, False, False, True)'}, 'n8AHRing': {'function': mordred.RingCount.RingCount(8, False, False, False, True), 'type': '2D', 'description': '8-membered aliphatic hetero ring count', 'module': 'RingCount', 'constructor': 'RingCount(8, False, False, False, True)'}, 'n9AHRing': {'function': mordred.RingCount.RingCount(9, False, False, False, True), 'type': '2D', 'description': '9-membered aliphatic hetero ring count', 'module': 'RingCount', 'constructor': 'RingCount(9, False, False, False, True)'}, 'n10AHRing': {'function': mordred.RingCount.RingCount(10, False, False, False, True), 'type': '2D', 'description': '10-membered aliphatic hetero ring count', 'module': 'RingCount', 'constructor': 'RingCount(10, False, False, False, True)'}, 'n11AHRing': {'function': mordred.RingCount.RingCount(11, False, False, False, True), 'type': '2D', 'description': '11-membered aliphatic hetero ring count', 'module': 'RingCount', 'constructor': 'RingCount(11, False, False, False, True)'}, 'n12AHRing': {'function': mordred.RingCount.RingCount(12, False, False, False, True), 'type': '2D', 'description': '12-membered aliphatic hetero ring count', 'module': 'RingCount', 'constructor': 'RingCount(12, False, False, False, True)'}, 'nG12AHRing': {'function': mordred.RingCount.RingCount(12, True, False, False, True), 'type': '2D', 'description': '12-or-greater-membered aliphatic hetero ring count', 'module': 'RingCount', 'constructor': 'RingCount(12, True, False, False, True)'}, 'nFRing': {'function': mordred.RingCount.RingCount(None, False, True, None, None), 'type': '2D', 'description': 'fused ring count', 'module': 'RingCount', 'constructor': 'RingCount(None, False, True, None, None)'}, 'n4FRing': {'function': mordred.RingCount.RingCount(4, False, True, None, None), 'type': '2D', 'description': '4-membered fused ring count', 'module': 'RingCount', 'constructor': 'RingCount(4, False, True, None, None)'}, 'n5FRing': {'function': mordred.RingCount.RingCount(5, False, True, None, None), 'type': '2D', 'description': '5-membered fused ring count', 'module': 'RingCount', 'constructor': 'RingCount(5, False, True, None, None)'}, 'n6FRing': {'function': mordred.RingCount.RingCount(6, False, True, None, None), 'type': '2D', 'description': '6-membered fused ring count', 'module': 'RingCount', 'constructor': 'RingCount(6, False, True, None, None)'}, 'n7FRing': {'function': mordred.RingCount.RingCount(7, False, True, None, None), 'type': '2D', 'description': '7-membered fused ring count', 'module': 'RingCount', 'constructor': 'RingCount(7, False, True, None, None)'}, 'n8FRing': {'function': mordred.RingCount.RingCount(8, False, True, None, None), 'type': '2D', 'description': '8-membered fused ring count', 'module': 'RingCount', 'constructor': 'RingCount(8, False, True, None, None)'}, 'n9FRing': {'function': mordred.RingCount.RingCount(9, False, True, None, None), 'type': '2D', 'description': '9-membered fused ring count', 'module': 'RingCount', 'constructor': 'RingCount(9, False, True, None, None)'}, 'n10FRing': {'function': mordred.RingCount.RingCount(10, False, True, None, None), 'type': '2D', 'description': '10-membered fused ring count', 'module': 'RingCount', 'constructor': 'RingCount(10, False, True, None, None)'}, 'n11FRing': {'function': mordred.RingCount.RingCount(11, False, True, None, None), 'type': '2D', 'description': '11-membered fused ring count', 'module': 'RingCount', 'constructor': 'RingCount(11, False, True, None, None)'}, 'n12FRing': {'function': mordred.RingCount.RingCount(12, False, True, None, None), 'type': '2D', 'description': '12-membered fused ring count', 'module': 'RingCount', 'constructor': 'RingCount(12, False, True, None, None)'}, 'nG12FRing': {'function': mordred.RingCount.RingCount(12, True, True, None, None), 'type': '2D', 'description': '12-or-greater-membered fused ring count', 'module': 'RingCount', 'constructor': 'RingCount(12, True, True, None, None)'}, 'nFHRing': {'function': mordred.RingCount.RingCount(None, False, True, None, True), 'type': '2D', 'description': 'fused hetero ring count', 'module': 'RingCount', 'constructor': 'RingCount(None, False, True, None, True)'}, 'n4FHRing': {'function': mordred.RingCount.RingCount(4, False, True, None, True), 'type': '2D', 'description': '4-membered fused hetero ring count', 'module': 'RingCount', 'constructor': 'RingCount(4, False, True, None, True)'}, 'n5FHRing': {'function': mordred.RingCount.RingCount(5, False, True, None, True), 'type': '2D', 'description': '5-membered fused hetero ring count', 'module': 'RingCount', 'constructor': 'RingCount(5, False, True, None, True)'}, 'n6FHRing': {'function': mordred.RingCount.RingCount(6, False, True, None, True), 'type': '2D', 'description': '6-membered fused hetero ring count', 'module': 'RingCount', 'constructor': 'RingCount(6, False, True, None, True)'}, 'n7FHRing': {'function': mordred.RingCount.RingCount(7, False, True, None, True), 'type': '2D', 'description': '7-membered fused hetero ring count', 'module': 'RingCount', 'constructor': 'RingCount(7, False, True, None, True)'}, 'n8FHRing': {'function': mordred.RingCount.RingCount(8, False, True, None, True), 'type': '2D', 'description': '8-membered fused hetero ring count', 'module': 'RingCount', 'constructor': 'RingCount(8, False, True, None, True)'}, 'n9FHRing': {'function': mordred.RingCount.RingCount(9, False, True, None, True), 'type': '2D', 'description': '9-membered fused hetero ring count', 'module': 'RingCount', 'constructor': 'RingCount(9, False, True, None, True)'}, 'n10FHRing': {'function': mordred.RingCount.RingCount(10, False, True, None, True), 'type': '2D', 'description': '10-membered fused hetero ring count', 'module': 'RingCount', 'constructor': 'RingCount(10, False, True, None, True)'}, 'n11FHRing': {'function': mordred.RingCount.RingCount(11, False, True, None, True), 'type': '2D', 'description': '11-membered fused hetero ring count', 'module': 'RingCount', 'constructor': 'RingCount(11, False, True, None, True)'}, 'n12FHRing': {'function': mordred.RingCount.RingCount(12, False, True, None, True), 'type': '2D', 'description': '12-membered fused hetero ring count', 'module': 'RingCount', 'constructor': 'RingCount(12, False, True, None, True)'}, 'nG12FHRing': {'function': mordred.RingCount.RingCount(12, True, True, None, True), 'type': '2D', 'description': '12-or-greater-membered fused hetero ring count', 'module': 'RingCount', 'constructor': 'RingCount(12, True, True, None, True)'}, 'nFaRing': {'function': mordred.RingCount.RingCount(None, False, True, True, None), 'type': '2D', 'description': 'aromatic fused ring count', 'module': 'RingCount', 'constructor': 'RingCount(None, False, True, True, None)'}, 'n4FaRing': {'function': mordred.RingCount.RingCount(4, False, True, True, None), 'type': '2D', 'description': '4-membered aromatic fused ring count', 'module': 'RingCount', 'constructor': 'RingCount(4, False, True, True, None)'}, 'n5FaRing': {'function': mordred.RingCount.RingCount(5, False, True, True, None), 'type': '2D', 'description': '5-membered aromatic fused ring count', 'module': 'RingCount', 'constructor': 'RingCount(5, False, True, True, None)'}, 'n6FaRing': {'function': mordred.RingCount.RingCount(6, False, True, True, None), 'type': '2D', 'description': '6-membered aromatic fused ring count', 'module': 'RingCount', 'constructor': 'RingCount(6, False, True, True, None)'}, 'n7FaRing': {'function': mordred.RingCount.RingCount(7, False, True, True, None), 'type': '2D', 'description': '7-membered aromatic fused ring count', 'module': 'RingCount', 'constructor': 'RingCount(7, False, True, True, None)'}, 'n8FaRing': {'function': mordred.RingCount.RingCount(8, False, True, True, None), 'type': '2D', 'description': '8-membered aromatic fused ring count', 'module': 'RingCount', 'constructor': 'RingCount(8, False, True, True, None)'}, 'n9FaRing': {'function': mordred.RingCount.RingCount(9, False, True, True, None), 'type': '2D', 'description': '9-membered aromatic fused ring count', 'module': 'RingCount', 'constructor': 'RingCount(9, False, True, True, None)'}, 'n10FaRing': {'function': mordred.RingCount.RingCount(10, False, True, True, None), 'type': '2D', 'description': '10-membered aromatic fused ring count', 'module': 'RingCount', 'constructor': 'RingCount(10, False, True, True, None)'}, 'n11FaRing': {'function': mordred.RingCount.RingCount(11, False, True, True, None), 'type': '2D', 'description': '11-membered aromatic fused ring count', 'module': 'RingCount', 'constructor': 'RingCount(11, False, True, True, None)'}, 'n12FaRing': {'function': mordred.RingCount.RingCount(12, False, True, True, None), 'type': '2D', 'description': '12-membered aromatic fused ring count', 'module': 'RingCount', 'constructor': 'RingCount(12, False, True, True, None)'}, 'nG12FaRing': {'function': mordred.RingCount.RingCount(12, True, True, True, None), 'type': '2D', 'description': '12-or-greater-membered aromatic fused ring count', 'module': 'RingCount', 'constructor': 'RingCount(12, True, True, True, None)'}, 'nFaHRing': {'function': mordred.RingCount.RingCount(None, False, True, True, True), 'type': '2D', 'description': 'aromatic fused hetero ring count', 'module': 'RingCount', 'constructor': 'RingCount(None, False, True, True, True)'}, 'n4FaHRing': {'function': mordred.RingCount.RingCount(4, False, True, True, True), 'type': '2D', 'description': '4-membered aromatic fused hetero ring count', 'module': 'RingCount', 'constructor': 'RingCount(4, False, True, True, True)'}, 'n5FaHRing': {'function': mordred.RingCount.RingCount(5, False, True, True, True), 'type': '2D', 'description': '5-membered aromatic fused hetero ring count', 'module': 'RingCount', 'constructor': 'RingCount(5, False, True, True, True)'}, 'n6FaHRing': {'function': mordred.RingCount.RingCount(6, False, True, True, True), 'type': '2D', 'description': '6-membered aromatic fused hetero ring count', 'module': 'RingCount', 'constructor': 'RingCount(6, False, True, True, True)'}, 'n7FaHRing': {'function': mordred.RingCount.RingCount(7, False, True, True, True), 'type': '2D', 'description': '7-membered aromatic fused hetero ring count', 'module': 'RingCount', 'constructor': 'RingCount(7, False, True, True, True)'}, 'n8FaHRing': {'function': mordred.RingCount.RingCount(8, False, True, True, True), 'type': '2D', 'description': '8-membered aromatic fused hetero ring count', 'module': 'RingCount', 'constructor': 'RingCount(8, False, True, True, True)'}, 'n9FaHRing': {'function': mordred.RingCount.RingCount(9, False, True, True, True), 'type': '2D', 'description': '9-membered aromatic fused hetero ring count', 'module': 'RingCount', 'constructor': 'RingCount(9, False, True, True, True)'}, 'n10FaHRing': {'function': mordred.RingCount.RingCount(10, False, True, True, True), 'type': '2D', 'description': '10-membered aromatic fused hetero ring count', 'module': 'RingCount', 'constructor': 'RingCount(10, False, True, True, True)'}, 'n11FaHRing': {'function': mordred.RingCount.RingCount(11, False, True, True, True), 'type': '2D', 'description': '11-membered aromatic fused hetero ring count', 'module': 'RingCount', 'constructor': 'RingCount(11, False, True, True, True)'}, 'n12FaHRing': {'function': mordred.RingCount.RingCount(12, False, True, True, True), 'type': '2D', 'description': '12-membered aromatic fused hetero ring count', 'module': 'RingCount', 'constructor': 'RingCount(12, False, True, True, True)'}, 'nG12FaHRing': {'function': mordred.RingCount.RingCount(12, True, True, True, True), 'type': '2D', 'description': '12-or-greater-membered aromatic fused hetero ring count', 'module': 'RingCount', 'constructor': 'RingCount(12, True, True, True, True)'}, 'nFARing': {'function': mordred.RingCount.RingCount(None, False, True, False, None), 'type': '2D', 'description': 'aliphatic fused ring count', 'module': 'RingCount', 'constructor': 'RingCount(None, False, True, False, None)'}, 'n4FARing': {'function': mordred.RingCount.RingCount(4, False, True, False, None), 'type': '2D', 'description': '4-membered aliphatic fused ring count', 'module': 'RingCount', 'constructor': 'RingCount(4, False, True, False, None)'}, 'n5FARing': {'function': mordred.RingCount.RingCount(5, False, True, False, None), 'type': '2D', 'description': '5-membered aliphatic fused ring count', 'module': 'RingCount', 'constructor': 'RingCount(5, False, True, False, None)'}, 'n6FARing': {'function': mordred.RingCount.RingCount(6, False, True, False, None), 'type': '2D', 'description': '6-membered aliphatic fused ring count', 'module': 'RingCount', 'constructor': 'RingCount(6, False, True, False, None)'}, 'n7FARing': {'function': mordred.RingCount.RingCount(7, False, True, False, None), 'type': '2D', 'description': '7-membered aliphatic fused ring count', 'module': 'RingCount', 'constructor': 'RingCount(7, False, True, False, None)'}, 'n8FARing': {'function': mordred.RingCount.RingCount(8, False, True, False, None), 'type': '2D', 'description': '8-membered aliphatic fused ring count', 'module': 'RingCount', 'constructor': 'RingCount(8, False, True, False, None)'}, 'n9FARing': {'function': mordred.RingCount.RingCount(9, False, True, False, None), 'type': '2D', 'description': '9-membered aliphatic fused ring count', 'module': 'RingCount', 'constructor': 'RingCount(9, False, True, False, None)'}, 'n10FARing': {'function': mordred.RingCount.RingCount(10, False, True, False, None), 'type': '2D', 'description': '10-membered aliphatic fused ring count', 'module': 'RingCount', 'constructor': 'RingCount(10, False, True, False, None)'}, 'n11FARing': {'function': mordred.RingCount.RingCount(11, False, True, False, None), 'type': '2D', 'description': '11-membered aliphatic fused ring count', 'module': 'RingCount', 'constructor': 'RingCount(11, False, True, False, None)'}, 'n12FARing': {'function': mordred.RingCount.RingCount(12, False, True, False, None), 'type': '2D', 'description': '12-membered aliphatic fused ring count', 'module': 'RingCount', 'constructor': 'RingCount(12, False, True, False, None)'}, 'nG12FARing': {'function': mordred.RingCount.RingCount(12, True, True, False, None), 'type': '2D', 'description': '12-or-greater-membered aliphatic fused ring count', 'module': 'RingCount', 'constructor': 'RingCount(12, True, True, False, None)'}, 'nFAHRing': {'function': mordred.RingCount.RingCount(None, False, True, False, True), 'type': '2D', 'description': 'aliphatic fused hetero ring count', 'module': 'RingCount', 'constructor': 'RingCount(None, False, True, False, True)'}, 'n4FAHRing': {'function': mordred.RingCount.RingCount(4, False, True, False, True), 'type': '2D', 'description': '4-membered aliphatic fused hetero ring count', 'module': 'RingCount', 'constructor': 'RingCount(4, False, True, False, True)'}, 'n5FAHRing': {'function': mordred.RingCount.RingCount(5, False, True, False, True), 'type': '2D', 'description': '5-membered aliphatic fused hetero ring count', 'module': 'RingCount', 'constructor': 'RingCount(5, False, True, False, True)'}, 'n6FAHRing': {'function': mordred.RingCount.RingCount(6, False, True, False, True), 'type': '2D', 'description': '6-membered aliphatic fused hetero ring count', 'module': 'RingCount', 'constructor': 'RingCount(6, False, True, False, True)'}, 'n7FAHRing': {'function': mordred.RingCount.RingCount(7, False, True, False, True), 'type': '2D', 'description': '7-membered aliphatic fused hetero ring count', 'module': 'RingCount', 'constructor': 'RingCount(7, False, True, False, True)'}, 'n8FAHRing': {'function': mordred.RingCount.RingCount(8, False, True, False, True), 'type': '2D', 'description': '8-membered aliphatic fused hetero ring count', 'module': 'RingCount', 'constructor': 'RingCount(8, False, True, False, True)'}, 'n9FAHRing': {'function': mordred.RingCount.RingCount(9, False, True, False, True), 'type': '2D', 'description': '9-membered aliphatic fused hetero ring count', 'module': 'RingCount', 'constructor': 'RingCount(9, False, True, False, True)'}, 'n10FAHRing': {'function': mordred.RingCount.RingCount(10, False, True, False, True), 'type': '2D', 'description': '10-membered aliphatic fused hetero ring count', 'module': 'RingCount', 'constructor': 'RingCount(10, False, True, False, True)'}, 'n11FAHRing': {'function': mordred.RingCount.RingCount(11, False, True, False, True), 'type': '2D', 'description': '11-membered aliphatic fused hetero ring count', 'module': 'RingCount', 'constructor': 'RingCount(11, False, True, False, True)'}, 'n12FAHRing': {'function': mordred.RingCount.RingCount(12, False, True, False, True), 'type': '2D', 'description': '12-membered aliphatic fused hetero ring count', 'module': 'RingCount', 'constructor': 'RingCount(12, False, True, False, True)'}, 'nG12FAHRing': {'function': mordred.RingCount.RingCount(12, True, True, False, True), 'type': '2D', 'description': '12-or-greater-membered aliphatic fused hetero ring count', 'module': 'RingCount', 'constructor': 'RingCount(12, True, True, False, True)'}, 'nRot': {'function': mordred.RotatableBond.RotatableBondsCount(), 'type': '2D', 'description': 'rotatable bonds count', 'module': 'RotatableBond', 'constructor': 'RotatableBondsCount()'}, 'RotRatio': {'function': mordred.RotatableBond.RotatableBondsRatio(), 'type': '2D', 'description': 'rotatable bonds ratio', 'module': 'RotatableBond', 'constructor': 'RotatableBondsRatio()'}, 'SLogP': {'function': mordred.SLogP.SLogP(), 'type': '2D', 'description': 'Wildman-Crippen LogP', 'module': 'SLogP', 'constructor': 'SLogP()'}, 'SMR': {'function': mordred.SLogP.SMR(), 'type': '2D', 'description': 'Wildman-Crippen MR', 'module': 'SLogP', 'constructor': 'SMR()'}, 'TopoPSA(NO)': {'function': mordred.TopoPSA.TopoPSA(True), 'type': '2D', 'description': 'topological polar surface area(use only nitrogen and oxygen)', 'module': 'TopoPSA', 'constructor': 'TopoPSA(True)'}, 'TopoPSA': {'function': mordred.TopoPSA.TopoPSA(False), 'type': '2D', 'description': 'topological polar surface area', 'module': 'TopoPSA', 'constructor': 'TopoPSA(False)'}, 'GGI1': {'function': mordred.TopologicalCharge.TopologicalCharge('raw', 1), 'type': '2D', 'description': '1-ordered raw topological charge', 'module': 'TopologicalCharge', 'constructor': 'TopologicalCharge("raw", 1)'}, 'GGI2': {'function': mordred.TopologicalCharge.TopologicalCharge('raw', 2), 'type': '2D', 'description': '2-ordered raw topological charge', 'module': 'TopologicalCharge', 'constructor': 'TopologicalCharge("raw", 2)'}, 'GGI3': {'function': mordred.TopologicalCharge.TopologicalCharge('raw', 3), 'type': '2D', 'description': '3-ordered raw topological charge', 'module': 'TopologicalCharge', 'constructor': 'TopologicalCharge("raw", 3)'}, 'GGI4': {'function': mordred.TopologicalCharge.TopologicalCharge('raw', 4), 'type': '2D', 'description': '4-ordered raw topological charge', 'module': 'TopologicalCharge', 'constructor': 'TopologicalCharge("raw", 4)'}, 'GGI5': {'function': mordred.TopologicalCharge.TopologicalCharge('raw', 5), 'type': '2D', 'description': '5-ordered raw topological charge', 'module': 'TopologicalCharge', 'constructor': 'TopologicalCharge("raw", 5)'}, 'GGI6': {'function': mordred.TopologicalCharge.TopologicalCharge('raw', 6), 'type': '2D', 'description': '6-ordered raw topological charge', 'module': 'TopologicalCharge', 'constructor': 'TopologicalCharge("raw", 6)'}, 'GGI7': {'function': mordred.TopologicalCharge.TopologicalCharge('raw', 7), 'type': '2D', 'description': '7-ordered raw topological charge', 'module': 'TopologicalCharge', 'constructor': 'TopologicalCharge("raw", 7)'}, 'GGI8': {'function': mordred.TopologicalCharge.TopologicalCharge('raw', 8), 'type': '2D', 'description': '8-ordered raw topological charge', 'module': 'TopologicalCharge', 'constructor': 'TopologicalCharge("raw", 8)'}, 'GGI9': {'function': mordred.TopologicalCharge.TopologicalCharge('raw', 9), 'type': '2D', 'description': '9-ordered raw topological charge', 'module': 'TopologicalCharge', 'constructor': 'TopologicalCharge("raw", 9)'}, 'GGI10': {'function': mordred.TopologicalCharge.TopologicalCharge('raw', 10), 'type': '2D', 'description': '10-ordered raw topological charge', 'module': 'TopologicalCharge', 'constructor': 'TopologicalCharge("raw", 10)'}, 'JGI1': {'function': mordred.TopologicalCharge.TopologicalCharge('mean', 1), 'type': '2D', 'description': '1-ordered mean topological charge', 'module': 'TopologicalCharge', 'constructor': 'TopologicalCharge("mean", 1)'}, 'JGI2': {'function': mordred.TopologicalCharge.TopologicalCharge('mean', 2), 'type': '2D', 'description': '2-ordered mean topological charge', 'module': 'TopologicalCharge', 'constructor': 'TopologicalCharge("mean", 2)'}, 'JGI3': {'function': mordred.TopologicalCharge.TopologicalCharge('mean', 3), 'type': '2D', 'description': '3-ordered mean topological charge', 'module': 'TopologicalCharge', 'constructor': 'TopologicalCharge("mean", 3)'}, 'JGI4': {'function': mordred.TopologicalCharge.TopologicalCharge('mean', 4), 'type': '2D', 'description': '4-ordered mean topological charge', 'module': 'TopologicalCharge', 'constructor': 'TopologicalCharge("mean", 4)'}, 'JGI5': {'function': mordred.TopologicalCharge.TopologicalCharge('mean', 5), 'type': '2D', 'description': '5-ordered mean topological charge', 'module': 'TopologicalCharge', 'constructor': 'TopologicalCharge("mean", 5)'}, 'JGI6': {'function': mordred.TopologicalCharge.TopologicalCharge('mean', 6), 'type': '2D', 'description': '6-ordered mean topological charge', 'module': 'TopologicalCharge', 'constructor': 'TopologicalCharge("mean", 6)'}, 'JGI7': {'function': mordred.TopologicalCharge.TopologicalCharge('mean', 7), 'type': '2D', 'description': '7-ordered mean topological charge', 'module': 'TopologicalCharge', 'constructor': 'TopologicalCharge("mean", 7)'}, 'JGI8': {'function': mordred.TopologicalCharge.TopologicalCharge('mean', 8), 'type': '2D', 'description': '8-ordered mean topological charge', 'module': 'TopologicalCharge', 'constructor': 'TopologicalCharge("mean", 8)'}, 'JGI9': {'function': mordred.TopologicalCharge.TopologicalCharge('mean', 9), 'type': '2D', 'description': '9-ordered mean topological charge', 'module': 'TopologicalCharge', 'constructor': 'TopologicalCharge("mean", 9)'}, 'JGI10': {'function': mordred.TopologicalCharge.TopologicalCharge('mean', 10), 'type': '2D', 'description': '10-ordered mean topological charge', 'module': 'TopologicalCharge', 'constructor': 'TopologicalCharge("mean", 10)'}, 'JGT10': {'function': mordred.TopologicalCharge.TopologicalCharge('global', 10), 'type': '2D', 'description': '10-ordered global topological charge', 'module': 'TopologicalCharge', 'constructor': 'TopologicalCharge("global", 10)'}, 'Diameter': {'function': mordred.TopologicalIndex.Diameter(), 'type': '2D', 'description': 'topological diameter', 'module': 'TopologicalIndex', 'constructor': 'Diameter()'}, 'Radius': {'function': mordred.TopologicalIndex.Radius(), 'type': '2D', 'description': 'topological radius', 'module': 'TopologicalIndex', 'constructor': 'Radius()'}, 'TopoShapeIndex': {'function': mordred.TopologicalIndex.TopologicalShapeIndex(), 'type': '2D', 'description': 'topological shape index', 'module': 'TopologicalIndex', 'constructor': 'TopologicalShapeIndex()'}, 'PetitjeanIndex': {'function': mordred.TopologicalIndex.PetitjeanIndex(), 'type': '2D', 'description': 'Petitjean index', 'module': 'TopologicalIndex', 'constructor': 'PetitjeanIndex()'}, 'Vabc': {'function': mordred.VdwVolumeABC.VdwVolumeABC(), 'type': '2D', 'description': 'ABC van der waals volume', 'module': 'VdwVolumeABC', 'constructor': 'VdwVolumeABC()'}, 'VAdjMat': {'function': mordred.VertexAdjacencyInformation.VertexAdjacencyInformation(), 'type': '2D', 'description': 'vertex adjacency information', 'module': 'VertexAdjacencyInformation', 'constructor': 'VertexAdjacencyInformation()'}, 'MWC01': {'function': mordred.WalkCount.WalkCount(1, False, False), 'type': '2D', 'description': 'walk count(leg-1)', 'module': 'WalkCount', 'constructor': 'WalkCount(1, False, False)'}, 'MWC02': {'function': mordred.WalkCount.WalkCount(2, False, False), 'type': '2D', 'description': 'walk count(leg-2)', 'module': 'WalkCount', 'constructor': 'WalkCount(2, False, False)'}, 'MWC03': {'function': mordred.WalkCount.WalkCount(3, False, False), 'type': '2D', 'description': 'walk count(leg-3)', 'module': 'WalkCount', 'constructor': 'WalkCount(3, False, False)'}, 'MWC04': {'function': mordred.WalkCount.WalkCount(4, False, False), 'type': '2D', 'description': 'walk count(leg-4)', 'module': 'WalkCount', 'constructor': 'WalkCount(4, False, False)'}, 'MWC05': {'function': mordred.WalkCount.WalkCount(5, False, False), 'type': '2D', 'description': 'walk count(leg-5)', 'module': 'WalkCount', 'constructor': 'WalkCount(5, False, False)'}, 'MWC06': {'function': mordred.WalkCount.WalkCount(6, False, False), 'type': '2D', 'description': 'walk count(leg-6)', 'module': 'WalkCount', 'constructor': 'WalkCount(6, False, False)'}, 'MWC07': {'function': mordred.WalkCount.WalkCount(7, False, False), 'type': '2D', 'description': 'walk count(leg-7)', 'module': 'WalkCount', 'constructor': 'WalkCount(7, False, False)'}, 'MWC08': {'function': mordred.WalkCount.WalkCount(8, False, False), 'type': '2D', 'description': 'walk count(leg-8)', 'module': 'WalkCount', 'constructor': 'WalkCount(8, False, False)'}, 'MWC09': {'function': mordred.WalkCount.WalkCount(9, False, False), 'type': '2D', 'description': 'walk count(leg-9)', 'module': 'WalkCount', 'constructor': 'WalkCount(9, False, False)'}, 'MWC10': {'function': mordred.WalkCount.WalkCount(10, False, False), 'type': '2D', 'description': 'walk count(leg-10)', 'module': 'WalkCount', 'constructor': 'WalkCount(10, False, False)'}, 'TMWC10': {'function': mordred.WalkCount.WalkCount(10, True, False), 'type': '2D', 'description': 'total walk count(leg-10)', 'module': 'WalkCount', 'constructor': 'WalkCount(10, True, False)'}, 'SRW02': {'function': mordred.WalkCount.WalkCount(2, False, True), 'type': '2D', 'description': 'walk count(leg-2, only self returning walk)', 'module': 'WalkCount', 'constructor': 'WalkCount(2, False, True)'}, 'SRW03': {'function': mordred.WalkCount.WalkCount(3, False, True), 'type': '2D', 'description': 'walk count(leg-3, only self returning walk)', 'module': 'WalkCount', 'constructor': 'WalkCount(3, False, True)'}, 'SRW04': {'function': mordred.WalkCount.WalkCount(4, False, True), 'type': '2D', 'description': 'walk count(leg-4, only self returning walk)', 'module': 'WalkCount', 'constructor': 'WalkCount(4, False, True)'}, 'SRW05': {'function': mordred.WalkCount.WalkCount(5, False, True), 'type': '2D', 'description': 'walk count(leg-5, only self returning walk)', 'module': 'WalkCount', 'constructor': 'WalkCount(5, False, True)'}, 'SRW06': {'function': mordred.WalkCount.WalkCount(6, False, True), 'type': '2D', 'description': 'walk count(leg-6, only self returning walk)', 'module': 'WalkCount', 'constructor': 'WalkCount(6, False, True)'}, 'SRW07': {'function': mordred.WalkCount.WalkCount(7, False, True), 'type': '2D', 'description': 'walk count(leg-7, only self returning walk)', 'module': 'WalkCount', 'constructor': 'WalkCount(7, False, True)'}, 'SRW08': {'function': mordred.WalkCount.WalkCount(8, False, True), 'type': '2D', 'description': 'walk count(leg-8, only self returning walk)', 'module': 'WalkCount', 'constructor': 'WalkCount(8, False, True)'}, 'SRW09': {'function': mordred.WalkCount.WalkCount(9, False, True), 'type': '2D', 'description': 'walk count(leg-9, only self returning walk)', 'module': 'WalkCount', 'constructor': 'WalkCount(9, False, True)'}, 'SRW10': {'function': mordred.WalkCount.WalkCount(10, False, True), 'type': '2D', 'description': 'walk count(leg-10, only self returning walk)', 'module': 'WalkCount', 'constructor': 'WalkCount(10, False, True)'}, 'TSRW10': {'function': mordred.WalkCount.WalkCount(10, True, True), 'type': '2D', 'description': 'total walk count(leg-10, only self returning walk)', 'module': 'WalkCount', 'constructor': 'WalkCount(10, True, True)'}, 'MW': {'function': mordred.Weight.Weight(True, False), 'type': '2D', 'description': 'exact molecular weight', 'module': 'Weight', 'constructor': 'Weight(True, False)'}, 'AMW': {'function': mordred.Weight.Weight(True, True), 'type': '2D', 'description': 'averaged exact molecular weight', 'module': 'Weight', 'constructor': 'Weight(True, True)'}, 'WPath': {'function': mordred.WienerIndex.WienerIndex(False), 'type': '2D', 'description': 'Wiener index', 'module': 'WienerIndex', 'constructor': 'WienerIndex(False)'}, 'WPol': {'function': mordred.WienerIndex.WienerIndex(True), 'type': '2D', 'description': 'Wiener polarity index', 'module': 'WienerIndex', 'constructor': 'WienerIndex(True)'}, 'Zagreb1': {'function': mordred.ZagrebIndex.ZagrebIndex(1, 1), 'type': '2D', 'description': 'Zagreb index(version 1)', 'module': 'ZagrebIndex', 'constructor': 'ZagrebIndex(1, 1)'}, 'Zagreb2': {'function': mordred.ZagrebIndex.ZagrebIndex(2, 1), 'type': '2D', 'description': 'Zagreb index(version 2)', 'module': 'ZagrebIndex', 'constructor': 'ZagrebIndex(2, 1)'}, 'mZagreb1': {'function': mordred.ZagrebIndex.ZagrebIndex(1, -1), 'type': '2D', 'description': 'modified Zagreb index(version 1)', 'module': 'ZagrebIndex', 'constructor': 'ZagrebIndex(1, -1)'}, 'mZagreb2': {'function': mordred.ZagrebIndex.ZagrebIndex(2, -1), 'type': '2D', 'description': 'modified Zagreb index(version 2)', 'module': 'ZagrebIndex', 'constructor': 'ZagrebIndex(2, -1)'}}

    def __init__(self, descriptors_list=list(descriptors_dict.keys())):
        self.descriptors_list = descriptors_list
    def fit(self, X, y=None):
        return self
    def transform(self, X, y=None):
        smiles_list=list(X.index)
        data=[self.calculate_descriptors(self.descriptors_list, smiles) for smiles in smiles_list]
        descriptors_df = pd.DataFrame(data=data,columns=self.descriptors_list, index=smiles_list)
        df=pd.concat([X, descriptors_df], axis=1)
        return df
    @staticmethod
    def calculate_descriptors(descriptors_list, smiles, descriptors_dict=descriptors_dict):
        descriptor_functions_list=[descriptors_dict[name]['function'] for name in descriptors_list]
        calc=Calculator(descriptor_functions_list)(Chem.MolFromSmiles(smiles))    
        return [calc.name[name] for name in descriptors_list]



class calculateFunctionalGroups(BaseEstimator, TransformerMixin):
    fg_dict={'Alkene': '[CH]=[CH]',
            'Acrylate':'C=C[CX3](=O)(O)',
            'Methacrylate': '[CH3]C(=[CH2])C(=O)OC',
            'Vinyl': '[CX3]=[CX3]',
            'Thiol': '[SX2H]',
            'Ester': '[CX3](=O)O',
            'Ketone': '[CX3](=O)[CX3]',
            'Ether': '[CH2]-O-[CH2]',
            'Hydroxyl': '[OX2H]',
            'CH3': '[CH3]',
            'Methyl': '[CH2]',
            'Ethyl': '[CX2]=[CX3]',
            'Methyl Sulfide': '[SX2]',
            'nSi': '[Si]'}

    def __init__(self, fg_list=list(fg_dict.keys())):
        self.fg_list = fg_list
    def fit(self, X, y=None):
        return self
    def transform(self, X, y=None):
        smiles_list=list(X.index)
        fg_df=self.count_functional_groups(smiles_list,self.fg_list)
        df=pd.concat([X, fg_df], axis=1)
        return df

    @staticmethod
    def search_substructure(structure_smart,molecule):
        count = molecule.GetSubstructMatches(Chem.MolFromSmarts(structure_smart))
        return len(count)

    def count_functional_groups(self, smiles_list,fg_list,fg_dict=fg_dict): 
        mol_list = [Chem.MolFromSmiles(smiles) for smiles in  smiles_list]
        count_array=[[ self.search_substructure(fg_dict[fg],mol) for fg in fg_list] for mol in mol_list]
        df=pd.DataFrame(data=count_array, index=smiles_list, columns=fg_list)
        return df




class featureEngineering(BaseEstimator, TransformerMixin):

    @staticmethod
    def check_string_sanity(expr):
        import ast
        nodes_whitelist = {'Add','Attribute','BinOp','Call','Constant','Div','Expr','Index','Load','Module','Mult','Name','Sub','Subscript'}
        names_whitelist={'X'}
        nodes_used=set([x.__class__.__name__ for x in ast.walk(ast.parse(expr))])
        names_used=set([x.id for x in ast.walk(ast.parse(expr)) if x.__class__.__name__ == 'Name'])
        bad_code=(nodes_used-nodes_whitelist) | (names_used-names_whitelist)
        if bad_code:
            return False
        return True

    def __init__(self, operations_dict):
        self.operations_dict = operations_dict
    def fit(self, X, y=None):
        return self
    def transform(self, X, y=None):

        df=pd.DataFrame(index=X.index)
        for (k,v) in self.operations_dict.items():
            operation_string=re.sub(r"""(['|"][\w\-]+['|"])""",r'X[\1]',v)                        
            check=self.check_string_sanity(operation_string)
            if check:
                df[k]=eval(operation_string)
        return df